@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -24,12 +24,13 @@ type constant_info =
       (Str.regexp * Str.regexp list * string list)
       option (*coccigrep/gitgrep tokens*) *
       Get_constants2.combine option
+type merge_vars = string array list * string array list
+val union_merge_vars : merge_vars -> merge_vars -> merge_vars
 val pre_engine : (filename * filename) -> cocci_info * constant_info
 val worth_trying : filename list -> constant_info -> bool
 val full_engine :
-  cocci_info -> filename list -> (filename * filename option) list
-val post_engine : cocci_info -> unit
-val has_finalize : cocci_info -> bool
+  cocci_info -> filename list -> (filename * filename option) list * merge_vars
+val post_engine : cocci_info -> merge_vars -> unit
 
 (* because of the #include "toto.c" and also because we may associate the
  * same C file to multiple drivers because they share code, we can
@@ -44,14 +45,16 @@ val check_duplicate_modif :
 val sp_of_file :
   filename (* coccifile *)  -> filename option (* isofile *) ->
   Ast_cocci.metavar list list * Ast_cocci.rule list *
+      Ast_cocci.rule list (*scripts to compile*) *
       Ast_cocci.meta_name list list list *
-      Ast_cocci.meta_name list list list *
+      (Ast_cocci.meta_name list * Ast_cocci.meta_name list) list list * (*pos*)
       (Ast_cocci.meta_name list list list (*used after list*) *
 	 (*fresh used after list*)
 	 Ast_cocci.meta_name list list list *
 	 (*fresh used after list seeds*)
 	 Ast_cocci.meta_name list list list) *
       Ast_cocci.meta_name list list list * constant_info *
-      bool (* format information needed for strings? *)
+      bool (* format information needed for strings? *) *
+      bool (* contains modif in any rule *)
 
 val normalize_path : string -> string
