@@ -1,13 +1,11 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
 
 open Common
 
-open Ograph_extended
-
 module F = Control_flow_c
 
 (*****************************************************************************)
@@ -78,7 +76,7 @@ let valid_positions binding = function
 		match b with
 		  Ast_c.MetaPosValList l ->
 		    List.exists
-		      (function (_,elem,_,_) -> !Flag.current_element = elem)
+		      (function (_,elem,_,_,_) -> !Flag.current_element = elem)
 		      l
 		| _ ->
 		    failwith "position variable should have a position binding"
@@ -100,7 +98,7 @@ let loop_nodes p check nodes =
     [] nodes
 
 let labels_for_ctl (dropped_isos : string list)
-                    (nodes : (nodei * F.node) list)
+                    (nodes : (F.G.key * F.node) list)
                     (binding : Lib_engine.metavars_binding)
                     p =
   show_or_not_predicate p;
@@ -188,6 +186,10 @@ let labels_for_ctl (dropped_isos : string list)
 	  (function node ->
 	    match F.unwrap node with F.AfterNode _ -> true | _ -> false)
 
+    | Lib_engine.GotoAfter ->
+	loop_nodes p
+	  (function node -> F.unwrap node = F.AfterNode F.GotoAfterNode)
+
     | Lib_engine.FallThrough ->
 	loop_nodes p
 	  (function node ->
@@ -213,7 +215,7 @@ let labels_for_ctl (dropped_isos : string list)
     | Lib_engine.Exit ->
 	loop_nodes p
 	  (function node ->
-	    match F.unwrap node with F.Exit -> true | _ -> false)
+	    match F.unwrap node with F.Exit | F.EndNode -> true | _ -> false)
 
     | Lib_engine.ErrorExit ->
 	loop_nodes p
@@ -281,11 +283,9 @@ let quick_labels_for_ctl dropped_isos nodes binding p =
 (* Some fix flow, for CTL, for unparse *)
 (*****************************************************************************)
 (* could erase info on nodes, and edge, because they are not used by rene *)
-let (control_flow_for_ctl: F.cflow -> ('a, 'b) ograph_mutable) =
+let (control_flow_for_ctl: F.cflow -> 'a F.G.ograph_mutable) =
  fun cflow -> cflow
 
-
-
 (* Just make the final node of the control flow loop over itself.
  * It seems that one hypothesis of the SAT algorithm is that each node has at
  * least a successor.
@@ -329,39 +329,37 @@ let fix_flow_ctl2 (flow : F.cflow) : F.cflow =
 
     !g#add_arc ((exitnodei, exitnodei), F.Direct);
 
-    if ((!g#successors   errornodei)#tolist) = [] &&
-       ((!g#predecessors errornodei)#tolist) = []
+    if (F.KeyEdgeSet.is_empty (!g#successors errornodei)) &&
+       (F.KeyEdgeSet.is_empty (!g#predecessors errornodei))
     then !g#del_node errornodei
     else !g#add_arc ((errornodei, errornodei), F.Direct);
    with Not_found -> ()
   );
 
-  let fake_nodes = !g#nodes#tolist +> List.filter (fun (nodei, node) ->
-    match F.unwrap node with
+  let pred nodei node = match F.unwrap node with
     | F.CaseNode _
     | F.Enter
     (*| F.Fake*) (* [endif], [endswitch], ... *)
       -> true
-    | _ -> false
-    ) in
+    | _ -> false in
 
-  fake_nodes +> List.iter (fun (nodei, node) -> F.remove_one_node nodei !g);
+  let fake_nodes = F.KeyMap.filter pred !g#nodes in
+
+  F.KeyMap.iter (fun nodei node -> F.remove_one_node nodei !g) fake_nodes;
 
   (* even when have deadcode, julia want loop over those nodes *)
-  !g#nodes#tolist +> List.iter (fun (nodei, node) ->
-    if (!g#predecessors nodei)#null
+  F.KeyMap.iter (fun nodei node ->
+    if (F.KeyEdgeSet.is_empty (!g#predecessors nodei))
     then begin
       let fakei = !g#add_node (F.mk_node F.Fake [] [] "DEADCODELOOP") in
       !g#add_arc ((fakei, nodei), F.Direct);
       !g#add_arc ((fakei, fakei), F.Direct);
     end
-  );
+  ) !g#nodes;
 
-  !g#nodes#tolist +> List.iter (fun (nodei, node) ->
-    assert (List.length ((!g#successors nodei)#tolist) >= 1);
-    (* no:  && List.length ((!g#predecessors nodei)#tolist) >= 1
-       because    the enter node at least have no predecessors *)
-    );
+  F.KeyMap.iter (fun nodei node ->
+    assert (F.KeyEdgeSet.cardinal (!g#successors nodei) >= 1);
+    ) !g#nodes;
 
   !g
 let fix_flow_ctl a =
@@ -377,10 +375,10 @@ let fix_flow_ctl a =
  *)
 let model_for_ctl dropped_isos cflow binding =
  let newflow = cflow (* old: fix_flow_ctl (control_flow_for_ctl cflow) *) in
- let labels = labels_for_ctl dropped_isos (newflow#nodes#tolist) binding  in
+ let labels = labels_for_ctl dropped_isos (F.KeyMap.bindings newflow#nodes) binding  in
  let quicklabels =
-   quick_labels_for_ctl dropped_isos (newflow#nodes#tolist) binding  in
- let states = List.map fst  newflow#nodes#tolist  in
+   quick_labels_for_ctl dropped_isos (F.KeyMap.bindings newflow#nodes) binding  in
+ let states = List.map fst  (F.KeyMap.bindings newflow#nodes) in
  newflow, labels, quicklabels, states
 
 
@@ -440,6 +438,10 @@ module ENV =
       |	(Lib_engine.NormalMetaVal(Ast_c.MetaTypeVal a),
 	 Lib_engine.NormalMetaVal(Ast_c.MetaTypeVal b)) ->
           C_vs_c.eq_type a b
+      | (Lib_engine.NormalMetaVal(Ast_c.MetaStmtVal(a,_,_)),
+	 Lib_engine.NormalMetaVal(Ast_c.MetaStmtVal(b,_,_))) -> a = b
+      | (Lib_engine.NormalMetaVal(Ast_c.MetaDeclVal(a,_)),
+	 Lib_engine.NormalMetaVal(Ast_c.MetaDeclVal(b,_))) -> a = b
       |	(Lib_engine.LabelVal(l1),Lib_engine.LabelVal(l2)) ->
 	  compatible_labels l1 l2
       |	_ -> v = v'
@@ -467,14 +469,24 @@ module ENV =
 
 module CFG =
   struct
-    type node = Ograph_extended.nodei
-    type cfg = (F.node, F.edge) Ograph_extended.ograph_mutable
-    let predecessors cfg n = List.map fst ((cfg#predecessors n)#tolist)
-    let successors   cfg n = List.map fst ((cfg#successors n)#tolist)
+    type node = F.G.key
+    type cfg = F.node F.G.ograph_mutable
+    let predecessors cfg n =
+      List.map fst (F.KeyEdgeSet.elements (cfg#predecessors n))
+    let successors   cfg n =
+      List.map fst (F.KeyEdgeSet.elements (cfg#successors n))
+    let direct_predecessors cfg n =
+      List.map fst
+	(List.filter (fun (a,c) -> c = F.Direct)
+	   (F.KeyEdgeSet.elements (cfg#predecessors n)))
+    let direct_successors   cfg n =
+      List.map fst
+	(List.filter (fun (a,c) -> c = F.Direct)
+	   (F.KeyEdgeSet.elements (cfg#successors n)))
     let extract_is_loop cfg n =
-      Control_flow_c.extract_is_loop (cfg#nodes#find n)
+      Control_flow_c.extract_is_loop (F.KeyMap.find n cfg#nodes)
     let print_node i = Format.print_string (string_of_int i)
-    let size cfg = cfg#nodes#length
+    let size cfg = F.KeyMap.cardinal cfg#nodes
 
     (* In ctl_engine, we use 'node' for the node but in the Ograph_extended
      * terminology, this 'node' is in fact an index to access the real
@@ -482,7 +494,7 @@ module CFG =
      * the 'Ograph_extended.nodei'.
      *)
     let print_graph cfg label border_colors fill_colors filename =
-      Ograph_extended.print_ograph_mutable_generic cfg label
+      F.G.print_ograph_mutable_generic cfg label
         (fun (nodei, (node: F.node)) ->
           (* the string julia wants to put ? *)
           let bc = try Some(List.assoc nodei border_colors) with _ -> None in
@@ -521,9 +533,9 @@ let metavars_binding_to_binding2 binding =
 
 let (satbis_to_trans_info:
   (int list *
-     (nodei * Lib_engine.metavars_binding2 * Lib_engine.predicate)) list ->
+     (F.G.key * Lib_engine.metavars_binding2 * Lib_engine.predicate)) list ->
   (int list *
-     (nodei * Lib_engine.metavars_binding * Ast_cocci.rule_elem)) list) =
+     (F.G.key * Lib_engine.metavars_binding * Ast_cocci.rule_elem)) list) =
   fun xs ->
     xs +> List.fold_left (fun prev (index,(nodei, binding2, pred)) ->
       match pred with
@@ -563,8 +575,8 @@ let strip env =
 	    Ast_c.MetaExprVal(Lib_parsing_c.al_inh_expr a,c,ty)
 	| Ast_c.MetaExprListVal a ->
 	    Ast_c.MetaExprListVal(Lib_parsing_c.al_inh_arguments a)
-	| Ast_c.MetaStmtVal(a,ty) ->
-	    Ast_c.MetaStmtVal(Lib_parsing_c.al_inh_statement a,ty)
+	| Ast_c.MetaStmtVal(a,orig,ty) ->
+	    Ast_c.MetaStmtVal(Lib_parsing_c.al_inh_statement a,orig,ty)
 	| Ast_c.MetaInitVal a ->
 	    Ast_c.MetaInitVal(Lib_parsing_c.al_inh_init a)
 	| Ast_c.MetaInitListVal a ->
