@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -75,6 +75,8 @@ module XMATCH = struct
 
   type ('a, 'b) matcher = 'a -> 'b  -> tin -> ('a * 'b) tout
 
+  let constraint_checker = ref (fun _ -> failwith "unbound constraint_checker")
+
   (* was >&&> *)
   let (>>=) m1 m2 = fun tin ->
     let xs = m1 tin in
@@ -135,6 +137,11 @@ module XMATCH = struct
     then m tin
     else fail tin
 
+  let mnot f res = fun tin ->
+    match f tin with
+      [] -> return res tin
+    | _ -> fail tin
+
 
   let mode = Cocci_vs_c.PatternMode
 
@@ -272,6 +279,11 @@ module XMATCH = struct
       Ast_c.AbstractLineTok pi -> true
     | _ -> false
 
+  let is_fake ii = (* likely an invisible comma *)
+    match Ast_c.pinfo_of_info ii with
+      Ast_c.FakeTok _ -> true
+    | _ -> false
+
   let distrf (ii_of_x_f) =
     fun mcode x -> fun tin ->
       let iis = ii_of_x_f x in
@@ -324,16 +336,32 @@ module XMATCH = struct
           | _ (* success *) -> fail tin in
     loop constraints
 
-  let check_constraints matcher constraints term = fun f tin ->
-    if matcher constraints term (function id -> tin.binding0 +> List.assoc id)
-    then f () tin (* success *)
-    else fail tin (* failure *)
+  let check_constraints ida idb constraints f =
+    (fun tin ->
+      !constraint_checker ida idb
+	(function id -> tin.binding0 +> List.assoc id)
+	constraints tin)
+      >>= (fun _ _ -> f ())
 
-  let check_pos_constraints pname constraints pvalu f tin =
-    let res =
-      Cocci_vs_c.satisfies_constraint constraints (pname, pvalu)
-	(fun name -> List.assoc name tin.binding0) in
-    if res then f () tin (* success *) else fail tin (* failure *)
+  let check_re_constraints pname constraints f =
+    (fun tin ->
+      let myrule = fst pname in
+      try
+	let pvalu = List.assoc pname tin.binding in
+	!constraint_checker pname pvalu
+	  (fun ((rl,_) as name) ->
+	    let env = if rl = myrule then tin.binding else tin.binding0 in
+	    List.assoc name env)
+	  constraints tin
+      with Not_found -> return ((),()) tin)
+      >>= (fun _ _ -> f ())
+
+  let check_pos_constraints pname pvalu constraints f =
+    (fun tin ->
+      !constraint_checker pname pvalu
+	(fun name -> List.assoc name tin.binding0)
+	constraints tin)
+      >>= (fun _ _ -> f)
 
   (* ------------------------------------------------------------------------*)
   (* Environment *)
@@ -409,12 +437,12 @@ module XMATCH = struct
 		   then Lib_parsing_c.al_arguments a
 		   else Lib_parsing_c.semi_al_arguments a))
 
-          | Ast_c.MetaDeclVal a ->
-	      success
-		(Ast_c.MetaDeclVal
-		   (if strip
-		   then Lib_parsing_c.al_declaration a
-		   else Lib_parsing_c.semi_al_declaration a))
+          | Ast_c.MetaDeclVal(a,original) ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_declaration a
+		else Lib_parsing_c.semi_al_declaration a in
+	      success(Ast_c.MetaDeclVal(stripped,original))
           | Ast_c.MetaFieldVal a ->
 	      success
 		(Ast_c.MetaFieldVal
@@ -427,12 +455,12 @@ module XMATCH = struct
 		   (if strip
 		   then Lib_parsing_c.al_fields a
 		   else Lib_parsing_c.semi_al_fields a))
-          | Ast_c.MetaStmtVal(a,ty) ->
+          | Ast_c.MetaStmtVal(a,original,ty) ->
 	      let stripped =
 		if strip
 		then Lib_parsing_c.al_statement a
 		else Lib_parsing_c.semi_al_statement a in
-	      success(Ast_c.MetaStmtVal(stripped,ty))
+	      success(Ast_c.MetaStmtVal(stripped,original,ty))
           | Ast_c.MetaStmtListVal(a,ty) ->
 	      let stripped =
 		if strip
@@ -504,25 +532,34 @@ module XMATCH = struct
     match Ast_cocci.get_pos_var ia with
       [] -> finish tin
     | positions ->
-	let pvalu = Ast_c.MetaPosValList(get_pvalu()) in
-	let rec loop tin = function
-	    [] -> finish tin
-	  | Ast_cocci.MetaPos(name,constraints,per,keep,inherited) :: rest ->
-	      let name' = Ast_cocci.unwrap_mcode name in
-	      check_pos_constraints name' constraints pvalu
-		(function () ->
-	    (* constraints are satisfied, now see if we are compatible
-	       with existing bindings *)
-		  function new_tin ->
-		    let x = Ast_cocci.unwrap_mcode name in
-		    let new_binding =
-		      check_add_metavars_binding false keep inherited
-			(x, pvalu) tin in
-		    (match  new_binding with
-		      Some binding -> loop {tin with binding = binding} rest
-		    | None -> fail tin))
-		tin in
-	loop tin positions
+	match get_pvalu() with
+	  None -> finish tin
+	| Some pvalu ->
+	    let pvalu =
+	      List.map
+		(function (fname,current_element,st,ed) ->
+		  (fname,current_element,
+		   Some(Lazy.force (!Flag.current_element_pos)),st,ed))
+		pvalu in
+	    let pvalu = Ast_c.MetaPosValList(pvalu) in
+	    let rec loop tin = function
+		[] -> finish tin
+	      | Ast_cocci.MetaPos(name,constraints,per,keep,inherited)::rest ->
+		  let name' = Ast_cocci.unwrap_mcode name in
+		  check_pos_constraints name' pvalu constraints
+		    (fun new_tin ->
+		      (* constraints are satisfied, now see if we are
+			 compatible with existing bindings *)
+		      let x = Ast_cocci.unwrap_mcode name in
+		      let new_binding =
+			check_add_metavars_binding false keep inherited
+			  (x, pvalu) tin in
+		      (match  new_binding with
+			Some binding ->
+			  loop {new_tin with binding = binding} rest
+		      | None -> fail tin))
+		    tin in
+	    loop tin positions
 
   let envf keep inherited = fun (k, valu, get_max_min) f tin ->
     let x = Ast_cocci.unwrap_mcode k in
@@ -531,8 +568,9 @@ module XMATCH = struct
 	let new_tin = {tin with binding = binding} in
 	pos_variables new_tin k
 	  (function _ ->
-	    let (file,current_element,min,max) = get_max_min() in
-	    [(file,current_element,min,max)])
+	    match get_max_min() with
+	      Some pos -> Some [pos]
+	    | None -> None)
 	  (f ())
     | None -> fail tin
 
@@ -583,7 +621,10 @@ module XMATCH = struct
       let posmck = Ast_cocci.FixPos (pos, pos) in
       let finish tin = tag_mck_pos_mcode ia posmck ib tin in
       pos_variables tin ia
-	(function _ -> [Lib_parsing_c.lin_col_by_pos [ib]])
+	(function _ ->
+	  if is_fake ib
+	  then None
+	  else Some [Lib_parsing_c.lin_col_by_pos [ib]])
 	finish
 
   let tokenf_mck mck ib = fun tin ->
