@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -30,13 +30,13 @@ let rec pp_binding_kind = function
   | Ast_c.MetaInitListVal      ini -> pp "<<initlist>>"
   | Ast_c.MetaTypeVal      typ ->
       Pretty_print_c.pp_type_simple typ
-  | Ast_c.MetaDeclVal      decl ->
+  | Ast_c.MetaDeclVal      (decl,_) ->
       Pretty_print_c.pp_decl_simple decl
   | Ast_c.MetaFieldVal      decl ->
       Pretty_print_c.pp_field_simple decl
   | Ast_c.MetaFieldListVal      decls ->
       List.iter Pretty_print_c.pp_field_simple decls
-  | Ast_c.MetaStmtVal      (statement,_) ->
+  | Ast_c.MetaStmtVal      (statement,_,_) ->
       Pretty_print_c.pp_statement_simple statement
   | Ast_c.MetaStmtListVal      (statxs,_) ->
       Pretty_print_c.pp_statement_seq_list_simple statxs
@@ -56,9 +56,16 @@ let rec pp_binding_kind = function
       pp (Printf.sprintf ("poss[%s]")
 	    (String.concat ", "
 	       (List.map
-		  (function (fl,ce,(minl,minc),(maxl,maxc)) ->
-		    Printf.sprintf "(%s,%s,(%d,%d),(%d,%d))"
-		      fl ce minl minc maxl maxc)
+		  (function
+		      (fl,ce,None,(minl,minc),(maxl,maxc)) ->
+			Printf.sprintf "(%s,%s,unknown,(%d,%d),(%d,%d))"
+			  fl ce minl minc maxl maxc
+		    | (fl,ce,Some((ceminl,ceminc),(cemaxl,cemaxc)),
+		       (minl,minc),(maxl,maxc)) ->
+			Printf.sprintf
+			   "(%s,%s,((%d,%d),(%d,%d)),(%d,%d),(%d,%d))"
+			   fl ce ceminl ceminc cemaxl cemaxc
+			   minl minc maxl maxc)
 		  l)))
   | Ast_c.MetaNoVal -> pp "no value"
 
@@ -100,6 +107,7 @@ let pp_predicate = function
   | EscTrueBranch -> pp "EscTrueBranch"
   | FalseBranch -> pp "FalseBranch"
   | After -> pp "After"
+  | GotoAfter -> pp "GotoAfter"
   | FallThrough -> pp "FallThrough"
   | LoopFallThrough -> pp "LoopFallThrough"
   | Return -> pp "Return"
