@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -56,6 +56,13 @@ module XTRANS = struct
 
   type ('a, 'b) matcher = 'a -> 'b  -> tin -> ('a * 'b) tout
 
+  let constraint_checker =
+    ref ((fun _ -> failwith "unbound constraint_checker") :
+	   (Ast_cocci.meta_name ->
+            Ast_c.metavar_binding_kind ->
+            (Ast_cocci.meta_name -> Ast_c.metavar_binding_kind) ->
+            Ast_cocci.constraints -> tin -> (unit * unit) tout))
+
   let (>>=) m f = fun tin ->
      match m tin with
      | None -> None
@@ -78,6 +85,11 @@ module XTRANS = struct
   let (>&&>) f m = fun tin ->
     if f tin then m tin else fail tin
 
+  let mnot f res = fun tin ->
+    match f tin with
+      None -> return res tin
+    | _ -> fail tin
+
   let optional_storage_flag f = fun tin ->
     f (tin.extra.optional_storage_iso) tin
 
@@ -116,10 +128,16 @@ module XTRANS = struct
 	    (v,Ast_c.MetaInitVal(Lib_parsing_c.real_al_init i))
 	| Ast_c.MetaInitListVal(is) ->
 	    (v,Ast_c.MetaInitListVal(Lib_parsing_c.real_al_inits is))
-	| Ast_c.MetaDeclVal(d) ->
-	    (v,Ast_c.MetaDeclVal(Lib_parsing_c.real_al_decl d))
-	| Ast_c.MetaStmtVal(s,ty) ->
-	    (v,Ast_c.MetaStmtVal(Lib_parsing_c.real_al_statement s,ty))
+	| Ast_c.MetaDeclVal(d,original) ->
+	    (v,
+	     Ast_c.MetaDeclVal
+	       (Lib_parsing_c.real_al_decl d,
+		Lib_parsing_c.real_al_decl_with_comments original))
+	| Ast_c.MetaStmtVal(s,original,ty) ->
+	    (v,
+	     Ast_c.MetaStmtVal
+	       (Lib_parsing_c.real_al_statement s,
+		Lib_parsing_c.real_al_statement_with_comments original,ty))
 	(* These don't contain local variables, but the cocci_tag field
 	   causes problems too.  Why is this not needd for other metavars? *)
 	| Ast_c.MetaAssignOpVal(b) ->
@@ -233,31 +251,32 @@ module XTRANS = struct
       match Ast_cocci.unwrap e with
 	Ast_cocci.MetaId(name,constraints,u,i) ->
           Ast_cocci.rewrap e
-	    (Ast_cocci.MetaId(name,Ast_cocci.IdNoConstraint,u,i))
+	    (Ast_cocci.MetaId(name,Ast_cocci.CstrTrue,u,i))
       | Ast_cocci.MetaFunc(name,constraints,u,i) ->
           Ast_cocci.rewrap e
-	    (Ast_cocci.MetaFunc(name,Ast_cocci.IdNoConstraint,u,i))
+	    (Ast_cocci.MetaFunc(name,Ast_cocci.CstrTrue,u,i))
       | Ast_cocci.MetaLocalFunc(name,constraints,u,i) ->
           Ast_cocci.rewrap e
-	    (Ast_cocci.MetaLocalFunc(name,Ast_cocci.IdNoConstraint,u,i))
+	    (Ast_cocci.MetaLocalFunc(name,Ast_cocci.CstrTrue,u,i))
       |  _ -> e in
     let expression r k e =
       let e = k e in
       match Ast_cocci.unwrap e with
 	Ast_cocci.MetaErr(name,constraints,u,i) ->
 	  Ast_cocci.rewrap e
-	    (Ast_cocci.MetaErr(name,Ast_cocci.NoConstraint,u,i))
-      | Ast_cocci.MetaExpr(name,constraints,u,ty,form,i) ->
+	    (Ast_cocci.MetaErr(name,Ast_cocci.CstrTrue,u,i))
+      | Ast_cocci.MetaExpr(name,constraints,u,ty,form,i,bitfield) ->
           Ast_cocci.rewrap e
-	    (Ast_cocci.MetaExpr(name,Ast_cocci.NoConstraint,u,ty,form,i))
+	    (Ast_cocci.MetaExpr(name,Ast_cocci.CstrTrue,u,ty,form,i,bitfield))
       | _ -> e in
     let fn = Visitor_ast.rebuilder
 	mcode mcode mcode mcode mcode mcode mcode mcode mcode
 	mcode mcode mcode mcode mcode
-	donothing donothing donothing donothing donothing
+	donothing donothing donothing donothing donothing donothing
 	ident expression donothing donothing donothing donothing
-	donothing donothing donothing donothing donothing
-	donothing donothing donothing donothing donothing donothing in
+	donothing donothing donothing donothing donothing donothing
+	donothing donothing donothing donothing donothing donothing
+	donothing donothing in
 
   fn.Visitor_ast.rebuilder_anything anything
 
@@ -734,7 +753,7 @@ module XTRANS = struct
   let envf keep inherited = fun (s, value, _) f tin ->
     let s = Ast_cocci.unwrap_mcode s in
     let v =
-      if keep = Type_cocci.Saved
+      if keep = Ast_cocci.Saved
       then (
         try Some (List.assoc s tin.binding)
         with Not_found ->
@@ -771,7 +790,8 @@ module XTRANS = struct
         else fail tin
 
 
-  let check_constraints matcher c id = fun f tin -> f () tin
+  let check_constraints ida idb constraints = fun f tin -> f () tin
+  let check_re_constraints ida constraints = fun f tin -> f () tin
   let check_constraints_ne matcher constraints exp = fun f tin -> f () tin
 
   (* ------------------------------------------------------------------------*)
@@ -813,7 +833,7 @@ let (transform2: string (* rule name *) -> string list (* dropped_isos *) ->
 
    xs +> List.fold_left (fun acc (index, (nodei, binding, rule_elem)) ->
       (* subtil: not cflow#nodes but acc#nodes *)
-      let node  = acc#nodes#assoc nodei in
+      let node  = F.KeyMap.find nodei acc#nodes in
 
       if !Flag.show_transinfo
       then pr2 (Printf.sprintf "transform one node: %d" nodei);
