@@ -4,7 +4,6 @@ open Common
 *)
 
 
-(**/**)
 (** Coccinelle modules accessible from an ocaml script.
 *)
 include Exposed_modules
@@ -16,6 +15,18 @@ type pos = {
   current_element : string;
   (** {b current_element} is the name of the function containing the
       matched position *)
+  current_element_line : int;
+  (** {b line} is the number of the line containing the first
+      character of the matched position *)
+  current_element_col : int;
+  (** {b col} is the column containing the first character of the
+      matched position *)
+  current_element_line_end : int;
+  (** {b line_end} is the number of the line containing the last
+      character of the matched position *)
+  current_element_col_end : int;
+  (** {b col_end} is the column containing the last character of the
+       matched position. *)
   file :string ;
   (** {b file} is the name of the file containing the matched
       position *)
@@ -47,6 +58,7 @@ type param_type =
   | Int of int
   | Param of Ast_c.parameterType
   | ParamList of Ast_c.parameterType Ast_c.wrap2 list
+  | DParamList of (string Ast_c.wrap) Ast_c.wrap2 list
   | Expr of Ast_c.expression
   | ExprList of Ast_c.argument Ast_c.wrap2 list
   | Decl of Ast_c.declaration
@@ -60,7 +72,6 @@ type param_type =
 
 (* Function table management *)
 
-(**/**)
 (**
    For internal use only.
 *)
@@ -68,6 +79,14 @@ let fcts :
  (string, param_type list -> Ast_c.metavar_binding_kind ref list -> unit)
     Hashtbl.t =
   Hashtbl.create 11 (* Use prime number *)
+
+let bool_fcts :
+ (string, param_type list -> bool) Hashtbl.t =
+  Hashtbl.create 11 (* Use prime number *)
+
+let variables_to_merge: (unit -> string array) ref = ref (fun () -> [| |])
+
+let merged_variables: string list array option ref = ref None
 (**/**)
 
 (* This code needs to be here because we need to call the type annotater *)
@@ -82,10 +101,10 @@ let no_format s =
    not clear why that would be useful. *)
 let (cstatement_of_string: string -> string -> Ast_c.statement) =
   fun env s ->
-  assert (no_format s);
+(*  assert (no_format s);*) (* format strings are not getting parsed *)
   let tmpfile = Common.new_temp_file "cocci_stmt_of_s" "c" in
   Common.write_file tmpfile (Printf.sprintf "void main() {\n%s\n%s\n}" env s);
-  let program = Parse_c.parse_c_and_cpp false tmpfile +> fst in
+  let program = Parse_c.parse_c_and_cpp false false tmpfile +> fst in
   let _ =
     Type_annoter_c.annotate_program !Type_annoter_c.initial_env
       (List.map fst program) in
@@ -100,14 +119,20 @@ let (cstatement_of_string: string -> string -> Ast_c.statement) =
 
 let (cexpression_of_string: string -> string -> Ast_c.expression) =
   fun env s ->
-  assert (no_format s);
+(*  assert (no_format s);*) (* format strings are not getting parsed *)
   let tmpfile = Common.new_temp_file "cocci_expr_of_s" "c" in
   Common.write_file tmpfile (Printf.sprintf "void main() {\n%s\n%s;\n}" env s);
-  let program = Parse_c.parse_c_and_cpp false tmpfile +> fst in
+  let program = Parse_c.parse_c_and_cpp false false tmpfile +> fst in
   let _ =
     Type_annoter_c.annotate_program !Type_annoter_c.initial_env
       (List.map fst program) in
-  program +> Common.find_some (fun (e,_) ->
+  let rec find_some p = function
+    | [] -> failwith (Printf.sprintf "make_expr: no expression found in: %s" s)
+    | x :: l ->
+	match p x with
+	| Some v -> v
+	| None -> find_some p l in
+  program +> find_some (fun (e,_) ->
     match e with
     | Ast_c.Definition ({Ast_c.f_body = compound},_) ->
 	(match List.rev compound with
@@ -118,24 +143,43 @@ let (cexpression_of_string: string -> string -> Ast_c.expression) =
 	| _ -> None)
     | _ -> None)
 
+let parse_failure str s f =
+  try f s
+  with e -> Printf.eprintf "failed to parse %s %s\n" str s; raise e
+
 let make_ident s = Ast_c.MetaIdVal(s)
 let make_expr s =
-  Ast_c.MetaExprVal(Lib_parsing_c.al_expr(cexpression_of_string "" s), [],
-		    Ast_c.WITHOUT_TYPES)
+  parse_failure "expression" s
+    (function s ->
+      Ast_c.MetaExprVal(Lib_parsing_c.al_expr(cexpression_of_string "" s), [],
+			Ast_c.WITHOUT_TYPES))
 let make_expr_with_env env s =
-  Ast_c.MetaExprVal(Lib_parsing_c.al_expr(cexpression_of_string env s), [],
-		    Ast_c.WITH_TYPES)
+  parse_failure "expression" s
+    (function s ->
+      Ast_c.MetaExprVal(Lib_parsing_c.al_expr(cexpression_of_string env s), [],
+			Ast_c.WITH_TYPES))
 let make_stmt s =
-  Ast_c.MetaStmtVal(Lib_parsing_c.al_statement(cstatement_of_string "" s),
-		    Ast_c.WITHOUT_TYPES)
+  parse_failure "statement" s
+    (function s ->
+      let s = Lib_parsing_c.al_statement(cstatement_of_string "" s) in
+      Ast_c.MetaStmtVal(s,s,Ast_c.WITHOUT_TYPES))
 let make_stmt_with_env env s =
-  Ast_c.MetaStmtVal(Lib_parsing_c.al_statement(cstatement_of_string env s),
-		    Ast_c.WITH_TYPES)
+  parse_failure "statement" s
+    (function s ->
+      let s = Lib_parsing_c.al_statement(cstatement_of_string env s) in
+      Ast_c.MetaStmtVal(s,s,Ast_c.WITH_TYPES))
 let make_type s =
-  Ast_c.MetaTypeVal(Lib_parsing_c.al_type(Parse_c.type_of_string s))
+  parse_failure "type" s
+    (function s ->
+      Ast_c.MetaTypeVal(Lib_parsing_c.al_type(Parse_c.type_of_string s)))
 let make_listlen i = Ast_c.MetaListlenVal i
+let make_full_position fl fn ce_startl ce_startc ce_endl ce_endc
+    startl startc endl endc =
+  Ast_c.MetaPosValList
+    [(fl, fn, Some ((ce_startl,ce_startc),(ce_endl,ce_endc)),
+      (startl, startc), (endl,endc))]
 let make_position fl fn startl startc endl endc =
-  Ast_c.MetaPosValList [(fl, fn, (startl, startc), (endl,endc))]
+  Ast_c.MetaPosValList [(fl, fn, None, (startl, startc), (endl,endc))]
 
 (* ---------------------------------------------------------------------- *)
 (* Match management *)
@@ -164,8 +208,9 @@ let exit () = exited := true
 
 let dir () = !Flag.dir
 
-let file () =
-  match !Flag.currentfile with Some f -> f | None -> failwith "no file"
+let files () = !Flag.currentfiles
+
+let cocci_version () = Config.version
 
 (* ---------------------------------------------------------------------- *)
 (* org mode *)
