@@ -5,8 +5,22 @@ let string_value = function
       let locs =
 	List.map
 	  (function
-	      (fname,current_element,(line,col),(line_end,col_end)) ->
+	      (fname,current_element,current_element_pos,
+	       (line,col),(line_end,col_end)) ->
+		 let (current_element_line,current_element_col,
+		      current_element_line_end,current_element_col_end) =
+		   match current_element_pos with
+		     Some
+		       ((current_element_line,current_element_col),
+			(current_element_line_end,current_element_col_end)) ->
+			  (current_element_line,current_element_col,
+			   current_element_line_end,current_element_col_end)
+		   | None -> (-1,-1,-1,-1) in
 		{ Coccilib.current_element = current_element;
+		  Coccilib.current_element_line = current_element_line;
+		  Coccilib.current_element_col = current_element_col;
+		  Coccilib.current_element_line_end = current_element_line_end;
+		  Coccilib.current_element_col_end = current_element_col_end;
 		  Coccilib.file = fname;
 		  Coccilib.line = line;
 		  Coccilib.col = col;
@@ -38,10 +52,10 @@ let ast_binding vl = function
       | Ast_c.MetaTypeVal ty -> Coccilib.Type ty
       | Ast_c.MetaInitVal init -> Coccilib.Init init
       | Ast_c.MetaInitListVal init -> Coccilib.InitList init
-      | Ast_c.MetaDeclVal decl -> Coccilib.Decl decl
+      | Ast_c.MetaDeclVal(decl,_) -> Coccilib.Decl decl
       | Ast_c.MetaFieldVal field -> Coccilib.Field field
       | Ast_c.MetaFieldListVal field -> Coccilib.FieldList field
-      | Ast_c.MetaStmtVal(stm,_) -> Coccilib.Stmt stm
+      | Ast_c.MetaStmtVal(stm,_,_) -> Coccilib.Stmt stm
       | Ast_c.MetaStmtListVal(stm,_) -> Coccilib.StmtList stm
       | Ast_c.MetaFragListVal frags -> Coccilib.FragList frags
       | Ast_c.MetaFmtVal fmt -> Coccilib.Fmt fmt
