@@ -1,3 +1,5 @@
+let ocaml_support = true
+
 (* Note: this module passes paths to other commands, but does not take
  * quoting into account. Thus, if these paths contain spaces, it's likely
  * that things go wrong.
@@ -6,7 +8,7 @@
 module Ast = Ast_cocci
 
 exception CompileFailure of string
-exception LinkFailure of string
+exception LinkFailure of string * string
 
 let ext = if Config.dynlink_is_native then ".cmxs" else ".cma"
 
@@ -134,7 +136,7 @@ let print_iteration_code o =
     String.concat "\n              | "
       (List.map
 	 (function x -> Printf.sprintf "%s -> \"%s\""
-	     (String.capitalize x) x)
+	     (Stdcompat.String.capitalize_ascii x) x)
 	 l) in
   let add_virt_rules_method =
     match !Iteration.parsed_virtual_rules with
@@ -180,8 +182,8 @@ let print_iteration_code o =
 let prepare_mvs o str = function
     [] -> true
   | metavars ->
-      let fn _ =
-	List.map
+      try
+	List.iter
 	  (function
 	      ((Some nm,None),("virtual",vname),_,init) ->
 		let vl =
@@ -195,13 +197,10 @@ let prepare_mvs o str = function
 			raise Not_found
 		    | Ast.MVInitString s -> s
 		    | Ast.MVInitPosList -> failwith "no virt positions") in
-		(nm,vl)
-	    | _ -> failwith "invalid metavar in initialize or finalize")
-	  metavars in
-      try
-	List.iter
-	  (function (nm,vl) -> Printf.fprintf o "let %s = \"%s\"\n" nm vl)
-	  (fn());
+		Printf.fprintf o "let %s = \"%s\"\n" nm vl
+	    | ((Some _, None), ("merge", _), _, _) -> ()
+	    | _ -> failwith (Printf.sprintf "invalid metavar in %s" str))
+	  metavars;
 	Printf.fprintf o "\n";
 	true
       with Not_found -> false
@@ -248,9 +247,9 @@ let prepare coccifile code =
       List.fold_left
 	(function ((init,final) as prev) ->
 	  function
-	      Ast_cocci.InitialScriptRule (name,"ocaml",deps,mvs,code) ->
+	      Ast_cocci.InitialScriptRule (name,"ocaml",deps,mvs,_pos,code) ->
 		(Common.union_set mvs init,final)
-	    | Ast_cocci.FinalScriptRule (name,"ocaml",deps,mvs,code) ->
+	    | Ast_cocci.FinalScriptRule (name,"ocaml",deps,mvs,_pos,code) ->
 		(init,Common.union_set mvs final)
 	    | _ -> prev)
 	([],[]) code in
@@ -260,7 +259,7 @@ let prepare coccifile code =
     List.fold_left
       (function prev ->
 	function
-	    Ast_cocci.InitialScriptRule (name,"ocaml",deps,mvs,code) ->
+	    Ast_cocci.InitialScriptRule (name,"ocaml",deps,mvs,_pos,code) ->
 	      code :: prev
 	  | _ -> prev)
       [] code in
@@ -269,8 +268,8 @@ let prepare coccifile code =
     List.fold_left
       (function prev ->
 	function
-	    Ast_cocci.FinalScriptRule (name,"ocaml",deps,mvs,code) ->
-	      (name,[],[],code) :: prev
+	    Ast_cocci.FinalScriptRule (name,"ocaml",deps,mvs,_pos,code) ->
+	      (name,mvs,code) :: prev
 	  | _ -> prev)
       [] code in
   let final_rules = List.rev final_rules in
@@ -278,31 +277,25 @@ let prepare coccifile code =
     List.fold_left
       (function prev ->
 	function
-	    Ast_cocci.ScriptRule (name,"ocaml",deps,mv,script_vars,code) ->
+	    Ast_cocci.ScriptRule (name,"ocaml",deps,mv,script_vars,_pos,code) ->
 	      (name,mv,script_vars,code) :: prev
 	  | _ -> prev)
       [] code in
   let other_rules = List.rev other_rules in
-  let add_constraint_rules prev (kind, name, script_name, lang, params, body) =
+  let add_constraint_rules prev
+      (posvar, self, (script_name, lang, params, pos, body)) =
     if lang = "ocaml" then
-      let ((r,nm) as self) = Ast.unwrap_mcode name in
+      let kind =
+	if posvar then Ast_cocci.MetaPosDecl (Ast_cocci.NONE, self)
+	else Ast_cocci.MetaIdDecl (Ast_cocci.NONE, self) in
       let self = (self, kind) in
       (script_name, self::params, body) :: prev
     else
       prev in
   let constraint_rules =
-    List.fold_left
-      (function prev ->
-	function
-	    Ast.CocciRule(_,_,code,_,_) ->
-	      List.fold_left
-		(fun accu toplevel ->
-		  List.fold_left add_constraint_rules prev
-		    (Parse_cocci.enumerate_constraint_scripts toplevel))
-		prev code
-	  | _ -> prev)
-      [] code in
+    List.fold_left add_constraint_rules [] !Data.constraint_scripts in
   if init_rules = [] && other_rules = [] && constraint_rules = []
+      && final_rules = []
   then None
   else
     begin
@@ -314,14 +307,15 @@ let prepare coccifile code =
       then
 	Printf.fprintf o "type __virtual_rules__ = %s\n\n"
 	  (String.concat " | "
-	     (List.map String.capitalize !Iteration.parsed_virtual_rules)));
+	     (List.map Stdcompat.String.capitalize_ascii
+                !Iteration.parsed_virtual_rules)));
       (if !Iteration.parsed_virtual_identifiers != []
       then
 	Printf.fprintf o "type __virtual_identifiers__ = %s\n\n"
 	  (String.concat " | "
 	     (List.map
 		(function x -> Printf.sprintf "%s" x)
-		(List.map String.capitalize
+		(List.map Stdcompat.String.capitalize_ascii
 		   !Iteration.parsed_virtual_identifiers))));
       print_iteration_code o;
       (* Virtual metavariables for initialize and finalize rules *)
@@ -340,6 +334,39 @@ let prepare coccifile code =
       (* finalizer *)
       (if generate_final
       then
+	let merge_vars, local_vars =
+	  Ast_cocci.prepare_merge_variables
+	    (fun (name, mvs, code) -> Some ((name, code), mvs))
+	    final_rules in
+	let print_local_vars o =
+	  for i = 0 to Array.length local_vars - 1 do
+	    if i > 0 then
+	      Printf.fprintf o "; ";
+	    Printf.fprintf o "Marshal.to_string %s []" local_vars.(i)
+	  done in
+	Printf.fprintf o "
+let () =
+    Coccilib.variables_to_merge := (fun () -> [| %t |])
+" print_local_vars;
+	let add_merge_vars ((name, code), (from_index, merge_names)) =
+	  let let_merge_vars = Buffer.create 17 in
+	  let let_merge_var index merge_name =
+	    let global_index = from_index + index in
+	    Printf.bprintf let_merge_vars "\
+      let (%s : 'a%d list), (_ : 'a%d) =
+	  (
+	    match !(Coccilib.merged_variables) with
+	    | None -> []
+	    | Some merged_variables ->
+		List.map (fun s -> Marshal.from_string s 0)
+		  merged_variables.(%d)
+	  ), %s in\n"
+	      merge_name index index global_index
+	      local_vars.(global_index) in
+	  Array.iteri let_merge_var merge_names;
+	  let preambule = Buffer.contents let_merge_vars in
+	  (name, [], [], preambule ^ code) in
+	let final_rules = List.map add_merge_vars merge_vars in
 	let rule_code = List.map prepare_rule final_rules in
 	Printf.fprintf o "%s" (String.concat "\n\n" rule_code));
       close_out o;
@@ -438,14 +465,16 @@ let filter_dep existing_deps (accld, accinc) dep =
     | "Queue" | "Random" | "Scanf" | "Set" | "Sort" | "Stack" | "StdLabels"
     | "Str" | "Stream"
     | "String" | "StringLabels" | "Sys" | "ThreadUnix" | "Unix" | "UnixLabels"
-    | "Weak"     -> (accld, accinc)
+    | "Weak"
+        (* Filter out Parmap as Parmap is linked with Coccinelle *)
+    | "Parmap" -> (accld, accinc)
     | "Dbm"      -> ("dbm"::accld, accinc)
     | "Graphics" -> ("graphics"::accld, accinc)
     | "Thread"   -> ("thread"::accld, accinc)
     | "Tk"       -> ("tk"::accld, accinc)
 
     | _ ->
-        let dep = String.uncapitalize dep in
+        let dep = Stdcompat.String.uncapitalize_ascii dep in
 	  (accld, dep::accinc)
 
 let get_dir p =
@@ -524,8 +553,9 @@ let load_obj obj =
   Dynlink.allow_unsafe_modules true;
   try Dynlink.loadfile obj
   with Dynlink.Error e ->
-    Common.pr2 (Dynlink.error_message e);
-    raise (LinkFailure obj)
+    let msg = Dynlink.error_message e in
+    Common.pr2 msg;
+    raise (LinkFailure (obj, msg))
 
 let load_lib (dir, name) =
   let obj = dir ^ "/" ^name ^ ext in
@@ -542,7 +572,7 @@ let load_file mlfile =
   let flags =
     Printf.sprintf
       "-g -I %s %s -I %s"
-      (sysdir ()) inc (Filename.dirname cmifile) in
+      (Filename.dirname cmifile) inc (sysdir ()) in
   let (obj, cmd) =
     if Config.dynlink_is_native
     then compile_native_cmd flags mlfile
