@@ -197,6 +197,8 @@ and fullType = typeQualifier * typeC
 
   | TypeName   of name * fullType option (* semantic: filled later *)
 
+  | FieldType of fullType * name * constExpression option
+
   | ParenType of fullType (* for unparser: *)
 
   (* gccext: TypeOfType below may seems useless; Why declare a
@@ -349,7 +351,7 @@ and expression = (expressionbis * exp_info ref (* semantic: *)) wrap3
 
   (* for C++: *)
   | New of (argument wrap2 (* , *) list) option * argument
-  | Delete of expression
+  | Delete of bool (* true if [] *) * expression
 
   (* CPP [defined] operator, e.g. #if defined(A) *)
   | Defined of name
@@ -570,6 +572,7 @@ and declaration =
          v_storage: storage;
          v_local: local_decl; (* cocci: *)
          v_attr: attribute list; (* gccext: *)
+         v_endattr: attribute list; (* gccext: *)
        }
      and v_init =
        NoInit | ValInit of info * initialiser
@@ -809,10 +812,10 @@ and metavars_binding = (Ast_cocci.meta_name, metavar_binding_kind) assoc
   | MetaTypeVal      of fullType
   | MetaInitVal      of initialiser
   | MetaInitListVal  of initialiser wrap2 list
-  | MetaDeclVal      of declaration
+  | MetaDeclVal      of declaration * declaration
   | MetaFieldVal     of field
   | MetaFieldListVal of field list
-  | MetaStmtVal      of statement * stripped
+  | MetaStmtVal      of statement * statement * stripped
   | MetaStmtListVal  of statement_sequencable list * stripped
   | MetaDParamListVal of (string wrap) wrap2 list
   | MetaFmtVal       of string_format
@@ -825,8 +828,9 @@ and metavars_binding = (Ast_cocci.meta_name, metavar_binding_kind) assoc
    * variables accessible via SmPL whereas the position can be one day
    * so I think it's better to put MetaPosVal here *)
   | MetaPosVal       of (Ast_cocci.fixpos * Ast_cocci.fixpos) (* max, min *)
-  | MetaPosValList   of
-      (Common.filename * string (*element*) * posl * posl) list (* min, max *)
+  | MetaPosValList   of (* current elem, min, max *)
+      (Common.filename * string (*element*) * (posl * posl) option *
+	 posl * posl) list
   | MetaListlenVal   of int
   | MetaNoVal
 
@@ -1163,9 +1167,12 @@ let real_al_info x =
     danger = ref NoDanger;
   }
 
-let al_comments x =
+let al_comments keep_comments x =
   let keep_cpp l =
-    List.filter (function (Token_c.TCommentCpp _,_) -> true | _ -> false) l in
+    if keep_comments
+    then l
+    else
+      List.filter (function (Token_c.TCommentCpp _,_) -> true | _ -> false) l in
   let al_com (x,i) =
     (x,{i with Common.charpos = magic_real_number;
 	 Common.line = magic_real_number;
@@ -1187,7 +1194,7 @@ let al_info_cpp tokenindex x =
 	 str = str_of_info x});
     cocci_tag = ref emptyAnnot;
     annots_tag = Token_annot.empty;
-    comments_tag = ref (al_comments !(x.comments_tag));
+    comments_tag = ref (al_comments false !(x.comments_tag));
     danger = ref NoDanger;
   }
 
@@ -1195,11 +1202,11 @@ let semi_al_info_cpp x =
   { x with
     cocci_tag = ref emptyAnnot;
     annots_tag = Token_annot.empty;
-    comments_tag = ref (al_comments !(x.comments_tag));
+    comments_tag = ref (al_comments false !(x.comments_tag));
     danger = ref NoDanger;
   }
 
-let real_al_info_cpp x =
+let real_al_info_cpp keep_comments x =
   { pinfo =
     (AbstractLineTok
        {charpos = magic_real_number;
@@ -1209,7 +1216,7 @@ let real_al_info_cpp x =
 	 str = str_of_info x});
     cocci_tag = ref emptyAnnot;
     annots_tag = Token_annot.empty;
-    comments_tag =  ref (al_comments !(x.comments_tag));
+    comments_tag =  ref (al_comments keep_comments !(x.comments_tag));
     danger = ref NoDanger;
   }
 
