@@ -18,10 +18,6 @@ open Common
 open Ast_c
 open Control_flow_c
 
-open Ograph_extended
-open Oassoc
-open Oassocb
-
 module Lib = Lib_parsing_c
 
 (*****************************************************************************)
@@ -96,6 +92,11 @@ let pinfo_of_ii ii = Ast_c.get_opi (List.hd ii).Ast_c.pinfo
  * the context point to close the good number of '}' . For instance
  * where there is a 'continue', we must close only until the for.
  *)
+
+type nodei = Control_flow_c.G.key 
+
+module StringMap = Map.Make (String)
+
 type braceinfo =
     (node * (after_type -> string -> nodei -> unit), node) Common.either
 
@@ -123,7 +124,7 @@ and xinfo =  {
   compound_caller: compound_caller;
 
   (* does not change recursively. Some kind of globals. *)
-  labels_assoc: (string, nodei) oassoc;
+  labels_assoc: nodei StringMap.t;
   exiti:      nodei option;
   errorexiti: nodei option;
 
@@ -147,7 +148,7 @@ let initial_info = {
   labels = [];
 
   (* don't change when recurse *)
-  labels_assoc = new oassocb [];
+  labels_assoc = StringMap.empty;
   exiti = None;
   errorexiti = None;
 }
@@ -157,7 +158,7 @@ let initial_info = {
 (* (Semi) Globals, Julia's style. *)
 (*****************************************************************************)
 (* global graph *)
-let g = ref (new ograph_mutable)
+let g = ref (new Control_flow_c.G.ograph_mutable)
 
 let counter_for_labels = ref 0
 let counter_for_braces = ref 0
@@ -180,7 +181,7 @@ let counter_for_switch = ref 0
 let compute_labels_and_create_them st =
 
   (* map C label to index number in graph *)
-  let (h: (string, nodei) oassoc ref) = ref (new oassocb []) in
+  let h = ref StringMap.empty in
 
   begin
     st +> Visitor_c.vk_statement { Visitor_c.default_visitor_c with
@@ -195,8 +196,8 @@ let compute_labels_and_create_them st =
             in
             begin
               (* the C label already exists ? *)
-              if (!h#haskey s) then raise (Error (DuplicatedLabel s));
-              h := !h#add (s, newi);
+              if (StringMap.mem s !h) then raise (Error (DuplicatedLabel s));
+              h := StringMap.add s newi !h;
               (* not k _st !!! otherwise in lbl1: lbl2: i++; we miss lbl2 *)
               k st;
             end
@@ -342,9 +343,9 @@ let rec aux_statement : (nodei option * xinfo) -> statement -> nodei option =
 	if xi.compound_caller = Statement
 	then
 	  (let afteri = !g +> add_node (AfterNode ty) lbl str in
-	  let a1 = ((newi, afteri), Direct) in
+	  let a1 = ((newi, afteri), Control) in
 	  !g#add_arc a1;
-	  let a2 = ((afteri, endi), Direct) in
+	  let a2 = ((afteri, endi), Control) in
 	  !g#add_arc a2;
 	  ret_afters := (afteri,a1,a2) :: !ret_afters) in
 
@@ -388,8 +389,8 @@ let rec aux_statement : (nodei option * xinfo) -> statement -> nodei option =
    (* ------------------------- *)
   | Labeled (Ast_c.Label (name, st)) ->
       let s = Ast_c.str_of_name name in
-      let ilabel = xi.labels_assoc#find s in
-      let node = mk_node (unwrap (!g#nodes#find ilabel)) lbl [] (s ^ ":") in
+      let ilabel = StringMap.find s xi.labels_assoc in
+      let node = mk_node (unwrap (KeyMap.find ilabel !g#nodes)) lbl [] (s ^ ":") in
       !g#replace_node (ilabel, node);
       !g +> add_arc_opt (starti, ilabel);
       aux_statement (Some ilabel, xi_lbl) st
@@ -406,7 +407,7 @@ let rec aux_statement : (nodei option * xinfo) -> statement -> nodei option =
      then Some newi
      else begin
        let ilabel =
-         try xi.labels_assoc#find s
+         try StringMap.find s xi.labels_assoc
          with Not_found ->
                   (* jump vers ErrorExit a la place ?
                    * pourquoi tant de "cant jump" ? pas detecté par gcc ?
@@ -565,16 +566,16 @@ let rec aux_statement : (nodei option * xinfo) -> statement -> nodei option =
 
              let afteri =
 	       !g +> add_node (AfterNode NormalAfterNode) lbl "[after]" in
-             !g#add_arc ((newswitchi, afteri),  Direct);
-             !g#add_arc ((afteri, newendswitch), Direct);
+             !g#add_arc ((newswitchi, afteri),  Control);
+             !g#add_arc ((afteri, newendswitch), Control);
 
 
              !g#add_arc ((finalthen, newendswitch), Direct);
              Some newendswitch
          | None ->
-             if (!g#predecessors newendswitch)#null
+             if (KeyEdgeSet.is_empty (!g#predecessors newendswitch))
              then begin
-                 assert ((!g#successors newendswitch)#null);
+                 assert (KeyEdgeSet.is_empty (!g#successors newendswitch));
                  !g#del_node newendswitch;
                  None
              end
@@ -582,8 +583,8 @@ let rec aux_statement : (nodei option * xinfo) -> statement -> nodei option =
 
                let afteri =
 		 !g +> add_node (AfterNode NormalAfterNode) lbl "[after]" in
-               !g#add_arc ((newswitchi, afteri),  Direct);
-               !g#add_arc ((afteri, newendswitch), Direct);
+               !g#add_arc ((newswitchi, afteri),  Control);
+               !g#add_arc ((afteri, newendswitch), Control);
 
 
                Some newendswitch
@@ -723,8 +724,8 @@ let rec aux_statement : (nodei option * xinfo) -> statement -> nodei option =
         !g +> add_node (EndStatement (Some iifakeend)) lbl "[enddowhile]" in
 
       let afteri = !g +> add_node (AfterNode NormalAfterNode) lbl "[after]" in
-      !g#add_arc ((doi,afteri), Direct);
-      !g#add_arc ((afteri,newfakeelse), Direct);
+      !g#add_arc ((doi,afteri), Control);
+      !g#add_arc ((afteri,newfakeelse), Control);
 
       let newxi = { xi_lbl with
          ctx = LoopInfo (taili, newfakeelse, xi_lbl.braces, lbl);
@@ -746,7 +747,7 @@ let rec aux_statement : (nodei option * xinfo) -> statement -> nodei option =
       let finalthen = aux_statement (Some doi, newxi) st in
       (match finalthen with
       | None ->
-          if (!g#predecessors taili)#null
+          if (KeyEdgeSet.is_empty (!g#predecessors taili))
           then raise (Error (DeadCode (Some (pinfo_of_ii ii))))
           else Some newfakeelse
       | Some finali ->
@@ -1012,8 +1013,8 @@ and mk_If (starti :nodei option) (labels :int list) (xi_lbl :xinfo)
       (* if -> [after] -> [endif] *)
       let afteri = !g +>
             add_node (AfterNode NormalAfterNode) labels "[after]" in
-      !g#add_arc ((newi, afteri), Direct);
-      !g#add_arc ((afteri, lasti), Direct);
+      !g#add_arc ((newi, afteri), Control);
+      !g#add_arc ((afteri, lasti), Control);
 
       (* for ErrorExit heuristic *)
       let newxi = { xi_lbl with  under_ifthen = true; } in
@@ -1060,9 +1061,9 @@ and mk_If (starti :nodei option) (labels :int list) (xi_lbl :xinfo)
       let mkafter ty str lasti = begin
             (* if -> [after] -> [endif] *)
             let afteri = !g +> add_node (AfterNode ty) labels str in
-            let a1 = ((newi, afteri), Direct) in
+            let a1 = ((newi, afteri), Control) in
             !g#add_arc a1;
-            let a2 = ((afteri, lasti), Direct) in
+            let a2 = ((afteri, lasti), Control) in
             !g#add_arc a2;
             ret_afters := (afteri,a1,a2) :: !ret_afters
         end in
@@ -1263,8 +1264,8 @@ and aux_statement_list starti (xi, newxi) statxs =
 	because then no statement might be there.
 	let afteri =
           !g +> add_node (AfterNode NormalAfterNode) newxi'.labels "[after]" in
-	!g#add_arc ((newi, afteri), Direct);
-	!g#add_arc ((afteri, taili), Direct);
+	!g#add_arc ((newi, afteri), Control);
+	!g#add_arc ((afteri, taili), Control);
 *)
 
         Some taili
@@ -1356,7 +1357,7 @@ let specialdeclmacro_to_stmt (s, args, ii) =
 let rec ast_to_control_flow e =
 
   (* globals (re)initialialisation *)
-  g := (new ograph_mutable);
+  g := (new Control_flow_c.G.ograph_mutable);
   counter_for_labels := 1;
   counter_for_braces := 0;
   counter_for_switch := 0;
@@ -1513,21 +1514,17 @@ let annotate_loop_nodes g =
   (* just for opti a little *)
   let already = Hashtbl.create 101 in
 
-  g +> Ograph_extended.dfs_iter_with_path firsti (fun xi path ->
+  g +> Control_flow_c.G.dfs_iter_with_path firsti (fun xi path ->
     Hashtbl.add already xi true;
-    let succ = g#successors xi in
-    let succ = succ#tolist in
-    succ +> List.iter (fun (yi,_edge) ->
+    let aux (yi, _) =
       if Hashtbl.mem already yi && List.mem yi (xi::path)
       then
-        let node = g#nodes#find yi in
+        let node = KeyMap.find yi g#nodes in
         let ((node2, nodeinfo), nodestr) = node in
         let node' = ((node2, {nodeinfo with is_loop = true}), (nodestr ^ "*"))
-        in g#replace_node (yi, node');
-    );
+        in g#replace_node (yi, node') in
+    KeyEdgeSet.iter aux (g#successors xi);
   );
-
-
   g
 
 
@@ -1546,16 +1543,11 @@ let annotate_loop_nodes g =
  * mais dans le corps d'une macro et donc on le voit pas :(
  *
  *)
-let deadcode_detection g =
+let deadcode_detection (g : Control_flow_c.cflow) =
 
-  g#nodes#iter (fun (k, node) ->
-    let pred = g#predecessors k in
-    if pred#null then
+  KeyMap.iter  (fun k node ->
+    if KeyEdgeSet.is_empty (g#predecessors k) then
       (match unwrap node with
-      (* old:
-       * | Enter -> ()
-       * | EndStatement _ -> pr2 "deadcode sur fake node, pas grave";
-       *)
       | TopNode -> ()
       | FunHeader _ -> ()
       | ErrorExit -> ()
@@ -1569,7 +1561,7 @@ let deadcode_detection g =
           | _ -> pr2 "CFG: orphan nodes, maybe something weird happened"
           )
       )
-  )
+  ) g#nodes
 
 (*------------------------------------------------------------------------*)
 (* special_cfg_braces: the check are really specific to the way we
@@ -1582,42 +1574,33 @@ let deadcode_detection g =
  *)
 
 let check_control_flow (g : cflow) : unit =
-
-  let nodes = g#nodes  in
   let starti = first_node g in
-  let visited = ref (new oassocb []) in
+  let visited = ref KeyMap.empty in
 
   let print_trace_error xs =  pr2 "PB with flow:";  Common.pr2_gen xs; in
 
   let rec dfs (nodei, (* Depth depth,*) startbraces,  trace)  =
     let trace2 = nodei::trace in
-    if !visited#haskey nodei
+    if KeyMap.mem nodei !visited
     then
       (* if loop back, just check that go back to a state where have same depth
          number *)
-      let (*(Depth depth2)*) startbraces2 = !visited#find nodei in
+      let (*(Depth depth2)*) startbraces2 = KeyMap.find nodei !visited in
       if  (*(depth = depth2)*) startbraces <> startbraces2
       then
         begin
-          pr2 (Printf.sprintf "PB with flow: the node %d has not same braces count"
-                 nodei);
+	  let msg =
+	    Printf.sprintf
+	      "PB with flow: the node %d has not same braces count"
+              nodei in
+          pr2 msg;
           print_trace_error trace2
         end
     else
       let children = g#successors nodei in
-      let _ = visited := !visited#add (nodei, (* Depth depth*) startbraces) in
-
-      (* old: good, but detect a missing } too late, only at the end
+      visited := KeyMap.add nodei startbraces !visited;
       let newdepth =
-        (match fst (nodes#find nodei) with
-        | StartBrace i -> Depth (depth + 1)
-        | EndBrace i   -> Depth (depth - 1)
-        | _ -> Depth depth
-        )
-      in
-      *)
-      let newdepth =
-        (match unwrap (nodes#find nodei),  startbraces with
+        (match unwrap (KeyMap.find nodei g#nodes),  startbraces with
         | SeqStart (_,i,_), xs  -> i::xs
         | SeqEnd (i,_), j::xs ->
             if i = j
@@ -1637,14 +1620,13 @@ let check_control_flow (g : cflow) : unit =
       in
 
 
-      if (children#tolist) = []
+      if KeyEdgeSet.is_empty children
       then
         if (* (depth = 0) *) startbraces <> []
         then print_trace_error trace2
       else
-        children#tolist +> List.iter (fun (nodei,_) ->
-          dfs (nodei, newdepth, trace2)
-        )
+        let aux (key,_) = dfs (key, newdepth, trace2) in
+        KeyEdgeSet.iter aux children
     in
 
   dfs (starti, (* Depth 0*) [], [])
@@ -1654,32 +1636,34 @@ let check_control_flow (g : cflow) : unit =
 (*****************************************************************************)
 
 let report_error error =
-  let error_from_info info =
-    Common.error_message_short info.file ("", info.charpos)
-  in
-  match error with
-  | DeadCode          infoopt ->
-      (match infoopt with
-      | None ->   pr2 "FLOW: deadcode detected, but cant trace back the place"
-      | Some info -> pr2 ("FLOW: deadcode detected: " ^ error_from_info info)
-      )
-  | CaseNoSwitch      info ->
-      pr2 ("FLOW: case without corresponding switch: " ^ error_from_info info)
-  | OnlyBreakInSwitch info ->
-      pr2 ("FLOW: only break are allowed in switch: " ^ error_from_info info)
-  | WeirdSwitch info ->
-      pr2 ("FLOW: weird switch: " ^ error_from_info info)
-  | NoEnclosingLoop   (info) ->
-      pr2 ("FLOW: can't find enclosing loop: " ^ error_from_info info)
-  | GotoCantFindLabel (s, info) ->
-      pr2 ("FLOW: cant jump to " ^ s ^ ": because we can't find this label")
-  | NoExit info ->
-      pr2 ("FLOW: can't find exit or error exit: " ^ error_from_info info)
-  | DuplicatedLabel s ->
-      pr2 ("FLOW: duplicate label " ^ s)
-  | NestedFunc  ->
-      pr2 ("FLOW: not handling yet nested function")
-  | ComputedGoto ->
-      pr2 ("FLOW: not handling computed goto yet")
-  | Define info ->
-      pr2 ("Unsupported form of #define: " ^ error_from_info info)
+  if !Flag_parsing_c.verbose_cfg
+  then
+    let error_from_info info =
+      Common.error_message_short info.file ("", info.charpos) in
+    match error with
+    | DeadCode          infoopt ->
+	(match infoopt with
+	| None -> pr2 "FLOW: deadcode detected, but cant trace back the place"
+	| Some info ->
+	    pr2 ("FLOW: deadcode detected: " ^ error_from_info info))
+    | CaseNoSwitch      info ->
+	pr2
+	  ("FLOW: case without corresponding switch: " ^ error_from_info info)
+    | OnlyBreakInSwitch info ->
+	pr2 ("FLOW: only break are allowed in switch: " ^ error_from_info info)
+    | WeirdSwitch info ->
+	pr2 ("FLOW: weird switch: " ^ error_from_info info)
+    | NoEnclosingLoop   (info) ->
+	pr2 ("FLOW: can't find enclosing loop: " ^ error_from_info info)
+    | GotoCantFindLabel (s, info) ->
+	pr2 ("FLOW: cant jump to " ^ s ^ ": because we can't find this label")
+    | NoExit info ->
+	pr2 ("FLOW: can't find exit or error exit: " ^ error_from_info info)
+    | DuplicatedLabel s ->
+	pr2 ("FLOW: duplicate label " ^ s)
+    | NestedFunc  ->
+	pr2 ("FLOW: not handling yet nested function")
+    | ComputedGoto ->
+	pr2 ("FLOW: not handling computed goto yet")
+    | Define info ->
+	pr2 ("Unsupported form of #define: " ^ error_from_info info)
