@@ -78,6 +78,8 @@ let al_init      x = Visitor_c.vk_ini_s       (strip_info_visitor()) x
 let al_inits     x = Visitor_c.vk_inis_s      (strip_info_visitor()) x
 let al_param     x = Visitor_c.vk_param_s     (strip_info_visitor()) x
 let al_params    x = Visitor_c.vk_params_s    (strip_info_visitor()) x
+let al_define_params x =
+  Visitor_c.vk_define_params_s (strip_info_visitor()) x
 let al_arguments x = Visitor_c.vk_arguments_s (strip_info_visitor()) x
 let al_fields    x = Visitor_c.vk_struct_fields_s (strip_info_visitor()) x
 let al_name      x = Visitor_c.vk_name_s      (strip_info_visitor()) x
@@ -175,6 +177,8 @@ let semi_al_init      = Visitor_c.vk_ini_s       semi_strip_info_visitor
 let semi_al_inits     = Visitor_c.vk_inis_s      semi_strip_info_visitor
 let semi_al_param     = Visitor_c.vk_param_s     semi_strip_info_visitor
 let semi_al_params    = Visitor_c.vk_params_s    semi_strip_info_visitor
+let semi_al_define_params =
+  Visitor_c.vk_define_params_s semi_strip_info_visitor
 let semi_al_arguments = Visitor_c.vk_arguments_s semi_strip_info_visitor
 let semi_al_string_format =
   Visitor_c.vk_string_format_s semi_strip_info_visitor
@@ -192,7 +196,7 @@ let semi_al_program =
 let real_strip_info_visitor _ =
   { Visitor_c.default_visitor_c_s with
     Visitor_c.kinfo_s = (fun (k,_) i ->
-      Ast_c.real_al_info_cpp i
+      Ast_c.real_al_info_cpp false i
     );
 
     Visitor_c.kexpr_s = (fun (k,_) e ->
@@ -228,6 +232,42 @@ let real_al_statement_seq_list x =
   Visitor_c.vk_statement_sequencable_list_s (real_strip_info_visitor()) x
 let real_al_def       x = Visitor_c.vk_toplevel_s (real_strip_info_visitor()) x
 
+
+
+
+let real_strip_info_visitor_with_comments _ =
+  { Visitor_c.default_visitor_c_s with
+    Visitor_c.kinfo_s = (fun (k,_) i ->
+      Ast_c.real_al_info_cpp true i
+    );
+
+    Visitor_c.kexpr_s = (fun (k,_) e ->
+      let (e', ty),ii' = k e in
+      (e', Ast_c.noType()), ii'
+    );
+
+(*
+    Visitor_c.ktype_s = (fun (k,_) ft ->
+      let ft' = k ft in
+      match Ast_c.unwrap_typeC ft' with
+      | Ast_c.TypeName (s,_typ) ->
+          Ast_c.TypeName (s, Ast_c.noTypedefDef()) +> Ast_c.rewrap_typeC ft'
+      | _ -> ft'
+
+    );
+*)
+
+  }
+
+let real_al_decl_with_comments x =
+  Visitor_c.vk_decl_s   (real_strip_info_visitor_with_comments()) x
+let real_al_statement_with_comments x =
+  Visitor_c.vk_statement_s (real_strip_info_visitor_with_comments()) x
+
+
+
+
+
 (*****************************************************************************)
 (* Extract infos *)
 (*****************************************************************************)
@@ -294,6 +334,23 @@ let max_min_ii_by_pos xs =
         maxii', minii'
       ) (x,x)
 
+(* avoid memory costs of prefiltering the list *)
+let max_min_ii_by_pos_filtered f xs =
+  let xs = Common.drop_until f xs in
+  match xs with
+  | [] -> failwith "empty list, max_min_ii_by_pos"
+  | [x] -> (x, x)
+  | x::xs ->
+      let pos_leq p1 p2 = (Ast_c.compare_pos p1 p2) = (-1) in
+      xs +> List.fold_left (fun ((maxii,minii) as acc) e ->
+	if f e
+	then
+          let maxii' = if pos_leq maxii e then e else maxii in
+          let minii' = if pos_leq e minii then e else minii in
+          maxii', minii'
+	else acc
+      ) (x,x)
+
 let info_to_fixpos ii =
   match Ast_c.pinfo_of_info ii with
     Ast_c.OriginTok pi -> Ast_cocci.Real pi.Common.charpos
@@ -310,8 +367,8 @@ let max_min_by_pos xs =
 
 let lin_col_by_pos xs =
   (* put min before max; no idea why they are backwards above *)
-  let non_fake = List.filter (function ii -> not (Ast_c.is_fake ii)) xs in
-  let (i2, i1) = max_min_ii_by_pos non_fake in
+  let (i2, i1) =
+    max_min_ii_by_pos_filtered (function ii -> not (Ast_c.is_fake ii)) xs in
   let posf x = Ast_c.col_of_info x in
   let mposf x = Ast_c.col_of_info x + String.length (Ast_c.str_of_info x) in
   (Ast_c.file_of_info i1,!Flag.current_element,
@@ -329,9 +386,8 @@ let min_pinfo_of_node node =
 
 let (range_of_origin_ii: Ast_c.info list -> (int * int) option) =
  fun ii ->
-  let ii = List.filter Ast_c.is_origintok ii in
   try
-    let (max, min) = max_min_ii_by_pos ii in
+    let (max, min) = max_min_ii_by_pos_filtered Ast_c.is_origintok ii in
     assert(Ast_c.is_origintok max);
     assert(Ast_c.is_origintok min);
     let strmax = Ast_c.str_of_info max in
