@@ -148,8 +148,6 @@ let commentized xs = xs +> Common.tail_map_filter (function
 	      [Token_c.CppAttr;Token_c.CppPassingNormal;Token_c.CppMacro]
             ||
             (s =~ "__.*")
-
-
         | 5 ->
 	    (match cppkind with
 	      Token_c.CppAttr | Token_c.CppPassingNormal
@@ -158,10 +156,6 @@ let commentized xs = xs +> Common.tail_map_filter (function
 	    | _ -> false)
             ||
             (s =~ "__.*")
-
-
-
-
         | _ -> failwith "not valid level passing number"
       in
       if legal_passing then None else Some (ii.Ast_c.pinfo)
@@ -176,7 +170,6 @@ let commentized xs = xs +> Common.tail_map_filter (function
             )
         *)
 
-
   | Parser_c.TCommentMisc ii
   | Parser_c.TAction ii
     ->
@@ -203,8 +196,6 @@ let count_lines_commentized xs =
     !count
   end
 
-
-
 let print_commentized xs =
   let line = ref (-1) in
   begin
@@ -231,9 +222,6 @@ let print_commentized xs =
     if ys<>[] then pr2 "";
   end
 
-
-
-
 (*****************************************************************************)
 (* Lexing only *)
 (*****************************************************************************)
@@ -266,26 +254,11 @@ let tokens2 file =
     | e -> raise e
  )
 
-(* The result of lexing can be large.  Just keep the result for the most
-  recent file *)
-let most_recent_file = ref ""
-let most_recent_res = ref []
-
 let time_lexing ?(profile=true) a =
   if profile
   then Common.profile_code_exclusif "LEXING" (fun () -> tokens2 a)
   else tokens2 a
-let tokens ?profile a =
-  if a = !most_recent_file
-  then !most_recent_res
-  else
-    begin
-      most_recent_file := a;
-      most_recent_res := [];
-      let res = time_lexing ?profile a in
-      most_recent_res := res;
-      res
-    end
+let tokens ?profile a = time_lexing ?profile a
 
 let tokens_of_string string =
   let lexbuf = Lexing.from_string string in
@@ -319,26 +292,6 @@ let parse file =
   let result = Parser_c.main Lexer_c.token lexbuf in
   result
 
-
-let parse_print_error file =
-  let chan = (open_in file) in
-  let lexbuf = Lexing.from_channel chan in
-
-  let error_msg () = Common.error_message file (lexbuf_to_strpos lexbuf) in
-  try
-    Parser_c.main Lexer_c.token lexbuf
-  with
-  | Lexer_c.Lexical s ->
-      failwith ("lexical error " ^s^ "\n =" ^  error_msg ())
-  | Parsing.Parse_error ->
-      failwith ("parse error \n = " ^ error_msg ())
-  | Semantic_c.Semantic (s, i) ->
-      failwith ("semantic error " ^ s ^ "\n =" ^ error_msg ())
-  | e -> raise e
-
-
-
-
 (*****************************************************************************)
 (* Parsing subelements, useful to debug parser *)
 (*****************************************************************************)
@@ -721,12 +674,13 @@ let get_one_elem ~pass tr =
 
       (* must keep here, before the code that adjusts the tr fields *)
       let line_error = TH.line_of_tok tr.current in
+      let col_error = TH.col_of_tok tr.current in
 
       let passed_before_error = tr.passed in
       let current = tr.current in
       (*  error recovery, go to next synchro point *)
       let (passed', rest') =
-        Parsing_recovery_c.find_next_synchro tr.rest tr.passed in
+	Parsing_recovery_c.find_next_synchro tr.rest tr.passed in
       tr.rest <- rest';
       tr.passed <- passed';
 
@@ -737,7 +691,7 @@ let get_one_elem ~pass tr =
 
 
       let info_of_bads = Common.map_eff_rev TH.info_of_tok tr.passed in
-      Right (info_of_bads,  line_error,
+      Right (info_of_bads, line_error, col_error,
             tr.passed, passed_before_error,
             current, e, pass)
   )
@@ -787,7 +741,7 @@ let candidate_macros_in_passed ~defs b =
 
 
 
-let find_optional_macro_to_expand2 ~defs toks =
+let find_optional_macro_to_expand2 ~defs pos toks =
 
   let defs = Common.hash_of_list defs in
 
@@ -808,7 +762,7 @@ let find_optional_macro_to_expand2 ~defs toks =
   ) in
 
   let tokens = toks in
-  Parsing_hacks.fix_tokens_cpp ~macro_defs:defs tokens
+  Parsing_hacks.fix_tokens_cpp ~macro_defs:defs pos tokens
 
   (* just calling apply_macro_defs and having a specialized version
    * of the code in fix_tokens_cpp is not enough as some work such
@@ -823,15 +777,15 @@ let find_optional_macro_to_expand2 ~defs toks =
   Cpp_token_c.apply_macro_defs
     ~msg_apply_known_macro:(fun s -> pr2 (spf "APPLYING: %s" s))
     ~msg_apply_known_macro_hint:(fun s -> pr2 "hint")
-    defs paren_grouped;
+    defs pos paren_grouped;
   (* because the before field is used by apply_macro_defs *)
   tokens2 := TV.rebuild_tokens_extented !tokens2;
   Parsing_hacks.insert_virtual_positions
     (!tokens2 +> Common.acc_map (fun x -> x.TV.tok))
   *)
-let find_optional_macro_to_expand ~defs a =
+let find_optional_macro_to_expand ~defs pos a =
     Common.profile_code "MACRO managment" (fun () ->
-      find_optional_macro_to_expand2 ~defs a)
+      find_optional_macro_to_expand2 ~defs pos a)
 
 (*****************************************************************************)
 (* Parsing #if guards *)
@@ -925,6 +879,16 @@ type program2 = toplevel2 list
       (string, Cpp_token_c.define_def) Hashtbl.t (* macro defs *)
    and toplevel2 = Ast_c.toplevel * info_item
 
+type 'a generic_parse_info = {
+  filename : string;
+  parse_trees : 'a; (* program2 or extended_program2 *)
+  statistics : Parsing_stat.parsing_stat;
+}
+
+type parse_info = program2 generic_parse_info
+
+type extended_parse_info = extended_program2 generic_parse_info
+
 let program_of_program2 xs =
   xs +> List.map fst
 
@@ -956,9 +920,98 @@ let with_program2_unit f program2 =
  * tokens_stat record and parsing_stat record.
  *)
 
-let parse_print_error_heuristic2 saved_typedefs saved_macros parse_strings
-    file =
+module StringSet : Set.S with type elt = string = Set.Make(String)
 
+module StringMap : Map.S with type key = string = Map.Make(String)
+
+let header_cache = Common.create_bounded_cache 0(*disabled 300*) ("",None)
+
+let tree_stack = ref []
+let seen_files = ref []
+
+let normalize file =
+  let pieces = Str.split_delim (Str.regexp "/") file in
+  let rec loop prev = function
+      [] -> String.concat "/" (List.rev prev)
+    | ".."::rest ->
+	(match prev with
+	  ".."::xs -> loop (".."::prev) rest
+	| x::xs -> loop xs rest
+	| _ -> loop (".."::prev) rest)
+    | x::rest -> loop (x::prev) rest in
+  loop [] pieces
+
+let rec _parse_print_error_heuristic2 saved_typedefs saved_macros
+  parse_strings cache file use_header_cache =
+  let file = normalize file in
+  if List.mem file !seen_files
+  then None (* Inclusion loop, not re-parsing *)
+  else begin
+    seen_files := file :: !seen_files;
+    let cached_result =
+      if false && use_header_cache
+      then
+	try
+	  (match Common.find_bounded_cache header_cache file with
+	    None -> failwith "Not possible"
+	  | Some (cached,cached_includes) ->
+	      List.iter
+		(function incl ->
+		  handle_include file incl
+		    (fun nonlocal header_filename ->
+		      _parse_print_error_heuristic2
+			saved_typedefs saved_macros parse_strings
+			nonlocal header_filename use_header_cache))
+		cached_includes;
+	      Some cached)
+	with Not_found -> None
+      else None in
+    match cached_result with
+      | None ->
+        let result =
+          _parse_print_error_heuristic2bis saved_typedefs saved_macros
+            parse_strings file use_header_cache in
+        (if false && use_header_cache && cache
+	then
+	  let my_includes =
+	    let my_includes = ref [] in
+	    let cpp (k,bigf) directive =
+	      match directive with
+		Ast_c.Include incl -> my_includes := incl :: !my_includes
+	      | _ -> () in
+	    let bigf =
+	      { Visitor_c.default_visitor_c with
+		Visitor_c.kcppdirective = cpp } in
+	    let (pgm,ty,defs) = result.parse_trees in
+	    Visitor_c.vk_program bigf (List.map fst pgm);
+	    !my_includes in
+	  Common.extend_bounded_cache header_cache file
+	    (Some (result,my_includes)));
+        tree_stack := result :: !tree_stack;
+        Some result
+      | Some result ->
+        tree_stack := result :: !tree_stack;
+        Some result
+  end
+
+
+and handle_include file wrapped_incl k =
+    let incl = Ast_c.unwrap wrapped_incl.Ast_c.i_include in
+    let parsing_style = Includes.get_parsing_style () in
+    if Includes.should_parse parsing_style file incl
+    then begin match Includes.resolve file parsing_style incl with
+      | Some header_filename when Common.lfile_exists header_filename ->
+	  (if !Flag_parsing_c.verbose_includes
+	  then pr2 ("including "^header_filename));
+	  let nonlocal =
+	    match incl with Ast_c.NonLocal _ -> true | _ -> false in
+          ignore (k nonlocal header_filename)
+      | _ -> ()
+    end
+
+
+and _parse_print_error_heuristic2bis saved_typedefs saved_macros
+  parse_strings file use_header_cache =
   let stat = Parsing_stat.default_stat file in
 
   (* -------------------------------------------------- *)
@@ -995,7 +1048,8 @@ let parse_print_error_heuristic2 saved_typedefs saved_macros parse_strings
                 then Parsing_hacks.fix_tokens_ifdef toks
                 else toks
     in
-  let toks = Parsing_hacks.fix_tokens_cpp ~macro_defs:!_defs_builtins toks in
+  let toks =
+    Parsing_hacks.fix_tokens_cpp ~macro_defs:!_defs_builtins [] toks in
   let toks =
     if parse_strings
     then Parsing_hacks.fix_tokens_strings toks
@@ -1026,8 +1080,18 @@ let parse_print_error_heuristic2 saved_typedefs saved_macros parse_strings
           get_one_elem ~pass:1 tr
         ) in
       match pass1 with
-      | Left e -> Left e
-      | Right (info,line_err, passed, passed_before_error, cur, exn, _) ->
+      | Left e ->
+        begin
+          match e with
+            | Ast_c.CppTop(Ast_c.Include incl) ->
+		handle_include file incl
+		  (fun nonlocal header_filename ->
+		    _parse_print_error_heuristic2
+		      saved_typedefs saved_macros parse_strings
+		      nonlocal header_filename use_header_cache)
+            | _ -> ()
+        end; Left e
+      | Right (info,line_err, _, passed, passed_before_error, cur, exn, _) ->
           if !Flag_parsing_c.disable_multi_pass
           then pass1
           else begin
@@ -1041,13 +1105,12 @@ let parse_print_error_heuristic2 saved_typedefs saved_macros parse_strings
 
             (match passx with
             | Left e -> passx
-            | Right (info,line_err,passed,passed_before_error,cur,exn,_) ->
+            | Right (info,line_err,col_err,passed,_,cur,exn,_) ->
                 let candidates =
                   candidate_macros_in_passed ~defs:macros passed
                 in
 
-
-                if is_define_passed passed || candidates=[]
+                if is_define_passed passed
                 then passx
                 else begin
                   (* todo factorize code *)
@@ -1055,28 +1118,36 @@ let parse_print_error_heuristic2 saved_typedefs saved_macros parse_strings
                   pr2_err "parsing pass3: try again";
                   let toks = List.rev passed @ tr.rest in
                   let toks' =
-                    find_optional_macro_to_expand ~defs:candidates toks in
+                    find_optional_macro_to_expand ~defs:candidates
+		      [(line_err,col_err)] toks in
                   let new_tr = mk_tokens_state toks' in
                   copy_tokens_state ~src:new_tr ~dst:tr;
                   let passx = get_one_elem ~pass:3 tr in
 
                   (match passx with
                   | Left e -> passx
-                  | Right (info,line_err,passed,passed_before_error,cur,exn,_) ->
-                      pr2_err "parsing pass4: try again";
+                  | Right(info,le1,ce1,passed,passed_before_error,cur,exn,_) ->
+		      if candidates = [] && line_err = le1 && col_err = ce1
+		      then passx (* nothing changed, so don't try again *)
+		      else
+			begin
+			  pr2_err "parsing pass4: try again";
 
-                      let candidates =
-                        candidate_macros_in_passed
-                          ~defs:macros passed
-                      in
+			  let candidates =
+                            candidate_macros_in_passed
+                              ~defs:macros passed
+			  in
 
-                      let toks = List.rev passed @ tr.rest in
-                      let toks' =
-                      find_optional_macro_to_expand ~defs:candidates toks in
-                      let new_tr = mk_tokens_state toks' in
-                      copy_tokens_state ~src:new_tr ~dst:tr;
-                      let passx = get_one_elem ~pass:4 tr in
-                      passx
+			  let toks = List.rev passed @ tr.rest in
+			  let toks' =
+			    find_optional_macro_to_expand ~defs:candidates
+			      (Common.nub [(line_err,col_err);(le1,ce1)])
+			      toks in
+			  let new_tr = mk_tokens_state toks' in
+			  copy_tokens_state ~src:new_tr ~dst:tr;
+			  let passx = get_one_elem ~pass:4 tr in
+			  passx
+			end
                   )
                  end
             )
@@ -1111,7 +1182,7 @@ let parse_print_error_heuristic2 saved_typedefs saved_macros parse_strings
       | Left e ->
           stat.Stat.correct <- stat.Stat.correct + diffline;
           e
-      | Right (info_of_bads, line_error, toks_of_bads,
+      | Right (info_of_bads, line_error, col_error, toks_of_bads,
               passed_before_error, cur, exn, pass) ->
 
           let was_define = is_define_passed tr.passed in
@@ -1199,34 +1270,55 @@ let parse_print_error_heuristic2 saved_typedefs saved_macros parse_strings
     let new_td = ref (Common.clone_scoped_h_env !LP._typedef) in
     Common.clean_scope_h new_td;
     (v, !new_td, macros) in
-  (v, stat)
+  { filename = file; parse_trees = v; statistics = stat }
 
+let parse_print_error_heuristic2 saved_typedefs saved_macros
+  parse_strings cache file use_header_cache =
+  tree_stack := [];
+  seen_files := [];
+  ignore
+    (_parse_print_error_heuristic2 saved_typedefs saved_macros
+      parse_strings cache file use_header_cache);
+  match !tree_stack with
+    | [] -> assert false
+    | tree::trees -> tree_stack := []; (tree, List.rev trees)
 
-let time_total_parsing a b c d =
-  let res =
-    Common.profile_code "TOTAL"
-      (fun () -> parse_print_error_heuristic2 a b c d) in
-  most_recent_file := ""; (* remove now useless lexer information *)
-  most_recent_res := [];
-  res
+let time_total_parsing a b c d e =
+  Common.profile_code "TOTAL"
+    (fun () -> parse_print_error_heuristic2 a b c d e)
 
-let parse_print_error_heuristic a b c d =
-  Common.profile_code "C parsing" (fun () -> time_total_parsing a b c d)
+let parse_print_error_heuristic a b c d e =
+  Common.profile_code "C parsing" (fun () -> time_total_parsing a b c d e)
 
 
 (* alias *)
-let parse_c_and_cpp parse_strings a =
-  let ((c,_,_),stat) = parse_print_error_heuristic None None parse_strings a in
-  (c,stat)
-let parse_c_and_cpp_keep_typedefs td macs parse_strings a =
-  parse_print_error_heuristic td macs parse_strings a
+let parse_c_and_cpp parse_strings cache a =
+  let v =
+    (* no saved typedefs or macros, so can't use header cache *)
+    fst (parse_print_error_heuristic None None parse_strings cache a false) in
+  let (c, _, _) = v.parse_trees in
+  (c, v.statistics)
+
+let parse_c_and_cpp_keep_typedefs td macs parse_strings cache a =
+  parse_print_error_heuristic td macs parse_strings cache a true
 
 (*****************************************************************************)
 (* Same but faster cos memoize stuff *)
 (*****************************************************************************)
-let parse_cache parse_strings file =
+let parse_cache typedefs parse_strings cache file has_changes =
+  (* Normally, if there are no changes to headers, we should not have to parse
+     then again.  The problem is that headers have effects, eg on inferred
+     typedefs, and we aren't storing and then reapplying those effects.
+     So we have to reparse the header files again, at great runtime cost, to
+     get those effects.  It also seems that when a header file is retrieved
+     from the cache, then there is no runtime processing, eg matching, of at
+     least the header files that it includes.  That is, for recursive includes,
+     need not only the AST of the header file itself, but also of what it
+     includes, and at least the latter is not coming in the cached case. *)
   if not !Flag_parsing_c.use_cache
-  then parse_print_error_heuristic None None parse_strings file
+  then
+    parse_print_error_heuristic typedefs None parse_strings cache file
+      (not has_changes || !Includes.include_headers_for_types)
   else
   let _ = pr2_once "TOFIX: use_cache is not sensitive to changes in the considered macros, include files, etc" in
   let need_no_changed_files =
@@ -1269,7 +1361,7 @@ let parse_cache parse_strings file =
 		()
 	  | _ -> ());
       (* recompute *)
-      parse_print_error_heuristic None None true file)
+      parse_print_error_heuristic None None true cache file false)
 
 
 
@@ -1287,7 +1379,7 @@ let (cstatement_of_string: string -> Ast_c.statement) = fun s ->
   assert (no_format s);
   let tmpfile = Common.new_temp_file "cocci_stmt_of_s" "c" in
   Common.write_file tmpfile ("void main() { \n" ^ s ^ "\n}");
-  let program = parse_c_and_cpp false tmpfile +> fst in
+  let program = fst (parse_c_and_cpp false false tmpfile) in
   program +> Common.find_some (fun (e,_) ->
     match e with
     | Ast_c.Definition ({Ast_c.f_body = [Ast_c.StmtElem st]},_) -> Some st
@@ -1298,7 +1390,7 @@ let (cexpression_of_string: string -> Ast_c.expression) = fun s ->
   assert (no_format s);
   let tmpfile = Common.new_temp_file "cocci_expr_of_s" "c" in
   Common.write_file tmpfile ("void main() { \n" ^ s ^ ";\n}");
-  let program = parse_c_and_cpp false tmpfile +> fst in
+  let program = fst (parse_c_and_cpp false false tmpfile) in
   program +> Common.find_some (fun (e,_) ->
     match e with
     | Ast_c.Definition ({Ast_c.f_body = compound},_) ->
