@@ -50,7 +50,7 @@ val parse_cache:
     (string, Lexer_parser.identkind) Common.scoped_h_env option (* typedefs *) ->
     bool (* true if format characters need to be parsed *) ->
     bool (* true if parsing results should be cached *) ->
-    Common.filename (*cfile*) ->
+    Common.filename (*cfile*) -> bool (* true if no changes *) ->
     extended_parse_info * extended_parse_info list
 
 
@@ -67,7 +67,6 @@ val tokens:      ?profile:bool -> Common.filename -> Parser_c.token list
 val tokens_of_string: string -> Parser_c.token list
 
 val parse:                        Common.filename -> Ast_c.program
-val parse_print_error:            Common.filename -> Ast_c.program
 val parse_gen: cpp:bool -> tos:bool
     -> ((Lexing.lexbuf -> Parser_c.token) -> Lexing.lexbuf -> 'a)
     -> string -> 'a
