@@ -169,7 +169,7 @@ let msg_macro_higher_order s =
       | _ -> false
       )
     )
-    (fun s -> pr2_cpp ("MACRO: found higher ordre macro : " ^ s))
+    (fun s -> pr2_cpp ("MACRO: found higher order macro : " ^ s))
     s
 
 
@@ -271,7 +271,11 @@ let is_macro s =
 let not_macro s =
   not (is_macro s)
 
-
+let is_macro_paren s rest = (* likely macro call *)
+  is_macro s &&
+    match rest with
+      TOPar _::TOPar _::_ -> true
+    | _ -> false
 
 
 (*****************************************************************************)
@@ -309,7 +313,7 @@ let forLOOKAHEAD = 30
  *
  * todo: use indentation instead of premier(statement) ?
  *)
-let rec is_really_foreach xs =
+let is_really_foreach xs =
   let rec is_foreach_aux = function
     | [] -> false, []
     | TCPar _::TOBrace _::xs -> true, xs
@@ -442,28 +446,19 @@ and define_line_2 acc line lastinfo = function
       (* should not happened, should meet EOF before *)
       pr2 "PB: WEIRD";
       List.rev ((mark_end_define lastinfo)::acc)
-  | token::tokens as all_tokens ->
+  | token::tokens ->
       let line' = TH.line_of_tok token in
       let info = TH.info_of_tok token in
       (match token with
-      | EOF ii ->
-          define_line_1 (token::(mark_end_define lastinfo)::acc) tokens
+      | EOF ii -> define_line_1 (token::(mark_end_define lastinfo)::acc) tokens
       | TCppEscapedNewline ii ->
           if (line' <> line) then pr2 "PB: WEIRD: not same line number";
           define_line_2 ((TCommentSpace ii)::acc) (line+1) info tokens
+      | TCommentNewline ii ->
+	  define_line_1 (token::(mark_end_define lastinfo)::acc) tokens
       | _ ->
-          if line' = line
-          then
-            define_line_2
-              (token::acc) (end_line_of_tok line' token) info tokens
-          else
-	    (* Put end of line token before the newline.  A newline at least
-	       must be there because the line changed and because we saw a
-	       #define previously to get to this function at all *)
-	    define_line_1
-	      ((List.hd acc)::(mark_end_define lastinfo::(List.tl acc)))
-	      all_tokens
-      )
+          define_line_2
+            (token::acc) (end_line_of_tok line' token) info tokens)
 
 (* for a comment, the end line is not the same as line_of_tok *)
 and end_line_of_tok default = function
@@ -535,12 +530,40 @@ let rec define_ident acc = function
           define_ident acc xs
       )
     | TPragma ii ->
-      let rec loop acc = function
-	  ((TDefEOL i1) as x) :: xs -> define_ident (x::acc) xs
-	| TCommentSpace i1::TIdent (s,i2)::xs ->
-	    let acc = (TCommentSpace i1) :: acc in
-	    let acc = (TIdentDefine (s,i2)) :: acc in
-            loop acc xs
+	let safe =
+	  let rec loop2 = function
+	      (TCPar _)::(TDefEOL _)::_ -> true
+	    | (TCPar _)::_ -> false
+	    | _::xs -> loop2 xs
+	    | _ -> false in
+	  let rec loop1a = function
+	      (TCommentSpace _)::xs -> loop1a xs
+	    | (TIdent _)::xs -> loop1a xs
+	    | t::xs when TH.str_of_tok t ==~ Common.regexp_alpha -> loop1a xs
+	    | (TDefEOL i1)::_ -> true
+	    | _ -> false in
+	  let rec loop1 = function
+	      (TCommentSpace _)::xs -> loop1 xs
+	    | (TOPar _)::xs -> loop2 xs
+	    | (TIdent _)::xs -> loop1a xs
+	    | t::xs when TH.str_of_tok t ==~ Common.regexp_alpha -> loop1a xs
+	    | (TDefEOL i1)::_ -> true
+	    | _ -> false in
+	  let rec loop = function
+	      (TCommentSpace _)::xs -> loop xs
+	    | (TIdent _)::xs -> loop1 xs
+	    | t::xs when TH.str_of_tok t ==~ Common.regexp_alpha -> loop1 xs
+	    | (TDefEOL i1)::_ -> true
+	    | _ -> false in
+	  loop tokens in
+	if safe
+	then
+	  let rec loop acc = function
+	      ((TDefEOL i1) as x) :: xs -> define_ident (x::acc) xs
+	    | TCommentSpace i1::TIdent (s,i2)::xs ->
+		let acc = (TCommentSpace i1) :: acc in
+		let acc = (TIdentDefine (s,i2)) :: acc in
+		loop acc xs
 
       (* bugfix: ident of macro (as well as params, cf below) can be tricky
        * note, do we need to subst in the body of the define ? no cos
@@ -549,20 +572,36 @@ let rec define_ident acc = function
        * body (it would be a recursive macro, which is forbidden).
        *)
 
-	| TCommentSpace i1::t::xs
-	  when TH.str_of_tok t ==~ Common.regexp_alpha
-	  ->
-            let s = TH.str_of_tok t in
-            let ii = TH.info_of_tok t in
-	    pr2 (spf "remapping: %s to an ident in pragma" s);
-            let acc = (TCommentSpace i1) :: acc in
-	    let acc = (TIdentDefine (s,ii)) :: acc in
-            define_ident acc xs
+	    | TCommentSpace i1::t::xs
+	      when TH.str_of_tok t ==~ Common.regexp_alpha
+	      ->
+		let s = TH.str_of_tok t in
+		let ii = TH.info_of_tok t in
+		pr2 (spf "remapping: %s to an ident in pragma" s);
+		let acc = (TCommentSpace i1) :: acc in
+		let acc = (TIdentDefine (s,ii)) :: acc in
+		loop acc xs
 
-	| xs ->
-            pr2 "WEIRD: weird #pragma";
-            define_ident acc xs in
-      loop acc tokens
+	    | xs ->
+		pr2 "WEIRD: weird #pragma";
+		define_ident acc xs in
+	  loop acc tokens
+	else
+	    (* just turn all tokens into identifiers and move on *)
+	  let rec loop acc = function
+              ((TDefEOL i1) as x) :: xs -> define_ident (x::acc) xs
+	    | ((TCommentSpace _) as t)::xs -> loop (t::acc) xs
+	    | TIdent (s,i2)::xs ->
+		let acc = (TIdentDefine (s,i2)) :: acc in
+		loop acc xs
+	    | t::xs ->
+		let s = TH.str_of_tok t in
+		let ii = TH.info_of_tok t in
+		pr2 (spf "remapping: %s to an ident in pragma" s);
+		let acc = (TIdentDefine (s,ii)) :: acc in
+		loop acc xs
+	    | _ -> failwith "bad pragma line" in
+	  loop acc tokens
     | _ -> define_ident acc tokens
     )
 
@@ -936,7 +975,7 @@ let rec find_ifdef_funheaders = function
 
 
 (* ?? *)
-let rec adjust_inifdef_include xs =
+let adjust_inifdef_include xs =
   xs +> List.iter (function
   | NotIfdefLine _ -> ()
   | Ifdef (xxs, info_ifdef_stmt) | Ifdefbool (_, xxs, info_ifdef_stmt) ->
@@ -954,7 +993,7 @@ let rec adjust_inifdef_include xs =
 
 
 
-let rec find_ifdef_cparen_else xs =
+let find_ifdef_cparen_else xs =
   let rec aux xs =
   xs +> List.iter (function
   | NotIfdefLine _ -> ()
@@ -1088,20 +1127,22 @@ let rec find_macro_paren xs =
   | [] -> ()
 
   (* attribute *)
-  | PToken ({tok = Tattribute _} as id)
+  | PToken ({tok = Tattribute ii} as id)
     ::Parenthised (xxs,info_parens)
     ::xs
      ->
-      pr2_cpp ("MACRO: __attribute detected ");
+      pr2_cpp (Printf.sprintf "MACRO: attribute %s detected "
+		 (Ast_c.str_of_info ii));
       [Parenthised (xxs, info_parens)] +>
         iter_token_paren (TV.set_as_comment Token_c.CppAttr);
       TV.set_as_comment Token_c.CppAttr id;
       find_macro_paren xs
 
-  | PToken ({tok = TattributeNoarg _} as id)
+  | PToken ({tok = TattributeNoarg ii} as id)
     ::xs
      ->
-      pr2_cpp ("MACRO: __attributenoarg detected ");
+      pr2_cpp (Printf.sprintf "MACRO: attributenoarg %s detected "
+		 (Ast_c.str_of_info ii));
       TV.set_as_comment Token_c.CppAttr id;
       find_macro_paren xs
 
@@ -1135,16 +1176,6 @@ let rec find_macro_paren xs =
 
 *)
 
-  (* storage attribute *)
-  | PToken ({tok = (Tstatic _ | Textern _)} as tok1)
-    ::PToken ({tok = TMacroAttr (s,i1)} as attr)::xs
-    ->
-      pr2_cpp ("storage attribute: " ^ s);
-      attr.tok <- TMacroAttrStorage (s,i1);
-      (* recurse, may have other storage attributes *)
-      find_macro_paren (PToken (tok1)::xs)
-
-
   (* stringification
    *
    * the order of the matching clause is important
@@ -1211,7 +1242,7 @@ let rec find_macro_paren xs =
 
 
 (* don't forget to recurse in each case *)
-let rec find_macro_lineparen xs =
+let rec find_macro_lineparen prev_line_end xs =
   match xs with
   | [] -> ()
 
@@ -1231,7 +1262,7 @@ let rec find_macro_lineparen xs =
       let info = TH.info_of_tok macro.tok in
       macro.tok <- TMacroDecl (Ast_c.str_of_info info, info);
 
-      find_macro_lineparen (xs)
+      find_macro_lineparen true (xs)
 
   (* the static const case *)
   | (Line
@@ -1260,7 +1291,7 @@ let rec find_macro_lineparen xs =
       *)
       const.tok <- TMacroDeclConst (TH.info_of_tok const.tok);
 
-      find_macro_lineparen (xs)
+      find_macro_lineparen true (xs)
 
 
   (* same but without trailing ';'
@@ -1281,7 +1312,7 @@ let rec find_macro_lineparen xs =
       let info = TH.info_of_tok macro.tok in
       macro.tok <- TMacroDecl (Ast_c.str_of_info info, info);
 
-      find_macro_lineparen (xs)
+      find_macro_lineparen true (xs)
 
 
   (* on multiple lines *)
@@ -1304,7 +1335,7 @@ let rec find_macro_lineparen xs =
       let info = TH.info_of_tok macro.tok in
       macro.tok <- TMacroDecl (Ast_c.str_of_info info, info);
 
-      find_macro_lineparen (xs)
+      find_macro_lineparen true (xs)
 
 
   | (Line (* initializer case *)
@@ -1322,7 +1353,7 @@ let rec find_macro_lineparen xs =
       macro.tok <- TMacroDecl (Ast_c.str_of_info info, info);
 
       (* continue with the rest of the line *)
-      find_macro_lineparen ((Line(rest))::xs)
+      find_macro_lineparen false ((Line(rest))::xs)
 
 
   | (Line (* multi-line initializer case *)
@@ -1344,7 +1375,7 @@ let rec find_macro_lineparen xs =
       macro.tok <- TMacroDecl (Ast_c.str_of_info info, info);
 
       (* continue with the rest of the line *)
-      find_macro_lineparen ((Line(rest))::xs)
+      find_macro_lineparen false ((Line(rest))::xs)
 
 
   (* linuxext: ex: DECLARE_BITMAP();
@@ -1372,7 +1403,7 @@ let rec find_macro_lineparen xs =
       let info = TH.info_of_tok macro.tok in
       macro.tok <- TMacroDecl (Ast_c.str_of_info info, info);
 
-      find_macro_lineparen (xs)
+      find_macro_lineparen true (xs)
 
 
   (* toplevel macros.
@@ -1420,7 +1451,7 @@ let rec find_macro_lineparen xs =
 
         end;
 
-       find_macro_lineparen (xs)
+       find_macro_lineparen true (xs)
 
 
 
@@ -1437,6 +1468,11 @@ let rec find_macro_lineparen xs =
           (PToken ({col = col2 } as other)::restline2
           ) as line2)
     ::xs
+    when
+      ((*Printf.eprintf "checking on %s: %b %b %b %b\n" s
+	 (ctx = InFunction) (ctx = InStruct) (ctx = NoContext)
+	 (ctx = InInitializer);*)
+      List.mem ctx [InFunction;InStruct;NoContext])
     (* when s ==~ regexp_macro *)
     ->
       (* This can give a false positive for K&R functions if the function
@@ -1478,12 +1514,17 @@ let rec find_macro_lineparen xs =
         if col1 = 0 then ()
         else begin
           msg_macro_noptvirg s;
-          macro.tok <- TMacroStmt (s, TH.info_of_tok macro.tok);
+	  (match ctx with
+	    InStruct ->
+              macro.tok <- TMacroDecl (s, TH.info_of_tok macro.tok)
+	  | InFunction | NoContext ->
+              macro.tok <- TMacroStmt (s, TH.info_of_tok macro.tok)
+	  | _ -> failwith "not possible");
           [Parenthised (xxs, info_parens)] +>
             iter_token_paren (TV.set_as_comment Token_c.CppMacro);
         end;
 
-      find_macro_lineparen (line2::xs)
+      find_macro_lineparen true (line2::xs)
 
   (* linuxext:? single macro
    * ex: LOCK
@@ -1499,7 +1540,9 @@ let rec find_macro_lineparen xs =
     ::(Line
           (PToken ({col = col2 } as other)::restline2
           ) as line2)
-    ::xs ->
+    ::xs when
+      (* MacroStmt doesn't make sense otherwise *)
+      List.mem ctx [InFunction;NoContext] && prev_line_end ->
     (* when s ==~ regexp_macro *)
 
       let condition =
@@ -1528,10 +1571,16 @@ let rec find_macro_lineparen xs =
         msg_macro_noptvirg_single s;
         macro.tok <- TMacroStmt (s, TH.info_of_tok macro.tok);
       end;
-      find_macro_lineparen (line2::xs)
+      find_macro_lineparen true (line2::xs)
 
-  | x::xs ->
-      find_macro_lineparen xs
+  | (Line line)::xs ->
+      let prev_line_end =
+	match List.rev line with
+	  (PToken {tok = TPtVirg _} | PToken {tok = TOBrace _}
+	| PToken {tok = TCBrace _} | PToken {tok = TDotDot _}) :: _ ->
+	    true
+	| _ -> false in
+      find_macro_lineparen prev_line_end xs
 
 
 
@@ -1539,7 +1588,7 @@ let rec find_macro_lineparen xs =
 (* define tobrace init *)
 (* ------------------------------------------------------------------------- *)
 
-let rec find_define_init_brace_paren xs =
+let find_define_init_brace_paren xs =
  let rec aux xs =
   match xs with
   | [] -> ()
@@ -1714,7 +1763,7 @@ let insert_virtual_positions l =
 
 (* ------------------------------------------------------------------------- *)
 
-let fix_tokens_cpp2 ~macro_defs tokens =
+let fix_tokens_cpp2 ~macro_defs err_pos tokens =
   let tokens2 = ref (tokens +> Common.acc_map TV.mk_token_extended) in
 
   begin
@@ -1758,7 +1807,7 @@ let fix_tokens_cpp2 ~macro_defs tokens =
     Cpp_token_c.apply_macro_defs
       ~msg_apply_known_macro
       ~msg_apply_known_macro_hint
-      macro_defs paren_grouped;
+      macro_defs err_pos paren_grouped;
     (* because the before field is used by apply_macro_defs *)
     tokens2 := TV.rebuild_tokens_extented !tokens2;
 
@@ -1779,7 +1828,7 @@ let fix_tokens_cpp2 ~macro_defs tokens =
     let line_paren_grouped = TV.mk_line_parenthised paren_grouped in
     find_define_init_brace_paren paren_grouped;
     find_string_macro_paren paren_grouped;
-    find_macro_lineparen    line_paren_grouped;
+    find_macro_lineparen    true line_paren_grouped;
     find_macro_paren        paren_grouped;
 
 
@@ -1794,8 +1843,9 @@ let fix_tokens_cpp2 ~macro_defs tokens =
 let time_hack1 ~macro_defs a =
   Common.profile_code_exclusif "HACK" (fun () -> fix_tokens_cpp2 ~macro_defs a)
 
-let fix_tokens_cpp ~macro_defs a =
-  Common.profile_code "C parsing.fix_cpp" (fun () -> time_hack1 ~macro_defs a)
+let fix_tokens_cpp ~macro_defs err_pos a =
+  Common.profile_code "C parsing.fix_cpp"
+    (fun () -> time_hack1 ~macro_defs err_pos a)
 
 
 
@@ -1845,6 +1895,8 @@ open Lexer_parser (* for the fields of lexer_hint type *)
 
 let not_struct_enum = function
   | (Parser_c.Tstruct _ | Parser_c.Tunion _ | Parser_c.Tenum _)::_ -> false
+  | (Parser_c.TIdent _)::
+    (Parser_c.Tstruct _ | Parser_c.Tunion _ | Parser_c.Tenum _)::_ -> false
   | _ -> true
 
 let pointer ?(followed_by=fun _ -> true)
@@ -1922,6 +1974,12 @@ let paren_before_comma l =
 let lookahead2 ~pass next before =
   match (next, before) with
 
+  | t::_, (TOPar il)::TIdent(s,i2)::(Tstatic _|Tconst _|Textern _)::_
+    when is_macro s && (LP.current_context() = LP.InParameter) ->
+      (* swap context, not really in a parameter list *)
+      LP.pop_context(); LP.push_context (LP.current_context());
+      t
+
   (* c++ hacks *)
   (* yy xx(   and in function *)
   | TOPar i1::_,              TIdent(s,i2)::TypedefIdent _::_
@@ -2067,8 +2125,8 @@ let lookahead2 ~pass next before =
         ->
 	  TIdent (s, i1)
   (* xx yy *)
-  | (TIdent (s, i1)::TIdent (s2, i2)::_  , _) when not_struct_enum before
-      && ok_typedef s
+  | (TIdent (s, i1)::TIdent (s2, i2)::rest  , _) when not_struct_enum before
+      && ok_typedef s && not (is_macro_paren s2 rest)
         ->
          (* && not_annot s2 BUT lead to false positive*)
       msg_typedef s i1 2; LP.add_typedef_root s;
@@ -2088,6 +2146,7 @@ let lookahead2 ~pass next before =
   | (TIdent (s, i1)::(((TComma _|TCPar _)::_) as rest) ,
      ((TComma _ |TOPar _)::_ as bef))
     when not_struct_enum before && (LP.current_context() = LP.InParameter)
+      && not (!Flag_parsing_c.prevent_kr) && LP.is_kr_possible()
       && k_and_r rest
       && not_has_type_before is_cparen rest
       && not_has_type_before is_oparen bef
@@ -2404,7 +2463,8 @@ let lookahead2 ~pass next before =
    *)
 
   (*  (xx) yy *)
-  | (TIdent (s, i1)::TCPar i2::(TIdent (_,i3)|TInt (_,i3))::_ ,
+  | (TIdent (s, i1)::TCPar i2::
+     (TIdent (_,i3)|TInt (_,i3)|TChar (_,i3)|TString (_,i3)|TFloat (_,i3))::_ ,
     (TOPar info)::x::_)
     when not (TH.is_stuff_taking_parenthized x) (* &&
       Ast_c.line_of_info i2 = Ast_c.line_of_info i3 - why useful?
@@ -2588,6 +2648,25 @@ let lookahead2 ~pass next before =
       end
       else TIdent (s, i1)
 
+  | (TIdent(s1,i1)::(TPtVirg(ii2)|TEq(ii2))::rest,
+     TIdent(s2,i2)::TIdent(s3,i3)::_)
+      when (LP.current_context () = LP.InTopLevel &&
+	s1 ==~ regexp_annot) ->
+	  msg_attribute s1;
+	  TMacroEndAttr (s1, i1)
+
+  | (TIdent(s1,i1)::(TPtVirg(ii2)|TEq(ii2))::rest,TCCro(i2)::_)
+      when LP.current_context () = LP.InTopLevel &&
+	s1 ==~ regexp_annot ->
+	  msg_attribute s1;
+	  TMacroEndAttr (s1, i1)
+
+  | (TIdent(s1,i1)::TOBrace(ii2)::rest,TCPar(i2)::_)
+      when LP.current_context () = LP.InTopLevel &&
+	s1 ==~ regexp_annot ->
+	  msg_attribute s1;
+	  TMacroAttr (s1, i1)
+
 (*  (* christia: here insert support for macros on top level *)
   | TIdent (s, ii) :: tl :: _, _ when
     can_be_on_top_level tl && LP.current_context () = InTopLevel ->
@@ -2667,7 +2746,7 @@ let group_ifdef : Ast_c.matching_tag
   remaining
 
 
-let rec cpp_ifdef_statementize (ast :toplevel list) :toplevel list =
+let cpp_ifdef_statementize (ast :toplevel list) :toplevel list =
   Visitor_c.vk_program_s { Visitor_c.default_visitor_c_s with
     Visitor_c.kstatementseq_list_s = (fun (k, bigf) xs ->
       let rec aux : statement_sequencable list -> statement_sequencable list
