@@ -27,11 +27,7 @@ let pr2, pr2_once = Common.mk_pr2_wrappers Flag_parsing_c.verbose_unparsing
 (* Types *)
 (*****************************************************************************)
 
-type type_with_ident =
-    (string * Ast_c.info) option ->
-    (Ast_c.storage * Ast_c.il) option ->
-    Ast_c.fullType ->
-    Ast_c.attribute list -> unit
+type type_with_ident = Ast_c.fullType -> (unit -> unit) -> unit
 
 type 'a printer = 'a -> unit
 
@@ -50,13 +46,17 @@ type pretty_printers = {
   init_list       : (Ast_c.initialiser wrap2 list) printer;
   param           : Ast_c.parameterType printer;
   paramlist       : (Ast_c.parameterType Ast_c.wrap2 list) printer;
+  dparamlist      : ((string Ast_c.wrap) Ast_c.wrap2 list) printer;
   ty              : Ast_c.fullType printer;
   type_with_ident : type_with_ident;
+  base_type       : Ast_c.fullType printer;
+  type_with_ident_rest : type_with_ident;
   toplevel        : Ast_c.toplevel printer;
   fragment        : Ast_c.string_fragment printer;
   fragment_list   : (Ast_c.string_fragment list) printer;
   format          : Ast_c.string_format printer;
-  flow            : Control_flow_c.node printer
+  flow            : Control_flow_c.node printer;
+  name            : Ast_c.name printer
 }
 
 
@@ -166,13 +166,16 @@ let mk_pretty_printers
     | New   (None, t),     [i1] -> pr_elem i1; pp_argument t
     | New   (Some ts, t),     [i1; i2; i3] ->
 	pr_elem i1; pr_elem i2; pp_arg_list ts; pr_elem i3; pp_argument t
-    | Delete(t),     [i1] -> pr_elem i1; pp_expression t
+    | Delete(false,t), [i1] -> pr_elem i1; pr_space(); pp_expression t
+    | Delete(true,t), [i1;i2;i3] ->
+	pr_elem i1; pr_space(); pr_elem i2; pr_elem i3; pr_space();
+	pp_expression t
 
     | Defined name, [i1] ->
-        pr_elem i1; (* defined *)
+        pr_elem i1; (* defined *) pr_space();
         pp_name name;
     | Defined name, [i1;i2;i3] ->
-        pr_elem i1; (* defined *)
+        pr_elem i1; (* defined *) pr_space();
         pr_elem i2; (* ( *)
         pp_name name;
         pr_elem i3; (* ) *)
@@ -183,7 +186,7 @@ let mk_pretty_printers
     | ArrayAccess (_,_) | RecordAccess (_,_) | RecordPtAccess (_,_)
     | SizeOfExpr (_) | SizeOfType (_) | Cast (_,_)
     | StatementExpr (_) | Constructor _
-    | ParenExpr (_) | New (_) | Delete (_)
+    | ParenExpr (_) | New (_) | Delete (_,_)
     | Defined (_)),_ -> raise (Impossible 95)
     );
 
@@ -212,7 +215,7 @@ let mk_pretty_printers
   and pp_arg_list es = pp_list pp_argument es
 
   and pp_argument argument =
-    let rec pp_action (ActMisc ii) = ii +> List.iter pr_elem in
+    let pp_action (ActMisc ii) = ii +> List.iter pr_elem in
     match argument with
     | Left e -> pp_expression e
     | Right weird ->
@@ -272,12 +275,12 @@ and pp_string_format (e,ii) =
 	pr_outdent(); pp_name name; pr_elem i2; pr_nl(); pp_statement st
     | Labeled (Case  (e, st)), [i1;i2] ->
 	pr_unindent();
-        pr_elem i1; pp_expression e; pr_elem i2; pr_nl(); pr_indent();
-	pp_statement st
+        pr_elem i1; pr_space(); pp_expression e; pr_elem i2; pr_nl();
+	pr_indent(); pp_statement st
     | Labeled (CaseRange  (e, e2, st)), [i1;i2;i3] ->
 	pr_unindent();
-        pr_elem i1; pp_expression e; pr_elem i2; pp_expression e2; pr_elem i3;
-	pr_nl(); pr_indent();
+        pr_elem i1; pr_space(); pp_expression e; pr_elem i2;
+	pp_expression e2; pr_elem i3; pr_nl(); pr_indent();
         pp_statement st
     | Labeled (Default st), [i1;i2] ->
 	pr_unindent(); pr_elem i1; pr_elem i2; pr_nl(); pr_indent();
@@ -335,8 +338,10 @@ and pp_string_format (e,ii) =
 	    ForExp (e1opt,il1) ->
               pp_statement (Ast_c.mk_st (ExprStatement e1opt) il1)
 	  | ForDecl decl -> pp_decl decl);
+	  pr_space();
           pp_statement (Ast_c.mk_st (ExprStatement e2opt) il2);
           assert (il3 = []);
+	  pr_space();
           pp_statement (Ast_c.mk_st (ExprStatement e3opt) il3);
           pr_elem i3;
           indent_if_needed st (function _ -> pp_statement st);
@@ -496,15 +501,15 @@ and pp_string_format (e,ii) =
    pp_decl
 *)
   and (pp_type_with_ident:
-	 (string * info) option -> (storage * il) option ->
-	   fullType -> attribute list ->
+	 (unit -> unit) option -> (storage * il) option ->
+	   fullType -> attribute list -> attribute list ->
 	     unit) =
-    fun ident sto ft attrs ->
+    fun ident sto ft attrs endattrs ->
       pp_base_type ft  sto;
       (match (ident, Ast_c.unwrap_typeC ft) with
 	(Some _,_) | (_,Pointer _) -> pr_space()
       |	_ -> ());
-      pp_type_with_ident_rest ident ft attrs
+      pp_type_with_ident_rest ident ft attrs endattrs
 
 
   and (pp_base_type: fullType -> (storage * il) option -> unit) =
@@ -549,13 +554,14 @@ and pp_string_format (e,ii) =
 
           (match sopt,iis with
           | Some s , [i1;i2;i3;i4] ->
-              pr_elem i1; pr_elem i2; pr_elem i3;
+              pr_elem i1; pr_space(); pr_elem i2; pr_space(); pr_elem i3;
           | None, [i1;i2;i3] ->
-              pr_elem i1; pr_elem i2;
+              pr_elem i1; pr_space(); pr_elem i2;
           | x -> raise (Impossible 101)
 	  );
 
-          fields +> List.iter pp_field;
+          fields +> List.iter (fun x -> pr_nl(); pr_indent(); pp_field x);
+	  pr_nl();
 
           (match sopt,iis with
           | Some s , [i1;i2;i3;i4] -> pr_elem i4
@@ -630,6 +636,9 @@ and pp_string_format (e,ii) =
             pr_elem (Ast_c.fakeInfo() +> Ast_c.rewrap_str "*}");
           end;
 
+      | (FieldType (t, _, _), iis) ->
+	  pp_base_type t sto
+
       | (TypeOfExpr (e), iis) ->
           print_sto_qu (sto, qu);
           (match iis with
@@ -673,9 +682,10 @@ and pp_string_format (e,ii) =
 		let identinfo =
                   match nameopt with
 		  | None -> None
-                  | Some name -> Some (get_s_and_info_of_name name)
+                  | Some name -> Some (function _ -> pp_name name)
                 in
-		pp_type_with_ident identinfo None typ Ast_c.noattr;
+		pp_type_with_ident identinfo None typ Ast_c.noattr
+		  Ast_c.noattr;
 
 	    | (BitField (nameopt, typ, iidot, expr)), iivirg ->
                       (* first var cannot have a preceding ',' *)
@@ -684,9 +694,8 @@ and pp_string_format (e,ii) =
 		| None ->
 		    pp_type typ;
 		| Some name ->
-                    let (s, is) = get_s_and_info_of_name name in
-		    pp_type_with_ident
-		      (Some (s, is)) None typ Ast_c.noattr;
+		    pp_type_with_ident (Some (function _ -> pp_name name))
+		      None typ Ast_c.noattr Ast_c.noattr;
 		    );
                 pr_elem iidot;
 		pp_expression expr
@@ -700,17 +709,18 @@ and pp_string_format (e,ii) =
 		  let identinfo =
 		    match nameopt with
 		    | None -> None
-		    | Some name -> Some (get_s_and_info_of_name name)
+		    | Some name -> Some (function _ -> pp_name name)
 		  in
-		  pp_type_with_ident_rest identinfo typ Ast_c.noattr
+		  pp_type_with_ident_rest identinfo typ
+		    Ast_c.noattr Ast_c.noattr
 
 	      | (BitField (nameopt, typ, iidot, expr)), iivirg ->
 		  iivirg +> List.iter pr_elem;
 		  (match nameopt with
 		  | Some name ->
-                      let (s,is) = get_s_and_info_of_name name in
 		      pp_type_with_ident_rest
-			(Some (s, is)) typ Ast_c.noattr;
+			(Some (function _ -> pp_name name))
+			typ Ast_c.noattr Ast_c.noattr;
 		      pr_elem iidot;
 		      pp_expression expr
 		  | None ->
@@ -756,14 +766,16 @@ and pp_string_format (e,ii) =
 
 (* used because of DeclList, in    int i,*j[23];  we don't print anymore the
    int before *j *)
-  and (pp_type_with_ident_rest: (string * info) option ->
-    fullType -> attribute list -> unit) =
+  and (pp_type_with_ident_rest: (unit -> unit) option ->
+    fullType -> attribute list -> attribute list -> unit) =
 
-    fun ident (((qu, iiqu), (ty, iity)) as fullt) attrs ->
+    fun ident (((qu, iiqu), (ty, iity)) as fullt) attrs endattrs ->
 
-      let print_ident ident = Common.do_option (fun (s, iis) ->
+      let print_ident ident = Common.do_option (fun f ->
         (* XXX attrs +> pp_attributes pr_elem pr_space; *)
-        pr_elem iis
+        f();
+	(if not(endattrs = []) then pr_space());
+        endattrs +> pp_attributes pr_elem pr_space
 	) ident
       in
 
@@ -777,6 +789,7 @@ and pp_string_format (e,ii) =
       | (EnumName  s, iis)                      -> print_ident ident
       | (Decimal _, iis)                        -> print_ident ident
       | (TypeName (_name,_typ), iis)            -> print_ident ident
+      | (FieldType (_typ,_,_), iis)             -> print_ident ident
       | (TypeOfExpr (e), iis)                   -> print_ident ident
       | (TypeOfType (e), iis)                   -> print_ident ident
 
@@ -789,7 +802,7 @@ and pp_string_format (e,ii) =
           (* bug: pp_type_with_ident_rest None t;      print_ident ident *)
           pr_elem i;
           iiqu +> List.iter pr_elem; (* le const est forcement apres le '*' *)
-          pp_type_with_ident_rest ident t attrs;
+          pp_type_with_ident_rest ident t attrs Ast_c.noattr;
 
       (* ugly special case ... todo? maybe sufficient in practice *)
       | (ParenType ttop, [i1;i2]) ->
@@ -806,7 +819,7 @@ and pp_string_format (e,ii) =
 		  pp_type_right (q2, mk_tybis (FunctionType t) ii3);
               | _ ->
                   pr2 "PB PARENTYPE ZARB, I forget about the ()";
-                  pp_type_with_ident_rest ident ttop attrs;
+                  pp_type_with_ident_rest ident ttop attrs Ast_c.noattr;
               )
           (* another ugly special case *)
           | _q1, (Array (eopt,t2 ), [iarray1;iarray2]) ->
@@ -826,16 +839,16 @@ and pp_string_format (e,ii) =
 		      pp_type_right (q3, mk_tybis (FunctionType t) iifunc)
                   | _ ->
                       pr2 "PB PARENTYPE ZARB, I forget about the ()";
-                      pp_type_with_ident_rest ident ttop attrs;
+                      pp_type_with_ident_rest ident ttop attrs Ast_c.noattr;
                   )
               | _ ->
                   pr2 "PB PARENTYPE ZARB, I forget about the ()";
-                  pp_type_with_ident_rest ident ttop attrs;
+                  pp_type_with_ident_rest ident ttop attrs Ast_c.noattr;
               )
           | _t ->
 
               pr2 "PB PARENTYPE ZARB, I forget about the ()";
-              pp_type_with_ident_rest ident ttop attrs;
+              pp_type_with_ident_rest ident ttop attrs Ast_c.noattr;
           )
 
 
@@ -883,7 +896,7 @@ and pp_string_format (e,ii) =
       | (EnumName  s, iis) -> ()
       | (Decimal(l,p), iis) -> ()
       | (TypeName (_name,_typ), iis) -> ()
-
+      | FieldType (_, _, _), _ -> ()
       | TypeOfType _, _ -> ()
       | TypeOfExpr _, _ -> ()
 
@@ -901,9 +914,8 @@ and pp_string_format (e,ii) =
     | None ->
         pp_type t
     | Some name ->
-        let (s,i1) = get_s_and_info_of_name name in
-	pp_type_with_ident
-          (Some (s, i1)) None t Ast_c.noattr
+	pp_type_with_ident (Some (function _ -> pp_name name))
+	  None t Ast_c.noattr Ast_c.noattr
 
 
 
@@ -941,6 +953,7 @@ and pp_string_format (e,ii) =
     | (EnumName  s, iis) -> ()
     | (Decimal(l,p), iis) -> ()
     | (TypeName (name,_typ), iis) -> ()
+    | (FieldType (_, _, _), _) -> ()
 
     | TypeOfType _, _ -> ()
     | TypeOfExpr _, _ -> ()
@@ -948,7 +961,13 @@ and pp_string_format (e,ii) =
     | (FunctionType _ | Array _ | Pointer _), _ -> raise (Impossible 111)
 
   and pp_type t =
-    pp_type_with_ident None None t Ast_c.noattr
+    pp_type_with_ident None None t Ast_c.noattr Ast_c.noattr
+  and pp_type_ident t i =
+    pp_type_with_ident (Some i) None t Ast_c.noattr Ast_c.noattr
+  and pp_type_ident_rest t i =
+    pp_type_with_ident_rest (Some i) t Ast_c.noattr Ast_c.noattr
+  and pp_base_type2 t =
+    pp_base_type t None
 
 (* ---------------------- *)
   and pp_decl = function
@@ -956,6 +975,7 @@ and pp_string_format (e,ii) =
                    v_type = returnType;
                    v_storage = storage;
                    v_attr = attrs;
+                   v_endattr = endattrs;
                   },[])::xs),
 	       iivirg::ifakestart::iisto) ->
 
@@ -967,13 +987,13 @@ and pp_string_format (e,ii) =
         (* handling the first var. Special case, we print the whole type *)
 	(match var with
 	| Some (name, iniopt) ->
-            let (s,iis) = get_s_and_info_of_name name in
 	    pp_type_with_ident
-	      (Some (s, iis)) (Some (storage, iisto))
-	      returnType attrs;
+	      (Some (function _ -> pp_name name)) (Some (storage, iisto))
+	      returnType attrs endattrs;
 	    (match iniopt with
 	      Ast_c.NoInit -> ()
-	    | Ast_c.ValInit(iini,init) -> pr_elem iini; pp_init init
+	    | Ast_c.ValInit(iini,init) ->
+		pr_space(); pr_elem iini; pr_space(); pp_init init
 	    | Ast_c.ConstrInit((init,[lp;rp])) ->
 		pr_elem lp; pp_arg_list init; pr_elem rp
 	    | Ast_c.ConstrInit _ -> raise (Impossible 112))
@@ -986,13 +1006,13 @@ and pp_string_format (e,ii) =
 	    v_type = returnType;
 	    v_storage = storage2;
 	    v_attr = attrs;
+	    v_endattr = endattrs;
 	  }, iivirg) ->
 
-            let (s,iis) = get_s_and_info_of_name name in
 	    assert (storage2 = storage);
 	    iivirg +> List.iter pr_elem;
-	    pp_type_with_ident_rest
-	      (Some (s, iis)) returnType attrs;
+	    pp_type_with_ident_rest (Some (function _ -> pp_name name))
+	      returnType attrs endattrs;
 	    (match iniopt with
 	      Ast_c.NoInit -> ()
 	    | Ast_c.ValInit(iini,init) -> pr_elem iini; pp_init init
@@ -1121,7 +1141,7 @@ and pp_init (init, iinit) =
         pr_elem ifakestart;
 
         pp_type_with_ident None (Some (sto, isto))
-          returnt Ast_c.noattr;
+          returnt Ast_c.noattr Ast_c.noattr;
 
         pp_attributes pr_elem pr_space attrs;
 	pr_space();
@@ -1148,10 +1168,10 @@ and pp_init (init, iinit) =
 
 	       (match b, s, ii_b_s with
                | false, Some s, [i1] ->
-		   pp_type_with_ident (Some (s, i1)) None t;
+		   pp_type_with_ident (Some (function _ -> pr_elem i1)) None t;
                | true, Some s, [i1;i2] ->
 		   pr_elem i1;
-		   pp_type_with_ident (Some (s, i2)) None t;
+		   pp_type_with_ident (Some (function _ -> pr_elem i2)) None t;
 
             (* in definition we have name for params, except when f(void) *)
                | _, None, _ -> raise Impossible
@@ -1192,8 +1212,8 @@ and pp_init (init, iinit) =
 	pr_elem i1; pr_space(); pr_elem i2
     | Define ((s,ii), (defkind, defval)) ->
 	let (idefine,iident,ieol) = Common.tuple_of_list3 ii in
-	pr_elem idefine;
-	pr_elem iident;
+	pr_elem idefine; pr_space();
+	pr_elem iident; pr_space();
 
 	let define_val = function
           | DefineExpr e -> pp_expression e
@@ -1223,11 +1243,7 @@ and pp_init (init, iinit) =
 	| DefineFunc (params, ii) ->
             let (i1,i2) = tuple_of_list2 ii in
             pr_elem i1;
-            params +> List.iter (fun ((s,iis), iicomma) ->
-              assert (List.length iicomma <= 1);
-              iicomma +> List.iter pr_elem;
-              iis +> List.iter pr_elem;
-            );
+	    pp_define_param_list params;
             pr_elem i2;
 	);
 	define_val defval;
@@ -1235,8 +1251,8 @@ and pp_init (init, iinit) =
 
     | Pragma ((s,ii), pragmainfo) ->
 	let (ipragma,iident,ieol) = Common.tuple_of_list3 ii in
-	pr_elem ipragma;
-	pr_elem iident;
+	pr_elem ipragma; pr_space();
+	pr_elem iident; pr_space();
 	pp_pragmainfo pragmainfo;
 	pr_elem ieol
 
@@ -1250,11 +1266,14 @@ and pp_init (init, iinit) =
 	pp_arg_list args;
         pr_elem irp
     | PragmaIdList(ids) ->
-	let rec loop = function
+	let loop = function
 	    [] -> ()
 	  | [id,_] -> pp_name id
 	  | (id,_)::rest -> pp_name id; pr_space() in
-	loop ids in
+	loop ids
+
+  and pp_define_param_list dparams =
+    pp_list (fun (s,iis) -> iis +> List.iter pr_elem) dparams in
 
   let rec pp_toplevel = function
     | Declaration decl -> pp_decl decl
@@ -1311,7 +1330,7 @@ and pp_init (init, iinit) =
          iif iicomma;
 	 );
       *)
-		       pr2 "Def";
+		       pr2 "FunHeader";
 
 
     | F.Decl decl ->
@@ -1321,15 +1340,15 @@ and pp_init (init, iinit) =
     | F.ExprStatement (st, (eopt, ii)) ->
 	pp_statement (Ast_c.mk_st (ExprStatement eopt) ii)
 
-    | F.IfHeader (_, (e,ii))
-    | F.SwitchHeader (_, (e,ii))
-    | F.WhileHeader (_, (e,ii))
+    | F.IfHeader (_, (e,ii)) -> pr2 "IfHeader"
+    | F.SwitchHeader (_, (e,ii)) -> pr2 "SwitchHeader"
+    | F.WhileHeader (_, (e,ii)) -> pr2 "WhileHeader"
     | F.DoWhileTail (e,ii) ->
         (*
            iif ii;
            vk_expr bigf e
         *)
-	pr2 "XXX";
+	pr2 "DoWhileTail"
 
 
     | F.ForHeader (_st, ((first, (e2opt,i2), (e3opt,i3)), ii)) ->
@@ -1340,28 +1359,28 @@ and pp_init (init, iinit) =
            e2opt +> do_option (vk_expr bigf);
            e3opt +> do_option (vk_expr bigf);
         *)
-	pr2 "XXX"
+	pr2 "ForHeader"
 
     | F.MacroIterHeader (_s, ((s,es), ii)) ->
         (*
            iif ii;
            vk_argument_list bigf es;
         *)
-	pr2 "XXX"
+	pr2 "MacroIterHeader"
 
 
     | F.ReturnExpr (_st, (e,ii)) ->
         (* iif ii; vk_expr bigf e*)
-	pr2 "XXX"
+	pr2 "ReturnExpr"
 
 
     | F.Case  (_st, (e,ii)) ->
       (* iif ii; vk_expr bigf e *)
-	pr2 "XXX"
+	pr2 "Case"
 
     | F.CaseRange (_st, ((e1, e2),ii)) ->
         (* iif ii; vk_expr bigf e1; vk_expr bigf e2 *)
-	pr2 "XXX"
+	pr2 "CaseRange"
 
 
 
@@ -1369,23 +1388,23 @@ and pp_init (init, iinit) =
 
     | F.DefineExpr e  ->
         (* vk_expr bigf e *)
-	pr2 "XXX"
+	pr2 "DefineExpr"
 
     | F.DefineType ft  ->
         (* vk_type bigf ft *)
-	pr2 "XXX"
+	pr2 "DefineType"
 
     | F.DefineHeader ((s,ii), (defkind))  ->
         (*
            iif ii;
            vk_define_kind bigf defkind;
         *)
-	pr2 "XXX"
+	pr2 "DefineHeader"
 
 
     | F.DefineDoWhileZeroHeader (((),ii)) ->
         (* iif ii *)
-	pr2 "XXX"
+	pr2 "DefineDoWhileZeroHeader"
 
     | F.PragmaHeader((s,ii), pragmainfo) ->
 	let (ipragma,iident,ieol) = Common.tuple_of_list3 ii in
@@ -1395,60 +1414,60 @@ and pp_init (init, iinit) =
 
     | F.Include {i_include = (s, ii);} ->
         (* iif ii; *)
-	pr2 "XXX"
+	pr2 "Include"
 
 
     | F.MacroTop (s, args, ii) ->
         (* iif ii;
            vk_argument_list bigf args *)
-	pr2 "XXX"
+	pr2 "MacroTop"
 
 
     | F.Break    (st,((),ii),fromswitch) ->
         (* iif ii *)
-	pr2 "XXX"
+	pr2 "Break"
     | F.Continue (st,((),ii)) ->
         (* iif ii *)
-	pr2 "XXX"
+	pr2 "Continue"
     | F.Default  (st,((),ii)) ->
         (* iif ii *)
-	pr2 "XXX"
+	pr2 "Default"
     | F.Return   (st,((),ii)) ->
         (* iif ii *)
-	pr2 "XXX"
+	pr2 "Return"
     | F.Goto  (st, name, ((),ii)) ->
         (* iif ii *)
-	pr2 "XXX"
+	pr2 "Goto"
     | F.Label (st, name, ((),ii)) ->
         (* iif ii *)
-	pr2 "XXX"
+	pr2 "Label"
     | F.EndStatement iopt ->
         (* do_option infof iopt *)
-	pr2 "XXX"
+	pr2 "EndStatement"
     | F.DoHeader (st, info) ->
         (* infof info *)
-	pr2 "XXX"
+	pr2 "DoHeader"
     | F.Else info ->
         (* infof info *)
-	pr2 "XXX"
+	pr2 "Else"
     | F.SeqEnd (i, info) ->
         (* infof info *)
-	pr2 "XXX"
+	pr2 "SeqEnd"
     | F.SeqStart (st, i, info) ->
         (* infof info *)
-	pr2 "XXX"
+	pr2 "SeqStart"
 
     | F.MacroStmt (st, ((),ii)) ->
         (* iif ii *)
-	pr2 "XXX"
+	pr2 "MacroStmt"
     | F.Asm (st, (asmbody,ii)) ->
         (*
            iif ii;
            vk_asmbody bigf asmbody
         *)
-	pr2 "XXX"
+	pr2 "Asm"
 
-    | F.Exec(st,(code,ii)) -> pr2 "XXX"
+    | F.Exec(st,(code,ii)) -> pr2 "Exec"
 
     | F.IfdefHeader (info) ->
 	pp_ifdef info
@@ -1458,17 +1477,24 @@ and pp_init (init, iinit) =
 	pp_ifdef info
 
     | F.IfdefIteHeader _ii ->
-        pr2 "XXX"
+        pr2 "IfdefIteHeader"
 
     | F.DefineTodo ->
-	pr2 "XXX"
+	pr2 "DefineTodo"
 
 
-    | (F.TopNode|F.EndNode|
-      F.ErrorExit|F.Exit|F.Enter|F.LoopFallThroughNode|F.FallThroughNode|
-      F.AfterNode _|F.FalseNode|F.TrueNode _|F.InLoopNode|
-      F.Fake) ->
-        pr2 "YYY" in
+    | F.TopNode -> pr2 "TopNode"
+    | F.EndNode -> pr2 "EndNode"
+    | F.ErrorExit -> pr2 "ErrorExit"
+    | F.Exit -> pr2 "Exit"
+    | F.Enter -> pr2 "Enter"
+    | F.LoopFallThroughNode -> pr2 "LoopFallThroughNode"
+    | F.FallThroughNode -> pr2 "FallThroughNode"
+    | F.AfterNode _ -> pr2 "AfterNode"
+    | F.FalseNode -> pr2 "FalseNode"
+    | F.TrueNode _ -> pr2 "TrueNode"
+    | F.InLoopNode -> pr2 "InLoopNode"
+    | F.Fake -> pr2 "Fake" in
 
 
   { expression = pp_expression;
@@ -1485,13 +1511,17 @@ and pp_init (init, iinit) =
     init_list  = pp_init_list;
     param      = pp_param;
     paramlist  = pp_param_list;
+    dparamlist = pp_define_param_list;
     ty         = pp_type;
-    type_with_ident = pp_type_with_ident;
+    type_with_ident = pp_type_ident;
+    type_with_ident_rest = pp_type_ident_rest;
+    base_type  = pp_base_type2;
     toplevel   = pp_toplevel;
     fragment   = pp_string_fragment;
     fragment_list = pp_string_fragment_list;
     format     = pp_string_format;
     flow       = pp_flow;
+    name       = pp_name;
   }
 
 (*****************************************************************************)
@@ -1548,6 +1578,7 @@ let pp_toplevel_simple   = ppc.toplevel
 let pp_string_fragment_simple = ppc.fragment
 let pp_string_format_simple = ppc.format
 let pp_flow_simple       = ppc.flow
+let pp_name              = ppc.name
 
 
 let pp_elem_sp ~pr_elem ~pr_space =
@@ -1597,6 +1628,9 @@ let pp_param_gen ~pr_elem ~pr_space =
 let pp_param_list_gen ~pr_elem ~pr_space =
   (pp_elem_sp pr_elem pr_space).paramlist
 
+let pp_define_param_list_gen ~pr_elem ~pr_space =
+  (pp_elem_sp pr_elem pr_space).dparamlist
+
 let pp_type_gen ~pr_elem ~pr_space =
   (pp_elem_sp pr_elem pr_space).ty
 
@@ -1630,10 +1664,20 @@ let string_of_toplevel top =
     pp_toplevel_simple top
   )
 
+let string_of_fullType t =
+  Common.format_to_string (fun () ->
+    pp_type_simple t
+  )
+
+let string_of_name name =
+  Common.format_to_string (fun () ->
+    pp_name name
+  )
+
 let (debug_info_of_node:
-       Ograph_extended.nodei -> Control_flow_c.cflow -> string) =
+       Control_flow_c.G.key -> Control_flow_c.cflow -> string) =
   fun nodei flow ->
-    let node = flow#nodes#assoc nodei in
+    let node = Control_flow_c.KeyMap.find nodei flow#nodes in
     let s = Common.format_to_string (fun () ->
       pp_flow_simple node
     ) in
