@@ -1,8 +1,4 @@
-type type_with_ident =
-    (string * Ast_c.info) option ->
-    (Ast_c.storage * Ast_c.il) option ->
-    Ast_c.fullType ->
-    Ast_c.attribute list -> unit
+type type_with_ident = Ast_c.fullType -> (unit -> unit) -> unit
 
 type 'a printer = 'a -> unit
 
@@ -21,13 +17,17 @@ type pretty_printers = {
   init_list       : (Ast_c.initialiser Ast_c.wrap2 list) printer;
   param           : Ast_c.parameterType printer;
   paramlist       : (Ast_c.parameterType Ast_c.wrap2 list) printer;
+  dparamlist      : ((string Ast_c.wrap) Ast_c.wrap2 list) printer;
   ty              : Ast_c.fullType printer;
   type_with_ident : type_with_ident;
+  base_type       : Ast_c.fullType printer;
+  type_with_ident_rest : type_with_ident;
   toplevel        : Ast_c.toplevel printer;
   fragment        : Ast_c.string_fragment printer;
   fragment_list   : (Ast_c.string_fragment list) printer;
   format          : Ast_c.string_format printer;
-  flow            : Control_flow_c.node printer
+  flow            : Control_flow_c.node printer;
+  name            : Ast_c.name printer
 }
 
 val mk_pretty_printers :
@@ -65,6 +65,9 @@ val pp_param_gen:  pr_elem:Ast_c.info printer -> pr_space: unit printer ->
   Ast_c.parameterType printer
 val pp_param_list_gen:  pr_elem:Ast_c.info printer -> pr_space: unit printer ->
   (Ast_c.parameterType Ast_c.wrap2 list) printer
+val pp_define_param_list_gen:
+    pr_elem:Ast_c.info printer -> pr_space: unit printer ->
+      ((string Ast_c.wrap) Ast_c.wrap2 list) printer
 val pp_type_gen:  pr_elem:Ast_c.info printer -> pr_space:unit printer ->
   Ast_c.fullType printer
 val pp_init_gen:  pr_elem:Ast_c.info printer -> pr_space:unit printer ->
@@ -96,10 +99,10 @@ val pp_string_fragment_simple:   Ast_c.string_fragment printer
 val pp_string_format_simple:     Ast_c.string_format printer
 
 val debug_info_of_node:
-  Ograph_extended.nodei -> Control_flow_c.cflow -> string
+  Control_flow_c.G.key -> Control_flow_c.cflow -> string
 
 val string_of_expression: Ast_c.expression -> string
-(** Normalized string representation of an [Ifdef] guard.
+(* Normalized string representation of an [Ifdef] guard.
  *
  * Ignored #if conditions (cf. [Gnone]) are treated as 0, which is consistent
  * with the way Coccinelle handles them.
@@ -108,3 +111,5 @@ val string_of_expression: Ast_c.expression -> string
  *)
 val string_of_ifdef_guard: Ast_c.ifdef_guard -> string
 val string_of_toplevel: Ast_c.toplevel -> string
+val string_of_fullType: Ast_c.fullType -> string
+val string_of_name: Ast_c.name -> string
