@@ -215,7 +215,7 @@ let is_ident_like = function
   | TUnknown _
 
   | TMacroAttr _
-  | TMacroAttrStorage _
+  | TMacroEndAttr _
   | TMacroStmt _
   | TMacroString _
   | TMacroDecl _
@@ -332,7 +332,7 @@ let info_of_tok = function
 
   | TMacroIdentBuilder             (s, i) -> i
   | TMacroAttr             (s, i) -> i
-  | TMacroAttrStorage             (s, i) -> i
+  | TMacroEndAttr             (s, i) -> i
   | TMacroStmt             (s, i) -> i
   | TMacroString             (s, i) -> i
   | TMacroDecl             (s, i) -> i
@@ -358,7 +358,7 @@ let info_of_tok = function
   | TEndif               (_, i) -> i
   | TIfdefBool           (b, _, i) -> i
   | TIfdefMisc           (b, _, i) -> i
-  | TIfdefVersion           (b, _, i) -> i
+  | TIfdefVersion        (b, _, i) -> i
 
   | TUifdef              (i) -> i
   | TUelseif             (i) -> i
@@ -444,7 +444,7 @@ let info_of_tok = function
   | Tsizeof              (i) -> i
   | Tasm                 (i) -> i
   | Tattribute           (i) -> i
-  | TattributeNoarg           (i) -> i
+  | TattributeNoarg      (i) -> i
   | Tinline              (i) -> i
   | Ttypeof              (i) -> i
   | Tnew                 (i) -> i
@@ -454,7 +454,7 @@ let info_of_tok = function
 
   | EOF                  (i) -> i
   | Tnamespace           (i) -> i
-
+  | TTODO _ -> failwith "fake token, should not occur"
 
 
 
@@ -506,7 +506,7 @@ let visitor_info_of_tok f = function
 
   | TMacroIdentBuilder             (s, i) -> TMacroIdentBuilder (s, f i)
   | TMacroAttr           (s, i)   -> TMacroAttr            (s, f i)
-  | TMacroAttrStorage           (s, i)   -> TMacroAttrStorage         (s, f i)
+  | TMacroEndAttr        (s, i)   -> TMacroEndAttr         (s, f i)
   | TMacroStmt           (s, i)   -> TMacroStmt            (s, f i)
   | TMacroString         (s, i)   -> TMacroString          (s, f i)
   | TMacroDecl           (s, i) -> TMacroDecl            (s, f i)
@@ -628,7 +628,7 @@ let visitor_info_of_tok f = function
   | TOParCplusplusInit   (i) -> TOParCplusplusInit   (f i)
   | EOF                  (i) -> EOF                  (f i)
   | Tnamespace           (i) -> Tnamespace           (f i)
-
+  | TTODO _ -> failwith "fake token, should not occur"
 
 (*****************************************************************************)
 (* Accessors *)
@@ -773,13 +773,13 @@ let string_of_token = function
   | TPragma _ -> "TPragma"
   | TCppDirectiveOther _ -> "TCppDirectiveOther"
   | TMacroAttr _ -> "TMacroAttr"
+  | TMacroEndAttr _ -> "TMacroEndAttr"
   | TMacroStmt _ -> "TMacroStmt"
   | TMacroIdentBuilder _ -> "TMacroIdentBuilder"
   | TMacroString _ -> "TMacroString"
   | TMacroDecl _ -> "TMacroDecl"
   | TMacroDeclConst _ -> "TMacroDeclConst"
   | TMacroIterator _ -> "TMacroIterator"
-  | TMacroAttrStorage _ -> "TMacroAttrStorage"
   | TCommentSkipTagStart _ -> "TCommentSkipTagStart"
   | TCommentSkipTagEnd _ -> "TCommentSkipTagEnd"
   | TCParEOL _ -> "TCParEOL"
@@ -787,6 +787,7 @@ let string_of_token = function
   | TCommentMisc _ -> "TCommentMisc"
   | TCommentCpp _ -> "TCommentCpp"
   | EOF _ -> "EOF"
+  | TTODO _ -> failwith "fake token, should not occur"
 
 let file_of_tok x = Ast_c.file_of_info (info_of_tok x)
 let pinfo_of_tok x = Ast_c.pinfo_of_info (info_of_tok x)
