@@ -397,6 +397,8 @@ let rec type_unfold_one_step ty env =
           ty
       )
 
+  | FieldType (t, _, _) -> type_unfold_one_step t env
+
   | ParenType t -> type_unfold_one_step t env
   | TypeOfExpr e ->
       pr2_once ("Type_annoter: not handling typeof");
@@ -473,6 +475,9 @@ let rec typedef_fix ty env =
             with Not_found ->
               ty))
 
+    | FieldType (t, a, b) ->
+	FieldType (typedef_fix t env, a, b) +> Ast_c.rewrap_typeC ty
+
   (* remove paren for better matching with typed metavar. kind of iso again *)
     | ParenType t ->
 	typedef_fix t env
@@ -535,7 +540,7 @@ let rec is_simple_expr expr =
 
 
 (* memoise unnanoted var, to avoid too much warning messages *)
-let _notyped_var = ref (Hashtbl.create 101)
+let _notyped_var = Hashtbl.create 101
 
 let new_scope() =
   match !_scoped_env with
@@ -603,7 +608,7 @@ let add_binding2 namedef warning =
   if !Flag_parsing_c.check_annotater then begin
     (match namedef with
     | VarOrFunc (s, typ) ->
-        if Hashtbl.mem !_notyped_var s
+        if Hashtbl.mem _notyped_var s
         then pr2 ("warning: found typing information for a variable that was" ^
                      "previously unknown:" ^ s);
     | _ -> ()
@@ -842,10 +847,10 @@ let annotater_expr_visitor_subpart = (fun (k,bigf) expr ->
                     if not (s =~ "[A-Z_]+") (* if macro then no warning *)
                     then
                       if !Flag_parsing_c.check_annotater then
-                        if not (Hashtbl.mem !_notyped_var s)
+                        if not (Hashtbl.mem _notyped_var s)
                         then begin
                           pr2 ("Type_annoter: no type found for: " ^ s);
-                          Hashtbl.add !_notyped_var s true;
+                          Hashtbl.add _notyped_var s true;
                         end
                         else ()
                       else
@@ -977,11 +982,15 @@ let annotater_expr_visitor_subpart = (fun (k,bigf) expr ->
     (* pad: julia wrote this ? *)
     | Unary (e, UnPlus) ->
         k expr; (* recurse to set the types-ref of sub expressions *)
-        make_info_def (type_of_s "int")
-          (* todo? can convert from unsigned to signed if UnMinus ? *)
+	(* No type inference.  If one cares about being int, one probably
+	   cares about what kind of int too *)
+        Ast_c.get_type_expr e
+
     | Unary (e, UnMinus) ->
         k expr; (* recurse to set the types-ref of sub expressions *)
-        make_info_def (type_of_s "int")
+	(* No type inference.  If one cares about being int, one probably
+	   cares about what kind of int too *)
+        Ast_c.get_type_expr e
 
     | SizeOfType _|SizeOfExpr _ ->
         k expr; (* recurse to set the types-ref of sub expressions *)
@@ -1019,7 +1028,7 @@ let annotater_expr_visitor_subpart = (fun (k,bigf) expr ->
 	pr2_once "Type annotater:not handling New";
 	Type_c.noTypeHere (* TODO *)
 
-    | Delete e ->
+    | Delete (box,e) ->
 	k expr;
 	pr2_once "Type annotater:not handling Delete";
 	Type_c.noTypeHere (* TODO *)
@@ -1256,7 +1265,6 @@ let rec visit_toplevel ~just_add_in_env ~depth elem =
 
     (* ------------------------------------------------------------ *)
     Visitor_c.ktoplevel = (fun (k, bigf) elem ->
-      _notyped_var := Hashtbl.create 100;
       match elem with
       | Definition def ->
           let {f_name = name;
@@ -1344,7 +1352,7 @@ let rec visit_toplevel ~just_add_in_env ~depth elem =
     );
   }
   in
-  if just_add_in_env
+  (if just_add_in_env
   then
     if depth > 1
     then Visitor_c.vk_toplevel bigf elem
@@ -1352,7 +1360,9 @@ let rec visit_toplevel ~just_add_in_env ~depth elem =
       Common.profile_code "TAC.annotate_only_included" (fun () ->
         Visitor_c.vk_toplevel bigf elem
       )
-  else Visitor_c.vk_toplevel bigf elem
+  else Visitor_c.vk_toplevel bigf elem);
+  Stdcompat.Hashtbl.reset _notyped_var
+
 
 (*****************************************************************************)
 (* Entry point *)
@@ -1360,21 +1370,21 @@ let rec visit_toplevel ~just_add_in_env ~depth elem =
 (* catch all the decl to grow the environment *)
 
 
-let rec (annotate_program2 :
+let (annotate_program2 :
   environment -> toplevel list -> (toplevel * environment Common.pair) list) =
  fun env prog ->
 
   (* globals (re)initialialisation *)
   _scoped_env := env;
-  _notyped_var := (Hashtbl.create 100);
-
-  prog +> List.map (fun elem ->
-    let beforeenv = !_scoped_env in
-    visit_toplevel ~just_add_in_env:false ~depth:0 elem;
-    let afterenv = !_scoped_env in
-    (elem, (beforeenv, afterenv))
-  )
 
+   let res =
+     prog +> List.map (fun elem ->
+       let beforeenv = !_scoped_env in
+       visit_toplevel ~just_add_in_env:false ~depth:0 elem;
+       let afterenv = !_scoped_env in
+       (elem, (beforeenv, afterenv))) in
+  Stdcompat.Hashtbl.reset _notyped_var;
+   res
 
 
 
@@ -1399,7 +1409,7 @@ let annotate_test_expressions prog =
 	    let nm = str_of_name i in
 	    if List.mem nm ["likely";"unlikely"]
 	    then propagate_test a
-		else ()
+	    else ()
 	| _ -> ())
     | _ -> () in
 
