@@ -263,7 +263,34 @@ let get_fakeInfo_and_tokens celem toks =
 
   let pr_space _ = () in (* use the spacing that is there already *)
 
-  Pretty_print_c.pp_program_gen pr_elem pr_space celem;
+  let printed_toks = ref [] in
+  let pr_get_elem tok = push2 tok printed_toks in
+
+  Pretty_print_c.pp_program_gen pr_get_elem pr_space celem;
+
+  (* sort tokens when possible *)
+  let is_origin info =
+    match Ast_c.pinfo_of_info info with
+    | Ast_c.OriginTok _ -> true
+    | _ -> false in (* don't know if the others have proper positions *)
+  let rec tok_group = function
+      [] -> []
+    | tok::rest ->
+	let (nonorig,origrest) =
+	  Common.span (function tok -> not (is_origin tok)) rest in
+	(tok::nonorig)::(tok_group origrest) in
+  let (front,rest) =
+    match tok_group (List.rev !printed_toks) with
+      front::rest when not(is_origin (List.hd front)) (*might not be orig*)
+	  -> (front,rest)
+    | all -> ([],all) in
+  let tcompare l1 l2 =
+    match (l1,l2) with
+      (t1::_,t2::_) ->
+	compare (Ast_c.info_to_fixpos t1) (Ast_c.info_to_fixpos t2)
+    | _ -> failwith "not possible" in
+  let printed_toks = List.concat (front :: (List.sort tcompare rest)) in
+  List.iter pr_elem printed_toks;
 
   if  (!toks_in <> [])
   then failwith "WEIRD: unparsing not finished";
@@ -352,15 +379,17 @@ let displace_fake_nodes toks =
 (* Tokens2 generation *)
 (*****************************************************************************)
 
-let comment2t2 = function
+let rec comment2t2 = function
   | (Token_c.TCommentCpp
   (* not sure iif the following list is exhaustive or complete *)
     (Token_c.CppAttr|Token_c.CppMacro|Token_c.CppPassingCosWouldGetError),
-    (info : Token_c.info)) ->
-    C2(info.Common.str,None)
-  | (Token_c.TCommentCpp x,(info : Token_c.info)) ->
-    C2("\n"^info.Common.str^"\n",None)
-  | x -> failwith (Printf.sprintf "unexpected comment %s" (Dumper.dump x))
+    (info : Token_c.info)) :: rest ->
+    C2(info.Common.str,None) :: comment2t2 rest
+  | (Token_c.TCommentCpp x,(info : Token_c.info)) :: rest ->
+    C2("\n"^info.Common.str^"\n",None) :: comment2t2 rest
+  | [(_,_)] | [] -> []
+  (* keep spacing within comments, if available *)
+  | (_,info)::rest -> C2(info.Common.str,None) :: comment2t2 rest
 
 let expand_mcode toks =
   let toks_out = ref [] in
@@ -420,17 +449,15 @@ let expand_mcode toks =
       push2 (Cocci2 (s,ln,col,rcol,hint)) toks_out  in
     let pr_c info =
       (match Ast_c.pinfo_of_info info with
-      | Ast_c.AbstractLineTok _ ->
-        push2 (C2 (Ast_c.str_of_info info,None)) toks_out
       |	Ast_c.FakeTok (s,_) ->
         push2 (C2 (s,None)) toks_out
       |	_ ->
-        Printf.fprintf stderr "line: %s\n" (Dumper.dump info);
-        failwith "not an abstract line"
+        push2 (C2 (Ast_c.str_of_info info,None)) toks_out
       );
       (* why nothing for mbefore? *)
       (Ast_c.get_comments_after info) +>
-      List.iter (fun x -> push2 (comment2t2 x) toks_out) in
+      comment2t2 +>
+      List.iter (fun x -> push2 x toks_out) in
 
     let pr_barrier ln col = (* marks a position, used around C code *)
       push2 (Cocci2 ("",ln,col,col,None)) toks_out in
@@ -506,6 +533,10 @@ let is_space = function
   | T2(Parser_c.TCommentSpace _,_b,_i,_h) -> true (* only whitespace *)
   | _ -> false
 
+let is_comment = function
+  | T2(Parser_c.TComment _,_b,_i,_h) -> true (* only whitespace *)
+  | _ -> false
+
 let is_comment_or_space = function
   | T2(Parser_c.TCommentSpace _,_b,_i,_h) -> true (* only whitespace *)
   | T2(Parser_c.TComment _,_b,_i,_h) -> true (* only whitespace *)
@@ -556,6 +587,9 @@ let is_fake2 = function Fake2 _ -> true | _ -> false
 let is_whitespace x =
   is_space x || is_newline_or_comment x
 
+let is_noncomment_whitespace x =
+  is_space x || is_newline x
+
 let is_whitespace_or_fake x =
   is_space x || is_newline_or_comment x || is_fake2 x
 
@@ -668,6 +702,15 @@ let remove_minus_and_between_and_expanded_and_fake1 xs =
     (* non-empty intersection of witness trees *)
     not ((inter_set index1 index2) = []) in
 
+  let exists_before_end fn lst =
+    let rec loop = function
+	[] -> false
+      | x::xs ->
+	  if fn x
+	  then List.exists (function x -> not (fn x)) xs
+	  else loop xs in
+    loop lst in
+
   (* new idea: collects regions not containing non-space context code
   if two adjacent minus tokens satisfy common_adj then delete
   all spaces, comments etc between them
@@ -685,7 +728,7 @@ let remove_minus_and_between_and_expanded_and_fake1 xs =
       let (minus_list,rest) = span_not_context (t1::xs) in
       let (minus_list,rest) = drop_trailing_plus minus_list rest in
       let (pre_minus_list,_) = span not_context_newline minus_list in
-      let contains_plus = List.exists is_plus pre_minus_list in
+      let contains_plus = exists_before_end is_plus pre_minus_list in
       let x =
         match List.rev minus_list with
         | (T2(Parser_c.TCommentNewline c,_b,_i,_h))::rest
@@ -699,7 +742,7 @@ let remove_minus_and_between_and_expanded_and_fake1 xs =
       ((T2(_,Min adj1,_,_)) as t1)::xs ->
       let (minus_list,rest) = span_not_context (t1::xs) in
       let (pre_minus_list,_) = span not_context_newline minus_list in
-      let contains_plus = List.exists is_plus pre_minus_list in
+      let contains_plus = exists_before_end is_plus pre_minus_list in
       let x =
         match List.rev minus_list with
         | (T2(Parser_c.TCommentNewline c,_b,_i,_h))::rest
@@ -711,7 +754,7 @@ let remove_minus_and_between_and_expanded_and_fake1 xs =
     | ((Fake2(_,Min adj1) | T2(_,Min adj1,_,_)) as t1)::xs ->
       let (minus_list,rest) = span_not_context (t1::xs) in
       let (pre_minus_list,_) = span not_context_newline minus_list in
-      let contains_plus = List.exists is_plus pre_minus_list in
+      let contains_plus = exists_before_end is_plus pre_minus_list in
       adjust_within_minus contains_plus minus_list
       @ adjust_around_minus rest
     | x::xs ->
@@ -729,6 +772,7 @@ let remove_minus_and_between_and_expanded_and_fake1 xs =
 	      let (pre_minus_list,_) = span not_context_newline rest in
 	      List.exists is_plus pre_minus_list
 	    else cp in
+
         if common_adj adj1 adj2
         || ((not cp || not newcp) &&
 	    List.for_all is_whitespace_or_fake not_minus_list)
@@ -827,7 +871,8 @@ let remove_minus_and_between_and_expanded_and_fake1 xs =
     | ((T2(_,Ctx,_,_)) as x)::((T2(_,Min adj,_,_)::_) as xs)
       when str_of_token2 x = "{" ->
       let (between_minus,rest) = span minus_or_comment_nonl xs in
-      let (newlines,rest) = span is_whitespace rest in
+      let (newlines,rest) = span is_noncomment_whitespace rest in
+
       let (drop_newlines,last_newline) =
         let rec loop = function
           | [] -> ([],[])
@@ -1132,7 +1177,7 @@ let check_danger toks =
     match (tok,get_danger tok) with
       (T2(tok,Min _,a,b),Some Ast_c.Danger) -> T2(tok,Ctx,a,b)
     | _ -> tok in
-  let rec unminus_initial_danger toks =
+  let unminus_initial_danger toks =
     let (front,rest) = span not_danger toks in
     let (dangers,rest) = span nonspace_danger rest in
     front @ (List.map unminus dangers) @ rest in
@@ -1206,21 +1251,38 @@ let paren_then_brace toks =
   search_paren toks
 
 let is_ident_like s = s ==~ regexp_alpha
+let is_int_like s = s ==~ regexp_int
+
+let isop s =
+  List.mem s
+    ["=";"+";"-";"*";"/";"%";"^";"+=";"-=";"*=";"/=";"%=";"^=";":";"?";
+      "==";"!=";"<";">";"<=";">=";
+      "<?";">?";"<<";">>";"<?=";">?=";"<<=";">>=";
+      "&&";"||";"&";"|"]
+
+let space_after = function
+    [] -> true
+  | x::xs -> is_newline x || is_space x || is_added_whitespace x      
 
 let rec drop_space_at_endline = function
   | [] -> []
   | [x] -> [x]
   | (C2 (" ",_)) ::
-    ((((T2(Parser_c.TCommentSpace _,Ctx,_,_)) | Cocci2("\n",_,_,_,_) |
+    ((Cocci2(str,_,_,_,_) :: _) as rest)
+    when not (str = "") && String.get str 0 = '\n' ->
+    (* when unparse_cocci doesn't know whether space is needed *)
+    drop_space_at_endline rest
+  | (C2 (" ",_)) ::
+    ((((T2(Parser_c.TCommentSpace _,Ctx,_,_)) |
     (T2(Parser_c.TCommentNewline _,Ctx,_,_))) :: _) as rest) ->
     (* when unparse_cocci doesn't know whether space is needed *)
     drop_space_at_endline rest
   | ((T2(Parser_c.TCommentSpace _,Ctx,_i,_h)) as a)::rest ->
     let (outer_spaces,rest) = span is_space rest in
     let minus_or_comment_or_space_nocpp = function
+      | (T2(Parser_c.TCommentNewline _,_,_i,_)) -> false
       | T2(_,Min adj,_,_) -> true
       | (T2(Parser_c.TCommentSpace _,Ctx,_i,_)) -> true
-      | (T2(Parser_c.TCommentNewline _,Ctx,_i,_)) -> false
       | x -> false in
     let (minus,rest) = span minus_or_comment_or_space_nocpp rest in
     let fail _ = a :: outer_spaces @ minus @ (drop_space_at_endline rest) in
@@ -1235,15 +1297,17 @@ let rec drop_space_at_endline = function
   | a :: rest ->
     a :: drop_space_at_endline rest
 
-(* if a removed ( is between two tokens, then add a space *)
+(* if a removed ( is between two tokens, but not just after (,
+   then add a space *)
 let rec paren_to_space = function
   | [] -> []
   | [x] -> [x]
   | [x;y] -> [x;y]
-  | ((T2(_,Ctx,_,_)) as a)::
-    ((T2(t,Min _,_,_)) as b)::
+  | ((T2(t1,Ctx,_,_)) as a)::
+    ((T2(t2,Min _,_,_)) as b)::
     ((T2(_,Ctx,_,_)) as c)::rest
-    when not (is_whitespace a) && TH.str_of_tok t = "(" ->
+    when not (is_whitespace a) && not(TH.str_of_tok t1 = "(") &&
+      TH.str_of_tok t2 = "(" ->
     a :: b :: (C2 (" ",None)) :: (paren_to_space (c :: rest))
   | a :: rest -> a :: (paren_to_space rest)
 
@@ -1267,10 +1331,13 @@ let rec add_space xs =
     x::C2 (String.make (lny-lnx) '\n', None)::
     C2 (String.make (lcoly-1) ' ', None):: (* -1 is for the + *)
     add_space (y::xs)
-  | ((T2(_,Ctx,_,_)) as x)::((Cocci2 _) as y)::xs -> (* add space on boundary *)
+  | ((T2(_,Ctx,_,_)) as x)::(((Cocci2 _) | C2 _) as y)::xs
+  | (((Cocci2 _) | C2 _) as x)::((T2(_,Ctx,_,_)) as y)::xs ->
+    (* add space on boundary *)
     let sx = str_of_token2 x in
     let sy = str_of_token2 y in
-    if is_ident_like sx && (is_ident_like sy || List.mem sy ["="])
+    if (is_ident_like sx || List.mem sx [")";"]"]) &&
+      (is_ident_like sy || (isop sy && space_after xs) || is_int_like sy)
     then x::C2(" ",None)::(add_space (y::xs))
     else x::(add_space (y::xs))
   | ((T2(_,Ctx,_,_)) as x)::((T2(_,Ctx,_,_)) as y)::xs -> (* don't touch *)
@@ -1356,6 +1423,7 @@ let add_newlines toks tabbing_unit =
   let iscomma tok = str_of_token2 tok = "," in
   let isspace tok = str_of_token2 tok = " " in
   let isnewline tok =
+    not(is_comment tok) &&
     let s = str_of_token2 tok in
     try let _ = Str.search_forward (Str.regexp "\n") s 0 in true
     with Not_found -> false in
@@ -1381,7 +1449,7 @@ let add_newlines toks tabbing_unit =
     let count = count + String.length s in
     match (stack,space_cell) with
       ([(indent, tu)], Some (start,space_cell)) ->
-	if count > Flag_parsing_c.max_width
+	if count > !Flag_parsing_c.max_width
 	then
 	  let indent =
 	    match tu with
@@ -1478,6 +1546,39 @@ let add_newlines toks tabbing_unit =
         let (newcount,newstack,newspacecell,newseencocci) =
           end_box stack space_cell count seen_cocci s in
         a :: loop newstack newspacecell newcount newseencocci false xs
+    | t1 :: t2 :: rest
+      when nonempty stack && iscomma t1 && isnewline t2 ->
+	let seen_cocci = seen_cocci || iscocci t1 || iscocci t2 in
+	let (stack,newcount,t2,newseencocci) =
+	  get_indent stack t2 seen_cocci rest in
+	let t1 =
+	  match t1 with
+	    Cocci2(s,line,lcol,rcol,Some(Unparse_cocci.SpaceOrNewline sp)) ->
+	      (* no need for extra space, because we have a newline *)
+	      Cocci2(s,line,lcol,rcol,None)
+	  | t1 -> t1 in
+	(match stack with
+	  [_] ->
+	    let _count =
+	      if seen_cocci
+	      then update_previous_space stack space_cell count ","
+	      else count+1 in
+	    t1::t2::loop stack None newcount newseencocci false rest
+	| _ -> t1::t2::loop stack None newcount newseencocci false rest)
+    | t1 :: t2 :: rest
+      when iscomma t1 && isnewline t2 ->
+	(* nothing to do in this case *)
+	let t1 =
+	  match t1 with
+	    Cocci2(s,line,lcol,rcol,Some(Unparse_cocci.SpaceOrNewline sp)) ->
+	      (* no need for extra space, because we have a newline *)
+	      Cocci2(s,line,lcol,rcol,None)
+	  | t1 -> t1 in
+	(* 0 for second arg of string_length - because of nl, value doesn't
+	   matter *)
+	let (newcount,sawnl) = string_length (str_of_token2 t2) 0 in
+	let newseencocci = seen_cocci && (not sawnl) in
+	t1::t2::loop stack None newcount newseencocci false rest
     | ((Cocci2(s,line,lcol,rcol,Some(Unparse_cocci.SpaceOrNewline sp))) as a)::
       xs ->
 	let xs =
@@ -1506,19 +1607,6 @@ let add_newlines toks tabbing_unit =
 	| _ ->
 	    t1::t2::
 	    loop stack space_cell (count+space_sz+1) seen_cocci false rest)
-    | t1 :: t2 :: rest
-      when nonempty stack && iscomma t1 && isnewline t2 ->
-	let seen_cocci = seen_cocci || iscocci t1 || iscocci t2 in
-	let (stack,newcount,t2,newseencocci) =
-	  get_indent stack t2 seen_cocci rest in
-	(match stack with
-	  [_] ->
-	    let _count =
-	      if seen_cocci
-	      then update_previous_space stack space_cell count ","
-	      else count+1 in
-	    t1::t2::loop stack None newcount newseencocci false rest
-	| _ -> t1::t2::loop stack None newcount newseencocci false rest)
     | t1 :: rest
       when nonempty stack && iscomma t1 ->
 	let seen_cocci = seen_cocci || iscocci t1 in
@@ -1550,10 +1638,15 @@ let add_newlines toks tabbing_unit =
 	(String.get s1 0 = '\"') && (String.get s2 0 = '\"') ->
         let count =
 	  update_previous_space stack space_cell count s1 in
+	let count_after_space = count + 1 in
 	let sp = ref " " in
 	let a = C2(s1,Some(Unparse_cocci.SpaceOrNewline sp)) in
+	let space_cell =
+          match stack with
+            [_] -> Some(count_after_space,sp)
+          | _ -> space_cell in
 	a ::
-	loop stack space_cell ((simple_string_length s1 count)+1) true false xs
+	loop stack space_cell count_after_space true false xs
     | Fake2 _ :: _ | Indent_cocci2 :: _
     | Unindent_cocci2 _::_ | EatSpace2::_ ->
       failwith "unexpected fake, indent, unindent, or eatspace"
@@ -1856,7 +1949,8 @@ let token_effect tok dmin dplus inparens inassn inbrace accumulator xs =
       let accumulator = sub1top op accumulator in
       (Other 3,dmin,dplus,inparens,(0,0),sub1 op inbrace,
        drop_zeroes op accumulator xs)
-  | (Tok(";"|","),op) when getval inparens op = 0 && getval inassn op <= 1 ->
+  | (Tok(";"|"{}"|","),op) (* {} is generated when removing an if branch *)
+    when getval inparens op = 0 && getval inassn op <= 1 ->
       (Other 4,dmin,dplus,inparens,(0,0),inbrace,drop_zeroes op accumulator xs)
   | (Tok ";",op) ->
       (Other 5,dmin,dplus,inparens,sub1 op inassn,inbrace,accumulator)
@@ -1998,16 +2092,22 @@ let untimes_before cur n tabbing_unit ctr =
   (if n < 0 then failwith (Printf.sprintf "n is %d\n" n));
   let tabbing_unit =
     match tabbing_unit with None -> !default_indent | Some tu -> tu in
-  let len = String.length tabbing_unit in
-  let tabbing_unit = Str.regexp_string tabbing_unit in
-  let rec loop cur = function
-      0 -> cur
-    | n ->
-	if Str.string_match tabbing_unit cur 0
-	then loop (String.sub cur len (String.length cur - len)) (n-1)
-	else (* no idea what to do, just drop the first character... *)
-	  loop (String.sub cur 1 (String.length cur - 1)) (n-1) in
-  loop cur n
+  let cur = List.rev(Common.list_of_string cur) in
+  let tu = Common.list_of_string tabbing_unit in
+  let rec loop cur tu =
+    match (cur,tu) with
+      (x,[]) -> Common.string_of_chars (List.rev x)
+    | ([],_) -> ""
+    | (x::xs,y::ys) ->
+	if x = y
+	then loop xs ys
+	else if x = ' ' && y = '\t'
+	then loop (x::xs) (' '::' '::' '::' '::' '::' '::' '::' '::ys)
+	else if x = '\t' && y = ' '
+	then loop (' '::' '::' '::' '::' '::' '::' '::' '::xs) (y::ys)
+	else (* give up, no idea what to do *)
+	  Common.string_of_chars (List.rev cur) in
+  loop cur tu
 
 (* Probably doesn't do a good job of parens.  Code in parens may be aligned
 by tabbing unit or may have extra space specific to the position of the
@@ -2083,7 +2183,7 @@ let context_search_in_maps n depth samedelta inparens past_minmap minmap tu t =
 
 (* Add newlines where needed around unindents.  Lets adjust_indentation
 adjust them if needed. *)
-let rec newlines_for_unindents xs =
+let newlines_for_unindents xs =
   let is_ctxnl =
     function T2(Parser_c.TCommentNewline _,_b,_i,_h) -> true | _ -> false in
   let is_plusnl =
@@ -2305,14 +2405,20 @@ let print_all_tokens2 pr xs =
     );
   else
     let to_whitespace s =
-      let r = Bytes.copy s in
-      for i = 1 to String.length r do
-        let c = String.get r (i-1) in
-        match c with
-        | ' ' | '\t' | '\r' | '\n' -> ()
-        | _ -> Bytes.set r (i-1) ' '
-      done;
-      r in
+      let len = String.length s in
+      let rec translate n =
+	if n = len
+	then ()
+	else
+	  begin
+	    (match String.get s n with
+            | '\t' -> pr "\t"
+	    | '\r' -> pr "\r"
+	    | '\n' -> pr "\n"
+            | _ -> pr " ");
+	    translate (n+1)
+	  end in
+      translate 0 in
     let hiding_level = ref 0 in
     let handle_token t =
       let s = str_of_token2 t in
@@ -2331,8 +2437,8 @@ let print_all_tokens2 pr xs =
           );
           hide_current
         | _ -> !hiding_level > 0 in
-      if hide_current then to_whitespace s else s in
-    xs +> List.iter (fun x -> pr (handle_token x))
+      if hide_current then to_whitespace s else pr s in
+    xs +> List.iter handle_token
 
 
 
@@ -2411,7 +2517,6 @@ let pp_program2 xs outfile  =
               (* phase2: can now start to filter and adjust *)
 	      let toks = check_danger toks in
               let toks = paren_then_brace toks in
-              let toks = drop_space_at_endline toks in
 	      (* have to annotate droppable spaces early, so that can create
 		 the right minus and plus maps in adjust indentation.  For
 		 the same reason, cannot actually remove the minus tokens. *)
@@ -2423,6 +2528,7 @@ let pp_program2 xs outfile  =
               let toks = adjust_after_paren toks in(*also before remove minus*)
               let toks = paren_to_space toks in
               let toks = drop_end_comma toks in
+              let toks = drop_space_at_endline toks in
               let toks = remove_minus_and_between_and_expanded_and_fake2 toks in
               (* assert Origin + Cocci + C and no minus *)
               let toks = add_space toks in
