@@ -36,12 +36,17 @@ exception CantBeInPlus
 type pos = Before | After | InPlace
 type nlhint = StartBox | EndBox | SpaceOrNewline of string ref
 
+let error name e str =
+  failwith
+    (Printf.sprintf "%s: line %d %s\n" (snd (Ast.unwrap_mcode name))
+       (Ast.get_line e) str)
+
 let get_string_info = function
     Ast.Noindent s | Ast.Indent s | Ast.Space s -> s
 
 let unknown = -1
 
-let rec do_all
+let do_all
     (env, pr, pr_celem, pr_cspace, pr_space, pr_arity, pr_barrier,
      indent, unindent, eatspace)
     generating xxs before =
@@ -266,17 +271,20 @@ let rec ident i =
     | Ast.MetaId(name,_,_,_) ->
 	handle_metavar name (function
 			       | (Ast_c.MetaIdVal id) -> print_text id
-			       | _ -> raise (Impossible 142)
+			       | _ -> error name i "identifier value expected"
 			    )
     | Ast.MetaFunc(name,_,_,_) ->
 	handle_metavar name (function
 			       | (Ast_c.MetaFuncVal id) -> print_text id
-			       | _ -> raise (Impossible 143)
+			       | _ ->
+				   error name i "function name value expected"
 			    )
     | Ast.MetaLocalFunc(name,_,_,_) ->
 	handle_metavar name (function
 			       | (Ast_c.MetaLocalFuncVal id) -> print_text id
-			       | _ -> raise (Impossible 144)
+			       | _ ->
+				   error name i
+				     "local function name value expected"
 			    )
 
     | Ast.AsIdent(id,asid) -> ident id
@@ -374,7 +382,7 @@ let rec expression e =
 	    if generating then minprio else failwith "unbound MetaBinary"
 	| Some (Ast_c.MetaBinaryOpVal bop) -> left_prec_of_c bop
 	| Some _ -> failwith "bad MetaBinary value") in
-  let rec right_prec_of op =
+  let right_prec_of op =
     let right_prec_of_arith op =
       match Ast.unwrap_mcode op with
 	Ast.Plus | Ast.Minus -> mulit
@@ -445,7 +453,7 @@ let rec expression e =
     | Ast_c.Constructor (t, init) -> unary
     | Ast_c.ParenExpr (e) -> primary
     | Ast_c.New (_, t) -> unary
-    | Ast_c.Delete(t) -> unary
+    | Ast_c.Delete(_,t) -> unary
     | Ast_c.Defined _ -> primary
   in
 
@@ -505,7 +513,7 @@ let rec expression e =
   | Ast.MetaErr(name,_,_,_) ->
       failwith "metaErr not handled"
 
-  | Ast.MetaExpr (name,_,_,_typedontcare,_formdontcare,_) ->
+  | Ast.MetaExpr (name,_,_,_typedontcare,_formdontcare,_,_bitfield) ->
       handle_metavar name (function
         | Ast_c.MetaExprVal ((((e, _), _) as exp),_,_) ->
 	    if prec_of_c e < prec then
@@ -516,16 +524,16 @@ let rec expression e =
 	      end
 	    else
               pretty_print_c.Pretty_print_c.expression exp
-        | _ -> raise (Impossible 145)
+        | _ -> error name e "expression value expected"
       )
 
-  | Ast.MetaExprList (name,_,_,_) ->
+  | Ast.MetaExprList (name,_,_,_,_) ->
       handle_metavar name (function
         | Ast_c.MetaExprListVal args ->
             pretty_print_c.Pretty_print_c.arg_list args
 	| Ast_c.MetaParamListVal _ ->
 	    failwith "have meta param list matching meta exp list\n";
-        | _ -> raise (Impossible 146)
+        | _ -> error name e "expression list value expected"
       )
 
   | Ast.AsExpr(expr,asexpr) -> loop expr prec
@@ -580,12 +588,12 @@ and string_fragment e =
       mcode print_string pct;
       string_format fmt
   | Ast.Strdots dots -> mcode print_string dots
-  | Ast.MetaFormatList(pct,name,lenname,_,_) ->
+  | Ast.MetaFormatList(pct,name,lenname,_,_,_) ->
       (*mcode print_string pct;*)
       handle_metavar name (function
 	  Ast_c.MetaFragListVal(frags) ->
 	    frags +> (List.iter pretty_print_c.Pretty_print_c.fragment)
-	| _ -> raise (Impossible 158))
+	| _ -> error name e "format list value expected")
 
 and string_format e =
   match Ast.unwrap e with
@@ -594,7 +602,7 @@ and string_format e =
       handle_metavar name (function
 	  Ast_c.MetaFmtVal fmt ->
 	    pretty_print_c.Pretty_print_c.format fmt
-	| _ -> raise (Impossible 157))
+	| _ -> error name e "format value expected")
 
 and unaryOp = function
     Ast.GetRef -> print_string "&"
@@ -613,7 +621,7 @@ and assignOp op =
       handle_metavar name (function
 	  Ast_c.MetaAssignOpVal aop ->
 	    pretty_print_c.Pretty_print_c.assignOp aop
-	| _ -> raise (Impossible 159))
+	| _ -> error name op "assignment operator value expected")
 
 and fixOp = function
     Ast.Dec -> print_string "--"
@@ -627,7 +635,7 @@ and binaryOp op =
       handle_metavar name (function
 	  Ast_c.MetaBinaryOpVal bop ->
 	    pretty_print_c.Pretty_print_c.binaryOp bop
-	| _ -> raise (Impossible 160))
+	| _ -> error name op "binary operator value expected")
 
 and arithOp eq op =
   let print_string s = if eq then print_string (s^"=") else print_string s in
@@ -675,11 +683,12 @@ and fullType ft =
       |	_ -> print_option_space (mcode const_vol) cv; typeC ty)
 
   | Ast.AsType(ty, asty) -> fullType ty
-  | Ast.DisjType _ -> failwith "can't be in plus"
+  | Ast.DisjType _ | Ast.ConjType _ -> raise CantBeInPlus
   | Ast.OptType(_) -> raise CantBeInPlus
 
 and print_function_pointer (ty,lp1,star,rp1,lp2,params,rp2) fn =
-  fullType ty; mcode print_string lp1; mcode print_string star; fn();
+  fullType ty; pr_space();
+  mcode print_string lp1; mcode print_string star; fn();
   mcode print_string rp1; mcode print_string lp1;
   parameter_list params; mcode print_string rp2
 
@@ -722,16 +731,26 @@ and typeC ty =
   | Ast.StructUnionDef(ty,lb,decls,rb) ->
       fullType ty; ft_space ty;
       mcode print_string lb;
-      dots_before_and_after force_newline annotated_decl decls;
+      dots_before_and_after force_newline annotated_field decls;
       mcode print_string rb
+  | Ast.TypeOfExpr(typeof,lp,exp,rp) ->
+      mcode print_string typeof;
+      mcode print_string_box lp;  expression exp; close_box();
+      mcode print_string rp
+  | Ast.TypeOfType(typeof,lp,ty,rp) ->
+      mcode print_string typeof;
+      mcode print_string_box lp; fullType ty; close_box();
+      mcode print_string rp
   | Ast.TypeName(name)-> mcode print_string name
-  | Ast.MetaType(name,_,_) ->
+  | Ast.MetaType(name,_,_,_) ->
       handle_metavar name  (function
           Ast_c.MetaTypeVal exp ->
             pretty_print_c.Pretty_print_c.ty exp
-        | _ -> raise (Impossible 147))
+        | _ -> error name ty "type value expected")
 
-and baseType = function
+and baseType ty = print_string (Ast.string_of_baseType ty ^ " ")
+
+(* function
     Ast.VoidType -> print_string "void"
   | Ast.CharType -> print_string "char"
   | Ast.ShortType -> print_string "short"
@@ -747,7 +766,7 @@ and baseType = function
   | Ast.SizeType -> print_string "size_t "
   | Ast.SSizeType -> print_string "ssize_t "
   | Ast.PtrDiffType -> print_string "ptrdiff_t "
-
+*)
 and structUnion = function
     Ast.Struct -> print_string "struct"
   | Ast.Union -> print_string "union"
@@ -779,7 +798,7 @@ and print_named_type ty id =
       (match Ast.unwrap ty1 with
 	Ast.FunctionPointer(ty,lp1,star,rp1,lp2,params,rp2) ->
 	  print_function_pointer (ty,lp1,star,rp1,lp2,params,rp2)
-	    (function _ -> pr_space(); ident id)
+	    (function _ -> id())
       | Ast.Array(_,_,_,_) ->
 	  let rec loop ty k =
 	    match Ast.unwrap ty with
@@ -794,13 +813,19 @@ and print_named_type ty id =
 			print_option expression size;
 			mcode print_string rb)
 		| _ -> failwith "complex array types not supported")
-	    | _ -> typeC ty; ty_space ty; ident id; k () in
+	    | _ -> typeC ty; ty_space ty; id(); k () in
 	  loop ty1 (function _ -> ())
+      | Ast.MetaType(name,_,_,_) ->
+	  handle_metavar name  (function
+              Ast_c.MetaTypeVal ty ->
+		pretty_print_c.Pretty_print_c.type_with_ident ty
+		  (function _ -> id())
+            | _ -> error name ty "type value expected")
     (*| should have a case here for pointer to array or function type
         that would put ( * ) around the variable.  This makes one wonder
         why we really need a special case for function pointer *)
-      | _ -> fullType ty; ft_space ty; ident id)
-  | _ -> fullType ty; ft_space ty; ident id
+      | _ -> fullType ty; ft_space ty; id())
+  | _ -> fullType ty; ft_space ty; id()
 
 and ty_space ty =
   match Ast.unwrap ty with
@@ -813,7 +838,7 @@ and ft_space ty =
       let isptr =
 	match Ast.unwrap ty with
 	  Ast.Pointer(_,_) -> true
-	| Ast.MetaType(name,_,_) ->
+	| Ast.MetaType(name,_,_,_) ->
 	    let (res,name_string,line,lcol,rcol) = lookup_metavar name in
 	    (match res with
 	      None ->
@@ -831,38 +856,29 @@ and ft_space ty =
 
 and declaration d =
   match Ast.unwrap d with
-    Ast.MetaDecl(name,_,_) ->
+    Ast.MetaDecl(name,_,_,_) ->
       handle_metavar name
 	(function
-	    Ast_c.MetaDeclVal d ->
+	    Ast_c.MetaDeclVal (_,d) ->
               pretty_print_c.Pretty_print_c.decl d
-          | _ -> raise (Impossible 148))
-  | Ast.MetaField(name,_,_) ->
-      handle_metavar name
-	(function
-	    Ast_c.MetaFieldVal f ->
-              pretty_print_c.Pretty_print_c.field f
-          | _ -> raise (Impossible 149))
-
-  | Ast.MetaFieldList(name,_,_,_) ->
-      handle_metavar name
-	(function
-	    Ast_c.MetaFieldListVal f ->
-	      print_between force_newline pretty_print_c.Pretty_print_c.field f
-          | _ -> raise (Impossible 150))
+          | _ -> error name d "declaration value expected")
 
   | Ast.AsDecl(decl,asdecl) -> declaration decl
 
-  | Ast.Init(stg,ty,id,eq,ini,sem) ->
+  | Ast.Init(stg,ty,id,attr,eq,ini,sem) ->
       print_option (mcode storage) stg;
       print_option (function _ -> pr_space()) stg;
-      print_named_type ty id;
+      print_named_type ty (fun _ -> ident id);
+      (if not (attr = []) then pr_space());
+      print_between pr_space (mcode print_string) attr;
       pr_space(); mcode print_string eq;
       pr_space(); initialiser true ini; mcode print_string sem
-  | Ast.UnInit(stg,ty,id,sem) ->
+  | Ast.UnInit(stg,ty,id,attr,sem) ->
       print_option (mcode storage) stg;
       print_option (function _ -> pr_space()) stg;
-      print_named_type ty id;
+      print_named_type ty (fun _ -> ident id);
+      (if not (attr = []) then pr_space());
+      print_between pr_space (mcode print_string) attr;
       mcode print_string sem
   | Ast.FunProto (fninfo,name,lp1,params,va,rp1,sem) ->
       List.iter print_fninfo fninfo;
@@ -892,32 +908,68 @@ and declaration d =
       pr_space(); initialiser true ini; mcode print_string sem
   | Ast.TyDecl(ty,sem) -> fullType ty; mcode print_string sem
   | Ast.Typedef(stg,ty,id,sem) ->
-      mcode print_string stg;
-      fullType ty; pr_space(); typeC id;
+      mcode print_string stg; pr_space();
+      print_named_type ty (fun _ -> typeC id);
       mcode print_string sem
-  | Ast.DisjDecl(_) -> raise CantBeInPlus
+  | Ast.DisjDecl(_) | Ast.ConjDecl(_) -> raise CantBeInPlus
   | Ast.OptDecl(decl) -> raise CantBeInPlus
 
 and annotated_decl d =
   match Ast.unwrap d with
     Ast.DElem(_,_,decl) -> declaration decl
-  | Ast.Ddots(_,_) -> raise CantBeInPlus
+
+(* --------------------------------------------------------------------- *)
+(* Field declaration *)
+
+and field d =
+  match Ast.unwrap d with
+    Ast.MetaField(name,_,_,_) ->
+      handle_metavar name
+	(function
+	    Ast_c.MetaFieldVal f ->
+	      pretty_print_c.Pretty_print_c.field f
+	  | _ -> error name d "field value expected")
+
+  | Ast.MetaFieldList(name,_,_,_,_) ->
+      handle_metavar name
+	(function
+	    Ast_c.MetaFieldListVal f ->
+	      print_between force_newline pretty_print_c.Pretty_print_c.field f
+	  | _ -> error name d "field list value expected")
+  | Ast.Field(ty,id,bf,sem) ->
+      begin
+	match id with
+	  None -> fullType ty
+	| Some id -> print_named_type ty (fun _ -> ident id)
+      end;
+      let bitfield (c, e) =
+	mcode print_string c;
+	expression e in
+      Common.do_option bitfield bf;
+      mcode print_string sem
+  | Ast.DisjField(_) | Ast.ConjField(_) -> raise CantBeInPlus
+  | Ast.OptField(decl) -> raise CantBeInPlus
+
+and annotated_field d =
+  match Ast.unwrap d with
+    Ast.FElem(_,_,decl) -> field decl
+  | Ast.Fdots(_,_) -> raise CantBeInPlus
 
 (* --------------------------------------------------------------------- *)
 (* Initialiser *)
 
 and initialiser nlcomma i =
   match Ast.unwrap i with
-    Ast.MetaInit(name,_,_) ->
+    Ast.MetaInit(name,_,_,_) ->
       handle_metavar name  (function
           Ast_c.MetaInitVal ini ->
             pretty_print_c.Pretty_print_c.init ini
-        | _ -> raise (Impossible 151))
-  | Ast.MetaInitList(name,_,_,_) ->
+        | _ -> error name i "initialiser value expected")
+  | Ast.MetaInitList(name,_,_,_,_) ->
       handle_metavar name  (function
           Ast_c.MetaInitListVal ini ->
 	    pretty_print_c.Pretty_print_c.init_list ini
-        | _ -> raise (Impossible 152))
+        | _ -> error name i "initialiser list value expected")
   | Ast.AsInit(init,asinit) -> initialiser nlcomma init
   | Ast.InitExpr(exp) -> expression exp
   | Ast.ArInitList(lb,initlist,rb) ->
@@ -976,21 +1028,21 @@ and designator = function
 and parameterTypeDef p =
   match Ast.unwrap p with
     Ast.VoidParam(ty) -> fullType ty
-  | Ast.Param(ty,Some id) -> print_named_type ty id
+  | Ast.Param(ty,Some id) -> print_named_type ty (fun _ -> ident id)
   | Ast.Param(ty,None) -> fullType ty
 
-  | Ast.MetaParam(name,_,_) ->
+  | Ast.MetaParam(name,_,_,_) ->
       handle_metavar name
 	(function
 	    Ast_c.MetaParamVal p ->
               pretty_print_c.Pretty_print_c.param p
-          | _ -> raise (Impossible 153))
-  | Ast.MetaParamList(name,_,_,_) ->
+          | _ -> error name p "parameter value expected")
+  | Ast.MetaParamList(name,_,_,_,_) ->
       handle_metavar name
 	(function
 	    Ast_c.MetaParamListVal p ->
               pretty_print_c.Pretty_print_c.paramlist p
-          | _ -> raise (Impossible 154))
+          | _ -> error name p "parameter list value expected")
 
   | Ast.AsParam(p,e) -> raise CantBeInPlus
 
@@ -1013,6 +1065,7 @@ let rec inc_file = function
       print_string ("\""^(String.concat "/" (List.map inc_elem elems))^"\"")
   | Ast.NonLocal(elems) ->
       print_string ("<"^(String.concat "/" (List.map inc_elem elems))^">")
+  | Ast.AnyInc -> failwith "should not be in generated code"
 
 and inc_elem = function
     Ast.IncPath s -> s
@@ -1105,6 +1158,8 @@ and rule_elem arity re =
   | Ast.TopInit(init) -> initialiser false init
   | Ast.Include(inc,s) ->
       mcode print_string inc; print_text " "; mcode inc_file s
+  | Ast.MetaInclude(inc,s) ->
+      mcode print_string inc; print_text " "; expression s
   | Ast.Undef(def,id) ->
       mcode print_string def; pr_space(); ident id
   | Ast.DefineHeader(def,id,params) ->
@@ -1127,20 +1182,20 @@ and rule_elem arity re =
 	 print_text "\n)")
       else raise CantBeInPlus
 
-  | Ast.MetaRuleElem(name,_,_) ->
+  | Ast.MetaRuleElem(name,_,_,_) ->
       raise (Impossible 155)
 
-  | Ast.MetaStmt(name,_,_,_) ->
+  | Ast.MetaStmt(name,_,_,_,_) ->
       handle_metavar name (function
-        | Ast_c.MetaStmtVal(stm,_) ->
+        | Ast_c.MetaStmtVal(_,stm,_) ->
             pretty_print_c.Pretty_print_c.statement stm
-        | _ -> raise (Impossible 156))
+        | _ -> error name re "statement value expected")
 
-  | Ast.MetaStmtList(name,_,_,_) ->
+  | Ast.MetaStmtList(name,_,_,_,_) ->
       handle_metavar name (function
         | Ast_c.MetaStmtListVal(statxs,_) ->
             pretty_print_c.Pretty_print_c.statement_seq_list statxs
-        | _ -> raise (Impossible 161))
+        | _ -> error name re "statement list value expected")
   | Ast.AsRe(re,asre) -> rule_elem arity re
 
 and pragmainfo pi =
@@ -1167,6 +1222,12 @@ and print_define_parameters params =
 and print_define_param param =
   match Ast.unwrap param with
     Ast.DParam(id) -> ident id
+  | Ast.MetaDParamList(name,_,_,_,_) ->
+      handle_metavar name
+	(function
+	    Ast_c.MetaDParamListVal p ->
+              pretty_print_c.Pretty_print_c.dparamlist p
+          | _ -> error name param "#define parameter value expected")
   | Ast.DPComma(comma) -> mcode print_string comma
   | Ast.DPdots(dots) -> mcode print_string dots
   | Ast.OptDParam(dp) -> print_text "?"; print_define_param dp
@@ -1183,14 +1244,14 @@ let indent_if_needed s f =
       Ast.Seq(lbrace,body,rbrace) -> true
     | Ast.Atomic s ->
 	(match Ast.unwrap s with
-	| Ast.MetaStmt(name,_,_,_) ->
+	| Ast.MetaStmt(name,_,_,_,_) ->
 	    let (res,name_string,line,lcol,rcol) = lookup_metavar name in
 	    (match res with
 	      None ->
 		failwith
 		  (Printf.sprintf "variable %s not known on SP line %d\n"
 		     name_string line)
-	    | Some (Ast_c.MetaStmtVal(stm,_)) ->
+	    | Some (Ast_c.MetaStmtVal(_,stm,_)) ->
 		(match Ast_c.unwrap stm with
 		  Ast_c.Compound _ -> true
 		| _ -> false)
@@ -1342,7 +1403,7 @@ in
 let if_open_brace  = function "{" | "else" -> true | _ -> false in
 
 (* boolean result indicates whether an indent is needed *)
-let rec pp_any = function
+let pp_any = function
   (* assert: normally there is only CONTEXT NOTHING tokens in any *)
     Ast.FullTypeTag(x) -> fullType x; false
   | Ast.BaseTypeTag(x) -> baseType x unknown unknown; false
@@ -1365,6 +1426,7 @@ let rec pp_any = function
 
   | Ast.InitTag(x) -> initialiser false x; false
   | Ast.DeclarationTag(x) -> declaration x; false
+  | Ast.FieldTag(x) -> field x; false
 
   | Ast.StorageTag(x) -> storage x unknown unknown; false
   | Ast.IncFileTag(x) -> inc_file x unknown unknown; false
@@ -1417,6 +1479,8 @@ let rec pp_any = function
   | Ast.ParamDotsTag(x) -> parameter_list x; false
   | Ast.StmtDotsTag(x) -> dots force_newline (statement "") x; false
   | Ast.AnnDeclDotsTag(x) -> dots force_newline annotated_decl x; false
+  | Ast.AnnFieldDotsTag(x) -> dots force_newline annotated_field x; false
+  | Ast.DefParDotsTag(x) -> dots (fun _ -> ()) print_define_param x; false
 
   | Ast.TypeCTag(x) -> typeC x; false
   | Ast.ParamTag(x) -> parameterTypeDef x; false
@@ -1519,13 +1583,16 @@ in
 	      match List.hd(List.rev x) with
 		Ast.Token(t,_) when List.mem t [";";"{";"}"] -> true
 	      |	Ast.Token(t,_) -> false
+	      | Ast.ParamDotsTag _ | Ast.ExprDotsTag _ ->
+		  (* should never be followed by a newline *)
+		  false
 	      |	_ -> true in
 	    loop newline_needed indent_needed xs in
       loop false false (x::xs);
       (* print a newline at the end, if needed *)
       newline_after()
 
-let rec pp_list_list_any (envs, pr, pr_celem, pr_cspace, pr_space, pr_arity,
+let pp_list_list_any (envs, pr, pr_celem, pr_cspace, pr_space, pr_arity,
 			  pr_barrier, indent, unindent, eatspace)
     generating xxs before =
   List.iter
