@@ -113,7 +113,7 @@ let get_function_name rule env =
 	(match Ast.unwrap args with
 	  [e] ->
 	    (match Ast.unwrap e with
-	      Ast.MetaExprList(nm,_,_,_) ->
+	      Ast.MetaExprList(nm,_,_,_,_) ->
 		(match (Ast.unwrap_mcode nm,Ast.get_mcodekind nm) with
 		  ((_,"ARGS"), Ast.PLUS _) ->
 		    (match Ast.unwrap fn with
@@ -133,9 +133,10 @@ let get_function_name rule env =
     (V.combiner bind option_default
       mcode mcode mcode mcode mcode mcode mcode mcode mcode
       mcode mcode mcode mcode mcode
-      donothing donothing donothing donothing donothing
+      donothing donothing donothing donothing donothing donothing
       donothing expression donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing
+      donothing donothing donothing donothing donothing donothing donothing
+      donothing
       donothing donothing donothing donothing donothing).V.combiner_top_level
       rule in
   match names with
@@ -238,18 +239,20 @@ let print_extra_typedefs pr env =
       | Ast_c.MetaExprListVal(args) -> Visitor_c.vk_argument_list bigf args
       | Ast_c.MetaParamVal(param) -> Visitor_c.vk_param bigf param
       | Ast_c.MetaParamListVal(params) -> Visitor_c.vk_param_list bigf params
+      | Ast_c.MetaDParamListVal(params) ->
+	  Visitor_c.vk_define_params bigf params
 
       | Ast_c.MetaTypeVal(ty) -> Visitor_c.vk_type bigf ty
       | Ast_c.MetaInitVal(ty) -> Visitor_c.vk_ini bigf ty
       | Ast_c.MetaInitListVal(ty) -> Visitor_c.vk_ini_list bigf ty
-      | Ast_c.MetaDeclVal(decl) -> Visitor_c.vk_decl bigf decl
+      | Ast_c.MetaDeclVal(decl,_) -> Visitor_c.vk_decl bigf decl
       | Ast_c.MetaFieldVal(field) -> Visitor_c.vk_struct_field bigf field
       | Ast_c.MetaFieldListVal(fields) ->
 	  Visitor_c.vk_struct_fields bigf fields
       | Ast_c.MetaFmtVal(fmt) -> Visitor_c.vk_string_format bigf fmt
       | Ast_c.MetaFragListVal(frags) ->
 	  Visitor_c.vk_string_fragments bigf frags
-      | Ast_c.MetaStmtVal(stm,_) -> Visitor_c.vk_statement bigf stm
+      | Ast_c.MetaStmtVal(stm,_,_) -> Visitor_c.vk_statement bigf stm
       | Ast_c.MetaStmtListVal(stms,_) ->
 	  Visitor_c.vk_statement_sequencable_list bigf stms
       | Ast_c.MetaPosVal _ | Ast_c.MetaPosValList _
@@ -291,6 +294,8 @@ let rename argids env =
 	   Ast_c.MetaParamVal(Visitor_c.vk_param_s bigf param)
        | Ast_c.MetaParamListVal(params) ->
 	   Ast_c.MetaParamListVal(Visitor_c.vk_params_s bigf params)
+       | Ast_c.MetaDParamListVal(params) ->
+	   Ast_c.MetaDParamListVal(Visitor_c.vk_define_params_s bigf params)
 
        | Ast_c.MetaTypeVal(ty) ->
 	   Ast_c.MetaTypeVal(Visitor_c.vk_type_s bigf ty)
@@ -298,8 +303,9 @@ let rename argids env =
 	   Ast_c.MetaInitVal(Visitor_c.vk_ini_s bigf ini)
        | Ast_c.MetaInitListVal(ini) ->
 	   Ast_c.MetaInitListVal(Visitor_c.vk_inis_s bigf ini)
-       | Ast_c.MetaDeclVal(stm) ->
-	   Ast_c.MetaDeclVal(Visitor_c.vk_decl_s bigf stm)
+       | Ast_c.MetaDeclVal(stm,original) ->
+	   Ast_c.MetaDeclVal(Visitor_c.vk_decl_s bigf stm,
+			     Visitor_c.vk_decl_s bigf original)
        | Ast_c.MetaFieldVal(stm) ->
 	   Ast_c.MetaFieldVal(Visitor_c.vk_struct_field_s bigf stm)
        | Ast_c.MetaFieldListVal(stm) ->
@@ -308,8 +314,9 @@ let rename argids env =
 	   Ast_c.MetaFmtVal(Visitor_c.vk_string_format_s bigf fmt)
        | Ast_c.MetaFragListVal(frags) ->
 	   Ast_c.MetaFragListVal(Visitor_c.vk_string_fragments_s bigf frags)
-       | Ast_c.MetaStmtVal(stm,ty) ->
-	   Ast_c.MetaStmtVal(Visitor_c.vk_statement_s bigf stm,ty)
+       | Ast_c.MetaStmtVal(stm,original,ty) ->
+	   Ast_c.MetaStmtVal(Visitor_c.vk_statement_s bigf stm,
+			     Visitor_c.vk_statement_s bigf original,ty)
        | Ast_c.MetaStmtListVal(stm,ty) ->
 	   Ast_c.MetaStmtListVal
 	     (Visitor_c.vk_statement_sequencable_list_s bigf stm,ty)
@@ -318,8 +325,9 @@ let rename argids env =
        | Ast_c.MetaNoVal -> failwith "referencing a metavar with no value"))
     env
 
-let print_one_type pr env = function
-    (Type_cocci.MetaType(name,keep,inherited)) as ty ->
+let print_one_type pr env ty =
+  match Common.map_option Ast.unwrap (Ast.typeC_of_fullType_opt ty) with
+    Some (Ast_cocci.MetaType(name,cstr,keep,inherited)) ->
       (try
 	match List.assoc name env with
 	  Ast_c.MetaTypeVal ty ->
@@ -328,8 +336,8 @@ let print_one_type pr env = function
 	      (function _ -> pr " ")
 	      ty
         | _ -> failwith "impossible"
-      with Not_found -> pr (Type_cocci.type2c ty))
-  | ty -> pr (Type_cocci.type2c ty)
+      with Not_found -> pr (Ast_cocci.string_of_fullType ty))
+  | _ -> pr (Ast_cocci.string_of_fullType ty)
 
 let print_types pr env = function
     None -> ()
@@ -344,7 +352,7 @@ let pp_len pr len =
   let pp_name (_,n) = pr n in
   match len with
     Ast.AnyLen -> ()
-  | Ast.MetaLen len -> pr "["; pp_name len; pr "]"
+  | Ast.MetaLen (len,_) -> pr "["; pp_name len; pr "]"
   | Ast.CstLen len -> pr "["; pr (string_of_int len); pr "]"
 
 let pp_meta_decl pr env decl =
@@ -375,6 +383,8 @@ let pp_meta_decl pr env decl =
       no_arity ar; pr "parameter "; pp_name name; pr ";\n"
   | Ast.MetaParamListDecl(ar, name, len) ->
       no_arity ar; pr "parameter list "; pp_len pr len; pp_name name; pr ";\n"
+  | Ast.MetaDParamListDecl(ar, name, len) ->
+      no_arity ar; pr "identifier list "; pp_len pr len; pp_name name; pr ";\n"
   | Ast.MetaBinaryOperatorDecl(ar, name) ->
       no_arity ar; pr "binary operator "; pp_name name; pr ";\n"
   | Ast.MetaAssignmentOperatorDecl(ar, name) ->
@@ -384,9 +394,9 @@ let pp_meta_decl pr env decl =
       pp_name name; pr ";\n"
   | Ast.MetaErrDecl(ar, name) ->
       no_arity ar; pr "error "; pp_name name; pr ";\n"
-  | Ast.MetaExpDecl(ar, name, None) ->
+  | Ast.MetaExpDecl(ar, name, None, _bitfield) ->
       no_arity ar; pr "expression "; pp_name name; pr ";\n"
-  | Ast.MetaExpDecl(ar, name, types) ->
+  | Ast.MetaExpDecl(ar, name, types, _bitfield) ->
       no_arity ar; print_types pr env types; pp_name name; pr ";\n"
   | Ast.MetaIdExpDecl(ar, name, types) ->
       no_arity ar; pr "idexpression ";
@@ -523,8 +533,9 @@ let pp_rule local_metas ast env srcfile =
 	  started_files := (outfile,res)::!started_files;
 	  res in
     print_check_rule pr function_name function_name_count header_req;
+    let env' = List.map (fun (a, b) -> (Ast_cocci.make_mcode a, b)) env in
     let args =
-      print_metavariables pr local_metas paramst env header_req
+      print_metavariables pr local_metas paramst env' header_req
 	function_name_count in
     let (argids,args) = List.split args in
     let env = rename argids env in
