@@ -340,7 +340,7 @@ let rec vk_expr = fun bigf expr ->
     | New  (Some ts, t)   ->
         vk_argument_list bigf ts;
 	vk_argument bigf t
-    | Delete e -> vk_expr bigf e
+    | Delete (_,e) -> vk_expr bigf e
     | Defined name -> vk_name bigf name
 
 
@@ -505,6 +505,7 @@ and vk_type = fun bigf t ->
     (* don't go in _typ *)
     | TypeName (name,_typ) ->
         vk_name bigf name
+    | FieldType (t, _, _) -> typef t
 
     | ParenType t -> typef t
     | TypeOfExpr e -> vk_expr bigf e
@@ -561,7 +562,8 @@ and vk_onedecl = fun bigf onedecl ->
       v_type = t;
       v_type_bis = tbis;
       v_storage = _sto;
-      v_attr = attrs})  ->
+      v_attr = attrs;
+      v_endattr = endattrs})  ->
 
     vk_type bigf t;
     (* don't go in tbis *)
@@ -572,7 +574,8 @@ and vk_onedecl = fun bigf onedecl ->
 	Ast_c.NoInit -> ()
       |	Ast_c.ValInit(iini,init) -> iif [iini]; vk_ini bigf init
       |	Ast_c.ConstrInit((init,ii)) -> iif ii; vk_argument_list bigf init)
-    )
+    );
+    endattrs +> List.iter (vk_attribute bigf)
   in f (k, bigf) onedecl
 
 and vk_ini = fun bigf ini ->
@@ -1198,7 +1201,7 @@ let rec vk_expr_s = fun bigf expr ->
       | New (Some ts, t) ->
 	  New (Some (ts +> List.map (fun (e,ii) ->
 	    vk_argument_s bigf e, iif ii)), vk_argument_s bigf t)
-      | Delete e      -> Delete (vk_expr_s bigf e)
+      | Delete (box,e) -> Delete (box,vk_expr_s bigf e)
       | Defined name  -> Defined (vk_name_s bigf name)
 
     in
@@ -1420,6 +1423,7 @@ and vk_type_s = fun bigf t ->
       | StructUnionName (s, structunion) -> StructUnionName (s, structunion)
       | EnumName  s -> EnumName  s
       | TypeName (name, typ) -> TypeName (vk_name_s bigf name, typ)
+      | FieldType (t, a, b) -> FieldType (typef t, a, b)
 
       | ParenType t -> ParenType (typef t)
       | TypeOfExpr e -> TypeOfExpr (vk_expr_s bigf e)
@@ -1471,7 +1475,8 @@ and vk_decl_s = fun bigf d ->
             v_type_bis = tbis;
             v_storage = sto;
             v_local= local;
-            v_attr = attrs}, iicomma) =
+            v_attr = attrs;
+            v_endattr = endattrs}, iicomma) =
     {v_namei =
       (var +> map_option (fun (name, iniopt) ->
         vk_name_s bigf name,
@@ -1490,6 +1495,7 @@ and vk_decl_s = fun bigf d ->
      v_storage = sto;
      v_local = local;
      v_attr = attrs +> List.map (vk_attribute_s bigf);
+     v_endattr = endattrs +> List.map (vk_attribute_s bigf);
     },
     iif iicomma
 
@@ -1794,6 +1800,10 @@ and vk_ii_s = fun bigf ii ->
   List.map (vk_info_s bigf) ii
 
 (* ------------------------------------------------------------------------ *)
+
+(* Note that pointers to the complete statement are not processed.  This is
+not needed by Coccinelle, because there is no "node" metavariable type in
+SmPL, so nodes are never compared. *)
 and vk_node_s = fun bigf node ->
   let iif ii = vk_ii_s bigf ii in
   let infof info = vk_info_s bigf info  in
