@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -41,22 +41,16 @@ let update_after pragmas (info,x) =
 
 let rec right_decl d =
   match Ast0.unwrap d with
-    Ast0.MetaDecl(name,pure) ->
+    Ast0.MetaDecl(name,cstr,pure) ->
       call_right right_mcode name d
-	(function name -> Ast0.MetaDecl(name,pure))
-  | Ast0.MetaField(name,pure) ->
-      call_right right_mcode name d
-	(function name -> Ast0.MetaField(name,pure))
-  | Ast0.MetaFieldList(name,lenname,pure) ->
-      call_right right_mcode name d
-	(function name -> Ast0.MetaFieldList(name,lenname,pure))
+	(function name -> Ast0.MetaDecl(name,cstr,pure))
   | Ast0.AsDecl(decl,asdecl) -> failwith "not possible"
-  | Ast0.Init(stg,ty,id,eq,ini,sem) ->
+  | Ast0.Init(stg,ty,id,attr,eq,ini,sem) ->
       call_right right_mcode sem d
-	(function sem -> Ast0.Init(stg,ty,id,eq,ini,sem))
-  | Ast0.UnInit(stg,ty,id,sem) ->
+	(function sem -> Ast0.Init(stg,ty,id,attr,eq,ini,sem))
+  | Ast0.UnInit(stg,ty,id,attr,sem) ->
       call_right right_mcode sem d
-	(function sem -> Ast0.UnInit(stg,ty,id,sem))
+	(function sem -> Ast0.UnInit(stg,ty,id,attr,sem))
   | Ast0.FunProto(fninfo,id,lp1,params,va,rp1,sem) ->
       call_right right_mcode sem d
 	(function ty -> Ast0.FunProto(fninfo,id,lp1,params,va,rp1,sem))
@@ -72,8 +66,8 @@ let rec right_decl d =
   | Ast0.Typedef(stg,ty,id,sem) ->
       call_right right_mcode sem d
 	(function sem -> Ast0.Typedef(stg,ty,id,sem))
-  | Ast0.DisjDecl(starter,decls,mids,ender) -> None
-  | Ast0.Ddots(dots,whencode) -> None
+  | Ast0.DisjDecl(starter,decls,mids,ender)
+  | Ast0.ConjDecl(starter,decls,mids,ender) -> None
   | Ast0.OptDecl(decl) ->
       call_right right_decl decl d (function decl -> Ast0.OptDecl(decl))
 
@@ -121,12 +115,12 @@ let rec right_statement s =
   | Ast0.Exec(exec,lang,exp,sem) ->
       call_right right_mcode sem s
 	(function sem -> Ast0.Exec(exec,lang,exp,sem))
-  | Ast0.MetaStmt(name,pure) ->
+  | Ast0.MetaStmt(name,cstr,pure) ->
       call_right right_mcode name s
-	(function name -> Ast0.MetaStmt(name,pure))
-  | Ast0.MetaStmtList(name,lenname,pure) ->
+	(function name -> Ast0.MetaStmt(name,cstr,pure))
+  | Ast0.MetaStmtList(name,lenname,cstr,pure) ->
       call_right right_mcode name s
-	(function name -> Ast0.MetaStmtList(name,lenname,pure))
+	(function name -> Ast0.MetaStmtList(name,lenname,cstr,pure))
   | Ast0.AsStmt(stm,asstm) -> failwith "not possible"
   | Ast0.Disj(starter,statement_dots_list,mids,ender)
   | Ast0.Conj(starter,statement_dots_list,mids,ender) -> None
@@ -141,6 +135,8 @@ let rec right_statement s =
   | Ast0.Include(inc,name) ->
       call_right right_mcode name s
 	(function name -> Ast0.Include(inc,name))
+  | Ast0.MetaInclude(inc,name) ->
+      None (* likely not correct, wait for pb in practice *)
   | Ast0.Undef(def,id) ->
       (* nothing available for ident, and not sure code can appear
 	 here anyway *)
@@ -191,12 +187,17 @@ let rec left_ty t =
   | Ast0.StructUnionDef(ty,lb,decls,rb) ->
       call_right left_ty ty t
 	(function ty -> Ast0.StructUnionDef(ty,lb,decls,rb))
+  | Ast0.TypeOfExpr(tf,lb,e,rb) ->
+      call_right left_mcode tf t (function tf -> Ast0.TypeOfExpr(tf,lb,e,rb))
+  | Ast0.TypeOfType(tf,lb,ty,rb) ->
+      call_right left_mcode tf t (function tf -> Ast0.TypeOfType(tf,lb,ty,rb))
   | Ast0.TypeName(name) ->
       call_right left_mcode name t (function name -> Ast0.TypeName(name))
-  | Ast0.MetaType(name,x) ->
-      call_right left_mcode name t (function name -> Ast0.MetaType(name,x))
+  | Ast0.MetaType(name,cstr,x) ->
+      call_right left_mcode name t (function name -> Ast0.MetaType(name,cstr,x))
   | Ast0.AsType(ty,asty) -> failwith "not possible"
-  | Ast0.DisjType(starter,types,mids,ender) -> None
+  | Ast0.DisjType(starter,types,mids,ender)
+  | Ast0.ConjType(starter,types,mids,ender) -> None
   | Ast0.OptType(ty) ->
       call_right left_ty ty t (function ty -> Ast0.OptType(ty))
 
@@ -237,28 +238,22 @@ let left_fundecl name fninfo =
 
 let rec left_decl decl =
   match Ast0.unwrap decl with
-    Ast0.MetaDecl(name,pure) ->
+    Ast0.MetaDecl(name,cstr,pure) ->
       call_right right_mcode name decl
-	(function name -> Ast0.MetaDecl(name,pure))
-  | Ast0.MetaField(name,pure) ->
-      call_right right_mcode name decl
-	(function name -> Ast0.MetaField(name,pure))
-  | Ast0.MetaFieldList(name,lenname,pure) ->
-      call_right right_mcode name decl
-	(function name -> Ast0.MetaFieldList(name,lenname,pure))
+	(function name -> Ast0.MetaDecl(name,cstr,pure))
   | Ast0.AsDecl(decl,asdecl) -> failwith "not possible"
-  | Ast0.Init(Some stg,ty,id,eq,ini,sem) ->
+  | Ast0.Init(Some stg,ty,id,attr,eq,ini,sem) ->
       call_right left_mcode stg decl
-	(function stg -> Ast0.Init(Some stg,ty,id,eq,ini,sem))
-  | Ast0.Init(None,ty,id,eq,ini,sem) ->
+	(function stg -> Ast0.Init(Some stg,ty,id,attr,eq,ini,sem))
+  | Ast0.Init(None,ty,id,attr,eq,ini,sem) ->
       call_right left_ty ty decl
-	(function ty -> Ast0.Init(None,ty,id,eq,ini,sem))
-  | Ast0.UnInit(Some stg,ty,id,sem) ->
+	(function ty -> Ast0.Init(None,ty,id,attr,eq,ini,sem))
+  | Ast0.UnInit(Some stg,ty,id,attr,sem) ->
       call_right left_mcode stg decl
-	(function stg -> Ast0.UnInit(Some stg,ty,id,sem))
-  | Ast0.UnInit(None,ty,id,sem) ->
+	(function stg -> Ast0.UnInit(Some stg,ty,id,attr,sem))
+  | Ast0.UnInit(None,ty,id,attr,sem) ->
       call_right left_ty ty decl
-	(function ty -> Ast0.UnInit(None,ty,id,sem))
+	(function ty -> Ast0.UnInit(None,ty,id,attr,sem))
   | Ast0.FunProto(fi,name,lp1,params,va,rp1,sem) ->
       (match fi with
 	[] ->
@@ -298,8 +293,8 @@ let rec left_decl decl =
   | Ast0.Typedef(stg,ty,id,sem) ->
       call_right left_mcode stg decl
 	(function stg -> Ast0.Typedef(stg,ty,id,sem))
-  | Ast0.DisjDecl(starter,decls,mids,ender) -> None
-  | Ast0.Ddots(dots,whencode) -> None
+  | Ast0.DisjDecl(starter,decls,mids,ender)
+  | Ast0.ConjDecl(starter,decls,mids,ender) -> None
   | Ast0.OptDecl(d) ->
       call_right left_decl d decl (function decl -> Ast0.OptDecl(decl))
 
