@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -85,9 +85,9 @@ and 'a dots = 'a list wrap
 
 and base_ident =
     Id            of string mcode
-  | MetaId        of Ast.meta_name mcode * Ast.general_constraint * Ast.seed * pure
-  | MetaFunc      of Ast.meta_name mcode * Ast.general_constraint * pure
-  | MetaLocalFunc of Ast.meta_name mcode * Ast.general_constraint * pure
+  | MetaId        of Ast.meta_name mcode * constraints * Ast.seed * pure
+  | MetaFunc      of Ast.meta_name mcode * constraints * pure
+  | MetaLocalFunc of Ast.meta_name mcode * constraints * pure
   | AsIdent       of ident * ident (* as ident, always metavar *)
   | DisjId        of string mcode * ident list *
                      string mcode list (* the |s *) * string mcode
@@ -131,9 +131,10 @@ and base_expression =
 	              initialiser
   | MetaErr        of Ast.meta_name mcode * constraints * pure
   | MetaExpr       of Ast.meta_name mcode * constraints *
-	              typeC list option * Ast.form * pure
+	typeC list option * Ast.form * pure *
+	listlen option (* bitfield *)
   | MetaExprList   of Ast.meta_name mcode (* only in arg lists *) *
-	              listlen * pure
+	              listlen * constraints * pure
   | AsExpr         of expression * expression (* as expr, always metavar *)
   | AsSExpr        of expression * statement (* as expr, always metavar *)
   | EComma         of string mcode (* only in arg lists *)
@@ -152,14 +153,10 @@ and base_expression =
 
 and expression = base_expression wrap
 
-and constraints =
-    NoConstraint
-  | NotIdCstrt     of Ast.general_constraint
-  | NotExpCstrt    of expression list
-  | SubExpCstrt    of Ast.meta_name list
+and constraints = expression Ast.generic_constraints
 
 and listlen =
-    MetaListLen of Ast.meta_name mcode
+    MetaListLen of Ast.meta_name mcode * constraints
   | CstListLen of int
   | AnyListLen
 
@@ -167,13 +164,14 @@ and base_string_fragment =
     ConstantFragment of string mcode
   | FormatFragment of string mcode (*%*) * string_format (* format *)
   | Strdots of string mcode
-  | MetaFormatList of string mcode (*%*) * Ast.meta_name mcode * listlen
+  | MetaFormatList of string mcode (*%*) * Ast.meta_name mcode *
+	constraints * listlen
 
 and string_fragment = base_string_fragment wrap
 
 and base_string_format =
     ConstantFormat of string mcode
-  | MetaFormat of Ast.meta_name mcode * Ast.general_constraint
+  | MetaFormat of Ast.meta_name mcode * constraints
 
 and string_format = base_string_format wrap
 
@@ -182,14 +180,14 @@ and string_format = base_string_format wrap
 and base_assignOp =
     SimpleAssign of simpleAssignOp mcode
   | OpAssign of Ast_cocci.arithOp mcode
-  | MetaAssign of Ast_cocci.meta_name mcode * Ast.general_constraint * pure
+  | MetaAssign of Ast_cocci.meta_name mcode * constraints * pure
 and simpleAssignOp = string
 and assignOp = base_assignOp wrap
 
 and base_binaryOp =
     Arith of Ast_cocci.arithOp mcode
   | Logical of Ast_cocci.logicalOp mcode
-  | MetaBinary of Ast_cocci.meta_name mcode * Ast.general_constraint * pure
+  | MetaBinary of Ast_cocci.meta_name mcode * constraints * pure
 and binaryOp = base_binaryOp wrap
 
 (* --------------------------------------------------------------------- *)
@@ -214,11 +212,17 @@ and base_typeC =
 	string mcode (* { *) * expression dots * string mcode (* } *)
   | StructUnionName of Ast.structUnion mcode * ident option (* name *)
   | StructUnionDef  of typeC (* either StructUnionName or metavar *) *
-	string mcode (* { *) * declaration dots * string mcode (* } *)
+	string mcode (* { *) * field dots * string mcode (* } *)
+  | TypeOfExpr      of string mcode (* sizeof *) * string mcode (* ( *) *
+                       expression * string mcode (* ) *)
+  | TypeOfType      of string mcode (* sizeof *) * string mcode (* ( *) *
+                       typeC * string mcode (* ) *)
   | TypeName        of string mcode
-  | MetaType        of Ast.meta_name mcode * pure
+  | MetaType        of Ast.meta_name mcode * constraints * pure
   | AsType          of typeC * typeC (* as type, always metavar *)
-  | DisjType        of string mcode * typeC list * (* only after iso *)
+  | DisjType        of string mcode * typeC list *
+                       string mcode list (* the |s *)  * string mcode
+  | ConjType        of string mcode * typeC list *
                        string mcode list (* the |s *)  * string mcode
   | OptType         of typeC
 
@@ -230,16 +234,15 @@ and typeC = base_typeC wrap
    split out into multiple declarations of a single variable each. *)
 
 and base_declaration =
-    MetaDecl of Ast.meta_name mcode * pure (* variables *)
+    MetaDecl of Ast.meta_name mcode * constraints * pure (* variables *)
     (* the following are kept separate from MetaDecls because ultimately
        they don't match the same thing at all.  Consider whether there
        should be a separate type for fields, as in the C AST *)
-  | MetaField of Ast.meta_name mcode * pure (* structure fields *)
-  | MetaFieldList of Ast.meta_name mcode * listlen * pure (*structure fields*)
   | AsDecl        of declaration * declaration
-  | Init of Ast.storage mcode option * typeC * ident * string mcode (*=*) *
-	initialiser * string mcode (*;*)
-  | UnInit of Ast.storage mcode option * typeC * ident * string mcode (* ; *)
+  | Init of Ast.storage mcode option * typeC * ident * attr list *
+	string mcode (*=*) * initialiser * string mcode (*;*)
+  | UnInit of Ast.storage mcode option * typeC * ident * attr list *
+	string mcode (* ; *)
   | FunProto of
 	fninfo list * ident (* name *) *
 	string mcode (* ( *) * parameter_list *
@@ -257,19 +260,38 @@ and base_declaration =
   | Typedef of string mcode (* typedef *) * typeC * typeC * string mcode (*;*)
   | DisjDecl   of string mcode * declaration list *
 	          string mcode list (* the |s *)  * string mcode
-  (* Ddots is for a structure declaration *)
-  | Ddots      of string mcode (* ... *) * (string mcode * string mcode *
-	          declaration) option (* whencode *)
+  | ConjDecl   of string mcode * declaration list *
+	          string mcode list (* the &s *)  * string mcode
   | OptDecl    of declaration
 
 and declaration = base_declaration wrap
 
+(* --------------------------------------------------------------------- *)
+(* Field declaration *)
+
+and base_field =
+  | MetaField  of Ast_cocci.meta_name mcode * constraints *
+	pure (* structure fields *)
+  | MetaFieldList of Ast_cocci.meta_name mcode * listlen * constraints * pure
+  | Field     of typeC * ident option * bitfield option * string mcode (* ; *)
+  | DisjField   of string mcode * field list * string mcode list *
+	          string mcode
+  | ConjField   of string mcode * field list * string mcode list *
+	          string mcode
+  | Fdots      of string mcode (* ... *) * (string mcode * string mcode *
+                  field) option (* whencode *)
+  | OptField    of field
+
+and bitfield = string mcode (* : *) * expression
+
+and field = base_field wrap
+
 (* --------------------------------------------------------------------- *)
 (* Initializers *)
 
 and base_initialiser =
-    MetaInit of Ast.meta_name mcode * pure
-  | MetaInitList of Ast.meta_name mcode * listlen * pure
+    MetaInit of Ast.meta_name mcode * constraints * pure
+  | MetaInitList of Ast.meta_name mcode * listlen * constraints * pure
   | AsInit of initialiser * initialiser (* as init, always metavar *)
   | InitExpr of expression
   | InitList of string mcode (*{*) * initialiser_list * string mcode (*}*) *
@@ -302,8 +324,8 @@ and initialiser_list = initialiser dots
 and base_parameterTypeDef =
     VoidParam     of typeC
   | Param         of typeC * ident option
-  | MetaParam     of Ast.meta_name mcode * pure
-  | MetaParamList of Ast.meta_name mcode * listlen * pure
+  | MetaParam     of Ast.meta_name mcode * constraints * pure
+  | MetaParamList of Ast.meta_name mcode * listlen * constraints * pure
   | AsParam       of parameterTypeDef * expression (* expr, always metavar *)
   | PComma        of string mcode
   | Pdots         of string mcode (* ... *)
@@ -318,7 +340,7 @@ and parameter_list = parameterTypeDef dots
 
 and base_define_param =
     DParam        of ident
-  | MetaDParamList of Ast.meta_name mcode * listlen * pure
+  | MetaDParamList of Ast.meta_name mcode * listlen * constraints * pure
   | DPComma       of string mcode
   | DPdots        of string mcode (* ... *)
   | OptDParam     of define_param
@@ -374,9 +396,9 @@ and base_statement =
 	             string mcode (* ; *)
   | Exec          of string mcode (* EXEC *) * string mcode (* language *) *
 	             exec_code dots * string mcode (* ; *)
-  | MetaStmt      of Ast.meta_name mcode * pure
+  | MetaStmt      of Ast.meta_name mcode * constraints * pure
   | MetaStmtList  of Ast.meta_name mcode(*only in statement lists*) * listlen *
-	             pure
+	             constraints * pure
   | AsStmt        of statement * statement (* as statement, always metavar *)
   | Exp           of expression  (* only in dotted statement lists *)
   | TopExp        of expression (* for macros body *)
@@ -400,6 +422,7 @@ and base_statement =
 	string mcode (* } *) *
 	(info * mcodekind) (* after the function decl *)
   | Include of string mcode (* #include *) * Ast.inc_file mcode (* file *)
+  | MetaInclude of string mcode (* #include *) * expression (* file *)
   | Undef of string mcode (* #define *) * ident (* name *)
   | Define of string mcode (* #define *) * ident (* name *) *
 	define_parameters (*params*) * statement dots
@@ -423,7 +446,9 @@ and fninfo =
     FStorage of Ast.storage mcode
   | FType of typeC
   | FInline of string mcode
-  | FAttr of string mcode
+  | FAttr of attr
+
+and attr = string mcode
 
 and ('a,'b) whencode =
     WhenNot of string mcode (* when *) * string mcode (* != *) * 'a
@@ -456,7 +481,7 @@ and exec_code = base_exec_code wrap
 (* Positions *)
 
 and meta_pos =
-    MetaPos of Ast.meta_name mcode * Ast.general_constraint *
+    MetaPos of Ast.meta_name mcode * constraints *
 	Ast.meta_collect
 
 (* --------------------------------------------------------------------- *)
@@ -483,28 +508,30 @@ and parsed_rule =
   | ScriptRule of string (* name *) * string * Ast.dependency *
 	(Ast.script_meta_name * Ast.meta_name * Ast.metavar * Ast.mvinit)
 	  list *
-	Ast.meta_name list (*script vars*) *
+	Ast.meta_name list (*script vars*) * Ast.script_position *
 	string
   | InitialScriptRule of string (* name *) * string * Ast.dependency *
 	(Ast.script_meta_name * Ast.meta_name * Ast.metavar * Ast.mvinit)
-	  list *
+	  list * Ast.script_position *
 	string
   | FinalScriptRule of string (* name *) * string * Ast.dependency *
 	(Ast.script_meta_name * Ast.meta_name * Ast.metavar * Ast.mvinit)
-	  list *
+	  list * Ast.script_position *
 	string (* no script vars possible here *)
 
 (* --------------------------------------------------------------------- *)
 
-and dependency =
+and dep =
     Dep of string (* rule applies for the current binding *)
-  | AntiDep of dependency (* rule doesn't apply for the current binding *)
+  | AntiDep of dep (* rule doesn't apply for the current binding *)
   | EverDep of string (* rule applies for some binding *)
   | NeverDep of string (* rule never applies for any binding *)
-  | AndDep of dependency * dependency
-  | OrDep of dependency * dependency
+  | AndDep of dep * dep
+  | OrDep of dep * dep
   | FileIn of string
-  | NoDep | FailDep
+
+and dependency =
+    NoDep | FailDep | ExistsDep of dep | ForallDep of dep
 
 (* --------------------------------------------------------------------- *)
 
@@ -514,6 +541,7 @@ and anything =
   | DotsParamTag of parameterTypeDef dots
   | DotsStmtTag of statement dots
   | DotsDeclTag of declaration dots
+  | DotsFieldTag of field dots
   | DotsCaseTag of case_line dots
   | DotsDefParamTag of define_param dots
   | IdentTag of ident
@@ -526,6 +554,7 @@ and anything =
   | ParamTag of parameterTypeDef
   | InitTag of initialiser
   | DeclTag of declaration
+  | FieldTag of field
   | StmtTag of statement
   | ForInfoTag of forinfo
   | CaseLineTag of case_line
@@ -544,6 +573,7 @@ let dotsParam x = DotsParamTag x
 let dotsInit x = DotsInitTag x
 let dotsStmt x = DotsStmtTag x
 let dotsDecl x = DotsDeclTag x
+let dotsField x = DotsFieldTag x
 let dotsCase x = DotsCaseTag x
 let dotsDefParam x = DotsDefParamTag x
 let ident x = IdentTag x
@@ -554,6 +584,7 @@ let typeC x = TypeCTag x
 let param x = ParamTag x
 let ini x = InitTag x
 let decl x = DeclTag x
+let field x = FieldTag x
 let stmt x = StmtTag x
 let forinfo x = ForInfoTag x
 let case_line x = CaseLineTag x
@@ -641,8 +672,8 @@ let get_iso x           = x.iso_info
 let set_iso x i = if !Flag.track_iso_usage then {x with iso_info = i} else x
 let set_mcode_data data (_,ar,info,mc,pos,adj) = (data,ar,info,mc,pos,adj)
 let get_rule_name = function
-  | CocciRule ((_,_,(_,_,_,nm,_)),_,_,_) | InitialScriptRule (nm,_,_,_,_)
-  | FinalScriptRule (nm,_,_,_,_) | ScriptRule (nm,_,_,_,_,_) -> nm
+  | CocciRule ((_,_,(_,_,_,nm,_)),_,_,_) | InitialScriptRule (nm,_,_,_,_,_)
+  | FinalScriptRule (nm,_,_,_,_,_) | ScriptRule (nm,_,_,_,_,_,_) -> nm
 
 (* --------------------------------------------------------------------- *)
 
@@ -657,24 +688,24 @@ let rec meta_pos_name = function
       | _ -> failwith "bad metavariable")
   | ExprTag(e) ->
       (match unwrap e with
-	MetaExpr(name,_constraints,_ty,_form,_pure) -> name
-      | MetaExprList(name,_len,_pure) -> name
+	MetaExpr(name,_constraints,_ty,_form,_pure,_bitfield) -> name
+      | MetaExprList(name,_len,_constraints,_pure) -> name
       | _ -> failwith "bad metavariable")
   | TypeCTag(t) ->
       (match unwrap t with
-	MetaType(name,_pure) -> name
+	MetaType(name,_cstr,_pure) -> name
       | _ -> failwith "bad metavariable")
   | DeclTag(d) ->
       (match unwrap d with
-	MetaDecl(name,_pure) -> name
+	MetaDecl(name,_constraints,_pure) -> name
       | _ -> failwith "bad metavariable")
   | InitTag(i) ->
       (match unwrap i with
-	MetaInit(name,_pure) -> name
+	MetaInit(name,_constraints,_pure) -> name
       | _ -> failwith "bad metavariable")
   | StmtTag(s) ->
       (match unwrap s with
-	MetaStmt(name,_pure) -> name
+	MetaStmt(name,_constraints,_pure) -> name
       | _ -> failwith "bad metavariable")
   | _ -> failwith "bad metavariable"
 
@@ -703,7 +734,7 @@ let rec meta_names_of_typeC ty =
   | Array (ty, _, _, _) -> meta_names_of_typeC ty
   | EnumName (_, Some ident)
   | StructUnionName(_, Some ident) -> meta_names_of_ident ident
-  | MetaType (tyname, _) -> [unwrap_mcode tyname]
+  | MetaType (tyname, _, _) -> [unwrap_mcode tyname]
   | Decimal (_, _, e1, _, e2, _) ->
       let mn1 = meta_names_of_expression e1 in
       let mn2 = Common.default [] meta_names_of_expression e2 in
@@ -713,7 +744,7 @@ let rec meta_names_of_typeC ty =
 let meta_pos_constraint_names = function
     ExprTag(e) ->
       (match unwrap e with
-	MetaExpr(_name,_constraints,ty,_form,_pure) ->
+	MetaExpr(_name,_constraints,ty,_form,_pure,_bitfield) ->
 	  (match ty with
 	    Some tylist ->
               List.fold_left (fun prev cur -> meta_names_of_typeC cur @ prev)
