@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -14,7 +14,7 @@ module Ast = Ast_cocci
 module V0 = Visitor_ast0
 module VT0 = Visitor_ast0_types
 
-let unitary = Type_cocci.Unitary
+let unitary = Ast.Unitary
 
 let ctr = ref 0
 let get_ctr _ =
@@ -175,9 +175,10 @@ let inline_mcodes =
     mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
     mcode mcode
     do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
+    do_nothing
     do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
     do_nothing do_nothing do_nothing do_nothing do_nothing
-    do_nothing do_nothing
+    do_nothing do_nothing do_nothing do_nothing
 
 (* --------------------------------------------------------------------- *)
 (* For function declarations.  Can't use the mcode at the root, because that
@@ -212,15 +213,25 @@ let check_allminus =
 
   let declaration r k e =
     match Ast0.unwrap e with
-      Ast0.DisjDecl(starter,decls,mids,ender) ->
+      Ast0.DisjDecl(starter,decls,mids,ender)
+    | Ast0.ConjDecl(starter,decls,mids,ender) ->
 	List.for_all r.VT0.combiner_rec_declaration decls
     | Ast0.AsDecl(decl,asdecl) -> k decl
     | _ -> k e in
 
+  let field r k e =
+    match Ast0.unwrap e with
+      Ast0.DisjField(starter,decls,mids,ender)
+    | Ast0.ConjField(starter,decls,mids,ender) ->
+	List.for_all r.VT0.combiner_rec_field decls
+    | _ -> k e in
+
   let typeC r k e =
     match Ast0.unwrap e with
-      Ast0.DisjType(starter,decls,mids,ender) ->
-	List.for_all r.VT0.combiner_rec_typeC decls
+      Ast0.DisjType(starter,type_list,mids,ender) ->
+	List.for_all r.VT0.combiner_rec_typeC type_list
+    | Ast0.ConjType(starter,type_list,mids,ender) ->
+	List.for_all r.VT0.combiner_rec_typeC type_list
     | Ast0.AsType(ty,asty) -> k ty
     | _ -> k e in
 
@@ -246,9 +257,9 @@ let check_allminus =
   V0.flat_combiner bind option_default
     mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
     mcode mcode mcode mcode
-    donothing donothing donothing donothing donothing donothing
-    ident expression donothing donothing typeC initialiser donothing declaration
-    statement donothing case_line donothing donothing
+    donothing donothing donothing donothing donothing donothing donothing
+    donothing ident expression donothing donothing typeC initialiser donothing
+    declaration field statement donothing case_line donothing donothing
 
 (* --------------------------------------------------------------------- *)
 (* --------------------------------------------------------------------- *)
@@ -304,13 +315,18 @@ let pos_mcode(term,_,info,mcodekind,pos,adj) =
   (* avoids a recursion problem *)
   (term,convert_info info,convert_mcodekind adj mcodekind,[])
 
+(* Postponed transformer declaration for constraints *)
+let constraints' = ref (fun _ -> failwith "unbound constraints")
+let constraints c = !constraints' c
+
 let mcode (term,_,info,mcodekind,pos,adj) =
   let pos =
     List.fold_left
       (function prev ->
 	function
-	    Ast0.MetaPosTag(Ast0.MetaPos(pos,constraints,per)) ->
-	      (Ast.MetaPos(pos_mcode pos,constraints,per,unitary,false))::prev
+	    Ast0.MetaPosTag(Ast0.MetaPos(pos,cstr,per)) ->
+	      let cstr' = constraints cstr in
+	      (Ast.MetaPos(pos_mcode pos,cstr',per,unitary,false))::prev
 	  | _ -> prev)
       [] !pos in
   (term,convert_info info,convert_mcodekind adj mcodekind,List.rev pos)
@@ -343,12 +359,15 @@ and ident i =
       Ast0.Id(name) -> Ast.Id(mcode name)
     | Ast0.DisjId(_,id_list,_,_) ->
 	Ast.DisjId(List.map ident id_list)
-    | Ast0.MetaId(name,constraints,_,_) ->
-	Ast.MetaId(mcode name,constraints,unitary,false)
-    | Ast0.MetaFunc(name,constraints,_) ->
-	Ast.MetaFunc(mcode name,constraints,unitary,false)
-    | Ast0.MetaLocalFunc(name,constraints,_) ->
-	Ast.MetaLocalFunc(mcode name,constraints,unitary,false)
+    | Ast0.MetaId(name,cstr,_,_) ->
+	let cstr' = constraints cstr in
+	Ast.MetaId(mcode name,cstr',unitary,false)
+    | Ast0.MetaFunc(name,cstr,_) ->
+	let cstr' = constraints cstr in
+	Ast.MetaFunc(mcode name,cstr',unitary,false)
+    | Ast0.MetaLocalFunc(name,cstr,_) ->
+	let cstr' = constraints cstr in
+	Ast.MetaLocalFunc(mcode name,cstr',unitary,false)
     | Ast0.AsIdent(id,asid) ->
 	Ast.AsIdent(ident id,ident asid)
     | Ast0.OptIdent(id) -> Ast.OptIdent(ident id))
@@ -416,10 +435,14 @@ and expression e =
 	Ast.Constructor(mcode lp,typeC allminus ty,mcode rp,initialiser init)
     | Ast0.MetaErr(name,cstrts,_)  ->
 	Ast.MetaErr(mcode name,constraints cstrts,unitary,false)
-    | Ast0.MetaExpr(name,cstrts,ty,form,_)  ->
-	Ast.MetaExpr(mcode name,constraints cstrts,unitary,ty,form,false)
-    | Ast0.MetaExprList(name,lenname,_) ->
-	Ast.MetaExprList(mcode name,do_lenname lenname,unitary,false)
+    | Ast0.MetaExpr(name,cstrts,ty,form,_,bitfield)  ->
+        let ty' = Common.map_option (List.map (typeC false)) ty in
+	let bitfield' = Common.map_option do_lenname bitfield in
+        Ast.MetaExpr
+	  (mcode name, constraints cstrts, unitary, ty', form, false, bitfield')
+    | Ast0.MetaExprList(name,lenname,cstr,_) ->
+	let cstr' = constraints cstr in
+	Ast.MetaExprList (mcode name,do_lenname lenname,cstr',unitary,false)
     | Ast0.AsExpr(expr,asexpr) ->
 	Ast.AsExpr(expression expr,expression asexpr)
     | Ast0.AsSExpr(expr,asstm) ->
@@ -450,22 +473,18 @@ and assignOp op =
     (match Ast0.unwrap op with
       Ast0.SimpleAssign op' -> Ast.SimpleAssign (mcode op')
     | Ast0.OpAssign op' -> Ast.OpAssign (mcode op')
-    | Ast0.MetaAssign(mv, c, _) ->
-      Ast.MetaAssign(mcode mv, assignOpconstraint c, unitary, false))
-and assignOpconstraint = function
-  | Ast0.AssignOpNoConstraint -> Ast.AssignOpNoConstraint
-  | Ast0.AssignOpInSet ops -> Ast.AssignOpInSet (List.map assignOp ops)
+    | Ast0.MetaAssign(mv, cstr, _) ->
+	let cstr' = constraints cstr in
+	Ast.MetaAssign(mcode mv, cstr', unitary, false))
 
 and binaryOp op =
   rewrap op no_isos
     (match Ast0.unwrap op with
       Ast0.Arith op' -> Ast.Arith (mcode op')
     | Ast0.Logical op' -> Ast.Logical (mcode op')
-    | Ast0.MetaBinary(mv, c, _) ->
-      Ast.MetaBinary(mcode mv, binaryOpconstraint c, unitary, false))
-and binaryOpconstraint = function
-  | Ast0.BinaryOpNoConstraint -> Ast.BinaryOpNoConstraint
-  | Ast0.BinaryOpInSet ops -> Ast.BinaryOpInSet (List.map binaryOp ops)
+    | Ast0.MetaBinary(mv, cstr, _) ->
+	let cstr' = constraints cstr in
+	Ast.MetaBinary(mcode mv, cstr', unitary, false))
 
 and expression_dots ed = dots expression ed
 
@@ -476,26 +495,21 @@ and string_fragment e =
     | Ast0.FormatFragment(pct,fmt) ->
 	Ast.FormatFragment(mcode pct, string_format fmt)
     | Ast0.Strdots dots -> Ast.Strdots (mcode dots)
-    | Ast0.MetaFormatList(pct,name,lenname) ->
+    | Ast0.MetaFormatList(pct,name,cstr,lenname) ->
 	Ast.MetaFormatList(mcode pct, mcode name, do_lenname lenname,
-			   unitary,false))
+			   constraints cstr,unitary,false))
 
 and string_format e =
   rewrap e no_isos
     (match Ast0.unwrap e with
       Ast0.ConstantFormat(str) -> Ast.ConstantFormat(mcode str)
-    | Ast0.MetaFormat(name,constraints) ->
-	Ast.MetaFormat(mcode name,constraints,unitary,false))
-
-and constraints c =
-  match c with
-      Ast0.NoConstraint        -> Ast.NoConstraint
-    | Ast0.NotIdCstrt   idctrt -> Ast.NotIdCstrt idctrt
-    | Ast0.NotExpCstrt  exps   -> Ast.NotExpCstrt (List.map expression exps)
-    | Ast0.SubExpCstrt  ids    -> Ast.SubExpCstrt ids
+    | Ast0.MetaFormat(name,cstr) ->
+	let cstr' = constraints cstr in
+	Ast.MetaFormat(mcode name,cstr',unitary,false))
 
 and do_lenname = function
-    Ast0.MetaListLen(nm) -> Ast.MetaListLen(mcode nm,unitary,false)
+    Ast0.MetaListLen(nm,cstr) ->
+      Ast.MetaListLen(mcode nm,constraints cstr,unitary,false)
   | Ast0.CstListLen n -> Ast.CstListLen n
   | Ast0.AnyListLen -> Ast.AnyListLen
 
@@ -533,10 +547,13 @@ and typeC allminus t =
     | Ast0.Array(_,_,_,_) | Ast0.Decimal(_,_,_,_,_,_)
     | Ast0.EnumName(_,_) | Ast0.StructUnionName(_,_)
     | Ast0.StructUnionDef(_,_,_,_) | Ast0.EnumDef(_,_,_,_)
-    | Ast0.TypeName(_) | Ast0.MetaType(_,_) ->
+    | Ast0.TypeOfExpr(_,_,_,_) | Ast0.TypeOfType(_,_,_,_)
+    | Ast0.TypeName(_) | Ast0.MetaType(_,_,_) ->
 	Ast.Type(allminus,None,rewrap t no_isos (base_typeC allminus t))
     | Ast0.DisjType(_,types,_,_) ->
 	Ast.DisjType(List.map (typeC allminus) types)
+    | Ast0.ConjType(_,types,_,_) ->
+	Ast.ConjType(List.map (typeC allminus) types)
     | Ast0.AsType(ty,asty) ->
 	Ast.AsType(typeC allminus ty,typeC allminus asty)
     | Ast0.OptType(ty) -> Ast.OptType(typeC allminus ty))
@@ -568,10 +585,16 @@ and base_typeC allminus t =
       Ast.StructUnionName(mcode kind,get_option ident name)
   | Ast0.StructUnionDef(ty,lb,decls,rb) ->
       Ast.StructUnionDef(typeC allminus ty,mcode lb,
-			 declaration_dots decls, mcode rb)
+			 field_dots decls, mcode rb)
+  | Ast0.TypeOfExpr(tf,lp,exp,rp) ->
+      Ast.TypeOfExpr(mcode tf,mcode lp,expression exp,mcode rp)
+  | Ast0.TypeOfType(tf,lp,ty,rp) ->
+      let allminus = check_allminus.VT0.combiner_rec_typeC t in
+      Ast.TypeOfType(mcode tf, mcode lp,typeC allminus ty,mcode rp)
   | Ast0.TypeName(name) -> Ast.TypeName(mcode name)
-  | Ast0.MetaType(name,_) ->
-      Ast.MetaType(mcode name,unitary,false)
+  | Ast0.MetaType(name,cstr,_) ->
+      let cstr' = constraints cstr in
+      Ast.MetaType(mcode name,cstr',unitary,false)
   | _ -> failwith "ast0toast: unexpected type"
 
 (* --------------------------------------------------------------------- *)
@@ -582,24 +605,24 @@ and base_typeC allminus t =
 and declaration d =
   rewrap d (do_isos (Ast0.get_iso d))
     (match Ast0.unwrap d with
-      Ast0.MetaDecl(name,_) -> Ast.MetaDecl(mcode name,unitary,false)
-    | Ast0.MetaField(name,_) -> Ast.MetaField(mcode name,unitary,false)
-    | Ast0.MetaFieldList(name,lenname,_) ->
-	Ast.MetaFieldList(mcode name,do_lenname lenname,unitary,false)
+      Ast0.MetaDecl(name,cstr,_) ->
+	Ast.MetaDecl(mcode name,constraints cstr,unitary,false)
     | Ast0.AsDecl(decl,asdecl) ->
 	Ast.AsDecl(declaration decl,declaration asdecl)
-    | Ast0.Init(stg,ty,id,eq,ini,sem) ->
+    | Ast0.Init(stg,ty,id,attr,eq,ini,sem) ->
 	let allminus = check_allminus.VT0.combiner_rec_declaration d in
 	let stg = get_option mcode stg in
 	let ty = typeC allminus ty in
 	let id = ident id in
+	let attr = List.map mcode attr in
 	let eq = mcode eq in
 	let ini = initialiser ini in
 	let sem = mcode sem in
-	Ast.Init(stg,ty,id,eq,ini,sem)
-    | Ast0.UnInit(stg,ty,id,sem) ->
+	Ast.Init(stg,ty,id,attr,eq,ini,sem)
+    | Ast0.UnInit(stg,ty,id,attr,sem) ->
 	let allminus = check_allminus.VT0.combiner_rec_declaration d in
-	Ast.UnInit(get_option mcode stg,typeC allminus ty,ident id,
+	let attr = List.map mcode attr in
+	Ast.UnInit(get_option mcode stg,typeC allminus ty,ident id,attr,
 		   mcode sem)
     | Ast0.FunProto(fi,name,lp,params,va,rp,sem) ->
 	  let fi = List.map fninfo fi in
@@ -642,19 +665,14 @@ and declaration d =
 	  Ast.Type(_,None,id) -> (* only MetaType or Id *)
 	    Ast.Typedef(mcode stg,typeC allminus ty,id,mcode sem)
 	| _ -> failwith "bad typedef")
-    | Ast0.Ddots(dots,whencode) -> failwith "should not be possible"
     | Ast0.DisjDecl(_,decls,_,_) -> Ast.DisjDecl(List.map declaration decls)
+    | Ast0.ConjDecl(_,decls,_,_) -> Ast.ConjDecl(List.map declaration decls)
     | Ast0.OptDecl(decl) -> Ast.OptDecl(declaration decl))
 
 and annotated_decl bef d =
   rewrap d (do_isos (Ast0.get_iso d))
     (match Ast0.unwrap d with
-      Ast0.Ddots(dots,whencode) ->
-	(* structure definitions only *)
-	let dots = mcode dots in
-	let whencode = get_option (fun (_,_,b) -> declaration b) whencode in
-	Ast.Ddots(dots,whencode)
-    | _ -> (* for decls where there is no bef information needed *)
+      _ -> (* for decls where there is no bef information needed *)
 	let bef =
 	  match bef with
 	    None -> (* fake, no change here *)
@@ -667,6 +685,48 @@ and annotated_decl bef d =
 
 and declaration_dots l = dots (annotated_decl None) l
 
+(* --------------------------------------------------------------------- *)
+(* Field declaration *)
+
+and field d =
+  rewrap d (do_isos (Ast0.get_iso d))
+    (match Ast0.unwrap d with
+      Ast0.MetaField(name,cstr,_) ->
+	Ast.MetaField(mcode name,constraints cstr,unitary,false)
+    | Ast0.MetaFieldList(name,lenname,cstr,_) ->
+	let cstr' = constraints cstr in
+	Ast.MetaFieldList(mcode name,do_lenname lenname,cstr',unitary,false)
+    | Ast0.Field(ty,id,bf,sem) ->
+	let allminus = check_allminus.VT0.combiner_rec_field d in
+	let bitfield (c, e) = (mcode c, expression e) in
+	let bf = Common.map_option bitfield bf in
+	Ast.Field(typeC allminus ty,Common.map_option ident id, bf, mcode sem)
+    | Ast0.Fdots(dots,whencode) -> failwith "should not be possible"
+    | Ast0.DisjField(_,decls,_,_) -> Ast.DisjField(List.map field decls)
+    | Ast0.ConjField(_,decls,_,_) -> Ast.ConjField(List.map field decls)
+    | Ast0.OptField(decl) -> Ast.OptField(field decl))
+
+and annotated_field bef d =
+  rewrap d (do_isos (Ast0.get_iso d))
+    (match Ast0.unwrap d with
+      Ast0.Fdots(dots,whencode) ->
+	(* structure definitions only *)
+	let dots = mcode dots in
+	let whencode = get_option (fun (_,_,b) -> field b) whencode in
+	Ast.Fdots(dots,whencode)
+    | _ -> (* for decls where there is no bef information needed *)
+	let bef =
+	  match bef with
+	    None -> (* fake, no change here *)
+	      let bot = Ast0.default_token_info in
+	      Ast0.CONTEXT (ref(Ast.NOTHING,bot,bot))
+	  | Some bef -> bef in
+	let allminus = check_allminus.VT0.combiner_rec_field d in
+	Ast.FElem(convert_allminus_mcodekind allminus bef,allminus,
+		  field d))
+
+and field_dots l = dots (annotated_field None) l
+
 (* --------------------------------------------------------------------- *)
 (* Initialiser *)
 
@@ -713,9 +773,11 @@ and strip_idots initlist =
 and initialiser i =
   rewrap i no_isos
     (match Ast0.unwrap i with
-      Ast0.MetaInit(name,_) -> Ast.MetaInit(mcode name,unitary,false)
-    | Ast0.MetaInitList(name,lenname,_) ->
-	Ast.MetaInitList(mcode name,do_lenname lenname,unitary,false)
+      Ast0.MetaInit(name,cstr,_) ->
+	Ast.MetaInit(mcode name,constraints cstr,unitary,false)
+    | Ast0.MetaInitList(name,lenname,cstr,_) ->
+	let cstr' = constraints cstr in
+	Ast.MetaInitList(mcode name,do_lenname lenname,cstr',unitary,false)
     | Ast0.AsInit(init,asinit) ->
 	Ast.AsInit(initialiser init,initialiser asinit)
     | Ast0.InitExpr(exp) -> Ast.InitExpr(expression exp)
@@ -756,10 +818,11 @@ and parameterTypeDef p =
     | Ast0.Param(ty,id) ->
 	let allminus = check_allminus.VT0.combiner_rec_parameter p in
 	Ast.Param(typeC allminus ty,get_option ident id)
-    | Ast0.MetaParam(name,_) ->
-	Ast.MetaParam(mcode name,unitary,false)
-    | Ast0.MetaParamList(name,lenname,_) ->
-	Ast.MetaParamList(mcode name,do_lenname lenname,unitary,false)
+    | Ast0.MetaParam(name,cstr,_) ->
+	Ast.MetaParam(mcode name,constraints cstr,unitary,false)
+    | Ast0.MetaParamList(name,lenname,cstr,_) ->
+	let cstr' = constraints cstr in
+	Ast.MetaParamList(mcode name,do_lenname lenname,cstr',unitary,false)
     | Ast0.AsParam(p,asexpr) ->
 	Ast.AsParam(parameterTypeDef p,expression asexpr)
     | Ast0.PComma(cm) -> Ast.PComma(mcode cm)
@@ -881,13 +944,15 @@ and statement s =
 	  Ast.Atomic
 	    (rewrap_rule_elem s
 	       (Ast.Exec(mcode exec,mcode lang,dots exec_code code,mcode sem)))
-      | Ast0.MetaStmt(name,_) ->
+      | Ast0.MetaStmt(name,cstr,_) ->
+	  let cstr' = constraints cstr in
 	  Ast.Atomic(rewrap_rule_elem s
-		       (Ast.MetaStmt(mcode name,unitary,seqible,false)))
-      | Ast0.MetaStmtList(name,lenname,_) ->
+		       (Ast.MetaStmt(mcode name,cstr',unitary,seqible,false)))
+      | Ast0.MetaStmtList(name,lenname,cstr,_) ->
+	  let cstr' = constraints cstr in
 	  Ast.Atomic(rewrap_rule_elem s
 		       (Ast.MetaStmtList(mcode name,do_lenname lenname,
-					 unitary,false)))
+					 cstr',unitary,false)))
       | Ast0.AsStmt(stmt,asstmt) ->
 	  Ast.AsStmt(statement seqible stmt,statement seqible asstmt)
       | Ast0.TopExp(exp) ->
@@ -948,6 +1013,9 @@ and statement s =
                       ([],[],[],convert_allminus_mcodekind allminus aft))
       |	Ast0.Include(inc,str) ->
 	  Ast.Atomic(rewrap_rule_elem s (Ast.Include(mcode inc,mcode str)))
+      |	Ast0.MetaInclude(inc,str) ->
+	  Ast.Atomic
+	    (rewrap_rule_elem s (Ast.MetaInclude(mcode inc,expression str)))
       |	Ast0.Undef(def,id) ->
 	  Ast.Atomic(rewrap_rule_elem s (Ast.Undef(mcode def,ident id)))
       | Ast0.Define(def,id,params,body) ->
@@ -961,34 +1029,6 @@ and statement s =
 		       (Ast.Pragma(mcode prg,ident id,pragmainfo body)))
       | Ast0.OptStm(stm) -> Ast.OptStm(statement seqible stm))
 
-  and pragmainfo pi =
-    rewrap pi no_isos
-      (match Ast0.unwrap pi with
-	Ast0.PragmaTuple(lp,args,rp) ->
-	  let lp = mcode lp in
-	  let args = dots expression args in
-	  let rp = mcode rp in
-	  Ast.PragmaTuple(lp,args,rp)
-      | Ast0.PragmaIdList(ids) -> Ast.PragmaIdList(dots ident ids)
-      | Ast0.PragmaDots (dots) -> Ast.PragmaDots (mcode dots))
-
-  and define_parameters p =
-    rewrap p no_isos
-      (match Ast0.unwrap p with
-	Ast0.NoParams -> Ast.NoParams
-      | Ast0.DParams(lp,params,rp) ->
-	  Ast.DParams(mcode lp,
-		      dots define_param params,
-		      mcode rp))
-
-  and define_param p =
-    rewrap p no_isos
-      (match Ast0.unwrap p with
-	Ast0.DParam(id) -> Ast.DParam(ident id)
-      | Ast0.DPComma(comma) -> Ast.DPComma(mcode comma)
-      | Ast0.DPdots(d) -> Ast.DPdots(mcode d)
-      | Ast0.OptDParam(dp) -> Ast.OptDParam(define_param dp))
-
   and whencode notfn alwaysfn = function
       Ast0.WhenNot (_,_,a) -> Ast.WhenNot (notfn a)
     | Ast0.WhenAlways (_,_,a) -> Ast.WhenAlways (alwaysfn a)
@@ -1070,6 +1110,37 @@ and statement s =
 
   statement Ast.Sequencible s
 
+and pragmainfo pi =
+  rewrap pi no_isos
+    (match Ast0.unwrap pi with
+      Ast0.PragmaTuple(lp,args,rp) ->
+	let lp = mcode lp in
+	let args = dots expression args in
+	let rp = mcode rp in
+	Ast.PragmaTuple(lp,args,rp)
+    | Ast0.PragmaIdList(ids) -> Ast.PragmaIdList(dots ident ids)
+    | Ast0.PragmaDots (dots) -> Ast.PragmaDots (mcode dots))
+
+and define_parameters p =
+  rewrap p no_isos
+    (match Ast0.unwrap p with
+      Ast0.NoParams -> Ast.NoParams
+    | Ast0.DParams(lp,params,rp) ->
+	Ast.DParams(mcode lp,define_param_dots params,mcode rp))
+
+and define_param p =
+  rewrap p no_isos
+    (match Ast0.unwrap p with
+      Ast0.DParam(id) -> Ast.DParam(ident id)
+    | Ast0.MetaDParamList(name,lenname,cstr,_) ->
+	let cstr' = constraints cstr in
+	Ast.MetaDParamList(mcode name,do_lenname lenname,cstr',unitary,false)
+    | Ast0.DPComma(comma) -> Ast.DPComma(mcode comma)
+    | Ast0.DPdots(d) -> Ast.DPdots(mcode d)
+    | Ast0.OptDParam(dp) -> Ast.OptDParam(define_param dp))
+
+and define_param_dots l = dots define_param l
+
 and forinfo fi =
   match Ast0.unwrap fi with
     Ast0.ForExp(exp1,sem1) ->
@@ -1126,7 +1197,9 @@ and anything = function
   | Ast0.DotsInitTag(d) -> failwith "not possible"
   | Ast0.DotsStmtTag(d) -> Ast.StmtDotsTag(statement_dots d)
   | Ast0.DotsDeclTag(d) -> Ast.AnnDeclDotsTag(declaration_dots d)
+  | Ast0.DotsFieldTag(d) -> Ast.AnnFieldDotsTag(field_dots d)
   | Ast0.DotsCaseTag(d) -> failwith "not possible"
+  | Ast0.DotsDefParamTag(d) -> Ast.DefParDotsTag(define_param_dots d)
   | Ast0.IdentTag(d) -> Ast.IdentTag(ident d)
   | Ast0.ExprTag(d) -> Ast.ExpressionTag(expression d)
   | Ast0.AssignOpTag d -> Ast.AssignOpTag(assignOp d)
@@ -1137,6 +1210,7 @@ and anything = function
   | Ast0.ParamTag(d) -> Ast.ParamTag(parameterTypeDef d)
   | Ast0.InitTag(d) -> Ast.InitTag(initialiser d)
   | Ast0.DeclTag(d) -> Ast.DeclarationTag(declaration d)
+  | Ast0.FieldTag(d) -> Ast.FieldTag(field d)
   | Ast0.StmtTag(d) -> Ast.StatementTag(statement d)
   | Ast0.ForInfoTag(d) -> Ast.ForInfoTag(forinfo d)
   | Ast0.CaseLineTag(d) -> Ast.CaseLineTag(case_line d)
@@ -1163,6 +1237,12 @@ and top_level t =
     | Ast0.ERRORWORDS(exps) -> Ast.ERRORWORDS(List.map expression exps)
     | Ast0.OTHER(_) | Ast0.TOPCODE(_) -> failwith "eliminated by top_level")
 
+let constraints c =
+  let cstr_expr = Some (fun e -> Ast.CstrExpr (expression e)) in
+  Ast.cstr_map { Ast.empty_cstr_transformer with Ast.cstr_expr } c
+
+let () = constraints' := constraints
+
 (* --------------------------------------------------------------------- *)
 (* Entry point for minus code *)
 
