@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -16,6 +16,7 @@ type info = { line : int; column : int;
 
 type line = int
 type meta_name = string * string
+type script_position = string (* filename *) * line (* line *)
 (* need to be careful about rewrapping, to avoid duplicating pos info
 currently, the pos info is always None until asttoctl2. *)
 type 'a wrap =
@@ -23,9 +24,11 @@ type 'a wrap =
       node_line : line;
       free_vars : meta_name list; (*free vars*)
       minus_free_vars : meta_name list; (*minus free vars*)
+      minus_nc_free_vars : meta_name list; (*minus free vars, excluding cstrs*)
       fresh_vars : (meta_name * seed) list; (*fresh vars*)
       inherited : meta_name list; (*inherited vars*)
       positive_inherited_positions : meta_name list;
+      constraints : (meta_name * constraints) list;
       saved_witness : meta_name list; (*witness vars*)
       bef_aft : dots_bef_aft;
       (* the following is for or expressions *)
@@ -73,8 +76,9 @@ and dots_bef_aft =
   | AddingBetweenDots of statement * int (*index of let var*)
   | DroppingBetweenDots of statement * int (*index of let var*)
 
-and inherited = Type_cocci.inherited
-and keep_binding = Type_cocci.keep_binding
+and inherited = bool (* true if inherited *)
+and keep_binding = Unitary (* need no info *)
+  | Nonunitary (* need an env entry *) | Saved (* need a witness *)
 and multi = bool (*true if a nest is one or more, false if it is zero or more*)
 
 and end_info =
@@ -95,26 +99,28 @@ and metavar =
   | MetaInitListDecl of arity * meta_name (* name *) * list_len (*len*)
   | MetaListlenDecl of meta_name (* name *)
   | MetaParamDecl of arity * meta_name (* name *)
+  | MetaParamListDecl of arity * meta_name (*name*) * list_len (*len*)
   | MetaBinaryOperatorDecl of arity * meta_name
   | MetaAssignmentOperatorDecl of arity * meta_name
-  | MetaParamListDecl of arity * meta_name (*name*) * list_len (*len*)
   | MetaConstDecl of
-      arity * meta_name (* name *) * Type_cocci.typeC list option
+      arity * meta_name (* name *) * fullType list option
   | MetaErrDecl of arity * meta_name (* name *)
   | MetaExpDecl of
-      arity * meta_name (* name *) * Type_cocci.typeC list option
+      arity * meta_name (* name *) * fullType list option *
+	list_len option (* bitfield *)
   | MetaIdExpDecl of
-      arity * meta_name (* name *) * Type_cocci.typeC list option
+      arity * meta_name (* name *) * fullType list option
   | MetaLocalIdExpDecl of
-      arity * meta_name (* name *) * Type_cocci.typeC list option
+      arity * meta_name (* name *) * fullType list option
   | MetaGlobalIdExpDecl of
-      arity * meta_name (* name *) * Type_cocci.typeC list option
+      arity * meta_name (* name *) * fullType list option
   | MetaExpListDecl of arity * meta_name (*name*) * list_len (*len*)
   | MetaDeclDecl of arity * meta_name (* name *)
   | MetaFieldDecl of arity * meta_name (* name *)
   | MetaFieldListDecl of arity * meta_name (* name *) * list_len (*len*)
   | MetaStmDecl of arity * meta_name (* name *)
   | MetaStmListDecl of arity * meta_name (* name *) * list_len (*len*)
+  | MetaDParamListDecl of arity * meta_name (* name *) * list_len (*len*)
   | MetaFuncDecl of arity * meta_name (* name *)
   | MetaLocalFuncDecl of arity * meta_name (* name *)
   | MetaPosDecl of arity * meta_name (* name *)
@@ -125,7 +131,7 @@ and metavar =
   | MetaIteratorDecl of arity * meta_name (* name *)
   | MetaScriptDecl of metavar option ref * meta_name (* name *)
 
-and list_len = AnyLen | MetaLen of meta_name | CstLen of int
+and list_len = AnyLen | MetaLen of meta_name * constraints | CstLen of int
 
 and seed = NoVal | StringSeed of string | ListSeed of seed_elem list
 and seed_elem = SeedString of string | SeedId of meta_name
@@ -141,9 +147,12 @@ and 'a dots = 'a list wrap
 
 and base_ident =
     Id            of string mcode
-  | MetaId        of meta_name mcode * idconstraint * keep_binding * inherited
-  | MetaFunc      of meta_name mcode * idconstraint * keep_binding * inherited
-  | MetaLocalFunc of meta_name mcode * idconstraint * keep_binding * inherited
+  | MetaId        of
+      meta_name mcode * constraints * keep_binding * inherited
+  | MetaFunc      of
+      meta_name mcode * constraints * keep_binding * inherited
+  | MetaLocalFunc of
+      meta_name mcode * constraints * keep_binding * inherited
   | AsIdent       of ident * ident (* as ident, always metavar *)
 
   | DisjId        of ident list
@@ -191,8 +200,9 @@ and base_expression =
   | MetaErr        of meta_name mcode * constraints * keep_binding *
 	              inherited
   | MetaExpr       of meta_name mcode * constraints * keep_binding *
-	              Type_cocci.typeC list option * form * inherited
-  | MetaExprList   of meta_name mcode * listlen * keep_binding *
+	fullType list option * form * inherited *
+	listlen option (* bitfield *)
+  | MetaExprList   of meta_name mcode * listlen * constraints * keep_binding *
                       inherited (* only in arg lists *)
   | AsExpr         of expression * expression (* as expr, always metavar *)
   | AsSExpr        of expression * rule_elem (* as expr, always metavar *)
@@ -213,30 +223,42 @@ and base_expression =
 
   | OptExp         of expression
 
-and constraints =
-    NoConstraint
-  | NotIdCstrt     of reconstraint
-  | NotExpCstrt    of expression list
-  | SubExpCstrt    of meta_name list
+and 'expression generic_constraints =
+    CstrFalse
+  | CstrTrue
+  | CstrAnd of 'expression generic_constraints list
+  | CstrOr of 'expression generic_constraints list
+  | CstrNot of 'expression generic_constraints
+  | CstrConstant of constant_constraint
+  | CstrOperator of operator_constraint
+  | CstrMeta_name of meta_name
+  | CstrRegexp of string * Regexp.regexp
+  | CstrScript of bool (*true if immediately evaluable*) * script_constraint
+  | CstrExpr of 'expression
+  | CstrSub of meta_name list
+  | CstrType of fullType
 
-(* Constraints on Meta-* Identifiers, Functions *)
-and idconstraint =
-    IdNoConstraint
-  | IdPosIdSet         of string list * meta_name list
-  | IdNegIdSet         of string list * meta_name list
-  | IdRegExpConstraint of reconstraint
+and operator_constraint =
+    CstrAssignOp of assignOp
+  | CstrBinaryOp of binaryOp
 
-and reconstraint =
-  | IdRegExp        of string * Regexp.regexp
-  | IdNotRegExp     of string * Regexp.regexp
+and constant_constraint =
+    CstrInt of int_constraint
+  | CstrString of string
 
-and assignOpconstraint =
-    AssignOpNoConstraint
-  | AssignOpInSet of assignOp list
+and int_constraint =
+    CstrIntEq of int
+  | CstrIntLeq of int
+  | CstrIntGeq of int
 
-and binaryOpconstraint =
-    BinaryOpNoConstraint
-  | BinaryOpInSet of binaryOp list
+and constraints = expression generic_constraints
+
+and script_constraint =
+      string (* name of generated function *) *
+	string (* language *) *
+	(meta_name * metavar) list (* params *) *
+	script_position *
+	string (* code *)
 
 (* ANY = int E; ID = idexpression int X; CONST = constant int X; *)
 and form = ANY | ID | LocalID | GlobalID | CONST (* form for MetaExp *)
@@ -244,7 +266,7 @@ and form = ANY | ID | LocalID | GlobalID | CONST (* form for MetaExp *)
 and expression = base_expression wrap
 
 and listlen =
-    MetaListLen of meta_name mcode * keep_binding * inherited
+    MetaListLen of meta_name mcode * constraints * keep_binding * inherited
   | CstListLen of int
   | AnyListLen
 
@@ -253,13 +275,14 @@ and base_string_fragment =
   | FormatFragment of string mcode (*%*) * string_format (* format *)
   | Strdots of string mcode
   | MetaFormatList of string mcode (*%*) * meta_name mcode * listlen *
-	keep_binding * inherited
+	constraints * keep_binding * inherited
 
 and string_fragment = base_string_fragment wrap
 
 and base_string_format =
     ConstantFormat of string mcode
-  | MetaFormat of meta_name mcode * idconstraint * keep_binding * inherited
+  | MetaFormat of
+      meta_name mcode * constraints * keep_binding * inherited
 
 and string_format = base_string_format wrap
 
@@ -268,7 +291,7 @@ and base_assignOp =
     SimpleAssign of simpleAssignOp mcode
   | OpAssign of arithOp mcode
   | MetaAssign of
-      meta_name mcode * assignOpconstraint * keep_binding * inherited
+      meta_name mcode * constraints * keep_binding * inherited
 and simpleAssignOp = string
 and assignOp = base_assignOp wrap
 and fixOp = Dec | Inc
@@ -277,7 +300,7 @@ and base_binaryOp =
     Arith of arithOp mcode
   | Logical of logicalOp mcode
   | MetaBinary of
-      meta_name mcode * binaryOpconstraint * keep_binding * inherited
+      meta_name mcode * constraints * keep_binding * inherited
 and binaryOp = base_binaryOp wrap
 and arithOp =
     Plus | Minus | Mul | Div | Mod | DecLeft | DecRight | And | Or | Xor
@@ -298,7 +321,8 @@ and base_fullType =
     Type            of bool (* true if all minus *) *
 	               const_vol mcode option * typeC
   | AsType          of fullType * fullType (* as type, always metavar *)
-  | DisjType        of fullType list (* only after iso *)
+  | DisjType        of fullType list
+  | ConjType        of fullType list
   | OptType         of fullType
 
 and base_typeC =
@@ -319,10 +343,15 @@ and base_typeC =
 	string mcode (* { *) * expression dots * string mcode (* } *)
   | StructUnionName of structUnion mcode * ident option (* name *)
   | StructUnionDef  of fullType (* either StructUnionName or metavar *) *
-	string mcode (* { *) * annotated_decl dots * string mcode (* } *)
+	string mcode (* { *) * annotated_field dots * string mcode (* } *)
+  | TypeOfExpr      of string mcode (* sizeof *) * string mcode (* ( *) *
+                       expression * string mcode (* ) *)
+  | TypeOfType      of string mcode (* sizeof *) * string mcode (* ( *) *
+                       fullType * string mcode (* ) *)
   | TypeName        of string mcode (* pad: should be 'of ident' ? *)
 
-  | MetaType        of meta_name mcode * keep_binding * inherited
+  | MetaType        of meta_name mcode * constraints * keep_binding *
+	inherited
 
 and fullType = base_fullType wrap
 and typeC = base_typeC wrap
@@ -331,6 +360,7 @@ and baseType = VoidType | CharType | ShortType | ShortIntType | IntType
 | DoubleType | LongDoubleType | FloatType
 | LongType | LongIntType | LongLongType | LongLongIntType
 | SizeType | SSizeType | PtrDiffType
+| BoolType | Unknown
 
 and structUnion = Struct | Union
 
@@ -344,9 +374,10 @@ and const_vol = Const | Volatile
    split out into multiple declarations of a single variable each. *)
 
 and base_declaration =
-    Init of storage mcode option * fullType * ident * string mcode (*=*) *
-	initialiser * string mcode (*;*)
-  | UnInit of storage mcode option * fullType * ident * string mcode (* ; *)
+    Init of storage mcode option * fullType * ident * attr list *
+	string mcode (*=*) * initialiser * string mcode (*;*)
+  | UnInit of storage mcode option * fullType * ident * attr list *
+	string mcode (* ; *)
   | FunProto of
 	fninfo list * ident (* name *) *
 	string mcode (* ( *) * parameter_list *
@@ -363,9 +394,8 @@ and base_declaration =
   | Typedef of string mcode (*typedef*) * fullType *
                typeC (* either TypeName or metavar *) * string mcode (*;*)
   | DisjDecl of declaration list
-  | MetaDecl of meta_name mcode * keep_binding * inherited
-  | MetaField of meta_name mcode * keep_binding * inherited
-  | MetaFieldList of meta_name mcode * listlen * keep_binding * inherited
+  | ConjDecl of declaration list
+  | MetaDecl of meta_name mcode * constraints * keep_binding * inherited
   | AsDecl        of declaration * declaration
 
   | OptDecl    of declaration
@@ -375,17 +405,39 @@ and declaration = base_declaration wrap
 and base_annotated_decl =
     DElem of mcodekind (* before the decl *) * bool (* true if all minus *) *
       declaration
-  (* Ddots is for a structure declaration *)
-  | Ddots    of string mcode (* ... *) * declaration option (* whencode *)
 
 and annotated_decl = base_annotated_decl wrap
 
+(* --------------------------------------------------------------------- *)
+(* Field declaration *)
+
+and base_field =
+    Field of fullType * ident option * bitfield option * string mcode (* ; *)
+  | DisjField of field list
+  | ConjField of field list
+  | OptField of field
+  | MetaField of meta_name mcode * constraints * keep_binding * inherited
+  | MetaFieldList of meta_name mcode * listlen * constraints * keep_binding *
+	inherited
+
+and bitfield = string mcode (* : *) * expression
+
+and field = base_field wrap
+
+and base_annotated_field =
+    FElem of mcodekind (* before the decl *) * bool (* true if all minus *) *
+      field
+  | Fdots    of string mcode (* ... *) * field option (* whencode *)
+
+and annotated_field = base_annotated_field wrap
+
 (* --------------------------------------------------------------------- *)
 (* Initializers *)
 
 and base_initialiser =
-    MetaInit of meta_name mcode * keep_binding * inherited
-  | MetaInitList of meta_name mcode * listlen * keep_binding * inherited
+    MetaInit of meta_name mcode * constraints * keep_binding * inherited
+  | MetaInitList of meta_name mcode * listlen * constraints * keep_binding *
+	inherited
   | AsInit of initialiser * initialiser (* as init, always metavar *)
   | InitExpr of expression
   | ArInitList of string mcode (*{*) * initialiser dots * string mcode (*}*)
@@ -418,8 +470,9 @@ and base_parameterTypeDef =
     VoidParam     of fullType
   | Param         of fullType * ident option
 
-  | MetaParam     of meta_name mcode * keep_binding * inherited
-  | MetaParamList of meta_name mcode * listlen * keep_binding * inherited
+  | MetaParam     of meta_name mcode * constraints * keep_binding * inherited
+  | MetaParamList of meta_name mcode * listlen * constraints * keep_binding *
+	inherited
 
   | AsParam       of parameterTypeDef * expression (* expr, always metavar *)
 
@@ -438,6 +491,8 @@ and parameter_list = parameterTypeDef dots
 
 and base_define_param =
     DParam        of ident
+  | MetaDParamList of meta_name mcode * listlen * constraints * keep_binding *
+	inherited
   | DPComma       of string mcode
   | DPdots        of string mcode (* ... *)
   | OptDParam     of define_param
@@ -457,7 +512,7 @@ and define_parameters = base_define_parameters wrap
 and meta_collect = PER | ALL
 
 and meta_pos =
-    MetaPos of meta_name mcode * meta_name list *
+    MetaPos of meta_name mcode * constraints *
       meta_collect * keep_binding * inherited
 
 (* --------------------------------------------------------------------- *)
@@ -508,10 +563,11 @@ and base_rule_elem =
   | Exec          of string mcode (* EXEC *) * string mcode (* language *) *
 	             exec_code dots * string mcode (* ; *)
 
-  | MetaRuleElem  of meta_name mcode * keep_binding * inherited
-  | MetaStmt      of meta_name mcode * keep_binding * metaStmtInfo *
-	             inherited
-  | MetaStmtList  of meta_name mcode * listlen * keep_binding * inherited
+  | MetaRuleElem  of meta_name mcode * constraints * keep_binding * inherited
+  | MetaStmt      of meta_name mcode * constraints * keep_binding *
+	metaStmtInfo * inherited
+  | MetaStmtList  of meta_name mcode * listlen * constraints * keep_binding *
+	inherited
 
   | Exp           of expression (* matches a subterm *)
   | TopExp        of expression (* for macros body, exp at top level,
@@ -520,6 +576,7 @@ and base_rule_elem =
   | TopId         of ident (* only at top level *)
   | TopInit       of initialiser (* only at top level *)
   | Include       of string mcode (*#include*) * inc_file mcode (*file *)
+  | MetaInclude   of string mcode (* #include *) * expression (* file *)
   | Undef         of string mcode (* #define *) * ident (* name *)
   | DefineHeader  of string mcode (* #define *) * ident (* name *) *
 	             define_parameters (*params*)
@@ -544,7 +601,9 @@ and fninfo =
     FStorage of storage mcode
   | FType of fullType
   | FInline of string mcode
-  | FAttr of string mcode
+  | FAttr of attr
+
+and attr = string mcode
 
 and metaStmtInfo =
     NotSequencible | SequencibleAfterDots of dots_whencode list | Sequencible
@@ -620,6 +679,7 @@ and exec_code = base_exec_code wrap
 and inc_file =
     Local of inc_elem list
   | NonLocal of inc_elem list
+  | AnyInc
 
 and inc_elem =
     IncPath of string
@@ -659,16 +719,16 @@ and rule =
       string (*language*) * dependency *
 	(script_meta_name * meta_name * metavar * mvinit)
 	  list (*inherited vars*) *
-	meta_name list (*script vars*) * string
+	meta_name list (*script vars*) * script_position * string
   | InitialScriptRule of  string (* name *) *
 	string (*language*) * dependency *
 	(script_meta_name * meta_name * metavar * mvinit)
-	  list (*virtual vars*) *
+	  list (*virtual vars*) * script_position *
 	string (*code*)
   | FinalScriptRule of  string (* name *) *
 	string (*language*) * dependency *
 	(script_meta_name * meta_name * metavar * mvinit)
-	  list (*virtual vars*) *
+	  list (*virtual vars*) * script_position *
 	string (*code*)
 
 and script_meta_name = string option (*string*) * string option (*ast*)
@@ -678,14 +738,18 @@ and mvinit =
   | MVInitString of string
   | MVInitPosList
 
-and dependency =
+and dep =
     Dep of string (* rule applies for the current binding *)
   | AntiDep of string (* rule doesn't apply for the current binding *)
   | EverDep of string (* rule applies for some binding *)
   | NeverDep of string (* rule never applies for any binding *)
-  | AndDep of dependency * dependency
-  | OrDep of dependency * dependency
-  | NoDep | FailDep
+  | AndDep of dep * dep
+  | OrDep of dep * dep
+  | FileIn of string
+  | NotFileIn of string
+
+and dependency =
+    NoDep | FailDep | ExistsDep of dep | ForallDep of dep
 
 and rule_with_metavars = metavar list * rule
 
@@ -706,6 +770,7 @@ and anything =
   | ArithOpTag          of arithOp
   | LogicalOpTag        of logicalOp
   | DeclarationTag      of declaration
+  | FieldTag            of field
   | InitTag             of initialiser
   | StorageTag          of storage
   | IncFileTag          of inc_file
@@ -722,6 +787,8 @@ and anything =
   | ParamDotsTag        of parameterTypeDef dots
   | StmtDotsTag         of statement dots
   | AnnDeclDotsTag      of annotated_decl dots
+  | AnnFieldDotsTag     of annotated_field dots
+  | DefParDotsTag       of define_param dots
   | TypeCTag            of typeC
   | ParamTag            of parameterTypeDef
   | SgrepStartTag       of string
@@ -757,9 +824,12 @@ let get_fvs x              = x.free_vars
 let set_fvs fvs x          = {x with free_vars = fvs}
 let get_mfvs x             = x.minus_free_vars
 let set_mfvs mfvs x        = {x with minus_free_vars = mfvs}
+let get_minus_nc_fvs x     = x.minus_nc_free_vars
 let get_fresh x            = x.fresh_vars
 let get_inherited x        = x.inherited
 let get_inherited_pos x    = x.positive_inherited_positions
+let get_constraints x      = x.constraints
+let add_constraint x c     = {x with constraints = c::x.constraints}
 let get_saved x            = x.saved_witness
 let get_dots_bef_aft x     = x.bef_aft
 let set_dots_bef_aft d x   = {x with bef_aft = d}
@@ -788,38 +858,39 @@ let get_wcfvs (whencode : ('a wrap, 'b wrap) whencode list) =
 (* --------------------------------------------------------------------- *)
 
 let get_meta_name = function
-    MetaMetaDecl(ar,nm) -> nm
-  | MetaIdDecl(ar,nm) -> nm
+    MetaMetaDecl(_ar,nm) -> nm
+  | MetaIdDecl(_ar,nm) -> nm
   | MetaFreshIdDecl(nm,seed) -> nm
-  | MetaTypeDecl(ar,nm) -> nm
-  | MetaInitDecl(ar,nm) -> nm
-  | MetaInitListDecl(ar,nm,nm1) -> nm
+  | MetaTypeDecl(_ar,nm) -> nm
+  | MetaInitDecl(_ar,nm) -> nm
+  | MetaInitListDecl(_ar,nm,nm1) -> nm
   | MetaListlenDecl(nm) -> nm
-  | MetaParamDecl(ar,nm) -> nm
-  | MetaParamListDecl(ar,nm,nm1) -> nm
+  | MetaParamDecl(_ar,nm) -> nm
+  | MetaParamListDecl(_ar,nm,nm1) -> nm
   | MetaBinaryOperatorDecl(_,name) -> name
   | MetaAssignmentOperatorDecl(_,name) -> name
-  | MetaConstDecl(ar,nm,ty) -> nm
-  | MetaErrDecl(ar,nm) -> nm
-  | MetaExpDecl(ar,nm,ty) -> nm
-  | MetaIdExpDecl(ar,nm,ty) -> nm
-  | MetaLocalIdExpDecl(ar,nm,ty) -> nm
-  | MetaGlobalIdExpDecl(ar,nm,ty) -> nm
-  | MetaExpListDecl(ar,nm,nm1) -> nm
-  | MetaDeclDecl(ar,nm) -> nm
-  | MetaFieldDecl(ar,nm) -> nm
-  | MetaFieldListDecl(ar,nm,nm1) -> nm
-  | MetaStmDecl(ar,nm) -> nm
-  | MetaStmListDecl(ar,nm,nm1) -> nm
-  | MetaFuncDecl(ar,nm) -> nm
-  | MetaLocalFuncDecl(ar,nm) -> nm
-  | MetaPosDecl(ar,nm) -> nm
-  | MetaFmtDecl(ar,nm) -> nm
-  | MetaFragListDecl(ar,nm,nm1) -> nm
-  | MetaAnalysisDecl(code,nm) -> nm
-  | MetaDeclarerDecl(ar,nm) -> nm
-  | MetaIteratorDecl(ar,nm) -> nm
-  | MetaScriptDecl(ar,nm) -> nm
+  | MetaConstDecl(_ar,nm,_ty) -> nm
+  | MetaErrDecl(_ar,nm) -> nm
+  | MetaExpDecl(_ar,nm,_ty,_bitfield) -> nm
+  | MetaIdExpDecl(_ar,nm,_ty) -> nm
+  | MetaLocalIdExpDecl(_ar,nm,_ty) -> nm
+  | MetaGlobalIdExpDecl(_ar,nm,_ty) -> nm
+  | MetaExpListDecl(_ar,nm,_nm1) -> nm
+  | MetaDeclDecl(_ar,nm) -> nm
+  | MetaFieldDecl(_ar,nm) -> nm
+  | MetaFieldListDecl(_ar,nm,_nm1) -> nm
+  | MetaStmDecl(_ar,nm) -> nm
+  | MetaStmListDecl(_ar,nm,nm1) -> nm
+  | MetaDParamListDecl(_ar,nm,nm1) -> nm
+  | MetaFuncDecl(_ar,nm) -> nm
+  | MetaLocalFuncDecl(_ar,nm) -> nm
+  | MetaPosDecl(_ar,nm) -> nm
+  | MetaFmtDecl(_ar,nm) -> nm
+  | MetaFragListDecl(_ar,nm,_nm1) -> nm
+  | MetaAnalysisDecl(_code,nm) -> nm
+  | MetaDeclarerDecl(_ar,nm) -> nm
+  | MetaIteratorDecl(_ar,nm) -> nm
+  | MetaScriptDecl(_ar,nm) -> nm
 
 (* --------------------------------------------------------------------- *)
 
@@ -840,6 +911,7 @@ and tag2c = function
   | ArithOpTag _   -> "ArithOpTag"
   | LogicalOpTag _ -> "LogicalOpTag"
   | DeclarationTag _ -> "DeclarationTag"
+  | FieldTag _ -> "FieldTag"
   | InitTag _      -> "InitTag"
   | StorageTag _   -> "StorageTag"
   | IncFileTag _   -> "IncFileTag"
@@ -856,6 +928,8 @@ and tag2c = function
   | ParamDotsTag _ -> "ParamDotsTag"
   | StmtDotsTag _ -> "StmtDotsTag"
   | AnnDeclDotsTag _ -> "AnnDeclDotsTag"
+  | AnnFieldDotsTag _ -> "AnnFieldDotsTag"
+  | DefParDotsTag _ -> "DefParDotsTag"
   | TypeCTag _ -> "TypeCTag"
   | ParamTag _ -> "ParamTag"
   | SgrepStartTag _ -> "SgrepStartTag"
@@ -871,9 +945,11 @@ let make_term x =
     node_line = 0;
     free_vars = [];
     minus_free_vars = [];
+    minus_nc_free_vars = [];
     fresh_vars = [];
     inherited = [];
     positive_inherited_positions = [];
+    constraints = [];
     saved_witness = [];
     bef_aft = NoDots;
     pos_info = None;
@@ -886,9 +962,11 @@ let make_inherited_term x inherited inh_pos =
     node_line = 0;
     free_vars = [];
     minus_free_vars = [];
+    minus_nc_free_vars = [];
     fresh_vars = [];
     inherited = inherited;
     positive_inherited_positions = inh_pos;
+    constraints = [];
     saved_witness = [];
     bef_aft = NoDots;
     pos_info = None;
@@ -896,16 +974,16 @@ let make_inherited_term x inherited inh_pos =
     safe_for_multi_decls = false;
     iso_info = [] }
 
-let make_meta_rule_elem s d (fvs,fresh,inh) =
+let make_meta_rule_elem s d c (fvs,fresh,inh) =
   let rule = "" in
   {(make_term
-      (MetaRuleElem(((rule,s),no_info,d,[]),Type_cocci.Unitary,false)))
+      (MetaRuleElem(((rule,s),no_info,d,[]),c,Unitary,false)))
   with free_vars = fvs; fresh_vars = fresh; inherited = inh}
 
-let make_meta_decl s d (fvs,fresh,inh) =
+let make_meta_decl s d c (fvs,fresh,inh) =
   let rule = "" in
   {(make_term
-      (MetaDecl(((rule,s),no_info,d,[]),Type_cocci.Unitary,false))) with
+      (MetaDecl(((rule,s),no_info,d,[]),c,Unitary,false))) with
     free_vars = fvs; fresh_vars = fresh; inherited = inh}
 
 let make_mcode x = (x,no_info,CONTEXT(NoPos,NOTHING),[])
@@ -915,6 +993,8 @@ let make_mcode x = (x,no_info,CONTEXT(NoPos,NOTHING),[])
 let equal_pos x y = x = y
 
 (* --------------------------------------------------------------------- *)
+(* used for error messages in type_infer and for printing type constraints in
+   pretty_print_cocci.ml *)
 
 let string_of_arithOp = function
   | Plus -> "+"
@@ -951,3 +1031,435 @@ let string_of_assignOp op = match (unwrap op) with
     let s = string_of_arithOp (unwrap_mcode op') in
     s ^ "="
   | MetaAssign _ -> "MetaAssign"
+
+let string_of_sign = function
+    Signed -> "signed"
+  | Unsigned -> "unsigned"
+
+let string_of_baseType = function
+    VoidType -> "void"
+  | CharType -> "char"
+  | ShortType -> "short"
+  | ShortIntType -> "short int"
+  | IntType -> "int"
+  | DoubleType -> "double"
+  | LongDoubleType -> "long double"
+  | FloatType -> "float"
+  | LongType -> "long"
+  | LongIntType -> "long int"
+  | LongLongType -> "long long"
+  | LongLongIntType -> "long long int"
+  | SizeType -> "size_t"
+  | SSizeType -> "ssize_t"
+  | PtrDiffType -> "ptrdiff_t"
+  | BoolType -> "bool"
+  | Unknown -> "unknown"
+
+let string_of_const_vol = function
+    Const -> "const"
+  | Volatile -> "volatile"
+
+let string_of_meta_name (_, name) = name
+
+let rec string_of_ident id =
+  match unwrap id with
+    Id id -> unwrap_mcode id
+  | MetaId (m, _, _, _)
+  | MetaFunc (m, _, _, _)
+  | MetaLocalFunc (m, _, _, _) -> string_of_meta_name (unwrap_mcode m)
+  | AsIdent (id, _) -> string_of_ident id
+  | OptIdent id -> string_of_ident id ^ "?"
+  | DisjId l -> String.concat "|" (List.map string_of_ident l)
+
+let string_of_expression e =
+  match unwrap e with
+    Ident id -> string_of_ident id
+  | _ -> "?"
+
+let string_of_structUnion = function
+    Struct -> "struct"
+  | Union -> "union"
+
+let rec string_of_typeC ty =
+  match unwrap ty with
+    BaseType (bt, _) -> string_of_baseType bt ^ " "
+  | SignedT (sign, ty') ->
+      let ssign = string_of_sign (unwrap_mcode sign) in
+      ssign ^ " " ^ Common.default "" string_of_typeC ty'
+  | Pointer (ty', _) ->
+      string_of_fullType ty' ^ "*"
+  | FunctionPointer (ty', _, _, _, _, _, _) ->
+      string_of_fullType ty' ^ "(*)()"
+  | Array (ty', _, _, _) ->
+      string_of_fullType ty' ^ "[]"
+  | Decimal(_, _, e0, _, e1, _) ->
+      let s0 = string_of_expression e0
+      and s1 = Common.default "?" string_of_expression e1 in
+      Printf.sprintf "decimal(%s,%s) " s0 s1
+  | EnumName (_, name) -> "enum " ^ (Common.default "?" string_of_ident name)
+  | StructUnionName (kind, name) ->
+      Printf.sprintf "%s %s"
+	(string_of_structUnion (unwrap_mcode kind))
+	(Common.default "?" string_of_ident name)
+  | EnumDef (ty', _, _, _)
+  | StructUnionDef (ty', _, _, _) -> string_of_fullType ty'
+  | TypeOfExpr(_,_,e,_) -> "typeof("^string_of_expression e^")"
+  | TypeOfType(_,_,t,_) -> "typeof("^string_of_fullType t^")"
+  | TypeName (name) -> unwrap_mcode name ^ " "
+  | MetaType (m, _, _, _) -> string_of_meta_name (unwrap_mcode m) ^ " "
+and string_of_fullType ty =
+  match unwrap ty with
+    Type (_, None, ty') -> string_of_typeC ty'
+  | Type (_, Some const_vol, ty') ->
+      string_of_const_vol (unwrap_mcode const_vol) ^ " " ^ string_of_typeC ty'
+  | AsType (ty', _) -> string_of_fullType ty'
+  | DisjType l -> String.concat "|" (List.map string_of_fullType l)
+  | ConjType l -> String.concat "&" (List.map string_of_fullType l)
+  | OptType ty -> string_of_fullType ty ^ "?"
+
+let typeC_of_fullType_opt ty =
+  match unwrap ty with
+    Type (_, None, ty') -> Some ty'
+  | _ -> None
+
+let ident_of_expression_opt expression =
+  match unwrap expression with
+    Ident ident -> Some ident
+  | _ -> None
+
+type 'a transformer = {
+    baseType: (baseType -> string mcode list -> 'a) option;
+    decimal: (string mcode -> string mcode -> expression ->
+      string mcode option -> expression option -> string mcode -> 'a) option;
+    enumName: (string mcode -> ident option -> 'a) option;
+    structUnionName: (structUnion mcode -> ident option -> 'a) option;
+    typeName: (string mcode -> 'a) option;
+    metaType: (meta_name mcode -> constraints -> keep_binding ->
+      inherited -> 'a) option
+  }
+
+let empty_transformer = {
+  baseType = None;
+  decimal = None;
+  enumName = None;
+  structUnionName = None;
+  typeName = None;
+  metaType = None
+}
+
+let rec fullType_map tr ty =
+  rewrap ty begin
+    match unwrap ty with
+      Type (a, b, ty') -> Type (a, b, typeC_map tr ty')
+    | AsType (ty0, ty1) ->
+        AsType (fullType_map tr ty0, fullType_map tr ty1)
+    | DisjType l -> DisjType (List.map (fullType_map tr) l)
+    | ConjType l -> ConjType (List.map (fullType_map tr) l)
+    | OptType ty' -> OptType (fullType_map tr ty')
+  end
+and typeC_map tr ty =
+  match unwrap ty with
+    BaseType (ty', s) ->
+      begin
+        match tr.baseType with
+          None -> ty
+        | Some f -> rewrap ty (f ty' s)
+      end
+  | Pointer (ty', s) -> rewrap ty (Pointer (fullType_map tr ty', s))
+  | FunctionPointer (ty, s0, s1, s2, s3, s4, s5) ->
+      rewrap ty (FunctionPointer (fullType_map tr ty, s0, s1, s2, s3, s4, s5))
+  | Array (ty', s0, s1, s2) ->
+      rewrap ty (Array (fullType_map tr ty', s0, s1, s2))
+  | EnumName (s0, ident) ->
+      begin
+        match tr.enumName with
+          None -> ty
+        | Some f -> rewrap ty (f s0 ident)
+      end
+  | StructUnionName (su, ident) ->
+      begin
+        match tr.structUnionName with
+          None -> ty
+        | Some f -> rewrap ty (f su ident)
+      end
+  | TypeOfExpr(_,_,_,_) -> ty
+  | TypeOfType(tf,lp,t,rp) ->
+      rewrap ty (TypeOfType(tf,lp,fullType_map tr t,rp))
+  | TypeName name ->
+      begin
+        match tr.typeName with
+          None -> ty
+        | Some f -> rewrap ty (f name)
+      end
+  | MetaType (name, cstr, keep, inherited) ->
+      begin
+        match tr.metaType with
+          None -> ty
+        | Some f -> rewrap ty (f name cstr keep inherited)
+      end
+  | Decimal (s0, s1, e0, s2, e1, s3) ->
+      begin
+        match tr.decimal with
+          None -> ty
+        | Some f -> rewrap ty (f s0 s1 e0 s2 e1 s3)
+      end
+  | SignedT (_, None) -> ty
+  | SignedT (sgn, Some ty') ->
+      rewrap ty (SignedT (sgn, Some (typeC_map tr ty')))
+  | EnumDef (ty', s0, e, s1) ->
+      rewrap ty (EnumDef (fullType_map tr ty', s0, e, s1))
+  | StructUnionDef (ty', s0, a, s1) ->
+      rewrap ty (StructUnionDef (fullType_map tr ty', s0, a, s1))
+
+let rec fullType_fold tr ty v =
+  match unwrap ty with
+    Type (_, _, ty') -> typeC_fold tr ty' v
+  | AsType (ty0, ty1) ->
+      let v' = fullType_fold tr ty0 v in
+      fullType_fold tr ty1 v'
+  | DisjType l -> List.fold_left (fun v' ty' -> fullType_fold tr ty' v') v l
+  | ConjType l -> List.fold_left (fun v' ty' -> fullType_fold tr ty' v') v l
+  | OptType ty' -> fullType_fold tr ty' v
+and typeC_fold tr ty v =
+  match unwrap ty with
+    BaseType (ty', s0) -> Common.default v (fun f -> f ty' s0 v) tr.baseType
+  | SignedT (_, None) -> v
+  | SignedT (_, Some ty') -> typeC_fold tr ty' v
+  | Pointer (ty', _)
+  | FunctionPointer (ty', _, _, _, _, _, _)
+  | Array (ty', _, _, _)
+  | EnumDef (ty', _, _, _)
+  | StructUnionDef (ty', _, _, _) -> fullType_fold tr ty' v
+  | Decimal (s0, s1, e0, s2, e1, s3) ->
+      Common.default v (fun f -> f s0 s1 e0 s2 e1 s3 v) tr.decimal
+  | EnumName (s0, ident) -> Common.default v (fun f -> f s0 ident v) tr.enumName
+  | StructUnionName (su, ident) ->
+      Common.default v (fun f -> f su ident v) tr.structUnionName
+  | TypeOfExpr(_,_,e,_) -> v
+  | TypeOfType(_,_,t,_) -> fullType_fold tr t v
+  | TypeName name -> Common.default v (fun f -> f name v) tr.typeName
+  | MetaType (name, cstr, keep, inherited) ->
+      Common.default v (fun f -> f name cstr keep inherited v) tr.metaType
+
+let fullType_iter tr ty =
+  fullType_fold {
+    baseType = Common.map_option (fun f ty' s0 () -> f ty' s0) tr.baseType;
+    decimal = Common.map_option
+      (fun f s0 s1 e0 s2 e1 s3 () -> f s0 s1 e0 s2 e1 s3) tr.decimal;
+    enumName = Common.map_option (fun f s0 ident () -> f s0 ident) tr.enumName;
+    structUnionName = Common.map_option
+      (fun f su ident () -> f su ident) tr.structUnionName;
+    typeName = Common.map_option (fun f name () -> f name) tr.typeName;
+    metaType = Common.map_option
+      (fun f name cstr keep inherited () -> f name cstr keep inherited)
+    tr.metaType
+  } ty ()
+
+let rec ident_fold_meta_names f ident v =
+  match unwrap ident with
+    Id _ -> v
+  | MetaId (tyname, _, _, _)
+  | MetaFunc (tyname, _, _, _)
+  | MetaLocalFunc (tyname, _, _, _) -> f (unwrap_mcode tyname) v
+  | AsIdent (ident0, ident1) ->
+      let v' = ident_fold_meta_names f ident0 v in
+      ident_fold_meta_names f ident1 v'
+  | DisjId l ->
+      List.fold_left (fun v' ident' -> ident_fold_meta_names f ident' v') v l
+  | OptIdent ident' -> ident_fold_meta_names f ident' v
+
+let expression_fold_ident f e v = f (Common.just (ident_of_expression_opt e)) v
+
+let fullType_fold_meta_names f ty v =
+  let enumOrStructUnionName _ ident v =
+    Common.default v (fun ident' -> ident_fold_meta_names f ident' v) ident in
+  fullType_fold { empty_transformer with
+    decimal = Some (fun _ _ e1 _ e2 _ v ->
+      let v' = expression_fold_ident (ident_fold_meta_names f) e1 v in
+      Common.default v'
+	(fun e -> expression_fold_ident (ident_fold_meta_names f) e v) e2);
+    enumName = Some enumOrStructUnionName;
+    structUnionName = Some enumOrStructUnionName;
+    metaType = Some (fun tyname _ _ _ v -> f (unwrap_mcode tyname) v)
+  } ty v
+
+let meta_names_of_fullType ty =
+  fullType_fold_meta_names
+    (fun meta_name meta_names -> meta_name :: meta_names) ty []
+
+let string_of_expression e =
+  match unwrap e with
+    Ident id ->
+      begin
+	match unwrap id with
+	  Id m -> Some (unwrap_mcode m)
+	| _ -> None
+      end
+  | _ -> None
+
+type ('expression, 'a) cstr_transformer = {
+    cstr_constant: (constant_constraint -> 'a) option;
+    cstr_operator: (operator_constraint -> 'a) option;
+    cstr_meta_name: (meta_name -> 'a) option;
+    cstr_regexp: (string -> Regexp.regexp -> 'a) option;
+    cstr_script: (bool * script_constraint -> 'a) option;
+    cstr_expr: ('expression -> 'a) option;
+    cstr_sub: (meta_name list -> 'a) option;
+    cstr_type: (fullType -> 'a) option;
+  }
+
+let empty_cstr_transformer = {
+  cstr_constant = None;
+  cstr_operator = None;
+  cstr_meta_name = None;
+  cstr_regexp = None;
+  cstr_script = None;
+  cstr_expr = None;
+  cstr_sub = None;
+  cstr_type = None;
+}
+
+let rec cstr_fold_sign pos neg c accu =
+  match c with
+    CstrFalse
+  | CstrTrue -> accu
+  | CstrAnd list | CstrOr list ->
+      List.fold_left (fun accu c' -> cstr_fold_sign pos neg c' accu) accu list
+  | CstrNot c' -> cstr_fold_sign neg pos c' accu
+  | CstrConstant c' ->
+      Common.default accu (fun f -> f c' accu) pos.cstr_constant
+  | CstrOperator c' ->
+      Common.default accu (fun f -> f c' accu) pos.cstr_operator
+  | CstrMeta_name mn ->
+      Common.default accu (fun f -> f mn accu) pos.cstr_meta_name
+  | CstrRegexp (s, re) ->
+      Common.default accu (fun f -> f s re accu) pos.cstr_regexp
+  | CstrScript(local,((_name,_lang,params,_pos,_code) as script_constraint)) ->
+      begin
+	match pos.cstr_script with
+	  None ->
+	    Common.default accu
+	      (fun f ->
+		List.fold_left (fun accu' (mv, _) -> f mv accu') accu params)
+	      pos.cstr_meta_name
+	| Some f -> f (local,script_constraint) accu
+      end
+  | CstrExpr e ->
+      Common.default accu (fun f -> f e accu) pos.cstr_expr
+  | CstrSub l ->
+      begin
+	match pos.cstr_sub with
+	  None ->
+	    Common.default accu
+	      (fun f ->
+		List.fold_left (fun accu' mv -> f mv accu') accu l)
+	      pos.cstr_meta_name
+	| Some f -> f l accu
+      end
+  | CstrType ty -> Common.default accu (fun f -> f ty accu) pos.cstr_type
+
+let cstr_fold transformer c accu =
+  cstr_fold_sign transformer transformer c accu
+
+let cstr_iter transformer c =
+  cstr_fold
+    { cstr_constant =
+      Common.map_option (fun f c' () -> f c') transformer.cstr_constant;
+      cstr_operator =
+      Common.map_option (fun f c' () -> f c') transformer.cstr_operator;
+      cstr_meta_name =
+      Common.map_option (fun f s () -> f s) transformer.cstr_meta_name;
+      cstr_regexp =
+      Common.map_option (fun f s re () -> f s re) transformer.cstr_regexp;
+      cstr_script =
+      Common.map_option (fun f s () -> f s) transformer.cstr_script;
+      cstr_expr =
+      Common.map_option (fun f s () -> f s) transformer.cstr_expr;
+      cstr_sub =
+      Common.map_option (fun f s () -> f s) transformer.cstr_sub;
+      cstr_type =
+      Common.map_option (fun f s () -> f s) transformer.cstr_type; } c ()
+
+let rec cstr_map transformer c =
+  match c with
+    CstrFalse -> CstrFalse
+  | CstrTrue -> CstrTrue
+  | CstrAnd list -> CstrAnd (List.map (cstr_map transformer) list)
+  | CstrOr list -> CstrOr (List.map (cstr_map transformer) list)
+  | CstrNot c' -> CstrNot (cstr_map transformer c')
+  | CstrConstant c' ->
+      Common.default (CstrConstant c') (fun f -> f c') transformer.cstr_constant
+  | CstrOperator c' ->
+      Common.default (CstrOperator c') (fun f -> f c') transformer.cstr_operator
+  | CstrMeta_name mn ->
+      Common.default (CstrMeta_name mn) (fun f -> f mn)
+	transformer.cstr_meta_name
+  | CstrRegexp (s, re) ->
+      Common.default (CstrRegexp (s, re)) (fun f -> f s re)
+	transformer.cstr_regexp
+  | CstrScript (local,script_constraint) ->
+      Common.default (CstrScript (local,script_constraint))
+	(fun f -> f (local,script_constraint))
+	transformer.cstr_script
+  | CstrExpr e ->
+      (* Untransformed expressions are discarded! *)
+      Common.default CstrTrue (fun f -> f e)
+	transformer.cstr_expr
+  | CstrSub l ->
+      Common.default (CstrSub l) (fun f -> f l)
+	transformer.cstr_sub
+  | CstrType ty ->
+      Common.default (CstrType ty) (fun f -> f ty)
+	transformer.cstr_type
+
+let rec cstr_push_not c =
+  match c with
+    CstrAnd list -> CstrAnd (List.map cstr_push_not list)
+  | CstrOr list -> CstrOr (List.map cstr_push_not list)
+  | CstrNot c' -> cstr_push_not_neg c'
+  | _ -> c
+and cstr_push_not_neg c =
+  match c with
+    CstrFalse -> CstrTrue
+  | CstrTrue -> CstrFalse
+  | CstrAnd list -> CstrOr (List.map cstr_push_not_neg list)
+  | CstrOr list -> CstrAnd (List.map cstr_push_not_neg list)
+  | CstrNot c' -> cstr_push_not c'
+  | _ -> CstrNot c
+
+let cstr_meta_names c =
+  cstr_fold
+    { empty_cstr_transformer with
+      cstr_meta_name = Some (fun mn accu -> mn :: accu)
+    } c []
+
+let cstr_pos_meta_names c =
+  cstr_fold_sign
+    { empty_cstr_transformer with
+      cstr_meta_name = Some (fun mn accu -> mn :: accu)
+    } empty_cstr_transformer c []
+
+let filter_merge_variables metavars =
+  let filter_var accu metavar =
+    match metavar with
+      ((Some merge_name, None), ("merge", local_name), _, _) ->
+	(merge_name, local_name) :: accu
+    | _ -> accu in
+  List.fold_left filter_var [] metavars
+
+let prepare_merge_variables final rules =
+  let (merge_names, local_names, _length) =
+    List.fold_left
+      (fun (merge_names, local_names, index) r ->
+	match final r with
+	  Some (rulename, mvs) ->
+	    let merge_vars = filter_merge_variables mvs in
+	    let merge_names' = Array.of_list (List.map fst merge_vars) in
+	    let local_names' = List.map snd merge_vars in
+	    let merge_names =
+	      (rulename, (index, merge_names')) :: merge_names in
+	    let local_names = List.rev_append local_names' local_names in
+	    (merge_names, local_names, index + Array.length merge_names')
+	| _ -> (merge_names, local_names, index))
+      ([], [], 0) rules in
+  (merge_names, Array.of_list (List.rev local_names))
