@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -15,14 +15,17 @@ type info = { line : int; column : int;
               whitespace : string }
 type line = int
 type meta_name = string * string
+type script_position = string (* filename *) * line (* line *)
 type 'a wrap =
     {node : 'a;
       node_line : line;
       free_vars : meta_name list; (*free vars*)
       minus_free_vars : meta_name list; (*minus free vars*)
+      minus_nc_free_vars : meta_name list; (*minus free vars, excluding cstrs*)
       fresh_vars : (meta_name * seed) list; (*fresh vars*)
       inherited : meta_name list; (*inherited vars*)
       positive_inherited_positions : meta_name list;
+      constraints : (meta_name * constraints) list;
       saved_witness : meta_name list; (*witness vars*)
       bef_aft : dots_bef_aft;
       pos_info : meta_name mcode option; (* pos info, try not to duplicate *)
@@ -58,8 +61,9 @@ and dots_bef_aft =
   | AddingBetweenDots of statement * int (*index of let var*)
   | DroppingBetweenDots of statement * int (*index of let var*)
 
-and inherited = Type_cocci.inherited
-and keep_binding = Type_cocci.keep_binding
+and inherited = bool (* true if inherited *)
+and keep_binding = Unitary (* need no info *)
+  | Nonunitary (* need an env entry *) | Saved (* need a witness *)
 and multi = bool (*true if a nest is one or more, false if it is zero or more*)
 
 and end_info =
@@ -80,26 +84,28 @@ and metavar =
   | MetaInitListDecl of arity * meta_name (* name *) * list_len (*len*)
   | MetaListlenDecl of meta_name (* name *)
   | MetaParamDecl of arity * meta_name (* name *)
+  | MetaParamListDecl of arity * meta_name (*name*) * list_len (*len*)
   | MetaBinaryOperatorDecl of arity * meta_name
   | MetaAssignmentOperatorDecl of arity * meta_name
-  | MetaParamListDecl of arity * meta_name (*name*) * list_len (*len*)
   | MetaConstDecl of
-      arity * meta_name (* name *) * Type_cocci.typeC list option
+      arity * meta_name (* name *) * fullType list option
   | MetaErrDecl of arity * meta_name (* name *)
   | MetaExpDecl of
-      arity * meta_name (* name *) * Type_cocci.typeC list option
+      arity * meta_name (* name *) * fullType list option *
+	list_len option (* bitfield *)
   | MetaIdExpDecl of
-      arity * meta_name (* name *) * Type_cocci.typeC list option
+      arity * meta_name (* name *) * fullType list option
   | MetaLocalIdExpDecl of
-      arity * meta_name (* name *) * Type_cocci.typeC list option
+      arity * meta_name (* name *) * fullType list option
   | MetaGlobalIdExpDecl of
-      arity * meta_name (* name *) * Type_cocci.typeC list option
+      arity * meta_name (* name *) * fullType list option
   | MetaExpListDecl of arity * meta_name (*name*) * list_len (*len*)
   | MetaDeclDecl of arity * meta_name (* name *)
   | MetaFieldDecl of arity * meta_name (* name *)
   | MetaFieldListDecl of arity * meta_name (* name *) * list_len (*len*)
   | MetaStmDecl of arity * meta_name (* name *)
   | MetaStmListDecl of arity * meta_name (* name *) * list_len (*len*)
+  | MetaDParamListDecl of arity * meta_name (* name *) * list_len (*len*)
   | MetaFuncDecl of arity * meta_name (* name *)
   | MetaLocalFuncDecl of arity * meta_name (* name *)
   | MetaPosDecl of arity * meta_name (* name *)
@@ -110,7 +116,7 @@ and metavar =
   | MetaIteratorDecl of arity * meta_name (* name *)
   | MetaScriptDecl of metavar option ref * meta_name (* name *)
 
-and list_len = AnyLen | MetaLen of meta_name | CstLen of int
+and list_len = AnyLen | MetaLen of meta_name * constraints | CstLen of int
 
 and seed = NoVal | StringSeed of string | ListSeed of seed_elem list
 and seed_elem = SeedString of string | SeedId of meta_name
@@ -126,9 +132,12 @@ and 'a dots = 'a list wrap
 
 and base_ident =
     Id            of string mcode
-  | MetaId        of meta_name mcode * idconstraint * keep_binding * inherited
-  | MetaFunc      of meta_name mcode * idconstraint * keep_binding * inherited
-  | MetaLocalFunc of meta_name mcode * idconstraint * keep_binding * inherited
+  | MetaId        of
+      meta_name mcode * constraints * keep_binding * inherited
+  | MetaFunc      of
+      meta_name mcode * constraints * keep_binding * inherited
+  | MetaLocalFunc of
+      meta_name mcode * constraints * keep_binding * inherited
   | AsIdent       of ident * ident (* as ident, always metavar *)
 
   | DisjId        of ident list
@@ -175,8 +184,9 @@ and base_expression =
   | MetaErr        of meta_name mcode * constraints * keep_binding *
 	              inherited
   | MetaExpr       of meta_name mcode * constraints * keep_binding *
-	              Type_cocci.typeC list option * form * inherited
-  | MetaExprList   of meta_name mcode * listlen *
+	fullType list option * form * inherited *
+	listlen option (* bitfield *)
+  | MetaExprList   of meta_name mcode * listlen * constraints *
 	              keep_binding * inherited (* only in arg lists *)
   | AsExpr         of expression * expression (* as expr, always metavar *)
   | AsSExpr        of expression * rule_elem (* as expr, always metavar *)
@@ -197,37 +207,49 @@ and base_expression =
 
   | OptExp         of expression
 
-and constraints =
-    NoConstraint
-  | NotIdCstrt     of reconstraint
-  | NotExpCstrt    of expression list
-  | SubExpCstrt    of meta_name list
+and 'expression generic_constraints =
+    CstrFalse
+  | CstrTrue
+  | CstrAnd of 'expression generic_constraints list
+  | CstrOr of 'expression generic_constraints list
+  | CstrNot of 'expression generic_constraints
+  | CstrConstant of constant_constraint
+  | CstrOperator of operator_constraint
+  | CstrMeta_name of meta_name
+  | CstrRegexp of string * Regexp.regexp
+  | CstrScript of bool (*true if immediately evaluable*) * script_constraint
+  | CstrExpr of 'expression
+  | CstrSub of meta_name list
+  | CstrType of fullType
 
-(* Constraints on Meta-* Identifiers, Functions *)
-and idconstraint =
-    IdNoConstraint
-  | IdPosIdSet         of string list * meta_name list
-  | IdNegIdSet         of string list * meta_name list
-  | IdRegExpConstraint of reconstraint
+and constant_constraint =
+    CstrInt of int_constraint
+  | CstrString of string
 
-and reconstraint =
-  | IdRegExp        of string * Regexp.regexp
-  | IdNotRegExp     of string * Regexp.regexp
+and int_constraint =
+    CstrIntEq of int
+  | CstrIntLeq of int
+  | CstrIntGeq of int
 
-and assignOpconstraint =
-    AssignOpNoConstraint
-  | AssignOpInSet of assignOp list
+and operator_constraint =
+    CstrAssignOp of assignOp
+  | CstrBinaryOp of binaryOp
 
-and binaryOpconstraint =
-    BinaryOpNoConstraint
-  | BinaryOpInSet of binaryOp list
+and constraints = expression generic_constraints
+
+and script_constraint =
+      string (* name of generated function *) *
+	string (* language *) *
+	(meta_name * metavar) list (* params *) *
+	script_position *
+	string (* code *)
 
 and form = ANY | ID | LocalID| GlobalID | CONST (* form for MetaExp *)
 
 and expression = base_expression wrap
 
 and listlen =
-    MetaListLen of meta_name mcode * keep_binding * inherited
+    MetaListLen of meta_name mcode * constraints * keep_binding * inherited
   | CstListLen of int
   | AnyListLen
 
@@ -236,13 +258,14 @@ and base_string_fragment =
   | FormatFragment of string mcode (*%*) * string_format (* format *)
   | Strdots of string mcode
   | MetaFormatList of string mcode (*%*) * meta_name mcode * listlen *
-	keep_binding * inherited
+	constraints * keep_binding * inherited
 
 and string_fragment = base_string_fragment wrap
 
 and base_string_format =
     ConstantFormat of string mcode
-  | MetaFormat of meta_name mcode * idconstraint * keep_binding * inherited
+  | MetaFormat of
+      meta_name mcode * constraints * keep_binding * inherited
 
 and string_format = base_string_format wrap
 
@@ -250,7 +273,8 @@ and  unaryOp = GetRef | GetRefLabel | DeRef | UnPlus |  UnMinus | Tilde | Not
 and  base_assignOp =
     SimpleAssign of simpleAssignOp mcode
   | OpAssign of arithOp mcode
-  | MetaAssign of meta_name mcode * assignOpconstraint * keep_binding * inherited
+  | MetaAssign of
+      meta_name mcode * constraints * keep_binding * inherited
 and simpleAssignOp = string
 and assignOp = base_assignOp wrap
 and  fixOp = Dec | Inc
@@ -258,7 +282,8 @@ and  fixOp = Dec | Inc
 and  base_binaryOp =
     Arith of arithOp mcode
   | Logical of logicalOp mcode
-  | MetaBinary of meta_name mcode * binaryOpconstraint * keep_binding * inherited
+  | MetaBinary of
+      meta_name mcode * constraints * keep_binding * inherited
 and binaryOp = base_binaryOp wrap
 and  arithOp =
     Plus | Minus | Mul | Div | Mod | DecLeft | DecRight | And | Or | Xor | Min | Max
@@ -278,7 +303,8 @@ and base_fullType =
     Type            of bool (* true if all minus *) *
                        const_vol mcode option * typeC
   | AsType          of fullType * fullType (* as type, always metavar *)
-  | DisjType        of fullType list (* only after iso *)
+  | DisjType        of fullType list
+  | ConjType        of fullType list
   | OptType         of fullType
 
 and base_typeC =
@@ -299,10 +325,15 @@ and base_typeC =
 	string mcode (* { *) * expression dots * string mcode (* } *)
   | StructUnionName of structUnion mcode * ident option (* name *)
   | StructUnionDef  of fullType (* either StructUnionName or metavar *) *
-	string mcode (* { *) * annotated_decl dots * string mcode (* } *)
+	string mcode (* { *) * annotated_field dots * string mcode (* } *)
+  | TypeOfExpr      of string mcode (* typeof *) * string mcode (* ( *) *
+                       expression * string mcode (* ) *)
+  | TypeOfType      of string mcode (* typeof *) * string mcode (* ( *) *
+                       fullType * string mcode (* ) *)
   | TypeName        of string mcode
 
-  | MetaType        of meta_name mcode * keep_binding * inherited
+  | MetaType        of meta_name mcode * constraints * keep_binding *
+	inherited
 
 and fullType = base_fullType wrap
 and typeC = base_typeC wrap
@@ -311,6 +342,7 @@ and baseType = VoidType | CharType | ShortType | ShortIntType | IntType
 | DoubleType | LongDoubleType | FloatType
 | LongType | LongIntType | LongLongType | LongLongIntType
 | SizeType | SSizeType | PtrDiffType
+| BoolType | Unknown
 
 and structUnion = Struct | Union
 
@@ -324,9 +356,10 @@ and const_vol = Const | Volatile
    split out into multiple declarations of a single variable each. *)
 
 and base_declaration =
-    Init of storage mcode option * fullType * ident * string mcode (*=*) *
-	initialiser * string mcode (*;*)
-  | UnInit of storage mcode option * fullType * ident * string mcode (* ; *)
+    Init of storage mcode option * fullType * ident * attr list *
+	string mcode (*=*) * initialiser * string mcode (*;*)
+  | UnInit of storage mcode option * fullType * ident * attr list *
+	string mcode (* ; *)
   | FunProto of
 	fninfo list * ident (* name *) *
 	string mcode (* ( *) * parameter_list *
@@ -342,10 +375,8 @@ and base_declaration =
         initialiser * string mcode (* ; *)
   | Typedef of string mcode (*typedef*) * fullType * typeC * string mcode (*;*)
   | DisjDecl   of declaration list
-
-  | MetaDecl of meta_name mcode * keep_binding * inherited
-  | MetaField of meta_name mcode * keep_binding * inherited
-  | MetaFieldList of meta_name mcode * listlen * keep_binding * inherited
+  | ConjDecl of declaration list
+  | MetaDecl of meta_name mcode * constraints * keep_binding * inherited
   | AsDecl        of declaration * declaration
 
   | OptDecl    of declaration
@@ -355,17 +386,39 @@ and declaration = base_declaration wrap
 and base_annotated_decl =
     DElem of mcodekind (* before the decl *) * bool (* true if all minus *) *
       declaration
-  (* Ddots is for a structure declaration *)
-  | Ddots    of string mcode (* ... *) * declaration option (* whencode *)
 
 and annotated_decl = base_annotated_decl wrap
 
+(* --------------------------------------------------------------------- *)
+(* Field declaration *)
+
+and base_field =
+    Field of fullType * ident option * bitfield option * string mcode (* ; *)
+  | DisjField of field list
+  | ConjField of field list
+  | OptField of field
+  | MetaField of meta_name mcode * constraints * keep_binding * inherited
+  | MetaFieldList of meta_name mcode * listlen * constraints * keep_binding *
+	inherited
+
+and bitfield = string mcode (* : *) * expression
+
+and field = base_field wrap
+
+and base_annotated_field =
+    FElem of mcodekind (* before the decl *) * bool (* true if all minus *) *
+      field
+  | Fdots    of string mcode (* ... *) * field option (* whencode *)
+
+and annotated_field = base_annotated_field wrap
+
 (* --------------------------------------------------------------------- *)
 (* Initializers *)
 
 and base_initialiser =
-    MetaInit of meta_name mcode * keep_binding * inherited
-  | MetaInitList of meta_name mcode * listlen * keep_binding * inherited
+    MetaInit of meta_name mcode * constraints * keep_binding * inherited
+  | MetaInitList of meta_name mcode * listlen * constraints * keep_binding
+	* inherited
   | AsInit of initialiser * initialiser (* as init, always metavar *)
   | InitExpr of expression
   | ArInitList of string mcode (*{*) * initialiser dots * string mcode (*}*)
@@ -397,8 +450,9 @@ and base_parameterTypeDef =
     VoidParam     of fullType
   | Param         of fullType * ident option
 
-  | MetaParam     of meta_name mcode * keep_binding * inherited
-  | MetaParamList of meta_name mcode * listlen * keep_binding * inherited
+  | MetaParam     of meta_name mcode * constraints * keep_binding * inherited
+  | MetaParamList of meta_name mcode * listlen * constraints * keep_binding *
+	inherited
 
   | AsParam       of parameterTypeDef * expression (* expr, always metavar *)
 
@@ -417,6 +471,8 @@ and parameter_list = parameterTypeDef dots
 
 and base_define_param =
     DParam        of ident
+  | MetaDParamList of meta_name mcode * listlen * constraints * keep_binding *
+	inherited
   | DPComma       of string mcode
   | DPdots        of string mcode (* ... *)
   | OptDParam     of define_param
@@ -436,7 +492,7 @@ and define_parameters = base_define_parameters wrap
 and meta_collect = PER | ALL
 
 and meta_pos =
-    MetaPos of meta_name mcode * meta_name list *
+    MetaPos of meta_name mcode * constraints *
 	meta_collect * keep_binding * inherited
 
 (* --------------------------------------------------------------------- *)
@@ -487,10 +543,11 @@ and base_rule_elem =
   | Exec          of string mcode (* EXEC *) * string mcode (* language *) *
 	             exec_code dots * string mcode (* ; *)
 
-  | MetaRuleElem  of meta_name mcode * keep_binding * inherited
-  | MetaStmt      of meta_name mcode * keep_binding * metaStmtInfo *
-	             inherited
-  | MetaStmtList  of meta_name mcode * listlen * keep_binding * inherited
+  | MetaRuleElem  of meta_name mcode * constraints * keep_binding * inherited
+  | MetaStmt      of meta_name mcode * constraints * keep_binding *
+	metaStmtInfo * inherited
+  | MetaStmtList  of meta_name mcode * listlen * constraints * keep_binding *
+	inherited
 
   | Exp           of expression
   | TopExp        of expression (* for macros body *)
@@ -498,6 +555,7 @@ and base_rule_elem =
   | TopId         of ident (* only at top level *)
   | TopInit       of initialiser (* only at top level *)
   | Include       of string mcode (*#include*) * inc_file mcode (*file *)
+  | MetaInclude   of string mcode (* #include *) * expression (* file *)
   | Undef         of string mcode (* #define *) * ident (* name *)
   | DefineHeader  of string mcode (* #define *) * ident (* name *) *
 	             define_parameters (*params*)
@@ -522,7 +580,9 @@ and fninfo =
     FStorage of storage mcode
   | FType of fullType
   | FInline of string mcode
-  | FAttr of string mcode
+  | FAttr of attr
+
+and attr = string mcode
 
 and metaStmtInfo =
     NotSequencible | SequencibleAfterDots of dots_whencode list | Sequencible
@@ -593,6 +653,7 @@ and exec_code = base_exec_code wrap
 and inc_file =
     Local of inc_elem list
   | NonLocal of inc_elem list
+  | AnyInc
 
 and inc_elem =
     IncPath of string
@@ -629,14 +690,14 @@ and rule =
   | ScriptRule of string (* name *) *
       string * dependency *
 	(script_meta_name * meta_name * metavar * mvinit) list *
-	meta_name list (*script vars*) * string
+	meta_name list (*script vars*) * script_position * string
   | InitialScriptRule of  string (* name *) * string * dependency *
 	(script_meta_name * meta_name * metavar * mvinit)
-	  list (*virtual vars*) *
+	  list (*virtual vars*) * script_position *
 	string
   | FinalScriptRule of  string (* name *) * string * dependency *
 	(script_meta_name * meta_name * metavar * mvinit)
-	  list (*virtual vars*) *
+	  list (*virtual vars*) * script_position *
 	string
 
 and script_meta_name = string option (*string*) * string option (*ast*)
@@ -645,15 +706,18 @@ and mvinit =
     NoMVInit
   | MVInitString of string
   | MVInitPosList
-
-and dependency =
+and dep =
     Dep of string (* rule applies for the current binding *)
   | AntiDep of string (* rule doesn't apply for the current binding *)
   | EverDep of string (* rule applies for some binding *)
   | NeverDep of string (* rule never applies for any binding *)
-  | AndDep of dependency * dependency
-  | OrDep of dependency * dependency
-  | NoDep | FailDep
+  | AndDep of dep * dep
+  | OrDep of dep * dep
+  | FileIn of string
+  | NotFileIn of string
+
+and dependency =
+    NoDep | FailDep | ExistsDep of dep | ForallDep of dep
 
 and rule_with_metavars = metavar list * rule
 
@@ -674,6 +738,7 @@ and anything =
   | ArithOpTag          of arithOp
   | LogicalOpTag        of logicalOp
   | DeclarationTag      of declaration
+  | FieldTag            of field
   | InitTag             of initialiser
   | StorageTag          of storage
   | IncFileTag          of inc_file
@@ -690,6 +755,8 @@ and anything =
   | ParamDotsTag        of parameterTypeDef dots
   | StmtDotsTag         of statement dots
   | AnnDeclDotsTag      of annotated_decl dots
+  | AnnFieldDotsTag     of annotated_field dots
+  | DefParDotsTag       of define_param dots
   | TypeCTag            of typeC
   | ParamTag            of parameterTypeDef
   | SgrepStartTag       of string
@@ -720,9 +787,12 @@ val get_wcfvs : ('a wrap,'b wrap) whencode list -> meta_name list
 val set_fvs : meta_name list -> 'a wrap -> 'a wrap
 val get_mfvs : 'a wrap -> meta_name list
 val set_mfvs : meta_name list -> 'a wrap -> 'a wrap
+val get_minus_nc_fvs : 'a wrap -> meta_name list
 val get_fresh : 'a wrap -> (meta_name * seed) list
 val get_inherited : 'a wrap -> meta_name list
 val get_inherited_pos : 'a wrap -> meta_name list
+val get_constraints : 'a wrap -> (meta_name * constraints) list
+val add_constraint : 'a wrap -> meta_name * constraints -> 'a wrap
 val get_saved : 'a wrap -> meta_name list
 val get_dots_bef_aft : statement -> dots_bef_aft
 val set_dots_bef_aft : dots_bef_aft -> statement -> statement
@@ -744,12 +814,12 @@ val tag2c : anything -> string
 val no_info : info
 
 val make_meta_rule_elem :
-    string -> mcodekind ->
+    string -> mcodekind -> constraints ->
       (meta_name list * (meta_name * seed) list * meta_name list) ->
       rule_elem
 
 val make_meta_decl :
-    string -> mcodekind ->
+    string -> mcodekind -> constraints ->
       (meta_name list * (meta_name * seed) list * meta_name list) ->
       declaration
 
@@ -764,3 +834,118 @@ val string_of_arithOp : arithOp -> string
 val string_of_logicalOp : logicalOp -> string
 val string_of_assignOp : assignOp -> string
 val string_of_binaryOp : binaryOp -> string
+
+val string_of_sign : sign -> string
+val string_of_baseType : baseType -> string
+val string_of_const_vol : const_vol -> string
+val string_of_structUnion : structUnion -> string
+val string_of_typeC : typeC -> string
+val string_of_fullType : fullType -> string
+
+val typeC_of_fullType_opt : fullType -> typeC option
+(**
+ * [typeC_of_fullType_opt ty] returns [Some ty'] iff [ty] is of the form
+ * [Type (_, false, ty')].
+ *)
+
+val ident_of_expression_opt : expression -> ident option
+(**
+ * [ident_of_expression_opt e] returns [Some id] iff [e] is of the form
+ * [Ident e] *)
+
+type 'a transformer = {
+    baseType: (baseType -> string mcode list -> 'a) option;
+    decimal: (string mcode -> string mcode -> expression ->
+      string mcode option -> expression option -> string mcode -> 'a) option;
+    enumName: (string mcode -> ident option -> 'a) option;
+    structUnionName: (structUnion mcode -> ident option -> 'a) option;
+    typeName: (string mcode -> 'a) option;
+    metaType: (meta_name mcode -> constraints -> keep_binding ->
+      inherited -> 'a) option
+  }
+(** ['a transformer] is an auxiliary record type for the iterators fullType_map,
+ * fullType_fold and fullType_iter.
+ * Each field corresponds to a constructor for a leaf in the AST of types.
+ *)
+
+val empty_transformer: 'a transformer
+(** The transformer with no function defined. *)
+
+
+val fullType_map: base_typeC transformer -> fullType -> fullType
+(**
+ * [fullType_map tr ty] returns [ty] where every subparts that match one of the
+ * kinds, which a function has been given for in [tr], has been substituted by
+ * the result of the corresponding function.
+ * Other subparts, for the kind of which no function has been provided, are
+ * preserved as is.
+ *)
+
+val fullType_fold: ('a -> 'a) transformer -> fullType -> 'a -> 'a
+(**
+  * [fullType_fold tr ty v] iterates the given functions to every subparts of
+  * [ty] that match one of the kinds, which a function has been given for in
+  * [tr], computing the iterated result value from the initial seed v.
+  *)
+
+val fullType_iter: unit transformer -> fullType -> unit
+(**
+  * [fullType_iter tr ty] applies the given functions to every subparts of [ty]
+  * that match one of the kinds, which a function has been given for in [tr].
+  *)
+
+val meta_names_of_fullType: fullType -> meta_name list
+(**
+ * [meta_names_of_fullType ty] enumerates all the meta names that occur in [ty].
+ *)
+
+val string_of_expression: expression -> string option
+(** [string_of_expression e] returns [Some id] if [e] is of the form
+    [Ident (Id id)], [None] otherwise.
+    Unquoted identifiers in constraints are represented as such expressions,
+    and associated strings are extracted when needed. *)
+
+type ('expression, 'a) cstr_transformer = {
+    cstr_constant: (constant_constraint -> 'a) option;
+    cstr_operator: (operator_constraint -> 'a) option;
+    cstr_meta_name: (meta_name -> 'a) option;
+    cstr_regexp: (string -> Regexp.regexp -> 'a) option;
+    cstr_script: (bool*script_constraint -> 'a) option;
+    cstr_expr: ('expression -> 'a) option;
+    cstr_sub: (meta_name list -> 'a) option;
+    cstr_type: (fullType -> 'a) option;
+  }
+
+val empty_cstr_transformer: ('expression, 'a) cstr_transformer
+
+val cstr_fold_sign:
+    ('expression, 'a -> 'a) cstr_transformer ->
+      ('expression, 'a -> 'a) cstr_transformer ->
+      'expression generic_constraints -> 'a -> 'a
+
+val cstr_fold: ('expression, 'a -> 'a) cstr_transformer ->
+  'expression generic_constraints -> 'a -> 'a
+
+val cstr_iter: ('expression, unit) cstr_transformer ->
+  'expression generic_constraints -> unit
+
+val cstr_map:
+    ('expression, 'expression' generic_constraints) cstr_transformer ->
+      'expression generic_constraints ->
+	'expression' generic_constraints
+(* Untransformed expressions are discarded! *)
+
+val cstr_push_not:
+    'expression generic_constraints -> 'expression generic_constraints
+
+val cstr_meta_names: 'expression generic_constraints -> meta_name list
+
+val cstr_pos_meta_names: 'expression generic_constraints -> meta_name list
+
+val filter_merge_variables:
+    (script_meta_name * meta_name * metavar * mvinit) list ->
+      (string * string) list
+
+val prepare_merge_variables:
+    ('a -> ('b * (script_meta_name * meta_name * metavar * mvinit) list) option)
+    -> 'a list -> ('b * (int * string array)) list * string array
