@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -15,7 +15,7 @@ module V0 = Visitor_ast0
 module VT0 = Visitor_ast0_types
 
 (* all fresh identifiers *)
-let fresh_table = (Hashtbl.create(50) : (Ast.meta_name, unit) Hashtbl.t)
+let fresh_table = (Hashtbl.create(51) : (Ast.meta_name, unit) Hashtbl.t)
 
 let warning s = Printf.fprintf stderr "warning: %s\n" s
 
@@ -59,7 +59,7 @@ type context = ID | FIELD | FN | GLOBAL
 
 (* heuristic for distinguishing ifdef variables from undeclared metavariables*)
 let is_ifdef name =
-  String.length name > 2 && String.uppercase name = name
+  String.length name > 2 && Stdcompat.String.uppercase_ascii name = name
 
 let rec ident context old_metas table minus i =
   match Ast0.unwrap i with
@@ -175,21 +175,20 @@ let rec expression context old_metas table minus e =
   | Ast0.TypeExp(ty) -> typeC old_metas table minus ty
   | Ast0.Constructor(lp,ty,rp,init) ->
       typeC old_metas table minus ty; initialiser old_metas table minus init
-  | Ast0.MetaExpr(name,_,Some tys,_,_) ->
+  | Ast0.MetaExpr(name,_,Some tys,_,_,bitfield) ->
       List.iter
 	(function x ->
 	  List.iter
 	    (function ty -> check_table table minus (promote ty))
-	    (get_type_name x))
+	    (Ast0.meta_names_of_typeC x))
 	tys;
-      check_table table minus name
-  | Ast0.MetaExpr(name,_,_,_,_) | Ast0.MetaErr(name,_,_) ->
-      check_table table minus name
-  | Ast0.MetaExprList(name,Ast0.MetaListLen lenname,_) ->
       check_table table minus name;
-      check_table table minus lenname
-  | Ast0.MetaExprList(name,_,_) ->
+      Common.do_option (check_len table minus) bitfield
+  | Ast0.MetaExpr(name,_,_,_,_,_) | Ast0.MetaErr(name,_,_) ->
       check_table table minus name
+  | Ast0.MetaExprList(name,len,_,_) ->
+      check_table table minus name;
+      check_len table minus len
   | Ast0.AsExpr(exp,asexp) -> failwith "not generated yet"
   | Ast0.AsSExpr(exp,asstm) -> failwith "not generated yet"
   | Ast0.DisjExpr(_,exps,_,_)
@@ -204,19 +203,10 @@ let rec expression context old_metas table minus e =
       expression ID old_metas table minus x
   | _ -> () (* no metavariable subterms *)
 
-and get_type_name = function
-    Type_cocci.ConstVol(_,ty) | Type_cocci.SignedT(_,Some ty)
-  | Type_cocci.Pointer(ty)
-  | Type_cocci.FunctionPointer(ty) | Type_cocci.Array(ty) -> get_type_name ty
-  | Type_cocci.EnumName(Type_cocci.MV(nm,_,_)) -> [nm]
-  | Type_cocci.StructUnionName(_,Type_cocci.MV(nm,_,_)) -> [nm]
-  | Type_cocci.MetaType(nm,_,_) -> [nm]
-  | Type_cocci.Decimal(nm1,nm2) ->
-      let get_name = function
-	  Type_cocci.MV(nm,_,_) -> [nm]
-	| _ -> [] in
-      (get_name nm1) @ (get_name nm2)
-  | _ -> []
+and check_len table minus len =
+  match len with
+    Ast0.MetaListLen (lenname, _) -> check_table table minus lenname
+  | _ -> ()
 
 (* --------------------------------------------------------------------- *)
 (* Types *)
@@ -236,10 +226,11 @@ and typeC old_metas table minus t =
   | Ast0.Decimal(dec,lp,length,comma,precision_opt,rp) ->
       expression ID old_metas table minus length;
       get_opt (expression ID old_metas table minus) precision_opt
-  | Ast0.MetaType(name,_) ->
+  | Ast0.MetaType(name,_,_) ->
       check_table table minus name
   | Ast0.AsType(ty,asty) -> failwith "not generated yet"
-  | Ast0.DisjType(_,types,_,_) ->
+  | Ast0.DisjType(_,types,_,_)
+  | Ast0.ConjType(_,types,_,_) ->
       List.iter (typeC old_metas table minus) types
   | Ast0.EnumName(en,Some id) -> ident GLOBAL old_metas table minus id
   | Ast0.EnumDef(ty,lb,ids,rb) ->
@@ -248,7 +239,7 @@ and typeC old_metas table minus t =
   | Ast0.StructUnionName(su,Some id) -> ident GLOBAL old_metas table minus id
   | Ast0.StructUnionDef(ty,lb,decls,rb) ->
       typeC old_metas table minus ty;
-      dots (declaration GLOBAL old_metas table minus) decls
+      dots (field GLOBAL old_metas table minus) decls
   | Ast0.OptType(ty) ->
       failwith "unexpected code"
   | _ -> () (* no metavariable subterms *)
@@ -260,15 +251,10 @@ and typeC old_metas table minus t =
 
 and declaration context old_metas table minus d =
   match Ast0.unwrap d with
-    Ast0.MetaDecl(name,_) | Ast0.MetaField(name,_) ->
-      check_table table minus name
-  | Ast0.MetaFieldList(name,Ast0.MetaListLen lenname,_) ->
-      check_table table minus name;
-      check_table table minus lenname
-  | Ast0.MetaFieldList(name,_,_) ->
+    Ast0.MetaDecl(name,_,_) ->
       check_table table minus name
   | Ast0.AsDecl(decl,asdecl) -> failwith "not generated yet"
-  | Ast0.Init(stg,ty,id,eq,ini,sem) ->
+  | Ast0.Init(stg,ty,id,attr,eq,ini,sem) ->
       typeC old_metas table minus ty;
       ident context old_metas table minus id;
       (match Ast0.unwrap ini with
@@ -281,7 +267,7 @@ and declaration context old_metas table minus d =
 	    failwith "complex initializer specification not allowed in - code"
 	  else*)
 	    initialiser old_metas table minus ini)
-  | Ast0.UnInit(stg,ty,id,sem) ->
+  | Ast0.UnInit(stg,ty,id,attr,sem) ->
       typeC old_metas table minus ty; ident context old_metas table minus id
   | Ast0.FunProto(fi,name,lp1,params,va,rp1,sem) ->
       ident FN old_metas table minus name;
@@ -300,25 +286,45 @@ and declaration context old_metas table minus d =
   | Ast0.Typedef(stg,ty,id,sem) ->
       typeC old_metas table minus ty;
       typeC old_metas table minus id
-  | Ast0.DisjDecl(_,decls,_,_) ->
+  | Ast0.DisjDecl(_,decls,_,_)
+  | Ast0.ConjDecl(_,decls,_,_) ->
       List.iter (declaration ID old_metas table minus) decls
-  | Ast0.Ddots(_,Some (_,_,x)) -> declaration ID old_metas table minus x
-  | Ast0.Ddots(_,None) -> ()
   | Ast0.OptDecl(_) ->
       failwith "unexpected code"
 
+(* --------------------------------------------------------------------- *)
+(* Field declaration *)
+
+and field context old_metas table minus d =
+  match Ast0.unwrap d with
+    Ast0.MetaField(name,_,_) ->
+      check_table table minus name
+  | Ast0.MetaFieldList(name,len,_,_) ->
+      check_table table minus name;
+      check_len table minus len
+  | Ast0.Field(ty,id,bf,sem) ->
+      typeC old_metas table minus ty;
+      Common.do_option (ident context old_metas table minus) id;
+      let bitfield (c, e) = expression context old_metas table minus e in
+      Common.do_option bitfield bf
+  | Ast0.DisjField(_,decls,_,_)
+  | Ast0.ConjField(_,decls,_,_) ->
+      List.iter (field ID old_metas table minus) decls
+  | Ast0.OptField(_) ->
+      failwith "unexpected code"
+  | Ast0.Fdots(_,Some (_,_,x)) -> field ID old_metas table minus x
+  | Ast0.Fdots(_,None) -> ()
+
 (* --------------------------------------------------------------------- *)
 (* Initialiser *)
 
 and initialiser old_metas table minus ini =
   match Ast0.unwrap ini with
-    Ast0.MetaInit(name,_) ->
+    Ast0.MetaInit(name,_,_) ->
       check_table table minus name
-  | Ast0.MetaInitList(name,Ast0.MetaListLen lenname,_) ->
+  | Ast0.MetaInitList(name,len,_,_) ->
       check_table table minus name;
-      check_table table minus lenname
-  | Ast0.MetaInitList(name,_,_) ->
-      check_table table minus name
+      check_len table minus len
   | Ast0.AsInit(ini,asini) -> failwith "not generated yet"
   | Ast0.InitExpr(exp) -> expression ID old_metas table minus exp
   | Ast0.InitList(lb,initlist,rb,ordered) ->
@@ -354,13 +360,11 @@ and parameterTypeDef old_metas table minus param =
     Ast0.Param(ty,id) ->
       get_opt (ident ID old_metas table minus) id;
       typeC old_metas table minus ty
-  | Ast0.MetaParam(name,_) ->
+  | Ast0.MetaParam(name,_,_) ->
       check_table table minus name
-  | Ast0.MetaParamList(name,Ast0.MetaListLen lenname,_) ->
+  | Ast0.MetaParamList(name,len,_,_) ->
       check_table table minus name;
-      check_table table minus lenname
-  | Ast0.MetaParamList(name,_,_) ->
-      check_table table minus name
+      check_len table minus len
   | _ -> () (* no metavariable subterms *)
 
 and parameter_list old_metas table minus =
@@ -375,11 +379,9 @@ and string_fragment old_metas table minus e =
   | Ast0.FormatFragment(pct,fmt) ->
       string_format old_metas table minus fmt
   | Ast0.Strdots dots -> ()
-  | Ast0.MetaFormatList(pct,name,Ast0.MetaListLen lenname) ->
+  | Ast0.MetaFormatList(pct,name,_,len) ->
       check_table table minus name;
-      check_table table minus lenname
-  | Ast0.MetaFormatList(pct,name,lenname) ->
-      check_table table minus name
+      check_len table minus len
 
 and string_format old_metas table minus e =
   match Ast0.unwrap e with
@@ -429,11 +431,10 @@ and statement old_metas table minus s =
   | Ast0.ReturnExpr(ret,exp,sem) -> expression ID old_metas table minus exp
   | Ast0.Exec(exec,lang,code,sem) ->
       dots (exec_code ID old_metas table minus) code
-  | Ast0.MetaStmt(name,_) ->     check_table table minus name
-  | Ast0.MetaStmtList(name,Ast0.MetaListLen lenname,_) ->
+  | Ast0.MetaStmt(name,_,_) ->     check_table table minus name
+  | Ast0.MetaStmtList(name,len,_,_) ->
       check_table table minus name;
-      check_table table minus lenname
-  | Ast0.MetaStmtList(name,_,_) -> check_table table minus name
+      check_len table minus len
   | Ast0.AsStmt(stm,asstm) -> failwith "not generated yet"
   | Ast0.Exp(exp) -> expression ID old_metas table minus exp
   | Ast0.TopExp(exp) -> expression ID old_metas table minus exp
@@ -460,6 +461,7 @@ and statement old_metas table minus s =
       parameter_list old_metas table minus params;
       dots (statement old_metas table minus) body
   | Ast0.Include(inc,s) -> () (* no metavariables possible *)
+  | Ast0.MetaInclude(inc,s) -> expression ID old_metas table minus s
   | Ast0.Undef(def,id) ->
       ident GLOBAL old_metas table minus id
   | Ast0.Define(def,id,params,body) ->
@@ -483,6 +485,9 @@ and pragmainfo old_metas table minus pi =
 and define_param old_metas table minus p =
   match Ast0.unwrap p with
     Ast0.DParam(id) -> ident GLOBAL old_metas table minus id
+  | Ast0.MetaDParamList(name,len,_,_) ->
+      check_table table minus name;
+      check_len table minus len
   | Ast0.DPComma(_) | Ast0.DPdots(_) ->
       () (* no metavariable subterms *)
   | Ast0.OptDParam(dp)    -> define_param old_metas table minus dp
@@ -556,7 +561,7 @@ let positions rname table rules =
 	rmcode name;
 	do_tynames var)
       (Ast0.get_pos x) in
-  let rec mcode x =
+  let mcode x =
     List.iter
       (function var ->
 	let name = Ast0.meta_pos_name var in
@@ -571,9 +576,10 @@ let positions rname table rules =
     V0.flat_combiner bind option_default
       mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
       mcode mcode mcode mcode
-      donothing donothing donothing donothing donothing donothing
       donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing donothing in
+      donothing donothing donothing donothing donothing donothing donothing
+      donothing donothing donothing donothing donothing donothing donothing
+      donothing in
 
   List.iter fn.VT0.combiner_rec_top_level rules
 
@@ -603,18 +609,28 @@ let dup_positions rules =
 
   let typeC r k e = (* not sure relevant because "only after iso" *)
     match Ast0.unwrap e with
-      Ast0.DisjType(_,types,_,_) ->
+      Ast0.DisjType(_,types,_,_)
+    | Ast0.ConjType(_,types,_,_) ->
 	List.fold_left Common.union_set option_default
 	  (List.map r.VT0.combiner_rec_typeC types)
     | _ -> k e in
 
   let declaration r k e =
     match Ast0.unwrap e with
-      Ast0.DisjDecl(_,decls,_,_) ->
+      Ast0.DisjDecl(_,decls,_,_)
+    | Ast0.ConjDecl(_,decls,_,_) ->
 	List.fold_left Common.union_set option_default
 	  (List.map r.VT0.combiner_rec_declaration decls)
     | _ -> k e in
 
+  let field r k e =
+    match Ast0.unwrap e with
+      Ast0.DisjField(_,decls,_,_)
+    | Ast0.ConjField(_,decls,_,_) ->
+	List.fold_left Common.union_set option_default
+	  (List.map r.VT0.combiner_rec_field decls)
+    | _ -> k e in
+
   let statement r k e =
     match Ast0.unwrap e with
       Ast0.Disj(_,stmts,_,_)
@@ -628,9 +644,9 @@ let dup_positions rules =
     V0.flat_combiner bind option_default
       mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
       mcode mcode mcode mcode
-      donothing donothing donothing donothing donothing donothing
-      donothing expression donothing donothing typeC donothing
-      donothing declaration statement
+      donothing donothing donothing donothing donothing donothing donothing
+      donothing donothing expression donothing donothing typeC donothing
+      donothing declaration field statement
       donothing donothing donothing donothing in
 
   let res =
@@ -672,6 +688,19 @@ let check_all_marked rname err table after_err =
 	    (Printf.sprintf "%s: %s %s not used %s" rname err name after_err))
     table
 
+let check_both_marked rname required optional =
+  Hashtbl.iter
+    (fun name ocell ->
+      if !ocell
+      then
+	let rcell = Hashtbl.find required name in
+	if not !rcell
+	then
+	  failwith
+	    (Printf.sprintf "error: %s: %s appears only in + code" rname
+	       (snd name)))
+    optional
+
 let check_meta rname old_metas inherited_metavars metavars minus plus =
   let old_metas =
     List.map (function (_,x) -> x) (List.map Ast.get_meta_name old_metas) in
@@ -684,9 +713,10 @@ let check_meta rname old_metas inherited_metavars metavars minus plus =
   let (ierr,iother) =
     List.partition (function Ast.MetaErrDecl(_,_) -> true | _ -> false)
       inherited_metavars in
-  let fresh_table = make_table fresh in
+  let frsh_table = make_table fresh in
   let err_table = make_table (err@ierr) in
   let other_table = make_table other in
+  let plus_table = make_table other in
   let iother_table = make_table iother in
 
   add_to_fresh_table fresh;
@@ -694,8 +724,10 @@ let check_meta rname old_metas inherited_metavars metavars minus plus =
   positions rname [iother_table;other_table] minus;
   dup_positions minus;
   check_all_marked rname "metavariable" other_table "in the - or context code";
-  rule old_metas [iother_table;fresh_table;err_table] false plus;
+  rule old_metas [iother_table;plus_table;frsh_table;err_table] false plus;
   check_all_marked rname "inherited metavariable" iother_table
     "in the -, +, or context code";
-  check_all_marked rname "metavariable" fresh_table "in the + code";
-  check_all_marked rname "error metavariable" err_table ""
+  check_all_marked rname "metavariable" frsh_table "in the + code";
+  check_all_marked rname "error metavariable" err_table "";
+  check_both_marked rname other_table plus_table;
+  Stdcompat.Hashtbl.reset fresh_table
