@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -445,16 +445,15 @@ let rec expression e =
       let name = normal_mcode name in
       let ln = promote_mcode name in
       mkres e (Ast0.MetaErr(name,a,b)) ln ln
-  | Ast0.MetaExpr(name,a,b,c,d) ->
+  | Ast0.MetaExpr(name,a,b,c,d,f) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
-      mkres e (Ast0.MetaExpr(name,a,b,c,d)) ln ln
-  | Ast0.MetaExprList(name,a,b) ->
+      mkres e (Ast0.MetaExpr(name,a,b,c,d,f)) ln ln
+  | Ast0.MetaExprList(name,a,b,c) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
-      mkres e (Ast0.MetaExprList(name,a,b)) ln ln
+      mkres e (Ast0.MetaExprList(name,a,b,c)) ln ln
   | Ast0.EComma(cm) ->
-      (*let cm = bad_mcode cm in*) (* why was this bad??? *)
       let cm = normal_mcode cm in
       let ln = promote_mcode cm in
       mkres e (Ast0.EComma(cm)) ln ln
@@ -511,7 +510,7 @@ and string_fragment e =
       let dots = normal_mcode dots in
       let ln = promote_mcode dots in
       mkres e (Ast0.Strdots dots) ln ln
-  | Ast0.MetaFormatList(pct,name,lenname) ->
+  | Ast0.MetaFormatList(pct,name,cstr,lenname) ->
       (* not sure what to do about the following comment... *)
       (* pct is particularly bad in this case, because it is ignored in
 	 the matching process.  The metavariable matches the complete format
@@ -520,7 +519,7 @@ and string_fragment e =
       let ln1 = promote_mcode pct in
       let name = normal_mcode name in
       let ln2 = promote_mcode name in
-      mkres e (Ast0.MetaFormatList(pct,name,lenname)) ln1 ln2
+      mkres e (Ast0.MetaFormatList(pct,name,cstr,lenname)) ln1 ln2
 
 and string_format e =
   match Ast0.unwrap e with
@@ -610,21 +609,38 @@ and typeC t =
       let ty = typeC ty in
       let lb = normal_mcode lb in
       let decls =
-	dots is_decl_dots (Some(promote_mcode lb)) declaration decls in
+	dots is_field_dots (Some(promote_mcode lb)) field decls in
       let rb = normal_mcode rb in
       mkres t (Ast0.StructUnionDef(ty,lb,decls,rb)) ty (promote_mcode rb)
+  | Ast0.TypeOfExpr(tf,lp,exp,rp) ->
+      let tf = normal_mcode tf in
+      let lp = normal_mcode lp in
+      let exp = expression exp in
+      let rp = normal_mcode rp in
+      mkres t (Ast0.TypeOfExpr(tf,lp,exp,rp))
+	(promote_mcode tf) (promote_mcode rp)
+  | Ast0.TypeOfType(tf,lp,ty,rp) ->
+      let tf = normal_mcode tf in
+      let lp = normal_mcode lp in
+      let rp = normal_mcode rp in
+      mkres t (Ast0.TypeOfType(tf,lp,typeC ty,rp))
+        (promote_mcode tf) (promote_mcode rp)
   | Ast0.TypeName(name) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
       mkres t (Ast0.TypeName(name)) ln ln
-  | Ast0.MetaType(name,a) ->
+  | Ast0.MetaType(name,cstr,a) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
-      mkres t (Ast0.MetaType(name,a)) ln ln
+      mkres t (Ast0.MetaType(name,cstr,a)) ln ln
   | Ast0.DisjType(starter,types,mids,ender) ->
       do_disj t starter types mids ender typeC
 	(fun starter types mids ender ->
 	  Ast0.DisjType(starter,types,mids,ender))
+  | Ast0.ConjType(starter,types,mids,ender) ->
+      do_disj t starter types mids ender typeC
+	(fun starter types mids ender ->
+	  Ast0.ConjType(starter,types,mids,ender))
   | Ast0.OptType(ty) ->
       let ty = typeC ty in mkres t (Ast0.OptType(ty)) ty ty
   | Ast0.AsType _ -> failwith "not possible"
@@ -634,48 +650,37 @@ and typeC t =
 (* Even if the Cocci program specifies a list of declarations, they are
    split out into multiple declarations of a single variable each. *)
 
-and is_decl_dots s =
-  match Ast0.unwrap s with
-    Ast0.Ddots(_,_) -> true
-  | _ -> false
-
 and declaration d =
   match Ast0.unwrap d with
-    Ast0.MetaDecl(name,a) ->
+    Ast0.MetaDecl(name,a,b) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
-      mkres d (Ast0.MetaDecl(name,a)) ln ln
-  | Ast0.MetaField(name,a) ->
-      let name = normal_mcode name in
-      let ln = promote_mcode name in
-      mkres d (Ast0.MetaField(name,a)) ln ln
-  | Ast0.MetaFieldList(name,a,b) ->
-      let name = normal_mcode name in
-      let ln = promote_mcode name in
-      mkres d (Ast0.MetaFieldList(name,a,b)) ln ln
-  | Ast0.Init(stg,ty,id,eq,exp,sem) ->
+      mkres d (Ast0.MetaDecl(name,a,b)) ln ln
+  | Ast0.Init(stg,ty,id,attr,eq,exp,sem) ->
       let ty = typeC ty in
       let id = ident id in
+      let attr = List.map normal_mcode attr in
       let eq = normal_mcode eq in
       let exp = initialiser exp in
       let sem = normal_mcode sem in
       (match stg with
 	None ->
-	  mkres d (Ast0.Init(stg,ty,id,eq,exp,sem)) ty (promote_mcode sem)
+	  mkres d (Ast0.Init(stg,ty,id,attr,eq,exp,sem)) ty (promote_mcode sem)
       | Some x ->
 	  let stg = Some (normal_mcode x) in
-	  mkres d (Ast0.Init(stg,ty,id,eq,exp,sem))
+	  mkres d (Ast0.Init(stg,ty,id,attr,eq,exp,sem))
 	    (promote_mcode x) (promote_mcode sem))
-  | Ast0.UnInit(stg,ty,id,sem) ->
+  | Ast0.UnInit(stg,ty,id,attr,sem) ->
       let ty = typeC ty in
       let id = ident id in
+      let attr = List.map normal_mcode attr in
       let sem = normal_mcode sem in
       (match stg with
 	None ->
-	  mkres d (Ast0.UnInit(stg,ty,id,sem)) ty (promote_mcode sem)
+	  mkres d (Ast0.UnInit(stg,ty,id,attr,sem)) ty (promote_mcode sem)
       | Some x ->
 	  let stg = Some (normal_mcode x) in
-	  mkres d (Ast0.UnInit(stg,ty,id,sem))
+	  mkres d (Ast0.UnInit(stg,ty,id,attr,sem))
 	    (promote_mcode x) (promote_mcode sem))
   | Ast0.FunProto(fninfo,name,lp1,params,va1,rp1,sem) ->
       let fninfo =
@@ -743,15 +748,56 @@ and declaration d =
       do_disj d starter decls mids ender declaration
 	(fun starter decls mids ender ->
 	  Ast0.DisjDecl(starter,decls,mids,ender))
-  | Ast0.Ddots(dots,whencode) ->
-      let dots = bad_mcode dots in
-      let ln = promote_mcode dots in
-      mkres d (Ast0.Ddots(dots,whencode)) ln ln
+  | Ast0.ConjDecl(starter,decls,mids,ender) ->
+      do_disj d starter decls mids ender declaration
+	(fun starter decls mids ender ->
+	  Ast0.ConjDecl(starter,decls,mids,ender))
   | Ast0.OptDecl(decl) ->
       let decl = declaration decl in
       mkres d (Ast0.OptDecl(declaration decl)) decl decl
   | Ast0.AsDecl _ -> failwith "not possible"
 
+(* --------------------------------------------------------------------- *)
+(* Field declaration *)
+
+and is_field_dots s =
+  match Ast0.unwrap s with
+    Ast0.Fdots(_,_) -> true
+  | _ -> false
+
+and field d =
+  match Ast0.unwrap d with
+     Ast0.MetaField(name,a,b) ->
+      let name = normal_mcode name in
+      let ln = promote_mcode name in
+      mkres d (Ast0.MetaField(name,a,b)) ln ln
+  | Ast0.MetaFieldList(name,a,b,c) ->
+      let name = normal_mcode name in
+      let ln = promote_mcode name in
+      mkres d (Ast0.MetaFieldList(name,a,b,c)) ln ln
+  | Ast0.Field(ty,id,bf,sem) ->
+      let ty = typeC ty in
+      let id = Common.map_option ident id in
+      let bitfield (c, e) = (normal_mcode c, expression e) in
+      let bf = Common.map_option bitfield bf in
+      let sem = normal_mcode sem in
+      mkres d (Ast0.Field(ty,id,bf,sem)) ty (promote_mcode sem)
+  | Ast0.DisjField(starter,decls,mids,ender) ->
+      do_disj d starter decls mids ender field
+	(fun starter decls mids ender ->
+	  Ast0.DisjField(starter,decls,mids,ender))
+  | Ast0.ConjField(starter,decls,mids,ender) ->
+      do_disj d starter decls mids ender field
+	(fun starter decls mids ender ->
+	  Ast0.ConjField(starter,decls,mids,ender))
+  | Ast0.OptField(decl) ->
+      let decl = field decl in
+      mkres d (Ast0.OptField(field decl)) decl decl
+  | Ast0.Fdots(dots,whencode) ->
+      let dots = bad_mcode dots in
+      let ln = promote_mcode dots in
+      mkres d (Ast0.Fdots(dots,whencode)) ln ln
+
 (* --------------------------------------------------------------------- *)
 (* Initializer *)
 
@@ -762,14 +808,14 @@ and is_init_dots i =
 
 and initialiser i =
   match Ast0.unwrap i with
-    Ast0.MetaInit(name,a) ->
+    Ast0.MetaInit(name,a,b) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
-      mkres i (Ast0.MetaInit(name,a)) ln ln
-  | Ast0.MetaInitList(name,a,b) ->
+      mkres i (Ast0.MetaInit(name,a,b)) ln ln
+  | Ast0.MetaInitList(name,a,b,c) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
-      mkres i (Ast0.MetaInitList(name,a,b)) ln ln
+      mkres i (Ast0.MetaInitList(name,a,b,c)) ln ln
   | Ast0.InitExpr(exp) ->
       let exp = expression exp in
       mkres i (Ast0.InitExpr(exp)) exp exp
@@ -841,16 +887,15 @@ and parameterTypeDef p =
       let ty = typeC ty in mkres p (Ast0.Param(ty,Some id)) ty id
   | Ast0.Param(ty,None) ->
       let ty = typeC ty in mkres p (Ast0.Param(ty,None)) ty ty
-  | Ast0.MetaParam(name,a) ->
+  | Ast0.MetaParam(name,a,b) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
-      mkres p (Ast0.MetaParam(name,a)) ln ln
-  | Ast0.MetaParamList(name,a,b) ->
+      mkres p (Ast0.MetaParam(name,a,b)) ln ln
+  | Ast0.MetaParamList(name,a,b,c) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
-      mkres p (Ast0.MetaParamList(name,a,b)) ln ln
+      mkres p (Ast0.MetaParamList(name,a,b,c)) ln ln
   | Ast0.PComma(cm) ->
-      (*let cm = bad_mcode cm in*) (* why was this bad??? *)
       let cm = normal_mcode cm in
       let ln = promote_mcode cm in
       mkres p (Ast0.PComma(cm)) ln ln
@@ -879,8 +924,11 @@ let rec define_param p =
   match Ast0.unwrap p with
     Ast0.DParam(id) ->
       let id = ident id in mkres p (Ast0.DParam(id)) id id
+  | Ast0.MetaDParamList(name,a,b,c) ->
+      let name = normal_mcode name in
+      let ln = promote_mcode name in
+      mkres p (Ast0.MetaDParamList(name,a,b,c)) ln ln
   | Ast0.DPComma(cm) ->
-      (*let cm = bad_mcode cm in*) (* why was this bad??? *)
       let cm = normal_mcode cm in
       let ln = promote_mcode cm in
       mkres p (Ast0.DPComma(cm)) ln ln
@@ -1076,18 +1124,18 @@ let rec statement s =
 	let sem = normal_mcode sem in
 	mkres s (Ast0.Exec(exec,lang,code,sem))
 	  (promote_mcode exec) (promote_mcode sem)
-    | Ast0.MetaStmt(name,a) ->
+    | Ast0.MetaStmt(name,a,b) ->
 	let name = normal_mcode name in
 	let ln = promote_mcode name in
-	mkres s (Ast0.MetaStmt(name,a)) ln ln
-    | Ast0.MetaStmtList(name,a,b) ->
+	mkres s (Ast0.MetaStmt(name,a,b)) ln ln
+    | Ast0.MetaStmtList(name,a,b,c) ->
 	(* This becomes ..., which becomes a metavar.  The metavar, with
 	   any attachments that this has, is then iterated over the code.
 	   If we attach anything here, then it will be duplicated for each
 	   statement in the list. *)
 	let name = bad_mcode name in
 	let ln = promote_mcode name in
-	mkres s (Ast0.MetaStmtList(name,a,b)) ln ln
+	mkres s (Ast0.MetaStmtList(name,a,b,c)) ln ln
     | Ast0.Exp(exp) ->
 	let exp = expression exp in
 	mkres s (Ast0.Exp(exp)) exp exp
@@ -1200,6 +1248,11 @@ let rec statement s =
 	let inc = normal_mcode inc in
 	let stm = normal_mcode stm in
 	mkres s (Ast0.Include(inc,stm)) (promote_mcode inc) (promote_mcode stm)
+    | Ast0.MetaInclude(inc,s) ->
+	let inc = normal_mcode inc in
+	let s = expression s in
+	mkres s (Ast0.MetaInclude(inc,s)) (promote_mcode inc) s
+
     | Ast0.Undef(def,id) ->
 	let def = normal_mcode def in
 	let id = ident id in
