@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -26,7 +26,9 @@ let set_mcodekind x mcodekind =
   | Ast0.DotsParamTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.DotsStmtTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.DotsDeclTag(d) -> Ast0.set_mcodekind d mcodekind
+  | Ast0.DotsFieldTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.DotsCaseTag(d) -> Ast0.set_mcodekind d mcodekind
+  | Ast0.DotsDefParamTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.IdentTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.ExprTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.AssignOpTag(d) -> Ast0.set_mcodekind d mcodekind
@@ -36,6 +38,7 @@ let set_mcodekind x mcodekind =
   | Ast0.TypeCTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.ParamTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.DeclTag(d) -> Ast0.set_mcodekind d mcodekind
+  | Ast0.FieldTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.InitTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.StmtTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.ForInfoTag(d) -> Ast0.set_mcodekind d mcodekind
@@ -56,7 +59,9 @@ let set_index x index =
   | Ast0.DotsParamTag(d) -> Ast0.set_index d index
   | Ast0.DotsStmtTag(d) -> Ast0.set_index d index
   | Ast0.DotsDeclTag(d) -> Ast0.set_index d index
+  | Ast0.DotsFieldTag(d) -> Ast0.set_index d index
   | Ast0.DotsCaseTag(d) -> Ast0.set_index d index
+  | Ast0.DotsDefParamTag(d) -> Ast0.set_index d index
   | Ast0.IdentTag(d) -> Ast0.set_index d index
   | Ast0.ExprTag(d) -> Ast0.set_index d index
   | Ast0.AssignOpTag(d) -> Ast0.set_index d index
@@ -67,6 +72,7 @@ let set_index x index =
   | Ast0.ParamTag(d) -> Ast0.set_index d index
   | Ast0.InitTag(d) -> Ast0.set_index d index
   | Ast0.DeclTag(d) -> Ast0.set_index d index
+  | Ast0.FieldTag(d) -> Ast0.set_index d index
   | Ast0.StmtTag(d) -> Ast0.set_index d index
   | Ast0.ForInfoTag(d) -> Ast0.set_index d index
   | Ast0.CaseLineTag(d) -> Ast0.set_index d index
@@ -85,7 +91,9 @@ let get_index = function
   | Ast0.DotsParamTag(d) -> Index.parameter_dots d
   | Ast0.DotsStmtTag(d) -> Index.statement_dots d
   | Ast0.DotsDeclTag(d) -> Index.declaration_dots d
+  | Ast0.DotsFieldTag(d) -> Index.field_dots d
   | Ast0.DotsCaseTag(d) -> Index.case_line_dots d
+  | Ast0.DotsDefParamTag(d) -> Index.define_param_dots d
   | Ast0.IdentTag(d) -> Index.ident d
   | Ast0.ExprTag(d) -> Index.expression d
   | Ast0.AssignOpTag(d) -> Index.assignOp d
@@ -96,6 +104,7 @@ let get_index = function
   | Ast0.ParamTag(d) -> Index.parameterTypeDef d
   | Ast0.InitTag(d) -> Index.initialiser d
   | Ast0.DeclTag(d) -> Index.declaration d
+  | Ast0.FieldTag(d) -> Index.field d
   | Ast0.StmtTag(d) -> Index.statement d
   | Ast0.ForInfoTag(d) -> Index.forinfo d
   | Ast0.CaseLineTag(d) -> Index.case_line d
@@ -165,8 +174,9 @@ let collect_plus_lines top =
       mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
       mcode mcode
       donothing donothing donothing donothing donothing donothing donothing
+      donothing
       donothing donothing donothing donothing donothing donothing donothing
-      statement donothing donothing donothing donothing in
+      donothing donothing statement donothing donothing donothing donothing in
   fn.VT0.combiner_rec_top_level top
 
 (* --------------------------------------------------------------------- *)
@@ -248,7 +258,17 @@ let option_default = (*Bind(Neutral,[],[],[],[],[])*)
   Recursor(Neutral,[],[],[])
 
 let contains_added_strings info =
-  not (info.Ast0.strings_before = []) || not (info.Ast0.strings_after = [])
+  let unsafe l =
+    List.exists
+      (fun (str,_) ->
+	match str with
+	  (Ast.Noindent s | Ast.Indent s) -> not (String.trim s = "")
+	| Ast.Space _ -> true (* adds a thing with space around it *))
+      l in
+  (* If this is true, eg if (x) return 0; will be broken up into
+     two rule elems.  Not sure why that is needed, but surely it is not needed
+     when it is just whitespace that is added. *)
+  unsafe info.Ast0.strings_before || unsafe info.Ast0.strings_after
 
 let mcode (_,_,info,mcodekind,pos,_) =
   let offset = info.Ast0.pos_info.Ast0.offset in
@@ -372,16 +392,15 @@ let classify is_minus all_marked table code =
 	  disj_cases e starter expr_list r.VT0.combiner_rec_expression ender
       |	_ -> k e) in
 
-  (* not clear why we have the next two cases, since DisjDecl and
-  DisjType shouldn't have been constructed yet, as they only come from isos *)
+  (* not clear why we have the next cases, since DisjDecl and
+  as it only comes from isos *)
   (* actually, DisjDecl now allowed in source struct decls *)
   let declaration r k e =
     compute_result Ast0.decl e
       (match Ast0.unwrap e with
-	Ast0.DisjDecl(starter,decls,_,ender) ->
+	Ast0.DisjDecl(starter,decls,_,ender)
+      | Ast0.ConjDecl(starter,decls,_,ender) ->
 	  disj_cases e starter decls r.VT0.combiner_rec_declaration ender
-      | Ast0.Ddots(dots,whencode) ->
-	  k (Ast0.rewrap e (Ast0.Ddots(dots,None)))
 	(* Need special cases for the following so that the type will be
 	   considered as a unit, rather than distributed around the
 	   declared variable.  This needs to be done because of the call to
@@ -391,16 +410,39 @@ let classify is_minus all_marked table code =
 	   term structure.  In (all?) other such cases, we visit the terms
 	   using rebuilder, which just visits the subterms, rather than
 	   reordering their components. *)
-      |	Ast0.Init(stg,ty,id,eq,ini,sem) ->
+      |	Ast0.Init(stg,ty,id,attr,eq,ini,sem) ->
 	  bind (match stg with Some stg -> mcode stg | _ -> option_default)
 	    (bind (r.VT0.combiner_rec_typeC ty)
 	       (bind (r.VT0.combiner_rec_ident id)
-		  (bind (mcode eq)
-		     (bind (r.VT0.combiner_rec_initialiser ini) (mcode sem)))))
-      | Ast0.UnInit(stg,ty,id,sem) ->
+                  (bind
+                     (List.fold_right bind (List.map mcode attr)
+			option_default)
+		     (bind (mcode eq)
+			(bind (r.VT0.combiner_rec_initialiser ini)
+			   (mcode sem))))))
+      | Ast0.UnInit(stg,ty,id,attr,sem) ->
 	  bind (match stg with Some stg -> mcode stg | _ -> option_default)
 	    (bind (r.VT0.combiner_rec_typeC ty)
-	       (bind (r.VT0.combiner_rec_ident id) (mcode sem)))
+	       (bind (r.VT0.combiner_rec_ident id)
+                  (bind
+                     (List.fold_right bind (List.map mcode attr)
+			option_default)
+		     (mcode sem))))
+      |	_ -> k e) in
+
+  let field r k e =
+    compute_result Ast0.field e
+      (match Ast0.unwrap e with
+	Ast0.DisjField(starter,decls,_,ender)
+      | Ast0.ConjField(starter,decls,_,ender) ->
+	  disj_cases e starter decls r.VT0.combiner_rec_field ender
+      | Ast0.Fdots(dots,whencode) ->
+	  k (Ast0.rewrap e (Ast0.Fdots(dots,None)))
+      | Ast0.Field(ty,id,bf,sem) ->
+	  let bitfield (_c, e) = r.VT0.combiner_rec_expression e in
+	  bind (r.VT0.combiner_rec_typeC ty)
+	    (bind (Common.default option_default r.VT0.combiner_rec_ident id)
+	       (bind (Common.default option_default bitfield bf) (mcode sem)))
       |	_ -> k e) in
 
   let param r k e =
@@ -414,7 +456,8 @@ let classify is_minus all_marked table code =
   let typeC r k e =
     compute_result Ast0.typeC e
       (match Ast0.unwrap e with
-	Ast0.DisjType(starter,types,_,ender) ->
+	Ast0.DisjType(starter,types,_,ender)
+      | Ast0.ConjType(starter,types,_,ender) ->
 	  disj_cases e starter types r.VT0.combiner_rec_typeC ender
       |	_ -> k e) in
 
@@ -481,8 +524,11 @@ let classify is_minus all_marked table code =
       mcode mcode mcode mcode mcode
       (do_nothing Ast0.dotsExpr) (do_nothing Ast0.dotsInit)
       (do_nothing Ast0.dotsParam) (do_nothing Ast0.dotsStmt)
-      (do_nothing Ast0.dotsDecl) (do_nothing Ast0.dotsCase)
-      ident expression (do_nothing Ast0.assignOp) (do_nothing Ast0.binaryOp) typeC initialiser param declaration
+      (do_nothing Ast0.dotsDecl) (do_nothing Ast0.dotsField)
+      (do_nothing Ast0.dotsCase)
+      (do_nothing Ast0.dotsDefParam)
+      ident expression (do_nothing Ast0.assignOp) (do_nothing Ast0.binaryOp)
+      typeC initialiser param declaration field
       statement (do_nothing Ast0.forinfo) case_line string_fragment
       (do_top Ast0.top) in
   combiner.VT0.combiner_rec_top_level code
@@ -519,7 +565,7 @@ let equal_option e1 e2 =
 let dots fn d1 d2 =
   List.length (Ast0.unwrap d1) = List.length (Ast0.unwrap d2)
 
-let rec equal_ident i1 i2 =
+let equal_ident i1 i2 =
   match (Ast0.unwrap i1,Ast0.unwrap i2) with
     (Ast0.Id(name1),Ast0.Id(name2)) -> equal_mcode name1 name2
   | (Ast0.MetaId(name1,_,_,_),Ast0.MetaId(name2,_,_,_)) ->
@@ -572,8 +618,8 @@ let rec equal_expression e1 e2 =
   | (Ast0.Constructor(lp1,_,rp1,_),Ast0.Constructor(lp2,_,rp2,_)) ->
       equal_mcode lp1 lp2 && equal_mcode rp1 rp2
   | (Ast0.MetaErr(name1,_,_),Ast0.MetaErr(name2,_,_))
-  | (Ast0.MetaExpr(name1,_,_,_,_),Ast0.MetaExpr(name2,_,_,_,_))
-  | (Ast0.MetaExprList(name1,_,_),Ast0.MetaExprList(name2,_,_)) ->
+  | (Ast0.MetaExpr(name1,_,_,_,_,_),Ast0.MetaExpr(name2,_,_,_,_,_))
+  | (Ast0.MetaExprList(name1,_,_,_),Ast0.MetaExprList(name2,_,_,_)) ->
       equal_mcode name1 name2
   | (Ast0.EComma(cm1),Ast0.EComma(cm2)) -> equal_mcode cm1 cm2
   | (Ast0.DisjExpr(starter1,_,mids1,ender1),
@@ -608,7 +654,7 @@ and binaryOp_equal op1 op2 =
       equal_mcode mv1 mv2
     | _ -> false
 
-let rec equal_typeC t1 t2 =
+let equal_typeC t1 t2 =
   match (Ast0.unwrap t1,Ast0.unwrap t2) with
     (Ast0.ConstVol(cv1,_),Ast0.ConstVol(cv2,_)) -> equal_mcode cv1 cv2
   | (Ast0.BaseType(ty1,stringsa),Ast0.BaseType(ty2,stringsb)) ->
@@ -632,14 +678,23 @@ let rec equal_typeC t1 t2 =
   | (Ast0.StructUnionDef(_,lb1,_,rb1),
      Ast0.StructUnionDef(_,lb2,_,rb2)) ->
        equal_mcode lb1 lb2 && equal_mcode rb1 rb2
+  | (Ast0.TypeOfExpr(tf1,lp1,_,rp1),Ast0.TypeOfExpr(tf2,lp2,_,rp2)) ->
+      equal_mcode tf1 tf2 && equal_mcode lp1 lp2 && equal_mcode rp1 rp2
+  | (Ast0.TypeOfType(tf1,lp1,_,rp1),Ast0.TypeOfType(tf2,lp2,_,rp2)) ->
+      equal_mcode tf1 tf2 && equal_mcode lp1 lp2 && equal_mcode rp1 rp2
   | (Ast0.TypeName(name1),Ast0.TypeName(name2)) -> equal_mcode name1 name2
-  | (Ast0.MetaType(name1,_),Ast0.MetaType(name2,_)) ->
+  | (Ast0.MetaType(name1,_,_),Ast0.MetaType(name2,_,_)) ->
       equal_mcode name1 name2
   | (Ast0.DisjType(starter1,_,mids1,ender1),
      Ast0.DisjType(starter2,_,mids2,ender2)) ->
        equal_mcode starter1 starter2 &&
        List.for_all2 equal_mcode mids1 mids2 &&
        equal_mcode ender1 ender2
+  | (Ast0.ConjType(starter1,_,mids1,ender1),
+     Ast0.ConjType(starter2,_,mids2,ender2)) ->
+       equal_mcode starter1 starter2 &&
+       List.for_all2 equal_mcode mids1 mids2 &&
+       equal_mcode ender1 ender2
   | (Ast0.OptType(_),Ast0.OptType(_)) -> true
   | _ -> false
 
@@ -653,14 +708,15 @@ let equal_fninfo x y =
 
 let equal_declaration d1 d2 =
   match (Ast0.unwrap d1,Ast0.unwrap d2) with
-    (Ast0.MetaDecl(name1,_),Ast0.MetaDecl(name2,_))
-  | (Ast0.MetaField(name1,_),Ast0.MetaField(name2,_))
-  | (Ast0.MetaFieldList(name1,_,_),Ast0.MetaFieldList(name2,_,_)) ->
+    (Ast0.MetaDecl(name1,_,_),Ast0.MetaDecl(name2,_,_)) ->
       equal_mcode name1 name2
-  | (Ast0.Init(stg1,_,_,eq1,_,sem1),Ast0.Init(stg2,_,_,eq2,_,sem2)) ->
-      equal_option stg1 stg2 && equal_mcode eq1 eq2 && equal_mcode sem1 sem2
-  | (Ast0.UnInit(stg1,_,_,sem1),Ast0.UnInit(stg2,_,_,sem2)) ->
-      equal_option stg1 stg2 && equal_mcode sem1 sem2
+  | (Ast0.Init(stg1,_,_,attr1,eq1,_,sem1),
+     Ast0.Init(stg2,_,_,attr2,eq2,_,sem2)) ->
+      equal_option stg1 stg2 && List.for_all2 equal_mcode attr1 attr2 &&
+      equal_mcode eq1 eq2 && equal_mcode sem1 sem2
+  | (Ast0.UnInit(stg1,_,_,attr1,sem1),Ast0.UnInit(stg2,_,_,attr2,sem2)) ->
+      equal_option stg1 stg2 && List.for_all2 equal_mcode attr1 attr2 &&
+      equal_mcode sem1 sem2
   | (Ast0.FunProto(fninfo1,name1,lp1,p1,va1,rp1,sem1),
      Ast0.FunProto(fninfo2,name2,lp2,p2,va2,rp2,sem2)) ->
        let equal_varargs va1 va2 = match (va1,va2) with
@@ -682,10 +738,29 @@ let equal_declaration d1 d2 =
        equal_mcode lp1 lp2 && equal_mcode rp1 rp2 && equal_mcode eq1 eq2
 	 && equal_mcode sem1 sem2
   | (Ast0.TyDecl(_,sem1),Ast0.TyDecl(_,sem2)) -> equal_mcode sem1 sem2
-  | (Ast0.Ddots(dots1,_),Ast0.Ddots(dots2,_)) -> equal_mcode dots1 dots2
   | (Ast0.OptDecl(_),Ast0.OptDecl(_)) -> true
   | (Ast0.DisjDecl(starter1,_,mids1,ender1),
-     Ast0.DisjDecl(starter2,_,mids2,ender2)) ->
+     Ast0.DisjDecl(starter2,_,mids2,ender2))
+  | (Ast0.ConjDecl(starter1,_,mids1,ender1),
+     Ast0.ConjDecl(starter2,_,mids2,ender2)) ->
+       equal_mcode starter1 starter2 &&
+       List.for_all2 equal_mcode mids1 mids2 &&
+       equal_mcode ender1 ender2
+  | _ -> false
+
+let equal_field d1 d2 =
+  match (Ast0.unwrap d1,Ast0.unwrap d2) with
+     (Ast0.MetaField(name1,_,_),Ast0.MetaField(name2,_,_))
+  | (Ast0.MetaFieldList(name1,_,_,_),Ast0.MetaFieldList(name2,_,_,_)) ->
+      equal_mcode name1 name2
+  | (Ast0.Field(_,_,_bf1,sem1),Ast0.Field(_,_,_bf2,sem2)) ->
+      equal_mcode sem1 sem2
+  | (Ast0.Fdots(dots1,_),Ast0.Fdots(dots2,_)) -> equal_mcode dots1 dots2
+  | (Ast0.OptField(_),Ast0.OptField(_)) -> true
+  | (Ast0.DisjField(starter1,_,mids1,ender1),
+     Ast0.DisjField(starter2,_,mids2,ender2))
+  | (Ast0.ConjField(starter1,_,mids1,ender1),
+     Ast0.ConjField(starter2,_,mids2,ender2)) ->
        equal_mcode starter1 starter2 &&
        List.for_all2 equal_mcode mids1 mids2 &&
        equal_mcode ender1 ender2
@@ -705,9 +780,9 @@ let equal_designator d1 d2 =
 
 let equal_initialiser i1 i2 =
   match (Ast0.unwrap i1,Ast0.unwrap i2) with
-    (Ast0.MetaInit(name1,_),Ast0.MetaInit(name2,_)) ->
+    (Ast0.MetaInit(name1,_,_),Ast0.MetaInit(name2,_,_)) ->
       equal_mcode name1 name2
-  | (Ast0.MetaInitList(name1,_,_),Ast0.MetaInitList(name2,_,_)) ->
+  | (Ast0.MetaInitList(name1,_,_,_),Ast0.MetaInitList(name2,_,_,_)) ->
       equal_mcode name1 name2
   | (Ast0.InitExpr(_),Ast0.InitExpr(_)) -> true
   | (Ast0.InitList(lb1,_,rb1,o1),Ast0.InitList(lb2,_,rb2,o2)) ->
@@ -729,15 +804,15 @@ let equal_parameterTypeDef p1 p2 =
   match (Ast0.unwrap p1,Ast0.unwrap p2) with
     (Ast0.VoidParam(_),Ast0.VoidParam(_)) -> true
   | (Ast0.Param(_,_),Ast0.Param(_,_)) -> true
-  | (Ast0.MetaParam(name1,_),Ast0.MetaParam(name2,_))
-  | (Ast0.MetaParamList(name1,_,_),Ast0.MetaParamList(name2,_,_)) ->
+  | (Ast0.MetaParam(name1,_,_),Ast0.MetaParam(name2,_,_))
+  | (Ast0.MetaParamList(name1,_,_,_),Ast0.MetaParamList(name2,_,_,_)) ->
       equal_mcode name1 name2
   | (Ast0.PComma(cm1),Ast0.PComma(cm2)) -> equal_mcode cm1 cm2
   | (Ast0.Pdots(dots1),Ast0.Pdots(dots2)) -> equal_mcode dots1 dots2
   | (Ast0.OptParam(_),Ast0.OptParam(_)) -> true
   | _ -> false
 
-let rec equal_statement s1 s2 =
+let equal_statement s1 s2 =
   match (Ast0.unwrap s1,Ast0.unwrap s2) with
     (Ast0.FunDecl(_,fninfo1,_,lp1,_,_,rp1,lbrace1,_,rbrace1,_),
      Ast0.FunDecl(_,fninfo2,_,lp2,_,_,rp2,lbrace2,_,rbrace2,_)) ->
@@ -794,13 +869,10 @@ let rec equal_statement s1 s2 =
   | (Ast0.Exec(exec1,lang1,_,sem1),Ast0.Exec(exec2,lang2,_,sem2)) ->
       equal_mcode exec1 exec2 && equal_mcode lang1 lang2 &&
       equal_mcode sem1 sem2
-  | (Ast0.MetaStmt(name1,_),Ast0.MetaStmt(name2,_))
-  | (Ast0.MetaStmtList(name1,_,_),Ast0.MetaStmtList(name2,_,_)) ->
+  | (Ast0.MetaStmt(name1,_,_),Ast0.MetaStmt(name2,_,_))
+  | (Ast0.MetaStmtList(name1,_,_,_),Ast0.MetaStmtList(name2,_,_,_)) ->
       equal_mcode name1 name2
-  | (Ast0.Disj(starter1,_,mids1,ender1),Ast0.Disj(starter2,_,mids2,ender2)) ->
-      equal_mcode starter1 starter2 &&
-      List.for_all2 equal_mcode mids1 mids2 &&
-      equal_mcode ender1 ender2
+  | (Ast0.Disj(starter1,_,mids1,ender1),Ast0.Disj(starter2,_,mids2,ender2))
   | (Ast0.Conj(starter1,_,mids1,ender1),Ast0.Conj(starter2,_,mids2,ender2)) ->
       equal_mcode starter1 starter2 &&
       List.for_all2 equal_mcode mids1 mids2 &&
@@ -815,6 +887,8 @@ let rec equal_statement s1 s2 =
   | (Ast0.Dots(d1,_),Ast0.Dots(d2,_)) -> equal_mcode d1 d2
   | (Ast0.Include(inc1,name1),Ast0.Include(inc2,name2)) ->
       equal_mcode inc1 inc2 && equal_mcode name1 name2
+  | (Ast0.MetaInclude(inc1,name1),Ast0.MetaInclude(inc2,name2)) ->
+      equal_mcode inc1 inc2
   | (Ast0.Undef(def1,_),Ast0.Undef(def2,_)) ->
       equal_mcode def1 def2
   | (Ast0.Define(def1,_,_,_),Ast0.Define(def2,_,_,_)) ->
@@ -838,7 +912,17 @@ let equal_case_line c1 c2 =
   | (Ast0.OptCase(_),Ast0.OptCase(_)) -> true
   | _ -> false
 
-let rec equal_top_level t1 t2 =
+let equal_define_param d1 d2 =
+  match (Ast0.unwrap d1,Ast0.unwrap d2) with
+    (Ast0.DParam _,Ast0.DParam _) -> true
+  | (Ast0.MetaDParamList(name1,_,_,_),Ast0.MetaDParamList(name2,_,_,_)) ->
+      equal_mcode name1 name2
+  | (Ast0.DPComma cm1,Ast0.DPComma cm2) -> equal_mcode cm1 cm2
+  | (Ast0.DPdots d1,Ast0.DPdots d2) -> equal_mcode d1 d2
+  | (Ast0.OptDParam(_),Ast0.OptDParam(_)) -> true
+  | _ -> false
+
+let equal_top_level t1 t2 =
   match (Ast0.unwrap t1,Ast0.unwrap t2) with
     (Ast0.NONDECL(_),Ast0.NONDECL(_)) -> true
   | (Ast0.FILEINFO(old_file1,new_file1),Ast0.FILEINFO(old_file2,new_file2)) ->
@@ -854,7 +938,10 @@ let root_equal e1 e2 =
       dots equal_parameterTypeDef d1 d2
   | (Ast0.DotsStmtTag(d1),Ast0.DotsStmtTag(d2)) -> dots equal_statement d1 d2
   | (Ast0.DotsDeclTag(d1),Ast0.DotsDeclTag(d2)) -> dots equal_declaration d1 d2
+  | (Ast0.DotsFieldTag(d1),Ast0.DotsFieldTag(d2)) -> dots equal_field d1 d2
   | (Ast0.DotsCaseTag(d1),Ast0.DotsCaseTag(d2)) -> dots equal_case_line d1 d2
+  | (Ast0.DotsDefParamTag(d1),Ast0.DotsDefParamTag(d2)) ->
+      dots equal_define_param d1 d2
   | (Ast0.IdentTag(i1),Ast0.IdentTag(i2)) -> equal_ident i1 i2
   | (Ast0.ExprTag(e1),Ast0.ExprTag(e2)) -> equal_expression e1 e2
   | (Ast0.ArgExprTag(d),_) -> failwith "not possible - iso only"
@@ -862,6 +949,7 @@ let root_equal e1 e2 =
   | (Ast0.ParamTag(p1),Ast0.ParamTag(p2)) -> equal_parameterTypeDef p1 p2
   | (Ast0.InitTag(d1),Ast0.InitTag(d2)) -> equal_initialiser d1 d2
   | (Ast0.DeclTag(d1),Ast0.DeclTag(d2)) -> equal_declaration d1 d2
+  | (Ast0.FieldTag(d1),Ast0.FieldTag(d2)) -> equal_field d1 d2
   | (Ast0.StmtTag(s1),Ast0.StmtTag(s2)) -> equal_statement s1 s2
   | (Ast0.TopTag(t1),Ast0.TopTag(t2)) -> equal_top_level t1 t2
   | (Ast0.IsoWhenTag(_),_) | (_,Ast0.IsoWhenTag(_))
@@ -898,14 +986,15 @@ let contextify_all =
   let bind x y = () in
   let option_default = () in
   let mcode x = () in
-  let do_nothing r k e = Ast0.set_mcodekind e (default_context()); k e in
+  let donothing r k e = Ast0.set_mcodekind e (default_context()); k e in
 
   V0.flat_combiner bind option_default
     mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
     mcode mcode
-    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
-    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
-    do_nothing do_nothing do_nothing do_nothing do_nothing
+    donothing donothing donothing donothing donothing donothing donothing
+    donothing donothing donothing donothing donothing donothing donothing
+    donothing donothing donothing donothing donothing donothing donothing
+    donothing
 
 let contextify_whencode =
   let bind x y = () in
