@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -76,63 +76,72 @@ let collect_refs include_constraints =
 	    if include_constraints
 	    then Ast.cstr_meta_names idconstraint
 	    else [] in
-	  bind (List.rev metas) [metaid name]
+	  bind metas [metaid name]
       | Ast.DisjId(ids) -> bind_disj (List.map k ids)
       | _ -> option_default) in
 
   let type_collect res ty = bind res (Ast.meta_names_of_fullType ty) in
 
+  let constraints cstr =
+    if include_constraints then Ast.cstr_pos_meta_names cstr
+    else [] in
+
+  let listlen l =
+    match l with
+      Ast.MetaListLen(lenname,cstr,_,_) ->
+	bind (constraints cstr) [metaid lenname]
+    | _ -> [] in
+
+  let listlen_option l =
+    match l with
+      None -> option_default
+    | Some l -> listlen l in
+
   let astfvexpr recursor k e =
     bind (k e)
       (match Ast.unwrap e with
-	Ast.MetaExpr(name,constraints,_,Some type_list,_,_) ->
-	  let types = List.fold_left type_collect option_default type_list in
+	Ast.MetaExpr(name,constraints,_,Some type_list,_,_,bitfield) ->
+	  let types =
+	    (* problem: if there are multiple types, then none in particular
+	       is needed *)
+	    if include_constraints || List.length type_list = 1
+	    then List.fold_left type_collect option_default type_list
+	    else [] in
 	  let extra =
 	    if include_constraints
-	    then
-	      match constraints with
-		Ast.SubExpCstrt l -> l
-	      |	_ -> []
+	    then Ast.cstr_pos_meta_names constraints
 	    else [] in
-	  bind extra (bind [metaid name] types)
-      | Ast.MetaErr(name,constraints,_,_)
-      | Ast.MetaExpr(name,constraints,_,_,_,_) ->
-	  let extra =
-	    if include_constraints
-	    then
-	      match constraints with
-		Ast.SubExpCstrt l -> l
-	      |	_ -> []
-	    else [] in
-	  bind extra [metaid name]
-      | Ast.MetaExprList(name,Ast.MetaListLen (lenname,_,_),_,_) ->
-	  [metaid name;metaid lenname]
-      | Ast.MetaExprList(name,_,_,_) -> [metaid name]
+	  let bitfield' = listlen_option bitfield in
+	  bind bitfield' (bind extra (bind [metaid name] types))
+      | Ast.MetaErr(name,cstr,_,_) ->
+	  bind (constraints cstr) [metaid name]
+      | Ast.MetaExpr(name,cstr,_,_,_,_,bitfield) ->
+	  let bitfield' = listlen_option bitfield in
+	  bind bitfield' (bind (constraints cstr) [metaid name])
+      | Ast.MetaExprList(name,len,cstr,_,_) ->
+	  bind (constraints cstr) (bind (listlen len) [metaid name])
       | Ast.DisjExpr(exps) -> bind_disj (List.map k exps)
       | _ -> option_default) in
 
   let astfvfrag recursor k ft =
     bind (k ft)
       (match Ast.unwrap ft with
-	Ast.MetaFormatList(pct,name,Ast.MetaListLen (lenname,_,_),_,_) ->
-	  [metaid name;metaid lenname]
-      |	Ast.MetaFormatList(pct,name,_,_,_) -> [metaid name]
+	Ast.MetaFormatList
+	  (pct,name,len,cstr,_,_) ->
+	  bind (constraints cstr) (bind (listlen len) [metaid name])
       | _ -> option_default) in
 
   let astfvfmt recursor k ft =
     bind (k ft)
       (match Ast.unwrap ft with
-	Ast.MetaFormat(name,_,_,_) ->
-	  (* constraint can only be a regexp, so no need to check
-	     include_constraints or check for Ast.IdNegIdSet *)
-	  [metaid name]
+	Ast.MetaFormat(name,cstr,_,_) ->
+	  bind (constraints cstr) [metaid name]
       | _ -> option_default) in
 
   let rec collect_assign_names aop =
     match Ast.unwrap aop with
-      Ast.MetaAssign(name,c,_,_) ->
-	if include_constraints then metaid name :: Ast.cstr_meta_names c
-	else [metaid name]
+      Ast.MetaAssign(name,cstr,_,_) ->
+	bind (constraints cstr) [metaid name]
     | _ -> option_default in
 
   let astfvassignop recursor k bop =
@@ -140,9 +149,8 @@ let collect_refs include_constraints =
 
   let rec collect_binary_names bop =
     match Ast.unwrap bop with
-      Ast.MetaBinary(name,c,_,_) ->
-	if include_constraints then metaid name :: Ast.cstr_meta_names c
-	else [metaid name]
+      Ast.MetaBinary(name,cstr,_,_) ->
+	bind (constraints cstr) [metaid name]
     | _ -> option_default in
 
   let astfvbinaryop recursor k bop =
@@ -151,14 +159,21 @@ let collect_refs include_constraints =
   let astfvdecls recursor k d =
     bind (k d)
       (match Ast.unwrap d with
-	Ast.MetaDecl(name,_,_) | Ast.MetaField(name,_,_) -> [metaid name]
-      | Ast.MetaFieldList(name,Ast.MetaListLen(lenname,_,_),_,_) ->
-	  [metaid name;metaid lenname]
-      | Ast.MetaFieldList(name,_,_,_) ->
-	  [metaid name]
+	Ast.MetaDecl(name,cstr,_,_) ->
+	  bind (constraints cstr) [metaid name]
       | Ast.DisjDecl(decls) -> bind_disj (List.map k decls)
       | _ -> option_default) in
 
+  let astfvfields recursor k d =
+    bind (k d)
+      (match Ast.unwrap d with
+	Ast.MetaField(name,cstr,_,_) ->
+	  bind (constraints cstr) [metaid name]
+      | Ast.MetaFieldList(name,len,cstr,_,_) ->
+	  bind (constraints cstr) (bind (listlen len) [metaid name])
+      | Ast.DisjField(decls) -> bind_disj (List.map k decls)
+      | _ -> option_default) in
+
   let astfvfullType recursor k ty =
     bind (k ty)
       (match Ast.unwrap ty with
@@ -168,33 +183,32 @@ let collect_refs include_constraints =
   let astfvtypeC recursor k ty =
     bind (k ty)
       (match Ast.unwrap ty with
-	Ast.MetaType(name,_,_) -> [metaid name]
+	Ast.MetaType(name,cstr,_,_) ->
+	  bind (constraints cstr) [metaid name]
       | _ -> option_default) in
 
   let astfvinit recursor k ty =
     bind (k ty)
       (match Ast.unwrap ty with
-	Ast.MetaInit(name,_,_) -> [metaid name]
-      | Ast.MetaInitList(name,Ast.MetaListLen(lenname,_,_),_,_) ->
-	  [metaid name;metaid lenname]
-      | Ast.MetaInitList(name,_,_,_) -> [metaid name]
+	Ast.MetaInit(name,cstr,_,_) -> [metaid name]
+      | Ast.MetaInitList(name,len,cstr,_,_) ->
+	  bind (constraints cstr) (bind (listlen len) [metaid name])
       | _ -> option_default) in
 
   let astfvparam recursor k p =
     bind (k p)
       (match Ast.unwrap p with
-	Ast.MetaParam(name,_,_) -> [metaid name]
-      | Ast.MetaParamList(name,Ast.MetaListLen(lenname,_,_),_,_) ->
-	  [metaid name;metaid lenname]
-      | Ast.MetaParamList(name,_,_,_) -> [metaid name]
+	Ast.MetaParam(name,cstr,_,_) ->
+	  bind (constraints cstr) [metaid name]
+      | Ast.MetaParamList(name,len,cstr,_,_) ->
+	  bind (constraints cstr) (bind (listlen len) [metaid name])
       | _ -> option_default) in
 
   let astfvdefine_param recursor k p =
     bind (k p)
       (match Ast.unwrap p with
-	Ast.MetaDParamList(name,Ast.MetaListLen(lenname,_,_),_,_) ->
-	  [metaid name;metaid lenname]
-      | Ast.MetaDParamList(name,_,_,_) -> [metaid name]
+	Ast.MetaDParamList(name,len,cstr,_,_) ->
+	  bind (constraints cstr) (bind (listlen len) [metaid name])
       | _ -> option_default) in
 
   let astfvrule_elem recursor k re =
@@ -202,11 +216,10 @@ let collect_refs include_constraints =
     bind (k re)
       (nub
 	 (match Ast.unwrap re with
-	   Ast.MetaRuleElem(name,_,_) | Ast.MetaStmt(name,_,_,_) ->
-	     [metaid name]
-	 | Ast.MetaStmtList(name,Ast.MetaListLen(lenname,_,_),_,_) ->
-	     [metaid name;metaid lenname]
-	 | Ast.MetaStmtList(name,_,_,_) -> [metaid name]
+	   Ast.MetaRuleElem(name,cstr,_,_) | Ast.MetaStmt(name,cstr,_,_,_) ->
+	     bind (constraints cstr) [metaid name]
+	 | Ast.MetaStmtList(name,len,cstr,_,_) ->
+	     bind (constraints cstr) (bind (listlen len) [metaid name])
 	 | _ -> option_default)) in
 
   let astfvstatement recursor k s =
@@ -234,10 +247,10 @@ let collect_refs include_constraints =
   V.combiner bind option_default
     mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
     mcode mcode mcode mcode
-    donothing donothing donothing donothing donothing
+    donothing donothing donothing donothing donothing donothing
     astfvident astfvexpr astfvfrag astfvfmt astfvassignop astfvbinaryop
     astfvfullType astfvtypeC astfvinit astfvparam astfvdefine_param
-    astfvdecls donothing
+    astfvdecls donothing astfvfields donothing
     astfvrule_elem astfvstatement donothing donothing donothing_a
 
 let collect_all_refs = collect_refs true
@@ -285,6 +298,7 @@ let collect_pos_positions =
     donothing donothing donothing donothing donothing
     donothing donothing donothing donothing donothing donothing donothing
     donothing donothing donothing donothing donothing donothing
+    donothing donothing donothing
     cprule_elem cpstmt donothing donothing donothing
 
 (* ---------------------------------------------------------------- *)
@@ -309,34 +323,40 @@ let collect_saved =
   let type_collect res ty =
     bind res (Common.set (Ast.meta_names_of_fullType ty)) in
 
+  let fvbitfield bitfield =
+    match bitfield with
+      Some (Ast.MetaListLen (lenname,_,_,_)) -> [metaid lenname]
+    | _ -> option_default in
+
   let astfvexpr recursor k e =
     let tymetas =
       match Ast.unwrap e with
-	Ast.MetaExpr(name,_,_,Some type_list,_,_) ->
-	  List.fold_left type_collect option_default type_list
+	Ast.MetaExpr(name,_,_,Some type_list,_,_,bitfield) ->
+	  List.fold_left type_collect (fvbitfield bitfield) type_list
       |	_ -> [] in
     let vars =
       bind (k e)
 	(match Ast.unwrap e with
-	  Ast.MetaErr(name,_,Ast.Saved,_) | Ast.MetaExpr(name,_,Ast.Saved,_,_,_)
-	  -> [metaid name]
-	| Ast.MetaExprList(name,Ast.MetaListLen (lenname,ls,_),ns,_) ->
+	  Ast.MetaErr(name,_,Ast.Saved,_) -> [metaid name]
+	| Ast.MetaExpr(name,_,Ast.Saved,_,_,_,bitfield) ->
+	    bind [metaid name] (fvbitfield bitfield)
+	| Ast.MetaExprList(name,Ast.MetaListLen (lenname,_,ls,_),_,ns,_) ->
 	    let namesaved =
 	      match ns with Ast.Saved -> [metaid name] | _ -> [] in
 	    let lensaved =
 	      match ls with Ast.Saved -> [metaid lenname] | _ -> [] in
 	    lensaved @ namesaved
-	| Ast.MetaExprList(name,_,Ast.Saved,_) -> [metaid name]
+	| Ast.MetaExprList(name,_,_,Ast.Saved,_) -> [metaid name]
 	| _ -> option_default) in
     bind tymetas vars in
 
   let astfvfrag recursor k ft =
     bind (k ft)
       (match Ast.unwrap ft with
-	Ast.MetaFormatList(pct,name,Ast.MetaListLen (lenname,_,_),
+	Ast.MetaFormatList(pct,name,Ast.MetaListLen (lenname,_,_,_),_,
 			   Ast.Saved,_) ->
 	  [metaid name;metaid lenname]
-      |	Ast.MetaFormatList(pct,name,_,Ast.Saved,_) -> [metaid name]
+      |	Ast.MetaFormatList(pct,name,_,_,Ast.Saved,_) -> [metaid name]
       | _ -> option_default) in
 
   let astfvfmt recursor k ft =
@@ -360,58 +380,67 @@ let collect_saved =
   let astfvtypeC recursor k ty =
     bind (k ty)
       (match Ast.unwrap ty with
-	Ast.MetaType(name,Ast.Saved,_) -> [metaid name]
+	Ast.MetaType(name,_,Ast.Saved,_) -> [metaid name]
       | _ -> option_default) in
 
   let astfvinit recursor k ty =
     bind (k ty)
       (match Ast.unwrap ty with
-	Ast.MetaInit(name,Ast.Saved,_) -> [metaid name]
-      |	Ast.MetaInitList(name,Ast.MetaListLen (lenname,ls,_),ns,_) ->
+	Ast.MetaInit(name,_,Ast.Saved,_) -> [metaid name]
+      |	Ast.MetaInitList(name,Ast.MetaListLen (lenname,_,ls,_),_,ns,_) ->
 	  let namesaved =
 	    match ns with Ast.Saved -> [metaid name] | _ -> [] in
 	  let lensaved =
 	    match ls with Ast.Saved -> [metaid lenname] | _ -> [] in
 	  lensaved @ namesaved
+      | Ast.MetaInitList(name,len,_,ns,_) ->
+	  (match ns with Ast.Saved -> [metaid name] | _ -> [])
       | _ -> option_default) in
 
   let astfvparam recursor k p =
     bind (k p)
       (match Ast.unwrap p with
-	Ast.MetaParam(name,Ast.Saved,_) -> [metaid name]
-      | Ast.MetaParamList(name,Ast.MetaListLen (lenname,ls,_),ns,_) ->
+	Ast.MetaParam(name,_,Ast.Saved,_) -> [metaid name]
+      | Ast.MetaParamList(name,Ast.MetaListLen (lenname,_,ls,_),_,ns,_) ->
 	  let namesaved =
 	    match ns with Ast.Saved -> [metaid name] | _ -> [] in
 	  let lensaved =
 	    match ls with Ast.Saved -> [metaid lenname] | _ -> [] in
 	  lensaved @ namesaved
-      | Ast.MetaParamList(name,_,Ast.Saved,_) -> [metaid name]
+      | Ast.MetaParamList(name,_,_,Ast.Saved,_) -> [metaid name]
       | _ -> option_default) in
 
   let astfvdefine_param recursor k p =
     bind (k p)
       (match Ast.unwrap p with
-	Ast.MetaDParamList(name,Ast.MetaListLen (lenname,ls,_),ns,_) ->
+	Ast.MetaDParamList(name,Ast.MetaListLen (lenname,_,ls,_),_,ns,_) ->
 	  let namesaved =
 	    match ns with Ast.Saved -> [metaid name] | _ -> [] in
 	  let lensaved =
 	    match ls with Ast.Saved -> [metaid lenname] | _ -> [] in
 	  lensaved @ namesaved
-      | Ast.MetaDParamList(name,_,Ast.Saved,_) -> [metaid name]
+      | Ast.MetaDParamList(name,_,_,Ast.Saved,_) -> [metaid name]
       | _ -> option_default) in
 
   let astfvdecls recursor k d =
     bind (k d)
       (match Ast.unwrap d with
-	Ast.MetaDecl(name,Ast.Saved,_) | Ast.MetaField(name,Ast.Saved,_) ->
+	Ast.MetaDecl(name,_,Ast.Saved,_) ->
 	  [metaid name]
-      | Ast.MetaFieldList(name,Ast.MetaListLen (lenname,ls,_),ns,_) ->
+      | _ -> option_default) in
+
+  let astfvfields recursor k d =
+    bind (k d)
+      (match Ast.unwrap d with
+	Ast.MetaField(name,_,Ast.Saved,_) ->
+	  [metaid name]
+      | Ast.MetaFieldList(name,Ast.MetaListLen (lenname,_,ls,_),_,ns,_) ->
 	  let namesaved =
 	    match ns with Ast.Saved -> [metaid name] | _ -> [] in
 	  let lensaved =
 	    match ls with Ast.Saved -> [metaid lenname] | _ -> [] in
 	  lensaved @ namesaved
-      | Ast.MetaFieldList(name,_,Ast.Saved,_) -> [metaid name]
+      | Ast.MetaFieldList(name,_,_,Ast.Saved,_) -> [metaid name]
       | _ -> option_default) in
 
   let astfvrule_elem recursor k re =
@@ -419,15 +448,15 @@ let collect_saved =
     bind (k re)
       (nub
 	 (match Ast.unwrap re with
-	   Ast.MetaRuleElem(name,Ast.Saved,_)
-	 | Ast.MetaStmt(name,Ast.Saved,_,_) -> [metaid name]
-	 | Ast.MetaStmtList(name,Ast.MetaListLen (lenname,ls,_),ns,_) ->
+	   Ast.MetaRuleElem(name,_,Ast.Saved,_)
+	 | Ast.MetaStmt(name,_,Ast.Saved,_,_) -> [metaid name]
+	 | Ast.MetaStmtList(name,Ast.MetaListLen (lenname,_,ls,_),_,ns,_) ->
 	     let namesaved =
 	       match ns with Ast.Saved -> [metaid name] | _ -> [] in
 	     let lensaved =
 	       match ls with Ast.Saved -> [metaid lenname] | _ -> [] in
 	     lensaved @ namesaved
-	 | Ast.MetaStmtList(name,_,Ast.Saved,_) -> [metaid name]
+	 | Ast.MetaStmtList(name,_,_,Ast.Saved,_) -> [metaid name]
 	 | _ -> option_default)) in
 
   let mcode r e =
@@ -441,10 +470,10 @@ let collect_saved =
   V.combiner bind option_default
     mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
     mcode mcode
-    donothing donothing donothing donothing donothing
+    donothing donothing donothing donothing donothing donothing
     astfvident astfvexpr astfvfrag astfvfmt astfvassign astfvbinary donothing
     astfvtypeC astfvinit astfvparam astfvdefine_param astfvdecls donothing
-    astfvrule_elem donothing donothing donothing donothing
+    astfvfields donothing astfvrule_elem donothing donothing donothing donothing
 
 (* ---------------------------------------------------------------- *)
 
@@ -508,7 +537,7 @@ let collect_fresh_seed metavars l =
 
 let collect_in_plus_term =
 
-  let bind x y = x @ y in
+  let bind x y = List.rev_append x y in
   let option_default = [] in
   let donothing r k e = k e in
 
@@ -519,8 +548,7 @@ let collect_in_plus_term =
 
   let annotated_decl decl =
     match Ast.unwrap decl with
-      Ast.DElem(bef,_,_) -> bef
-    | _ -> failwith "not possible" in
+      Ast.DElem(bef,_,_) -> bef in
 
   let astfvrule_elem recursor k re =
     match Ast.unwrap re with
@@ -567,8 +595,8 @@ let collect_in_plus_term =
     mcode mcode donothing donothing
     donothing donothing donothing donothing donothing donothing donothing
     donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing astfvrule_elem astfvstatement donothing donothing
-    donothing
+    donothing donothing donothing donothing donothing astfvrule_elem
+    astfvstatement donothing donothing donothing
 
 let collect_in_plus metavars minirules =
   nub
@@ -669,9 +697,9 @@ let classify_variables metavar_decls minirules used_after =
       structUnionName = Some (fun su ident ->
         let ident' = Common.map_option classify_ident ident in
         Ast.StructUnionName (su, ident'));
-      metaType = Some (fun name _ _ ->
+      metaType = Some (fun name cstr _ _ ->
         let (unitary,inherited) = classify (Ast.unwrap_mcode name,(),(),[]) in
-        Ast.MetaType (name,unitary,inherited))
+        Ast.MetaType (name,cstr,unitary,inherited))
     } ty in
 
   let expression r k e =
@@ -680,37 +708,46 @@ let classify_variables metavar_decls minirules used_after =
       Ast.MetaErr(name,constraints,_,_) ->
 	let (unitary,inherited) = classify name in
 	Ast.rewrap e (Ast.MetaErr(name,constraints,unitary,inherited))
-    | Ast.MetaExpr(name,constraints,_,ty,form,_) ->
+    | Ast.MetaExpr(name,constraints,_,ty,form,_,bitfield) ->
 	let (unitary,inherited) = classify name in
 	let ty = get_option (List.map type_infos) ty in
-	Ast.rewrap e (Ast.MetaExpr(name,constraints,unitary,ty,form,inherited))
-    | Ast.MetaExprList(name,Ast.MetaListLen(lenname,_,_),_,_) ->
+	let bitfield' =
+	  match bitfield with
+	    Some (Ast.MetaListLen (lenname, cstr, _, _)) ->
+	      let (lenunitary,leninherited) = classify lenname in
+	      Some (Ast.MetaListLen (lenname, cstr, lenunitary, leninherited))
+	  | _ -> bitfield in
+	Ast.rewrap e
+	  (Ast.MetaExpr(name,constraints,unitary,ty,form,inherited,bitfield'))
+    | Ast.MetaExprList(name,Ast.MetaListLen(lenname,cstr,_,_),cstr',_,_) ->
 	let (unitary,inherited) = classify name in
 	let (lenunitary,leninherited) = classify lenname in
 	Ast.rewrap e
 	  (Ast.MetaExprList
 	     (name,
-	      Ast.MetaListLen(lenname,lenunitary,leninherited),
-	      unitary,inherited))
-    | Ast.MetaExprList(name,lenname,_,_) ->
+	      Ast.MetaListLen(lenname,cstr,lenunitary,leninherited),
+	      cstr',unitary,inherited))
+    | Ast.MetaExprList(name,lenname,cstr,_,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap e (Ast.MetaExprList(name,lenname,unitary,inherited))
+	Ast.rewrap e (Ast.MetaExprList(name,lenname,cstr,unitary,inherited))
     | _ -> e in
 
   let string_fragment r k ft =
     let ft = k ft in
     match Ast.unwrap ft with
-      Ast.MetaFormatList(pct,name,Ast.MetaListLen (lenname,_,_),_,_) ->
+      Ast.MetaFormatList
+	(pct,name,Ast.MetaListLen (lenname,cstr,_,_),cstr',_,_) ->
 	let (unitary,inherited) = classify name in
 	let (lenunitary,leninherited) = classify lenname in
 	Ast.rewrap ft
 	  (Ast.MetaFormatList
 	     (pct,name,
-	      Ast.MetaListLen(lenname,lenunitary,leninherited),
-	      unitary,inherited))
-    | Ast.MetaFormatList(pct,name,lenname,_,_) ->
+	      Ast.MetaListLen(lenname,cstr,lenunitary,leninherited),
+	      cstr',unitary,inherited))
+    | Ast.MetaFormatList(pct,name,lenname,cstr,_,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap ft (Ast.MetaFormatList(pct,name,lenname,unitary,inherited))
+	Ast.rewrap ft
+	  (Ast.MetaFormatList(pct,name,lenname,cstr,unitary,inherited))
     | _ -> ft in
 
   let string_format r k ft =
@@ -740,108 +777,113 @@ let classify_variables metavar_decls minirules used_after =
   let typeC r k e =
     let e = k e in
     match Ast.unwrap e with
-      Ast.MetaType(name,_,_) ->
+      Ast.MetaType(name,cstr,_,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap e (Ast.MetaType(name,unitary,inherited))
+	Ast.rewrap e (Ast.MetaType(name,cstr,unitary,inherited))
     | _ -> e in
 
   let init r k e =
     let e = k e in
     match Ast.unwrap e with
-      Ast.MetaInit(name,_,_) ->
+      Ast.MetaInit(name,cstr,_,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap e (Ast.MetaInit(name,unitary,inherited))
-    | Ast.MetaInitList(name,Ast.MetaListLen (lenname,_,_),_,_) ->
+	Ast.rewrap e (Ast.MetaInit(name,cstr,unitary,inherited))
+    | Ast.MetaInitList(name,Ast.MetaListLen (lenname,cstr,_,_),cstr',_,_) ->
 	let (unitary,inherited) = classify name in
 	let (lenunitary,leninherited) = classify lenname in
 	Ast.rewrap e
 	  (Ast.MetaInitList
-	     (name,Ast.MetaListLen(lenname,lenunitary,leninherited),
-	      unitary,inherited))
-    | Ast.MetaInitList(name,lenname,_,_) ->
+	     (name,Ast.MetaListLen(lenname,cstr,lenunitary,leninherited),
+	      cstr',unitary,inherited))
+    | Ast.MetaInitList(name,lenname,cstr,_,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap e (Ast.MetaInitList(name,lenname,unitary,inherited))
+	Ast.rewrap e (Ast.MetaInitList(name,lenname,cstr,unitary,inherited))
     | _ -> e in
 
   let param r k e =
     let e = k e in
     match Ast.unwrap e with
-      Ast.MetaParam(name,_,_) ->
+      Ast.MetaParam(name,cstr,_,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap e (Ast.MetaParam(name,unitary,inherited))
-    | Ast.MetaParamList(name,Ast.MetaListLen (lenname,_,_),_,_) ->
+	Ast.rewrap e (Ast.MetaParam(name,cstr,unitary,inherited))
+    | Ast.MetaParamList(name,Ast.MetaListLen (lenname,cstr,_,_),cstr',_,_) ->
 	let (unitary,inherited) = classify name in
 	let (lenunitary,leninherited) = classify lenname in
 	Ast.rewrap e
 	  (Ast.MetaParamList
-	     (name,Ast.MetaListLen(lenname,lenunitary,leninherited),
-	      unitary,inherited))
-    | Ast.MetaParamList(name,lenname,_,_) ->
+	     (name,Ast.MetaListLen(lenname,cstr,lenunitary,leninherited),
+	      cstr',unitary,inherited))
+    | Ast.MetaParamList(name,lenname,cstr,_,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap e (Ast.MetaParamList(name,lenname,unitary,inherited))
+	Ast.rewrap e (Ast.MetaParamList(name,lenname,cstr,unitary,inherited))
     | _ -> e in
 
   let define_param r k e =
     let e = k e in
     match Ast.unwrap e with
-      Ast.MetaDParamList(name,Ast.MetaListLen (lenname,_,_),_,_) ->
+      Ast.MetaDParamList(name,Ast.MetaListLen (lenname,cstr,_,_),cstr',_,_) ->
 	let (unitary,inherited) = classify name in
 	let (lenunitary,leninherited) = classify lenname in
 	Ast.rewrap e
 	  (Ast.MetaDParamList
-	     (name,Ast.MetaListLen(lenname,lenunitary,leninherited),
-	      unitary,inherited))
-    | Ast.MetaDParamList(name,lenname,_,_) ->
+	     (name,Ast.MetaListLen(lenname,cstr,lenunitary,leninherited),
+	      cstr',unitary,inherited))
+    | Ast.MetaDParamList(name,lenname,cstr,_,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap e (Ast.MetaDParamList(name,lenname,unitary,inherited))
+	Ast.rewrap e (Ast.MetaDParamList(name,lenname,cstr,unitary,inherited))
     | _ -> e in
 
   let decl r k e =
     let e = k e in
     match Ast.unwrap e with
-      Ast.MetaDecl(name,_,_) ->
+      Ast.MetaDecl(name,cstr,_,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap e (Ast.MetaDecl(name,unitary,inherited))
-    | Ast.MetaField(name,_,_) ->
+	Ast.rewrap e (Ast.MetaDecl(name,cstr,unitary,inherited))
+    | _ -> e in
+
+  let field r k e =
+    let e = k e in
+    match Ast.unwrap e with
+      Ast.MetaField(name,cstr,_,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap e (Ast.MetaField(name,unitary,inherited))
-    | Ast.MetaFieldList(name,Ast.MetaListLen (lenname,_,_),_,_) ->
+	Ast.rewrap e (Ast.MetaField(name,cstr,unitary,inherited))
+    | Ast.MetaFieldList(name,Ast.MetaListLen (lenname,cstr,_,_),cstr',_,_) ->
 	let (unitary,inherited) = classify name in
 	let (lenunitary,leninherited) = classify lenname in
 	Ast.rewrap e
 	  (Ast.MetaFieldList
-	     (name,Ast.MetaListLen(lenname,lenunitary,leninherited),
-	      unitary,inherited))
-    | Ast.MetaFieldList(name,lenname,_,_) ->
+	     (name,Ast.MetaListLen(lenname,cstr,lenunitary,leninherited),
+	      cstr',unitary,inherited))
+    | Ast.MetaFieldList(name,lenname,cstr,_,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap e (Ast.MetaFieldList(name,lenname,unitary,inherited))
+	Ast.rewrap e (Ast.MetaFieldList(name,lenname,cstr,unitary,inherited))
     | _ -> e in
 
   let rule_elem r k e =
     let e = k e in
     match Ast.unwrap e with
-      Ast.MetaStmt(name,_,msi,_) ->
+      Ast.MetaStmt(name,cstr,_,msi,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap e (Ast.MetaStmt(name,unitary,msi,inherited))
-    | Ast.MetaStmtList(name,Ast.MetaListLen (lenname,_,_),_,_) ->
+	Ast.rewrap e (Ast.MetaStmt(name,cstr,unitary,msi,inherited))
+    | Ast.MetaStmtList(name,Ast.MetaListLen (lenname,cstr,_,_),cstr',_,_) ->
 	let (unitary,inherited) = classify name in
 	let (lenunitary,leninherited) = classify lenname in
 	Ast.rewrap e
 	  (Ast.MetaStmtList
-	     (name,Ast.MetaListLen(lenname,lenunitary,leninherited),
-	      unitary,inherited))
-    | Ast.MetaStmtList(name,lenname,_,_) ->
+	     (name,Ast.MetaListLen(lenname,cstr,lenunitary,leninherited),
+	      cstr',unitary,inherited))
+    | Ast.MetaStmtList(name,lenname,cstr,_,_) ->
 	let (unitary,inherited) = classify name in
-	Ast.rewrap e (Ast.MetaStmtList(name,lenname,unitary,inherited))
+	Ast.rewrap e (Ast.MetaStmtList(name,lenname,cstr,unitary,inherited))
     | _ -> e in
 
   let fn = V.rebuilder
       mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
       mcode mcode
-      donothing donothing donothing donothing donothing
+      donothing donothing donothing donothing donothing donothing
       ident expression string_fragment string_format assignop binaryop
       donothing typeC
-      init param define_param decl donothing rule_elem
+      init param define_param decl donothing field donothing rule_elem
       donothing donothing donothing donothing in
 
   List.map fn.V.rebuilder_top_level minirules
@@ -911,6 +953,7 @@ let astfvs metavars bound =
     | Ast.TopId _ -> "TopId"
     | Ast.TopInit _ -> "TopInit"
     | Ast.Include _ -> "Include"
+    | Ast.MetaInclude _ -> "MetaInclude"
     | Ast.Undef _ -> "Undef"
     | Ast.DefineHeader _ -> "DefineHeader"
     | Ast.Pragma _ -> "Pragma"
@@ -942,6 +985,7 @@ let astfvs metavars bound =
     {(k re) with
       Ast.free_vars = matched;
       Ast.minus_free_vars = munbound;
+      Ast.minus_nc_free_vars = minus_nc_free;
       Ast.fresh_vars = fresh;
       Ast.inherited = inherited;
       Ast.positive_inherited_positions = inherited_pos;
@@ -1022,9 +1066,10 @@ let astfvs metavars bound =
   V.rebuilder
     mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
     mcode mcode
-    donothing donothing astfvstatement_dots donothing donothing
+    donothing donothing astfvstatement_dots donothing donothing donothing
     donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing donothing
+    donothing donothing donothing donothing donothing donothing donothing
+    donothing
     astfvrule_elem astfvstatement astfvcase_line astfvtoplevel donothing
 
 (*
@@ -1045,12 +1090,12 @@ let collect_astfvs rules =
       [] -> []
     | (metavars, rule)::rules ->
         match rule with
-          Ast.ScriptRule (_,_,_,_,script_vars,_) ->
+          Ast.ScriptRule (_,_,_,_,script_vars,_,_) ->
 	    (* why are metavars in rule, but outside for cocci rule??? *)
             let bound = script_vars @ bound in
 	    rule::(loop bound rules)
-        | Ast.InitialScriptRule (_,_,_,_,_)
-	| Ast.FinalScriptRule (_,_,_,_,_) ->
+        | Ast.InitialScriptRule (_,_,_,_,_,_)
+	| Ast.FinalScriptRule (_,_,_,_,_,_) ->
 	    (* bound stays as is because init/finalize provides no names, so
 	       inheritance by others is not possible *)
 	    rule::(loop bound rules)
@@ -1077,11 +1122,12 @@ let get_neg_pos_list (_,rule) used_after_list =
      Common.union_set ap1 ap2) in
   let option_default = ([],[],[]) in
   let metaid (x,_,_,_) = x in
-  let get_neg_pos_constraints constraints =
-    Ast.cstr_fold
+  let get_neg_pos_constraints =
+    Ast.cstr_fold_sign
       { Ast.empty_cstr_transformer with
-	Ast.cstr_meta_name = Some (fun mn accu -> mn :: accu);
-	cstr_script = Some (fun _ accu -> accu) } constraints [] in
+	Ast.cstr_meta_name = Some (fun mn (pos, neg) -> (mn :: pos, neg)) }
+      { Ast.empty_cstr_transformer with
+	Ast.cstr_meta_name = Some (fun mn (pos, neg) -> (pos, mn :: neg)) } in
   let mcode r mc =
     List.fold_left
       (function (pos_vars,neg_vars,all_vars) ->
@@ -1090,13 +1136,15 @@ let get_neg_pos_list (_,rule) used_after_list =
 	   all_vars are position variables with the annotation ALL.
 	   These sets can overlap. *)
 	(function
-	    Ast.MetaPos(name,constraints,Ast.PER,_,_) ->
-	      let constraint_vars = get_neg_pos_constraints constraints in
-	      ((metaid name)::pos_vars,constraint_vars@neg_vars,all_vars)
-	  | Ast.MetaPos(name,constraints,Ast.ALL,_,_) ->
-	      let constraint_vars = get_neg_pos_constraints constraints in
-	      let name = metaid name in
-	      (name::pos_vars,constraint_vars@neg_vars,name::all_vars)))
+	    Ast.MetaPos(name,constraints,collect,_,_) ->
+	      let name' = metaid name in
+	      let pos_vars', neg_vars' =
+		get_neg_pos_constraints constraints (pos_vars, neg_vars) in
+	      let (pos_vars'', all_vars') =
+		match collect with
+		  Ast.PER -> (name' :: pos_vars', all_vars)
+		| Ast.ALL -> (pos_vars', name' :: all_vars) in
+	      (pos_vars'', neg_vars',all_vars')))
       option_default (Ast.get_pos_var mc) in
   let v =
     V.combiner bind option_default
@@ -1104,18 +1152,25 @@ let get_neg_pos_list (_,rule) used_after_list =
     mcode mcode
     donothing donothing donothing donothing donothing donothing donothing
     donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing in
+    donothing donothing donothing donothing donothing donothing donothing
+    donothing donothing donothing donothing donothing in
   match rule with
-    Ast.CocciRule(_,_,minirules,_,_) ->
+    Ast.CocciRule(rule_name,_,minirules,_,_) ->
       List.map
 	(function toplevel ->
 	  let (positions,neg_positions,all_positions) =
 	    v.V.combiner_top_level toplevel in
-	  (if List.exists (function p -> List.mem p neg_positions) positions
-	  then
+	  (try
+	    let (badr,badnm) =
+	      List.find (function p -> List.mem p neg_positions) positions in
 	    failwith
-	      "a variable cannot be used both as a position and a constraint");
+	      (Printf.sprintf
+		 "Variable %s in %s cannot be used as both a position and a constraint"
+		 (if badr = rule_name
+		 then badnm
+		 else Printf.sprintf "%s.%s" badr badnm)
+		 rule_name)
+	      with Not_found -> ());
 	  (neg_positions, all_positions))
 	minirules
   | Ast.ScriptRule _ | Ast.InitialScriptRule _ | Ast.FinalScriptRule _ ->
@@ -1133,7 +1188,7 @@ are inherited, ie used but not defined.  These are accumulated back to
 their point of definition. *)
 
 
-let collect_top_level_used_after metavar_rule_list =
+let collect_top_level_used_after metavar_rule_list dropped_rules =
   let drop_virt = List.filter (function ("virtual",_) -> false | _ -> true) in
   let (used_after,used_after_lists) =
     List.fold_right
@@ -1141,17 +1196,17 @@ let collect_top_level_used_after metavar_rule_list =
 	function (used_after,used_after_lists) ->
 	  let locally_defined =
             match r with
-              Ast.ScriptRule (_,_,_,_,free_vars,_) -> free_vars
+              Ast.ScriptRule (_,_,_,_,free_vars,_,_) -> free_vars
 	    | _ -> List.map Ast.get_meta_name metavar_list in
 	  let continue_propagation =
 	    List.filter (function x -> not(List.mem x locally_defined))
 	      used_after in
 	  let free_vars =
             match r with
-              Ast.ScriptRule (_,_,_,mv,_,_) ->
+              Ast.ScriptRule (_,_,_,mv,_,_,_) ->
                 drop_virt(List.map (function (_,(r,v),_,_) -> (r,v)) mv)
-            | Ast.InitialScriptRule (_,_,_,mv,_)
-	    | Ast.FinalScriptRule (_,_,_,mv,_) ->
+            | Ast.InitialScriptRule (_,_,_,mv,_,_)
+	    | Ast.FinalScriptRule (_,_,_,mv,_,_) ->
 		(* only virtual identifiers *)
 		[]
             | Ast.CocciRule (_,_,rule,_,_) ->
@@ -1164,6 +1219,10 @@ let collect_top_level_used_after metavar_rule_list =
 	  (Common.union_set inherited continue_propagation,
 	   used_after::used_after_lists))
       metavar_rule_list ([],[]) in
+  let used_after =
+    List.filter
+      (function (rule,_) -> not(List.mem rule dropped_rules))
+      used_after in
   match used_after with
     [] -> used_after_lists
   | _ ->
@@ -1254,15 +1313,16 @@ let collect_local_used_after metavars minirules used_after =
 
 
 
-let collect_used_after metavar_rule_list =
-  let used_after_lists = collect_top_level_used_after metavar_rule_list in
+let collect_used_after metavar_rule_list dropped_rules =
+  let used_after_lists =
+    collect_top_level_used_after metavar_rule_list dropped_rules in
   List.map2
     (function (metavars,r) ->
       function used_after ->
         match r with
-          Ast.ScriptRule (_,_,_,_,_,_) (* no minirules, so nothing to do? *)
-	| Ast.InitialScriptRule (_,_,_,_,_)
-	| Ast.FinalScriptRule (_,_,_,_,_) ->
+          Ast.ScriptRule (_,_,_,_,_,_,_) (* no minirules, so nothing to do? *)
+	| Ast.InitialScriptRule (_,_,_,_,_,_)
+	| Ast.FinalScriptRule (_,_,_,_,_,_) ->
 	    ([], [used_after], [[]], [])
         | Ast.CocciRule (name, rule_info, minirules, _,_) ->
           collect_local_used_after metavars minirules used_after
@@ -1276,11 +1336,11 @@ let rec split4 = function
 (* ---------------------------------------------------------------- *)
 (* entry point *)
 
-let free_vars rules =
+let free_vars rules dropped_rules =
   let metavars = List.map (function (mv,rule) -> mv) rules in
   let (fvs_lists,used_after_matched_lists,
        fresh_used_after_lists,fresh_used_after_lists_seeds) =
-    split4 (collect_used_after rules) in
+    split4 (collect_used_after rules dropped_rules) in
   let neg_pos_lists =
     List.map2 get_neg_pos_list rules used_after_matched_lists in
   let positions_list = (* for all rules, assume all positions are used after *)
