@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -9,9 +9,11 @@ bindings.  These are combined in ctlcocciintegration, ie after the CTL
 generation. *)
 
 val free_vars : Ast_cocci.rule_with_metavars list ->
+  string list (* droped rules *) ->
   (Ast_cocci.metavar list list) * (Ast_cocci.rule list) *
     (((Ast_cocci.meta_name list) list) list) (*fvs of the rule*) *
-    (((Ast_cocci.meta_name list) list) list) (*negated position vars*) *
+    (((Ast_cocci.meta_name list * Ast_cocci.meta_name list)
+	list) list) (*negated position vars*) *
     ((((Ast_cocci.meta_name list) list) list) (*used after list*) *
        (((Ast_cocci.meta_name list) list) list) (*fresh used after list*) *
        (((Ast_cocci.meta_name list) list)list)(*fresh used after list seeds*))*
