@@ -1,12 +1,11 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
 
 module Ast = Ast_cocci
 module V = Visitor_ast
-module TC = Type_cocci
 
 (* Issues:
 
@@ -37,7 +36,15 @@ type combine =
     And of combine list | Or of combine list | Not of combine
   | Elem of string | False | True
 
-let false_on_top_err = "False should not be in the final result.  Perhaps your rule doesn't contain any +/-/* code, or you have a failed dependency.  If the problem is not clear, try the option --debug-parse-cocci."
+let false_on_top_err virt =
+  Printf.sprintf
+    "No rules apply.  Perhaps your semantic patch doesn't contain any +/-/* code, or you have a failed dependency.  If the problem is not clear, try --debug-parse-cocci%s."
+    (match virt with
+      [] -> ""
+    | x::_ ->
+	Printf.sprintf
+	  " or check whether any virtual rules (e.g., %s) should be defined"
+	  x)
 
 let rec dep2c = function
     And l -> Printf.sprintf "(%s)" (String.concat "&" (List.map dep2c l))
@@ -94,7 +101,7 @@ let reduce_glimpse x =
   let res = List.sort compare res in
   List.map (function (_,x) -> x) res
 
-let interpret_glimpse strict x =
+let interpret_glimpse strict x virt =
   let rec loop = function
       Elem x -> x
     | Not x -> failwith "not unexpected in glimpse arg"
@@ -108,17 +115,17 @@ let interpret_glimpse strict x =
 	else "True"
     | False ->
 	if strict
-	then failwith false_on_top_err
+	then failwith (false_on_top_err virt)
 	else "False" in
   match x with
     True -> None
   | False when strict ->
-      failwith false_on_top_err
+      failwith (false_on_top_err virt)
   | _ ->
       Some (if strict then List.map loop (x::reduce_glimpse x) else [loop x])
 
 (* grep only does or *)
-let interpret_grep strict x =
+let interpret_grep strict x virt =
   let add x l = if List.mem x l then l else x :: l in
   let rec loop collected = function
       Elem x -> add x collected
@@ -134,15 +141,17 @@ let interpret_grep strict x =
 	else add "True" collected
     | False ->
 	if strict
-	then failwith false_on_top_err
+	then failwith (false_on_top_err virt)
 	else add "False" collected in
   match x with
     True -> None
   | False when strict ->
-      failwith false_on_top_err
+      failwith (false_on_top_err virt)
   | _ -> Some (loop [] x)
 
-let interpret_cocci_git_grep strict x =
+let max_cnf = 5
+
+let interpret_cocci_git_grep strict x virt =
   (* convert to cnf *)
   let subset l1 l2 = List.for_all (fun e1 -> List.mem e1 l2) l1 in
   let opt_union_set longer shorter =
@@ -160,24 +169,32 @@ let interpret_cocci_git_grep strict x =
       Elem x -> [[x]]
     | Not x -> failwith "not unexpected in coccigrep arg"
     | And l ->
-	List.fold_left opt_union_set [] (List.map cnf l)
+	(match List.map cnf l with
+	  [] -> failwith "and should not be empty"
+	| x ::xs -> List.fold_left opt_union_set x xs)
     | Or l ->
 	let l = List.map cnf l in
-	(match l with
-	  fst::rest ->
-	    List.fold_left
-	      (function prev ->
-		function cur ->
-		  List.fold_left opt_union_set
-		    []
-		    (List.map (fun x -> List.map (Common.union_set x) prev)
-		       cur))
-	      fst rest
-	| [] -> [[]]) (* false *)
+	let icount =
+	  List.length
+	    (List.filter (function [] | [_] -> false | _ -> true) l) in
+	if icount > max_cnf
+	then [] (* true *)
+	else
+	  (match l with
+	    fst::rest ->
+	      List.fold_left
+		(function prev ->
+		  function cur ->
+		    List.fold_left opt_union_set
+		      []
+		      (List.map (fun x -> List.map (Common.union_set x) prev)
+			 cur))
+		fst rest
+	  | [] -> [[]]) (* false *)
     | True -> []
     | False ->
 	if strict
-	then failwith false_on_top_err
+	then failwith (false_on_top_err virt)
 	else [[]] in
   let optimize (l : string list list) =
     let l = List.map (function clause -> (List.length clause, clause)) l in
@@ -196,7 +213,7 @@ let interpret_cocci_git_grep strict x =
   match x with
     True -> None
   | False when strict ->
-      failwith false_on_top_err
+      failwith (false_on_top_err virt)
   | _ ->
       let orify l = Str.regexp (String.concat "\\|" (List.map wordify l)) in
       let res1 = orify (atoms [] x) in (* all atoms *)
@@ -213,10 +230,9 @@ let interpret_cocci_git_grep strict x =
 	  res in
       Some (res1,res2,res3)
 
-let combine2c x =
-  match interpret_glimpse false x with
-    None -> "None"
-  | Some x -> String.concat " || " x
+let interpret_idutils = function
+    True -> None
+  | x -> Some x
 
 let norm = function
     And l -> And (List.sort compare l)
@@ -238,7 +254,7 @@ let intersect l1 l2 = List.filter (function l1e -> List.mem l1e l2) l1
 
 let minus_set l1 l2 = List.filter (function l1e -> not (List.mem l1e l2)) l1
 
-let rec insert x l = merge [x] l
+let insert x l = merge [x] l
 
 let rec build_and x y =
   if x = y
@@ -301,7 +317,7 @@ and build_or x y =
 let keep x = Elem x
 let drop x = True
 
-let do_get_constants constants keywords env neg_pos =
+let do_get_constants constants keywords env (neg_pos,_) =
   let donothing r k e = k e in
   let option_default = True in
   let bad_default = False in
@@ -346,6 +362,27 @@ let do_get_constants constants keywords env neg_pos =
 		    prev l)
 	  option_default fresh in
 
+  let rec cstr r k c =
+    match c with
+      Ast.CstrFalse -> False
+    | Ast.CstrTrue -> True
+    | Ast.CstrAnd list ->
+	List.fold_left (fun accu c -> build_and accu (cstr r k c)) True list
+    | Ast.CstrOr list ->
+	List.fold_left (fun accu c -> build_or accu (cstr r k c)) False list
+    | Ast.CstrConstant (Ast.CstrString s) -> True
+    | Ast.CstrMeta_name mv -> inherited mv
+    | Ast.CstrScript (_, (_name, _lang, params, _pos, _code)) ->
+	List.fold_left (fun accu (mv, _) -> build_and accu (inherited mv)) True
+	  params
+    | Ast.CstrSub list ->
+	List.fold_left (fun accu mv -> build_or accu (inherited mv)) False list
+    | Ast.CstrExpr e -> r.V.combiner_expression e
+    | Ast.CstrType ty -> r.V.combiner_fullType ty
+    | Ast.CstrNot _
+    | Ast.CstrConstant (Ast.CstrInt _)
+    | Ast.CstrRegexp _ | Ast.CstrOperator _ -> True in
+
   let ident r k i =
     match Ast.unwrap i with
       Ast.Id(name) ->
@@ -353,30 +390,39 @@ let do_get_constants constants keywords env neg_pos =
 	  (match Ast.unwrap_mcode name with
 	    "NULL" -> keywords "NULL"
 	  | nm -> constants nm)
-    | Ast.MetaId(name,Ast.IdPosIdSet(strs,mids),_,_)
-    | Ast.MetaFunc(name,Ast.IdPosIdSet(strs,mids),_,_)
-    | Ast.MetaLocalFunc(name,Ast.IdPosIdSet(strs,mids),_,_) ->
-	let cur =
-	  build_or
-	    (disj_union_all (List.map constants strs))
-	    (disj_union_all (List.map inherited mids)) in
-	bind (k i) (bind (minherited name) cur)
-    | Ast.MetaId(name,_,_,_) | Ast.MetaFunc(name,_,_,_)
-    | Ast.MetaLocalFunc(name,_,_,_) ->
-	bind (k i) (minherited name)
+    | Ast.MetaId(name,c,_,_)
+    | Ast.MetaFunc(name,c,_,_)
+    | Ast.MetaLocalFunc(name,c,_,_) ->
+	bind (cstr r k (Ast.cstr_push_not c)) (bind (k i) (minherited name))
     | Ast.DisjId(ids) -> disj_union_all (List.map r.V.combiner_ident ids)
     | _ -> k i in
 
-  let rec type_collect res = function
-      TC.ConstVol(_,ty) | TC.Pointer(ty) | TC.FunctionPointer(ty)
-    | TC.Array(ty) -> type_collect res ty
-    | TC.Decimal _ -> build_or res (keywords "decimal")
-    | TC.MetaType(tyname,_,_) ->
-	build_or res (inherited tyname)
-    | TC.TypeName(s) -> build_or res (constants s)
-    | TC.EnumName(TC.Name s) -> build_or res (constants s)
-    | TC.StructUnionName(_,TC.Name s) -> build_or res (constants s)
-    | ty -> res in
+  let type_collect ty =
+    let add x res = build_and res x in
+    let add_ident ident =
+      match Ast.unwrap ident with
+        Ast.Id name -> add (constants (Ast.unwrap_mcode name))
+      | Ast.MetaId(name,_,_,_) -> add (minherited name)
+      | _ -> Common.id in
+    let enumOrStructUnionName _ ident res =
+      Common.default res (fun ident' -> add_ident ident' res) ident in
+    let pieces ty res =
+      Ast.fullType_fold
+	{ Ast.empty_transformer with
+	  Ast.decimal = Some (fun _ _ _ _ _ _ -> add (keywords "decimal"));
+	  metaType =
+	  Some (fun tyname _ _ _ -> add (inherited (Ast.unwrap_mcode tyname)));
+	  typeName =
+	  Some(fun tyname -> add (constants (Ast.unwrap_mcode tyname)));
+	  enumName = Some enumOrStructUnionName;
+	  structUnionName = Some enumOrStructUnionName
+	} ty res in
+    let rec loop ty =
+      match Ast.unwrap ty with
+	Ast.DisjType l ->
+	  List.fold_left (fun prev ty -> build_or prev (loop ty)) False l
+      | _ -> pieces ty option_default in
+    loop ty in
 
   (* no point to do anything special for records because glimpse is
      word-oriented *)
@@ -423,14 +469,19 @@ let do_get_constants constants keywords env neg_pos =
 	    Some strs -> constants (String.concat "" (List.rev strs))
 	  | None ->  option_default)
 	*)
-    | Ast.MetaExpr(name,_,_,Some type_list,_,_) ->
-	let types = List.fold_left type_collect option_default type_list in
+    | Ast.MetaExpr(name,_,_,Some type_list,_,_,_) ->
+	let types =
+	  match type_list with
+	    [] -> True (* no constraint *)
+	  | _ -> (* at least one constraint must be satisfied *)
+	      List.fold_left (fun prev ty -> build_or (type_collect ty) prev)
+		(type_collect(List.hd type_list)) (List.tl type_list) in
 	bind (k e) (bind (minherited name) types)
-    | Ast.MetaErr(name,_,_,_) | Ast.MetaExpr(name,_,_,_,_,_) ->
+    | Ast.MetaErr(name,_,_,_) | Ast.MetaExpr(name,_,_,_,_,_,_) ->
 	bind (k e) (minherited name)
-    | Ast.MetaExprList(name,Ast.MetaListLen (lenname,_,_),_,_) ->
+    | Ast.MetaExprList(name,Ast.MetaListLen (lenname,_,_,_),_,_,_) ->
 	bind (k e) (bind (minherited name) (minherited lenname))
-    | Ast.MetaExprList(name,_,_,_) -> minherited name
+    | Ast.MetaExprList(name,_,_,_,_) -> minherited name
     | Ast.SizeOfExpr(sizeof,exp) -> bind (keywords "sizeof") (k e)
     | Ast.SizeOfType(sizeof,lp,ty,rp) -> bind (keywords "sizeof") (k e)
     | Ast.NestExpr(starter,expr_dots,ender,wc,false) -> option_default
@@ -445,9 +496,9 @@ let do_get_constants constants keywords env neg_pos =
   (* cases for metavariabes *)
   let string_fragment r k ft =
     match Ast.unwrap ft with
-      Ast.MetaFormatList(pct,name,Ast.MetaListLen (lenname,_,_),_,_) ->
+      Ast.MetaFormatList(pct,name,Ast.MetaListLen (lenname,_,_,_),_,_,_) ->
 	bind (k ft) (bind (minherited name) (minherited lenname))
-    | Ast.MetaFormatList(pct,name,_,_,_) -> bind (k ft) (minherited name)
+    | Ast.MetaFormatList(pct,name,_,_,_,_) -> bind (k ft) (minherited name)
     | _ -> k ft in
 
   let string_format r k ft =
@@ -462,37 +513,39 @@ let do_get_constants constants keywords env neg_pos =
     | Ast.OptType(ty) -> option_default
     | _ -> k ft in
 
-  let baseType = function
-      Ast.VoidType -> keywords "void"
-    | Ast.CharType -> keywords "char"
-    | Ast.ShortType -> keywords "short"
-    | Ast.ShortIntType -> keywords "short"
-    | Ast.IntType -> keywords "int"
-    | Ast.DoubleType -> keywords "double"
-    | Ast.LongDoubleType -> keywords "double"
-    | Ast.FloatType -> keywords "float"
-    | Ast.LongType | Ast.LongLongType
-    | Ast.LongIntType | Ast.LongLongIntType -> keywords "long"
-    | Ast.SizeType -> keywords "size_t"
-    | Ast.SSizeType -> keywords "ssize_t"
-    | Ast.PtrDiffType -> keywords "ptrdiff_t" in
+  let baseType ty = keywords (Ast.string_of_baseType ty) in
 
   let typeC r k ty =
     match Ast.unwrap ty with
       Ast.BaseType(ty1,strings) -> bind (k ty) (baseType ty1)
+    | Ast.TypeOfExpr(sizeof,lp,e,rp) -> bind (keywords "typeof") (k ty)
+    | Ast.TypeOfType(sizeof,lp,t,rp) -> bind (keywords "typeof") (k ty)
     | Ast.TypeName(name) -> bind (k ty) (constants (Ast.unwrap_mcode name))
-    | Ast.MetaType(name,_,_) -> bind (minherited name) (k ty)
+    | Ast.MetaType(name,_,_,_) -> bind (minherited name) (k ty)
     | _ -> k ty in
 
   let declaration r k d =
     match Ast.unwrap d with
-      Ast.MetaDecl(name,_,_) | Ast.MetaField(name,_,_) ->
+      Ast.MetaDecl(name,_,_,_) ->
 	bind (k d) (minherited name)
-    | Ast.MetaFieldList(name,Ast.MetaListLen(lenname,_,_),_,_) ->
-	bind (minherited name) (bind (minherited lenname) (k d))
     | Ast.DisjDecl(decls) ->
 	disj_union_all (List.map r.V.combiner_declaration decls)
     | Ast.OptDecl(decl) -> option_default
+    (* need things with explicit names too *)
+    | Ast.Init(_,_,_,attr,_,_,_) | Ast.UnInit(_,_,_,attr,_) ->
+	List.fold_left bind (k d)
+	  (List.map (fun attr -> constants (Ast.unwrap_mcode attr)) attr)
+    | _ -> k d in
+
+  let field r k d =
+    match Ast.unwrap d with
+      Ast.MetaField(name,_,_,_) ->
+	bind (k d) (minherited name)
+    | Ast.MetaFieldList(name,Ast.MetaListLen(lenname,_,_,_),_,_,_) ->
+	bind (minherited name) (bind (minherited lenname) (k d))
+    | Ast.DisjField(decls) ->
+	disj_union_all (List.map r.V.combiner_field decls)
+    | Ast.OptField(decl) -> option_default
     | _ -> k d in
 
   let initialiser r k i =
@@ -503,17 +556,26 @@ let do_get_constants constants keywords env neg_pos =
   let parameter r k p =
     match Ast.unwrap p with
       Ast.OptParam(param) -> option_default
-    | Ast.MetaParam(name,_,_) -> bind (k p) (minherited name)
-    | Ast.MetaParamList(name,Ast.MetaListLen(lenname,_,_),_,_) ->
+    | Ast.MetaParam(name,_,_,_) -> bind (k p) (minherited name)
+    | Ast.MetaParamList(name,Ast.MetaListLen(lenname,_,_,_),_,_,_) ->
 	bind (minherited name) (bind (minherited lenname) (k p))
-    | Ast.MetaParamList(name,_,_,_) -> bind (k p) (minherited name)
+    | Ast.MetaParamList(name,_,_,_,_) -> bind (k p) (minherited name)
+    | _ -> k p in
+
+  let define_parameter r k p =
+    match Ast.unwrap p with
+      Ast.MetaDParamList(name,Ast.MetaListLen(lenname,_,_,_),_,_,_) ->
+	bind (minherited name) (bind (minherited lenname) (k p))
+    | Ast.MetaDParamList(name,_,_,_,_) -> bind (k p) (minherited name)
     | _ -> k p in
 
   let rule_elem r k re =
     bind (fresh_info re)
     (match Ast.unwrap re with
-      Ast.MetaRuleElem(name,_,_) | Ast.MetaStmt(name,_,_,_)
-    | Ast.MetaStmtList(name,_,_,_) -> bind (minherited name) (k re)
+      Ast.MetaStmtList(name,Ast.MetaListLen (lenname,_,_,_),_,_,_) ->
+	bind (minherited name) (bind (minherited lenname) (k re))
+    | Ast.MetaRuleElem(name,_,_,_) | Ast.MetaStmt(name,_,_,_,_)
+    | Ast.MetaStmtList(name,_,_,_,_) -> bind (minherited name) (k re)
     | Ast.WhileHeader(whl,lp,exp,rp) ->
 	bind (keywords "while") (k re)
     | Ast.WhileTail(whl,lp,exp,rp,sem) ->
@@ -533,7 +595,8 @@ let do_get_constants constants keywords env neg_pos =
     | Ast.Include(inc,s) ->
 	bind (k re)
 	  (match Ast.unwrap_mcode s with
-	    Ast.Local l | Ast.NonLocal l ->
+	    Ast.AnyInc -> True
+	  | Ast.Local l | Ast.NonLocal l ->
 	      let strings =
 		List.fold_left
 		  (function prev ->
@@ -566,36 +629,27 @@ let do_get_constants constants keywords env neg_pos =
   V.combiner bind option_default
     mcode mcode mcode mcode mcode mcode mcode mcode mcode
     mcode mcode mcode mcode mcode
-    donothing donothing donothing donothing donothing
+    donothing donothing donothing donothing donothing donothing
     ident expression string_fragment string_format donothing donothing
-    fullType typeC initialiser parameter declaration donothing
-    rule_elem statement donothing donothing donothing
+    fullType typeC initialiser parameter define_parameter declaration donothing
+    field donothing rule_elem statement donothing donothing donothing
 
 (* ------------------------------------------------------------------------ *)
 
 (* true means the rule should be analyzed, false means it should be ignored *)
-let rec dependencies env = function
-    Ast.Dep s -> (try List.assoc s env with Not_found -> False)
-  | Ast.AntiDep s -> True
-  | Ast.EverDep s -> (try List.assoc s env with Not_found -> False)
-  | Ast.NeverDep s -> True
-  | Ast.AndDep (d1,d2) -> build_and (dependencies env d1) (dependencies env d2)
-  | Ast.OrDep (d1,d2) -> build_or (dependencies env d1) (dependencies env d2)
-  | Ast.NoDep -> True
-  | Ast.FailDep -> False
-
-(* used with + items to find inconsistencies *)
-let rec exact_dependencies = function
-    Ast.Dep s -> Elem s
-  | Ast.AntiDep s -> Not (Elem s)
-  | Ast.EverDep s -> Elem s
-  | Ast.NeverDep s -> Not (Elem s)
-  | Ast.AndDep (d1,d2) ->
-      build_and (exact_dependencies d1) (exact_dependencies d2)
-  | Ast.OrDep (d1,d2) ->
-      build_or (exact_dependencies d1) (exact_dependencies d2)
-  | Ast.NoDep -> True
+let rec dependencies env d =
+  let rec loop = function
+      Ast.Dep s -> (try List.assoc s env with Not_found -> False)
+    | Ast.AntiDep s -> True
+    | Ast.EverDep s -> (try List.assoc s env with Not_found -> False)
+    | Ast.NeverDep s -> True
+    | Ast.AndDep (d1,d2) -> build_and (loop d1) (loop d2)
+    | Ast.OrDep (d1,d2) -> build_or (loop d1) (loop d2)
+    | Ast.FileIn _ | Ast.NotFileIn _ -> True in
+  match d with
+    Ast.NoDep -> True
   | Ast.FailDep -> False
+  | Ast.ExistsDep d | Ast.ForallDep d -> loop d
 
 (* ------------------------------------------------------------------------ *)
 
@@ -621,8 +675,7 @@ let all_context =
 
   let annotated_decl decl =
     match Ast.unwrap decl with
-      Ast.DElem(bef,_,_) -> bef
-    | _ -> failwith "not possible" in
+      Ast.DElem(bef,_,_) -> bef in
 
   let rule_elem r k e =
     match Ast.unwrap e with
@@ -645,9 +698,9 @@ let all_context =
     mcode mcode mcode mcode mcode mcode mcode mcode mcode
     mcode mcode mcode mcode mcode
     donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing donothing
-    initialiser donothing donothing donothing rule_elem statement
-    donothing donothing donothing
+    donothing donothing donothing donothing donothing donothing donothing
+    initialiser donothing donothing donothing donothing donothing donothing
+    rule_elem statement donothing donothing donothing
 
 (* ------------------------------------------------------------------------ *)
 
@@ -657,7 +710,7 @@ the thing, so no point to keep track of what is added by earlier rules.
 The situation is something like a -> b v (b & c).  We don't actually need
 both b and c, but if we don't have b, then the only way that we can get it is
 fro the first rule matching, in which case the formula is already true. *)
-let rule_fn nm tls exact_dependencies env neg_pos =
+let rule_fn nm tls env neg_pos =
   (* tls seems like it is supposed to relate to multiple minirules.  If we
      were to actually allow that, then the following could be inefficient,
      because it could run sat on the same rule name (x) more than once. *)
@@ -686,15 +739,20 @@ let rule_fn nm tls exact_dependencies env neg_pos =
 	| x -> build_or x rest_info)
     False (List.combine tls neg_pos)
 
-let debug_deps nm deps res =
+let debug_deps nm deps res from_code =
   if !Flag_parsing_cocci.debug_parse_cocci
   then
     begin
       Printf.fprintf stderr "Rule: %s\n" nm;
-      Printf.fprintf stderr "Dependecies: %s\n"
+      Printf.fprintf stderr "Dependencies: %s\n"
 	(Common.format_to_string
-	   (function _ -> Pretty_print_cocci.dep true deps));
-      Printf.fprintf stderr "Result: %s\n\n" (dep2c res)
+	   (function _ -> Pretty_print_cocci.dependency deps));
+      Printf.fprintf stderr "Result: %s\n" (dep2c res);
+      (match from_code with
+	Some deps ->
+	  Printf.fprintf stderr "Result_from_code: %s\n" (dep2c deps)
+      | None -> ());
+      Printf.fprintf stderr "\n";
     end
 
 let run rules neg_pos_vars =
@@ -702,39 +760,52 @@ let run rules neg_pos_vars =
     List.fold_left
       (function (rest_info,env,locals(*dom of env*)) ->
         function
-	    (Ast.ScriptRule (nm,_,deps,mv,_,_),_) ->
+	    (Ast.ScriptRule (nm,_,deps,mv,_,_,_),_) ->
 	      let extra_deps =
 		List.fold_left
 		  (function prev ->
-		    function (_,(rule,_),_,_) ->
-		      if rule = "virtual"
-		      then prev
-		      else Ast.AndDep (Ast.Dep rule,prev))
-		  deps mv in
+		    function
+			(_,("virtual",_),_,_) -> prev
+		      | (_,(rule,_),_,Ast.NoMVInit) -> Ast.Dep rule :: prev
+		      | (_,(rule,_),_,_) ->
+			  (* default initializer, so no dependency *)
+			  prev)
+		  [] mv in
+	      let extra_deps =
+		match extra_deps with
+		  [] -> deps
+		| x::xs ->
+		    let extra_deps =
+		      List.fold_left (fun prev x -> Ast.AndDep(x,prev)) x xs in
+		    match deps with
+		      Ast.NoDep -> Ast.ExistsDep(extra_deps)
+		    | Ast.FailDep -> Ast.FailDep
+		    | Ast.ExistsDep d ->
+			Ast.ExistsDep(Ast.AndDep(d,extra_deps))
+		    | Ast.ForallDep d ->
+			Ast.ForallDep(Ast.AndDep(d,extra_deps)) in
 	      let dependencies = dependencies env extra_deps in
-	      debug_deps nm extra_deps dependencies;
+	      debug_deps nm extra_deps dependencies None;
 	      (match dependencies with
 		False ->
 		  (rest_info, (nm,True)::env, nm::locals)
 	      | dependencies ->
 		  (build_or dependencies rest_info, env, locals))
-          | (Ast.InitialScriptRule (_,_,deps,_,_),_)
-	  | (Ast.FinalScriptRule (_,_,deps,_,_),_) ->
+          | (Ast.InitialScriptRule (_,_,deps,_,_,_),_)
+	  | (Ast.FinalScriptRule (_,_,deps,_,_,_),_) ->
 	      (* initialize and finalize dependencies are irrelevant to
 		 get_constants *)
 	      (* only possible metavariables are virtual *)
 	      (rest_info, env, locals)
           | (Ast.CocciRule (nm,(dep,_,_),cur,_,_),neg_pos_vars) ->
 	      let dependencies = dependencies env dep in
-	      let exact_dependencies = exact_dependencies dep in
-	      let cur_info =
-		rule_fn nm cur exact_dependencies ((nm,True)::env)
-		  neg_pos_vars in
-	      debug_deps nm dep dependencies;
 	      (match dependencies with
 		False -> (rest_info,env,locals)
 	      | dependencies ->
+		  let cur_info =
+		    rule_fn nm cur ((nm,True)::env) neg_pos_vars in
 		  let re_cur_info = build_and dependencies cur_info in
+		  debug_deps nm dep dependencies (Some cur_info);
 		  if List.for_all all_context.V.combiner_top_level cur
 		  then (rest_info,(nm,re_cur_info)::env,nm::locals)
 		  else
@@ -756,20 +827,21 @@ then the CNF regexp for more refined scanning.  git grep uses the second
 CNF representation.
 4. An arbitrary formula, usable by the support for idutils *)
 
-let get_constants rules neg_pos_vars =
+let get_constants rules neg_pos_vars virt =
   if !Flag.worth_trying_opt
   then
     begin
     let res = run rules neg_pos_vars in
-    let grep = interpret_grep true res in (* useful because in string form *)
-    let coccigrep = interpret_cocci_git_grep true res in
+    let grep =
+      interpret_grep true res virt in (* useful because in string form *)
+    let coccigrep = interpret_cocci_git_grep true res virt in
     match !Flag.scanner with
       Flag.NoScanner ->
 	(grep,None,coccigrep,None)
     | Flag.Glimpse ->
-	(grep,interpret_glimpse true res,coccigrep,None)
+	(grep,interpret_glimpse true res virt,coccigrep,None)
     | Flag.IdUtils ->
-	(grep,None,coccigrep,Some res)
+	(grep,None,coccigrep,interpret_idutils res)
     | Flag.CocciGrep | Flag.GitGrep -> (grep,None,coccigrep,None)
     end
   else (None,None,None,None)
