@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -74,12 +74,14 @@ it *)
       mcode mcode
       (donothing Ast0.dotsExpr) (donothing Ast0.dotsInit)
       (donothing Ast0.dotsParam) (donothing Ast0.dotsStmt)
-      (donothing Ast0.dotsDecl) (donothing Ast0.dotsCase)
+      (donothing Ast0.dotsDecl) (donothing Ast0.dotsField)
+      (donothing Ast0.dotsCase)
       (donothing Ast0.dotsDefParam)
       (donothing Ast0.ident) expression  (donothing Ast0.assignOp)
       (donothing Ast0.binaryOp)
       (donothing Ast0.typeC) initialiser
-      (donothing Ast0.param) (donothing Ast0.decl) statement
+      (donothing Ast0.param) (donothing Ast0.decl)
+      (donothing Ast0.field) statement
       (donothing Ast0.forinfo) (donothing Ast0.case_line)
       (donothing Ast0.string_fragment) topfn in
   res.VT0.combiner_rec_top_level e
@@ -98,7 +100,7 @@ declaration *)
 type minus_join_point = Favored | Unfavored | Toplevel | Decl
 
 (* Maps the index of a node to the indices of the mcodes it contains *)
-let root_token_table = (Hashtbl.create(50) : (int, int list) Hashtbl.t)
+let root_token_table = (Hashtbl.create(51) : (int, int list) Hashtbl.t)
 
 let create_root_token_table minus =
   Hashtbl.iter
@@ -111,6 +113,7 @@ let create_root_token_table minus =
 	  | Ast0.DotsParamTag(d) -> Ast0.get_index d
 	  | Ast0.DotsStmtTag(d) -> Ast0.get_index d
 	  | Ast0.DotsDeclTag(d) -> Ast0.get_index d
+	  | Ast0.DotsFieldTag(d) -> Ast0.get_index d
 	  | Ast0.DotsCaseTag(d) -> Ast0.get_index d
 	  | Ast0.DotsDefParamTag(d) -> Ast0.get_index d
 	  | Ast0.IdentTag(d) -> Ast0.get_index d
@@ -123,6 +126,7 @@ let create_root_token_table minus =
 	  | Ast0.ParamTag(d) -> Ast0.get_index d
 	  | Ast0.InitTag(d) -> Ast0.get_index d
 	  | Ast0.DeclTag(d) -> Ast0.get_index d
+	  | Ast0.FieldTag(d) -> Ast0.get_index d
 	  | Ast0.StmtTag(d) -> Ast0.get_index d
 	  | Ast0.ForInfoTag(d) -> Ast0.get_index d
 	  | Ast0.CaseLineTag(d) -> Ast0.get_index d
@@ -214,6 +218,7 @@ bind to that; not good for isomorphisms *)
 
   let sdots r k d = dots r.VT0.combiner_rec_statement k d in
   let ddots r k d = dots r.VT0.combiner_rec_declaration k d in
+  let fdots r k d = dots r.VT0.combiner_rec_field k d in
   let cdots r k d = dots r.VT0.combiner_rec_case_line k d in
 
   (* a case for everything that has a Opt *)
@@ -306,6 +311,13 @@ bind to that; not good for isomorphisms *)
 	r.VT0.combiner_rec_declaration d
     | _ -> do_nothing r k e in
 
+  let field r k e =
+    match Ast0.unwrap e with
+      Ast0.OptField d ->
+	(* put the + code on the thing, not on the opt *)
+	r.VT0.combiner_rec_field d
+    | _ -> do_nothing r k e in
+
   let initialiser r k e =
     match Ast0.unwrap e with
       Ast0.Idots(d,whencode) -> mcode d (* ignore whencode *)
@@ -333,9 +345,9 @@ bind to that; not good for isomorphisms *)
   V0.flat_combiner bind option_default
     mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
     mcode mcode
-    edots idots pdots sdots ddots cdots dpdots
+    edots idots pdots sdots ddots fdots cdots dpdots
     ident expression do_nothing do_nothing
-    typeC initialiser param decl statement forinfo
+    typeC initialiser param decl field statement forinfo
     case_line do_nothing do_top
 
 
@@ -359,6 +371,9 @@ let call_collect_minus context_nodes :
       | Ast0.DotsDeclTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_declaration_dots e)
+      | Ast0.DotsFieldTag(e) ->
+	  (Ast0.get_index e,
+	   (collect_minus_join_points e).VT0.combiner_rec_field_dots e)
       | Ast0.DotsCaseTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_case_line_dots e)
@@ -391,6 +406,9 @@ let call_collect_minus context_nodes :
       | Ast0.DeclTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_declaration e)
+      | Ast0.FieldTag(e) ->
+	  (Ast0.get_index e,
+	   (collect_minus_join_points e).VT0.combiner_rec_field e)
       | Ast0.StmtTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_statement e)
@@ -446,15 +464,18 @@ let verify l =
     l
 
 let process_minus minus =
-  Hashtbl.clear root_token_table;
   create_root_token_table minus;
-  List.concat
-    (List.map
-       (function x ->
-	 let res = call_collect_minus (collect_context x) in
-	 verify res;
-	 res)
-       minus)
+  let res =
+    List.concat
+      (List.map
+	 (function x ->
+	   let res = call_collect_minus (collect_context x) in
+	   verify res;
+	   res)
+	 minus) in
+  Stdcompat.Hashtbl.reset root_token_table;
+  res
+
 (* --------------------------------------------------------------------- *)
 (* --------------------------------------------------------------------- *)
 (* collect the plus tokens *)
@@ -475,6 +496,7 @@ let mk_logicalOp x        = Ast.LogicalOpTag x
 let mk_arithOp x          = Ast.ArithOpTag x
 let mk_logicalOp x        = Ast.LogicalOpTag x
 let mk_declaration x      = Ast.DeclarationTag (Ast0toast.declaration x)
+let mk_field x            = Ast.FieldTag (Ast0toast.field x)
 let mk_topdeclaration x   = Ast.DeclarationTag (Ast0toast.declaration x)
 let mk_storage x          = Ast.StorageTag x
 let mk_inc_file x         = Ast.IncFileTag x
@@ -491,6 +513,7 @@ let mk_exprdots x  = Ast.ExprDotsTag (Ast0toast.expression_dots x)
 let mk_paramdots x = Ast.ParamDotsTag (Ast0toast.parameter_list x)
 let mk_stmtdots x  = Ast.StmtDotsTag (Ast0toast.statement_dots x)
 let mk_decldots x  = Ast.AnnDeclDotsTag (Ast0toast.declaration_dots x)
+let mk_fielddots x  = Ast.AnnFieldDotsTag (Ast0toast.field_dots x)
 let mk_casedots x  = failwith "+ case lines not supported"
 let mk_defpardots x= Ast.DefParDotsTag (Ast0toast.define_param_dots x)
 let mk_typeC x     = Ast.FullTypeTag (Ast0toast.typeC false x)
@@ -597,11 +620,12 @@ let collect_plus_nodes root =
     (mcode mk_storage) (mcode mk_inc_file)
     (do_nothing mk_exprdots) initdots
     (do_nothing mk_paramdots) stmt_dots (do_nothing mk_decldots)
+    (do_nothing mk_fielddots)
     (do_nothing mk_casedots) (do_nothing mk_defpardots)
     (do_nothing mk_ident) (do_nothing mk_expression)
     (do_nothing mk_assignOp) (do_nothing mk_binaryOp)
     (do_nothing mk_typeC) (do_nothing mk_init) (do_nothing mk_param)
-    (do_nothing mk_declaration)
+    (do_nothing mk_declaration) (do_nothing mk_field)
     stmt (do_nothing mk_forinfo) (do_nothing mk_case_line)
     (do_nothing mk_string_fragment) toplevel
 
@@ -625,6 +649,9 @@ let call_collect_plus context_nodes :
       | Ast0.DotsDeclTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_declaration_dots e)
+      | Ast0.DotsFieldTag(e) ->
+	  (Ast0.get_index e,
+	   (collect_plus_nodes e).VT0.combiner_rec_field_dots e)
       | Ast0.DotsCaseTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_case_line_dots e)
@@ -657,6 +684,9 @@ let call_collect_plus context_nodes :
       | Ast0.DeclTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_declaration e)
+      | Ast0.FieldTag(e) ->
+	  (Ast0.get_index e,
+	   (collect_plus_nodes e).VT0.combiner_rec_field e)
       | Ast0.StmtTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_statement e)
@@ -1146,7 +1176,8 @@ let reevaluate_contextness =
       mcode mcode
       donothing donothing donothing donothing donothing donothing donothing
       donothing donothing donothing donothing donothing
-      donothing donothing donothing stmt donothing donothing donothing
+      donothing donothing donothing donothing donothing stmt donothing
+      donothing donothing
       donothing in
   res.VT0.combiner_rec_top_level
 
