@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -13,23 +13,21 @@ let contains_string_constant = ref false
 (* types for metavariable tokens *)
 type info = Ast.meta_name * Ast0.pure * Data.clt
 type midinfo =
-    Ast.meta_name * Data.iconstraints * Ast.seed * Ast0.pure * Data.clt
-type idinfo = Ast.meta_name * Data.iconstraints * Ast0.pure * Data.clt
-type assignOpinfo =
-    Ast.meta_name * Ast0_cocci.assignOpconstraint * Ast0.pure * Data.clt
-type binaryOpinfo =
-    Ast.meta_name * Ast0_cocci.binaryOpconstraint * Ast0.pure * Data.clt
-type expinfo = Ast.meta_name * Data.econstraints * Ast0.pure * Data.clt
+    Ast.meta_name * Ast0.constraints * Ast.seed * Ast0.pure * Data.clt
+type cstrinfo = Ast.meta_name * Ast0.constraints * Ast0.pure * Data.clt
+type assignOpinfo = Ast.meta_name * Ast0.constraints * Ast0.pure * Data.clt
+type binaryOpinfo = Ast.meta_name * Ast0.constraints * Ast0.pure * Data.clt
+type expinfo = Ast.meta_name * Ast0.constraints * Ast0.pure * Data.clt
 type tyinfo = Ast.meta_name * Ast0.typeC list * Ast0.pure * Data.clt
-type list_info = Ast.meta_name * Ast.list_len * Ast0.pure * Data.clt
+type list_info = Ast.meta_name * Ast.list_len * Ast0.constraints * Ast0.pure *
+      Data.clt
 type typed_expinfo =
-    Ast.meta_name * Data.econstraints * Ast0.pure *
-      Type_cocci.typeC list option * Data.clt
-type pos_info = Ast.meta_name * Data.pconstraints * Ast.meta_collect * Data.clt
-
-let get_option fn = function
-    None -> None
-  | Some x -> Some (fn x)
+    Ast.meta_name * Ast0.constraints * Ast0.pure *
+      Ast0.typeC list option * Data.clt
+type typed_expinfo_bitfield =
+    Ast.meta_name * Ast0.constraints * Ast0.pure *
+      Ast0.typeC list option * Data.clt * Ast.list_len option
+type pos_info = Ast.meta_name * Ast0.constraints * Ast.meta_collect * Data.clt
 
 let make_info line logical_line logical_line_end offset col strbef straft 
     isSymbol ws =
@@ -120,15 +118,15 @@ let mkidots str (dot,whencode) =
     "..." -> Ast0.wrap(Ast0.Idots(clt2mcode str dot, whencode))
   | _ -> failwith "cannot happen"
 
-let mkddots str (dot,whencode) =
+let mkfdots str (dot,whencode) =
   match (str,whencode) with
-    ("...",None) -> Ast0.wrap(Ast0.Ddots(clt2mcode str dot, None))
-  | ("...",Some [w]) -> Ast0.wrap(Ast0.Ddots(clt2mcode str dot, Some w))
+    ("...",None) -> Ast0.wrap(Ast0.Fdots(clt2mcode str dot, None))
+  | ("...",Some [w]) -> Ast0.wrap(Ast0.Fdots(clt2mcode str dot, Some w))
   | _ -> failwith "cannot happen"
 
-let mkddots_one str (dot,whencode) =
+let mkfdots_one str (dot,whencode) =
   match str with
-    "..." -> Ast0.wrap(Ast0.Ddots(clt2mcode str dot, whencode))
+    "..." -> Ast0.wrap(Ast0.Fdots(clt2mcode str dot, whencode))
   | _ -> failwith "cannot happen"
 
 let mkpdots str dot =
@@ -160,7 +158,7 @@ let pointerify ty m =
 
 let ty_pointerify ty m =
   List.fold_left
-    (function inner -> function cur -> Type_cocci.Pointer(inner))
+    (fun inner cur -> Ast0.wrap (Ast0.Pointer (inner, Ast0.make_mcode "")))
     ty m
 
 let arrayify ty ar =
@@ -189,7 +187,7 @@ let iso_adjust first_fn fn first rest =
 
 let lookup rule name =
   try
-    let info = Hashtbl.find Data.all_metadecls rule in
+    let info = !(Hashtbl.find Data.all_metadecls rule) in
     List.find (function mv -> Ast.get_meta_name mv = (rule,name)) info
   with
     Not_found ->
@@ -209,6 +207,10 @@ let check_meta_tyopt type_irrelevant v =
     raise
       (Semantic_cocci.Semantic
 	 ("incompatible inheritance declaration "^name)) in
+  let type_equal ty0 ty1 =
+    type_irrelevant ||
+    Common.equal_option (Common.equal_list Unify_ast.fullType_compatible)
+      ty0 ty1 in
   match v with
     Ast.MetaMetaDecl(Ast.NONE,(rule,name)) ->
       (match meta_lookup rule name v with
@@ -217,7 +219,7 @@ let check_meta_tyopt type_irrelevant v =
   | Ast.MetaIdDecl(Ast.NONE,(rule,name)) ->
       (match meta_lookup rule name v with
 	Ast.MetaIdDecl(_,_) | Ast.MetaFreshIdDecl(_,_) -> ()
-      | _ -> fail name)
+      | x -> fail name)
   | Ast.MetaFreshIdDecl((rule,name),seed) ->
       raise
 	(Semantic_cocci.Semantic
@@ -246,25 +248,37 @@ let check_meta_tyopt type_irrelevant v =
       (match meta_lookup rule name v with
 	Ast.MetaParamListDecl(_,_,_) -> ()
       | _ -> fail name)
+  | Ast.MetaBinaryOperatorDecl(Ast.NONE,(rule,name)) ->
+      (match meta_lookup rule name v with
+	Ast.MetaBinaryOperatorDecl(_,_) -> ()
+      | _ -> fail name)
+  | Ast.MetaAssignmentOperatorDecl(Ast.NONE,(rule,name)) ->
+      (match meta_lookup rule name v with
+	Ast.MetaAssignmentOperatorDecl(_,_) -> ()
+      | _ -> fail name)
   | Ast.MetaConstDecl(Ast.NONE,(rule,name),ty) ->
       (match meta_lookup rule name v with
-	Ast.MetaConstDecl(_,_,ty1) when type_irrelevant || ty = ty1 -> ()
+	Ast.MetaConstDecl(_,_,ty1) when type_equal ty ty1 -> ()
       | _ -> fail name)
   | Ast.MetaErrDecl(Ast.NONE,(rule,name)) ->
       (match meta_lookup rule name v with
 	Ast.MetaErrDecl(_,_) -> ()
       | _ -> fail name)
-  | Ast.MetaExpDecl(Ast.NONE,(rule,name),ty) ->
+  | Ast.MetaExpDecl(Ast.NONE,(rule,name),ty,_) ->
       (match meta_lookup rule name v with
-	Ast.MetaExpDecl(_,_,ty1) when type_irrelevant || ty = ty1 -> ()
+	Ast.MetaExpDecl(_,_,ty1,_) when type_equal ty ty1 -> ()
       | _ -> fail name)
   | Ast.MetaIdExpDecl(Ast.NONE,(rule,name),ty) ->
       (match meta_lookup rule name v with
-	Ast.MetaIdExpDecl(_,_,ty1) when type_irrelevant || ty = ty1 -> ()
+	Ast.MetaIdExpDecl(_,_,ty1) when type_equal ty ty1 -> ()
       | _ -> fail name)
   | Ast.MetaLocalIdExpDecl(Ast.NONE,(rule,name),ty) ->
       (match meta_lookup rule name v with
-	Ast.MetaLocalIdExpDecl(_,_,ty1) when type_irrelevant || ty = ty1 -> ()
+	Ast.MetaLocalIdExpDecl(_,_,ty1) when type_equal ty ty1 -> ()
+      | _ -> fail name)
+  | Ast.MetaGlobalIdExpDecl(Ast.NONE,(rule,name),ty) ->
+      (match meta_lookup rule name v with
+	Ast.MetaGlobalIdExpDecl(_,_,ty1) when type_equal ty ty1 -> ()
       | _ -> fail name)
   | Ast.MetaExpListDecl(Ast.NONE,(rule,name),len_name) ->
       (match meta_lookup rule name v with
@@ -338,6 +352,11 @@ let check_meta_tyopt type_irrelevant v =
 
 let check_meta m = check_meta_tyopt false m
 
+let check_inherited_constraint_without_type meta_name =
+  match meta_name with
+    None, _ -> failwith "constraint must be an inherited variable"
+  | Some rule, name -> rule, name
+
 let check_inherited_constraint meta_name fn =
   match meta_name with
     (None,_) -> failwith "constraint must be an inherited variable"
@@ -410,19 +429,21 @@ let create_metadec_ty ar ispure kindfn ids current_rule =
 let create_len_metadec ar ispure kindfn lenid ids current_rule =
   let (lendec,lenname) =
     match lenid with
-      Common.Left lenid ->
+      Common.Left (lenid,cstr) ->
 	let lendec =
 	  create_metadec Ast.NONE Ast0.Impure
 	    (fun _ name _ check_meta -> check_meta(Ast.MetaListlenDecl(name)))
 	    [lenid] current_rule in
+	let cstr = Ast0toast.constraints cstr in
 	let lenname =
 	  match lendec with
-	    [Common.Left (Ast.MetaListlenDecl(x))] -> Ast.MetaLen x
-	  | [Common.Right (Ast.MetaListlenDecl(x))] -> Ast.MetaLen x
+	    [Common.Left (Ast.MetaListlenDecl(x))] -> Ast.MetaLen (x,cstr)
+	  | [Common.Right (Ast.MetaListlenDecl(x))] -> Ast.MetaLen (x,cstr)
 	  | _ -> failwith "unexpected length declaration" in
 	(lendec,lenname)
     | Common.Right n -> ([],Ast.CstLen n) in
-  lendec@(create_metadec ar ispure (kindfn lenname) ids current_rule)
+  lendec@
+  create_metadec_with_constraints ar ispure (kindfn lenname) ids current_rule
 
 (* ---------------------------------------------------------------------- *)
 
@@ -434,34 +455,37 @@ let str2inc s =
 (* declarations and statements *)
 
 let meta_decl name =
-  let (nm,pure,clt) = name in
-  Ast0.wrap(Ast0.MetaDecl(clt2mcode nm clt,pure))
+  let (nm,cstr,pure,clt) = name in
+  Ast0.wrap(Ast0.MetaDecl(clt2mcode nm clt,cstr,pure))
 
 let meta_field name =
-  let (nm,pure,clt) = name in
-  Ast0.wrap(Ast0.MetaField(clt2mcode nm clt,pure))
+  let (nm,cstr,pure,clt) = name in
+  Ast0.wrap(Ast0.MetaField(clt2mcode nm clt,cstr,pure))
+
+let dolen clt = function
+    Ast.AnyLen -> Ast0.AnyListLen
+  | Ast.MetaLen (nm,cstr) ->
+      let cstr' =
+	(* Discard expression! (from Ast to Ast0) *)
+	Ast.cstr_map Ast.empty_cstr_transformer cstr in
+      Ast0.MetaListLen(clt2mcode nm clt,cstr')
+  | Ast.CstLen n -> Ast0.CstListLen n
 
 let meta_field_list name =
-  let (nm,lenname,pure,clt) = name in
-  let lenname =
-    match lenname with
-      Ast.AnyLen -> Ast0.AnyListLen
-    | Ast.MetaLen nm -> Ast0.MetaListLen(clt2mcode nm clt)
-    | Ast.CstLen n -> Ast0.CstListLen n in
-  Ast0.wrap(Ast0.MetaFieldList(clt2mcode nm clt,lenname,pure))
+  let (nm,lenname,cstr,pure,clt) = name in
+  Ast0.wrap(Ast0.MetaFieldList(clt2mcode nm clt,dolen clt lenname,cstr,pure))
 
 let meta_stm name =
-  let (nm,pure,clt) = name in
-  Ast0.wrap(Ast0.MetaStmt(clt2mcode nm clt,pure))
+  let (nm,cstr,pure,clt) = name in
+  Ast0.wrap(Ast0.MetaStmt(clt2mcode nm clt,cstr,pure))
 
 let meta_stm_list name =
-  let (nm,lenname,pure,clt) = name in
-  let lenname =
-    match lenname with
-      Ast.AnyLen -> Ast0.AnyListLen
-    | Ast.MetaLen nm -> Ast0.MetaListLen(clt2mcode nm clt)
-    | Ast.CstLen n -> Ast0.CstListLen n in
-  Ast0.wrap(Ast0.MetaStmtList(clt2mcode nm clt,lenname,pure))
+  let (nm,lenname,cstr,pure,clt) = name in
+  Ast0.wrap(Ast0.MetaStmtList(clt2mcode nm clt,dolen clt lenname,cstr,pure))
+
+let meta_dparam_list name =
+  let (nm,lenname,cstr,pure,clt) = name in
+  Ast0.wrap(Ast0.MetaDParamList(clt2mcode nm clt,dolen clt lenname,cstr,pure))
 
 let exp_stm exp pv =
   Ast0.wrap(Ast0.ExprStatement (exp, clt2mcode ";" pv))
@@ -539,15 +563,14 @@ let seq lb s rb =
 let check_rule_name = function
     Some nm ->
       let n = id2name nm in
-      (try let _ =  Hashtbl.find Data.all_metadecls n in
-      raise (Semantic_cocci.Semantic ("repeated rule name"))
-      with Not_found -> Some n)
+      if Hashtbl.mem Data.all_metadecls n
+      then raise (Semantic_cocci.Semantic ("repeated rule name"))
+      else Some n
   | None -> None
 
 let make_iso_rule_name_result n =
-  (try let _ =  Hashtbl.find Data.all_metadecls n in
-  raise (Semantic_cocci.Semantic ("repeated rule name"))
-  with Not_found -> ());
+  (if Hashtbl.mem Data.all_metadecls n
+  then raise (Semantic_cocci.Semantic ("repeated rule name")));
   Ast.CocciRulename
     (Some n,Ast.NoDep,[],[],Ast.Undetermined,Ast.AnyP (*discarded*))
 
@@ -568,9 +591,13 @@ let fix_dependencies d =
 	Ast.AndDep(loop inverted d1,loop inverted d2)
     | Ast0.OrDep(d1,d2) ->
 	Ast.OrDep(loop inverted d1,loop inverted d2)
-    | Ast0.NoDep -> Ast.NoDep
-    | Ast0.FailDep -> Ast.FailDep in
-  loop false d
+    | Ast0.FileIn s when inverted -> Ast.NotFileIn s
+    | Ast0.FileIn s -> Ast.FileIn s in
+  match d with
+    Ast0.NoDep -> Ast.NoDep
+  | Ast0.FailDep -> Ast.FailDep
+  | Ast0.ExistsDep d -> Ast.ExistsDep (loop false d)
+  | Ast0.ForallDep d -> Ast.ForallDep (loop false d)
 
 let make_cocci_rule_name_result nm d i a e ee =
   Ast.CocciRulename (check_rule_name nm,fix_dependencies d,i,a,e,ee)
@@ -627,11 +654,16 @@ let string_metavariables str clt =
     MFmt(Ast0.wrap(Ast0.MetaFormat(clt2mcode name clt,constraints)))
   with Not_found ->
     try
-      let (name,lenname) = List.assoc str !Data.format_list_metavariables in
+      let (name,lenname,constraints) =
+	List.assoc str !Data.format_list_metavariables in
       let lenname =
 	match lenname with
 	  Ast.AnyLen -> Ast0.AnyListLen
-	| Ast.MetaLen nm -> Ast0.MetaListLen(clt2mcode nm clt)
+	| Ast.MetaLen (nm,cstr) ->
+	    let cstr' =
+	      (* Discard expression! (from Ast to Ast0) *)
+	      Ast.cstr_map Ast.empty_cstr_transformer cstr in
+	    Ast0.MetaListLen(clt2mcode nm clt,cstr')
 	| Ast.CstLen n ->
 	    if n < 1
 	    then
@@ -641,7 +673,8 @@ let string_metavariables str clt =
 	    else Ast0.CstListLen n in
       MFrag
 	(fun pct ->
-	  Ast0.wrap(Ast0.MetaFormatList(pct,clt2mcode name clt,lenname)))
+	  Ast0.wrap(
+	  Ast0.MetaFormatList(pct,clt2mcode name clt,constraints,lenname)))
     with Not_found -> failwith "bad metavariable in string"
 
 let pct_split str =
@@ -781,7 +814,7 @@ let not_format_string str clt =
   Ast0.wrap(Ast0.Constant (clt2mcode (Ast.String str) clt))
 
 let nometas str =
-  match Str.split (Str.regexp "@") str with
+  match Str.split_delim (Str.regexp "@") str with
     before::within::after::_ -> false (* need at least %@d@ *)
   | _ -> true
 
