@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -19,7 +19,7 @@ module V = Visitor_ast
 (* ------------------------------------------------------------------------- *)
 (* check if everything is removed, with no additions allowed *)
 
-let all_removed =
+let all_removed_recursor =
   let bind x y = x && y in
   let option_default = true in
   let do_nothing r k e = k e in
@@ -29,15 +29,20 @@ let all_removed =
     | Ast.MINUS(_,_,_,_) -> false
     | Ast.PLUS _ -> failwith "not possible"
     | Ast.CONTEXT(_,info) -> false in
-  let recursor =
-    V.combiner bind option_default
-      mcode mcode mcode mcode mcode mcode mcode mcode mcode
-      mcode mcode mcode mcode mcode
-      do_nothing do_nothing do_nothing do_nothing do_nothing
-      do_nothing do_nothing do_nothing do_nothing do_nothing
-      do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
-      do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing in
-  recursor.V.combiner_declaration
+  V.combiner bind option_default
+    mcode mcode mcode mcode mcode mcode mcode mcode mcode
+    mcode mcode mcode mcode mcode
+    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
+    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
+    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
+    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
+    do_nothing do_nothing
+
+let all_removed_decl =
+  all_removed_recursor.V.combiner_declaration
+
+let all_removed_field =
+  all_removed_recursor.V.combiner_field
 
 (* ------------------------------------------------------------------------- *)
 
@@ -47,14 +52,21 @@ let mcode _ (_,_,kind,_) =
   | Ast.PLUS _ -> failwith "not possible"
   | Ast.CONTEXT(_,info) -> not (info = Ast.NOTHING)
 
+(* The mcode is a fake one before a decl or field.  Only need to check
+for additions here.  Replacement will affect decl or field instead. *)
+let add_on_mcode _ (_,_,kind,_) =
+  match kind with
+    Ast.MINUS(_,_,_,_) -> false
+  | Ast.PLUS _ -> failwith "not possible"
+  | Ast.CONTEXT(_,info) -> not (info = Ast.NOTHING)
+
 let contains_modif =
   let bind x y = x || y in
   let option_default = false in
   let do_nothing r k e = k e in
   let annotated_decl decl =
     match Ast.unwrap decl with
-      Ast.DElem(bef,_,_) -> bef
-    | _ -> failwith "not possible" in
+      Ast.DElem(bef,_,_) -> bef in
   let rule_elem r k re =
     (* Very obscure how this could arise.  Array type contains an expression
        and the expression could contain a statement. *)
@@ -76,30 +88,71 @@ let contains_modif =
     V.combiner bind option_default
       mcode mcode mcode mcode mcode mcode mcode mcode mcode
       mcode mcode mcode mcode mcode
-      do_nothing do_nothing do_nothing do_nothing do_nothing
+      do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
+      do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
+      do_nothing do_nothing init do_nothing
       do_nothing do_nothing do_nothing do_nothing
-      do_nothing do_nothing do_nothing do_nothing init do_nothing do_nothing
       do_nothing rule_elem do_nothing do_nothing do_nothing do_nothing in
   recursor.V.combiner_fullType
 
 let decl r k e =
   let e = k e in
-  if all_removed e
+  if all_removed_decl e
   then {e with Ast.safe_for_multi_decls = true}
   else
     match Ast.unwrap e with
-      Ast.Init(stg,ty,_,_,_,sem)
-    | Ast.UnInit(stg,ty,_,sem) ->
+      Ast.Init(stg,ty,_,attr,_,_,sem)
+    | Ast.UnInit(stg,ty,_,attr,sem) ->
 	let stg_modif =
 	  match stg with
 	    Some stg -> mcode () stg
 	  | None -> false in
+	let attr_modif = List.exists (mcode ()) attr in
 	let ft_modif = contains_modif ty in
 	let sem_modif = mcode () sem in
-	if not(stg_modif || ft_modif || sem_modif)
+	if not(stg_modif || attr_modif || ft_modif || sem_modif)
 	then {e with Ast.safe_for_multi_decls = true}
 	else e
     | _ -> e
+
+let anndecl r k e =
+  let e = k e in
+  match Ast.unwrap e with
+    Ast.DElem(bef,allminus,decl) ->
+      let bef_modif = add_on_mcode () ((),(),bef,[]) in
+      if bef_modif && decl.Ast.safe_for_multi_decls
+      then (* not actually safe *)
+	Ast.rewrap e
+	  (Ast.DElem(bef,allminus,
+		     {decl with Ast.safe_for_multi_decls = false}))
+      else e
+
+let field r k e =
+  let e = k e in
+  if all_removed_field e
+  then {e with Ast.safe_for_multi_decls = true}
+  else
+    match Ast.unwrap e with
+      Ast.Field(ty,_,_bf,sem) ->
+	let ft_modif = contains_modif ty in
+	let sem_modif = mcode () sem in
+	if not(ft_modif || sem_modif)
+	then {e with Ast.safe_for_multi_decls = true}
+	else e
+    | _ -> e
+
+let annfield r k e =
+  let e = k e in
+  match Ast.unwrap e with
+    Ast.FElem(bef,allminus,fld) ->
+      let bef_modif = add_on_mcode () ((),(),bef,[]) in
+      if bef_modif && fld.Ast.safe_for_multi_decls
+      then (* not actually safe *)
+	Ast.rewrap e
+	  (Ast.FElem(bef,allminus,{fld with Ast.safe_for_multi_decls = false}))
+      else e
+  | _ -> e
+
 let mcode e = e
 let donothing r k e = k e
 
@@ -108,9 +161,9 @@ let process =
       mcode mcode mcode mcode mcode mcode mcode mcode mcode
       mcode mcode mcode mcode mcode
       donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing donothing
-      donothing donothing decl donothing donothing
-      donothing donothing donothing donothing in
+      donothing donothing donothing donothing donothing donothing donothing
+      donothing donothing donothing decl anndecl field annfield
+      donothing donothing donothing donothing donothing in
   List.map fn.V.rebuilder_top_level
 
 let safe_for_multi_decls rules =
