@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -142,8 +142,8 @@ let rec left_expression e =
   | Ast0.TypeExp(ty) -> left_typeC ty
   | Ast0.Constructor(lp,ty,rp,init) -> modif_before_mcode lp
   | Ast0.MetaErr(name,_,_) -> modif_before_mcode name
-  | Ast0.MetaExpr(name,_,ty,_,_) -> modif_before_mcode name
-  | Ast0.MetaExprList(name,_,_) -> modif_before_mcode name
+  | Ast0.MetaExpr(name,_,ty,_,_,_bitfield) -> modif_before_mcode name
+  | Ast0.MetaExprList(name,_,_,_) -> modif_before_mcode name
   | Ast0.EComma(cm) -> modif_before_mcode cm
   | Ast0.DisjExpr(_,exp_list,_,_) -> List.exists left_expression exp_list
   | Ast0.ConjExpr(_,exp_list,_,_) -> List.exists left_expression exp_list
@@ -171,9 +171,12 @@ and left_typeC t =
   | Ast0.EnumDef(ty,lb,ids,rb) -> left_typeC ty
   | Ast0.StructUnionName(kind,name) -> modif_before_mcode kind
   | Ast0.StructUnionDef(ty,lb,decls,rb) -> left_typeC ty
+  | Ast0.TypeOfExpr(tf,_,_,_) -> modif_before_mcode tf
+  | Ast0.TypeOfType(tf,_,_,_) -> modif_before_mcode tf
   | Ast0.TypeName(name) -> modif_before_mcode name
-  | Ast0.MetaType(name,_) -> modif_before_mcode name
+  | Ast0.MetaType(name,_,_) -> modif_before_mcode name
   | Ast0.DisjType(lp,types,mids,rp) -> List.exists left_typeC types
+  | Ast0.ConjType(lp,types,mids,rp) -> List.exists left_typeC types
   | Ast0.OptType(ty) -> left_typeC ty
   | Ast0.AsType _ -> failwith "not possible"
 
@@ -185,13 +188,11 @@ and left_typeC t =
 and left_declaration d =
   modif_before d ||
   match Ast0.unwrap d with
-    Ast0.MetaDecl(name,_) | Ast0.MetaField(name,_)
-  | Ast0.MetaFieldList(name,_,_) ->
-      modif_before_mcode name
-  | Ast0.Init(Some stg,ty,id,eq,ini,sem) -> modif_before_mcode stg
-  | Ast0.Init(None,ty,id,eq,ini,sem) -> left_typeC ty
-  | Ast0.UnInit(Some stg,ty,id,sem) -> modif_before_mcode stg
-  | Ast0.UnInit(None,ty,id,sem) -> left_typeC ty
+    Ast0.MetaDecl(name,_,_) -> modif_before_mcode name
+  | Ast0.Init(Some stg,ty,id,attr,eq,ini,sem) -> modif_before_mcode stg
+  | Ast0.Init(None,ty,id,attr,eq,ini,sem) -> left_typeC ty
+  | Ast0.UnInit(Some stg,ty,id,attr,sem) -> modif_before_mcode stg
+  | Ast0.UnInit(None,ty,id,attr,sem) -> left_typeC ty
   | Ast0.FunProto(fninfo,name,lp1,params,va,rp1,sem) ->
       (* should not be nested in anything anyway *)
       false
@@ -203,28 +204,53 @@ and left_declaration d =
   | Ast0.TyDecl(ty,sem) -> left_typeC ty
   | Ast0.Typedef(stg,ty,id,sem) -> modif_before_mcode stg
   | Ast0.DisjDecl(_,decls,_,_) -> List.exists left_declaration decls
-  | Ast0.Ddots(dots,_) -> false
+  | Ast0.ConjDecl(_,decls,_,_) -> List.exists left_declaration decls
   | Ast0.OptDecl(decl) -> left_declaration decl
   | Ast0.AsDecl _ -> failwith "not possible"
 
 and right_declaration d =
   modif_before d ||
   match Ast0.unwrap d with
-    Ast0.MetaDecl(name,_) | Ast0.MetaField(name,_)
-  | Ast0.MetaFieldList(name,_,_) ->
-      modif_before_mcode name
-  | Ast0.Init(_,ty,id,eq,ini,sem) -> modif_after_mcode sem
-  | Ast0.UnInit(_,ty,id,sem) -> modif_after_mcode sem
+    Ast0.MetaDecl(name,_,_) -> modif_before_mcode name
+  | Ast0.Init(_,ty,id,eq,ini,attr,sem) -> modif_after_mcode sem
+  | Ast0.UnInit(_,ty,id,attr,sem) -> modif_after_mcode sem
   | Ast0.FunProto(fninfo,name,lp1,params,va,rp1,sem) -> modif_after_mcode sem
   | Ast0.MacroDecl(_,name,lp,args,rp,sem) -> modif_after_mcode sem
   | Ast0.MacroDeclInit(_,name,lp,args,rp,eq,ini,sem) -> modif_after_mcode sem
   | Ast0.TyDecl(ty,sem) -> modif_after_mcode sem
   | Ast0.Typedef(stg,ty,id,sem) -> modif_after_mcode sem
   | Ast0.DisjDecl(_,decls,_,_) -> List.exists right_declaration decls
-  | Ast0.Ddots(dots,_) -> false
+  | Ast0.ConjDecl(_,decls,_,_) -> List.exists right_declaration decls
   | Ast0.OptDecl(decl) -> right_declaration decl
   | Ast0.AsDecl _ -> failwith "not possible"
 
+(* --------------------------------------------------------------------- *)
+(* Field declaration *)
+
+and left_field d =
+  modif_before d ||
+  match Ast0.unwrap d with
+    Ast0.MetaField(name,_,_)
+  | Ast0.MetaFieldList(name,_,_,_) ->
+      modif_before_mcode name
+  | Ast0.Field(ty,id,_bf,sem) -> left_typeC ty
+  | Ast0.DisjField(_,decls,_,_) -> List.exists left_field decls
+  | Ast0.ConjField(_,decls,_,_) -> List.exists left_field decls
+  | Ast0.OptField(decl) -> left_field decl
+  | Ast0.Fdots(dots,_) -> false
+
+and right_field d =
+  modif_before d ||
+  match Ast0.unwrap d with
+    Ast0.MetaField(name,_,_)
+  | Ast0.MetaFieldList(name,_,_,_) ->
+      modif_before_mcode name
+  | Ast0.Field(ty,id,_bf,sem) -> modif_after_mcode sem
+  | Ast0.DisjField(_,decls,_,_) -> List.exists right_field decls
+  | Ast0.ConjField(_,decls,_,_) -> List.exists right_field decls
+  | Ast0.OptField(decl) -> right_field decl
+  | Ast0.Fdots(dots,_) -> false
+
 (* --------------------------------------------------------------------- *)
 (* Top-level code *)
 (* These functions seem to be never used
@@ -267,6 +293,7 @@ and left_statement s =
   | Ast0.TopInit(init) -> false (* can only be replaced by an init *)
   | Ast0.Dots(d,whn) -> false
   | Ast0.Include(inc,s) -> modif_before_mcode inc
+  | Ast0.MetaInclude(inc,s) -> modif_before_mcode inc
   | Ast0.Undef(def,id) -> modif_before_mcode def
   | Ast0.Define(def,id,params,body) -> modif_before_mcode def
   | Ast0.Pragma(prg,id,body) -> modif_before_mcode prg
@@ -311,6 +338,7 @@ and right_statement s =
   | Ast0.TopInit(init) -> false (* can only be replaced by an init *)
   | Ast0.Dots(d,whn) -> false
   | Ast0.Include(inc,s) -> modif_after_mcode s
+  | Ast0.MetaInclude(inc,s) -> false (* irrelevant, not a substatement *)
   | Ast0.Undef(def,id) -> right_ident id
   | Ast0.Define(def,id,params,body) -> right_dots right_statement body
   | Ast0.Pragma(prg,id,body) -> right_pragma body -- not defined, b/c not used
@@ -374,9 +402,10 @@ and contains_only_minus =
   let expression r k e =
     mcodekind (Ast0.get_mcodekind e) &&
     match Ast0.unwrap e with
-      Ast0.DisjExpr(starter,expr_list,mids,ender)
-    | Ast0.ConjExpr(starter,expr_list,mids,ender) ->
+      Ast0.DisjExpr(starter,expr_list,mids,ender) ->
 	List.for_all r.VT0.combiner_rec_expression expr_list
+    | Ast0.ConjExpr(starter,expr_list,mids,ender) ->
+	List.exists r.VT0.combiner_rec_expression expr_list
     | _ -> k e in
 
   let declaration r k e =
@@ -384,6 +413,17 @@ and contains_only_minus =
     match Ast0.unwrap e with
       Ast0.DisjDecl(starter,decls,mids,ender) ->
 	List.for_all r.VT0.combiner_rec_declaration decls
+    | Ast0.ConjDecl(starter,decls,mids,ender) ->
+	List.exists r.VT0.combiner_rec_declaration decls
+    | _ -> k e in
+
+  let field r k e =
+    mcodekind (Ast0.get_mcodekind e) &&
+    match Ast0.unwrap e with
+      Ast0.DisjField(starter,decls,mids,ender) ->
+	List.for_all r.VT0.combiner_rec_field decls
+    | Ast0.ConjField(starter,decls,mids,ender) ->
+	List.exists r.VT0.combiner_rec_field decls
     | _ -> k e in
 
   let typeC r k e =
@@ -391,6 +431,8 @@ and contains_only_minus =
     match Ast0.unwrap e with
       Ast0.DisjType(starter,types,mids,ender) ->
 	List.for_all r.VT0.combiner_rec_typeC types
+    | Ast0.ConjType(starter,types,mids,ender) ->
+	List.exists r.VT0.combiner_rec_typeC types
     | _ -> k e in
 
   let statement r k e =
@@ -398,6 +440,8 @@ and contains_only_minus =
     match Ast0.unwrap e with
       Ast0.Disj(starter,statement_dots_list,mids,ender) ->
 	List.for_all r.VT0.combiner_rec_statement_dots statement_dots_list
+    | Ast0.Conj(starter,statement_dots_list,mids,ender) ->
+	List.exists r.VT0.combiner_rec_statement_dots statement_dots_list
     | _ -> k e in
 
   let case_line r k e =
@@ -410,9 +454,9 @@ and contains_only_minus =
   V0.flat_combiner bind option_default
     mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
     mcode mcode mcode
-    dots dots dots dots dots dots dots
+    dots dots dots dots dots dots dots dots
     identifier expression donothing donothing typeC donothing donothing
-    declaration statement donothing case_line donothing donothing
+    declaration field statement donothing case_line donothing donothing
 
 
 (* needs a special case when there is a Disj or an empty DOTS *)
@@ -595,8 +639,8 @@ let rec statement dots_before dots_after s =
   | Ast0.Return(ret,sem) -> do_one s
   | Ast0.ReturnExpr(ret,exp,sem) -> do_one s
   | Ast0.Exec(exec,lang,code,sem) -> do_one s
-  | Ast0.MetaStmt(name,_) -> do_one s
-  | Ast0.MetaStmtList(name,_,_) -> do_one s
+  | Ast0.MetaStmt(name,_,_) -> do_one s
+  | Ast0.MetaStmtList(name,_,_,_) -> do_one s
   | Ast0.Disj(starter,statement_dots_list,mids,ender) ->
       Ast0.rewrap s
 	(Ast0.Disj(starter,
@@ -627,6 +671,7 @@ let rec statement dots_before dots_after s =
   | Ast0.TopInit(init) -> s
   | Ast0.Dots(d,whn) -> s
   | Ast0.Include(inc,string) -> s (* doesn't affect the need for braces *)
+  | Ast0.MetaInclude(inc,name) -> s (* doesn't affect the need for braces *)
   | Ast0.Undef(def,id) -> s (* same as include *)
   | Ast0.Define(def,id,params,body) -> s (* same as include *)
   | Ast0.Pragma(prg,id,body) -> s (* same as include *)
