@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -93,11 +93,12 @@ let rec is_int_type_unwrap = function
   | Ast0.BaseType (Ast.LongType, _)
   | Ast0.BaseType (Ast.ShortType, _)
   | Ast0.BaseType (Ast.SizeType, _)
-  | Ast0.MetaType(_,_)
+  | Ast0.MetaType(_,_,_)
+  | Ast0.TypeOfExpr _
   | Ast0.TypeName _
   | Ast0.EnumName _
   | Ast0.Signed(_,None) -> true
-  | Ast0.Signed(_,Some ty) -> is_int_type ty
+  | Ast0.Signed(_,Some ty) | Ast0.TypeOfType(_,_,ty,_) -> is_int_type ty
   | _ -> false
 and is_int_type ty = is_int_type_unwrap (Ast0.unwrap ty)
 
@@ -248,7 +249,7 @@ let rec propagate_types env =
 		 None -> None
 	       | Some (Ast0.Array(ty, _, _, _)) -> Some ty
 	       | Some (Ast0.Pointer(ty, _)) -> Some ty
-	       | Some (Ast0.MetaType(_,_)) -> None
+	       | Some (Ast0.MetaType(_,_,_)) -> None
 	       | Some x -> err exp1 (Ast0.wrap x) "ill-typed array reference")
 	      (* pad: should handle structure one day and look 'field' in environment *)
 	| Ast0.RecordAccess(exp,pt,field) ->
@@ -259,7 +260,7 @@ let rec propagate_types env =
 	       | Some (Ast0.StructUnionName(_,_)) -> None
 	       | Some (Ast0.TypeName(s)) ->
 			  None
-	       | Some (Ast0.MetaType(_,_)) -> None
+	       | Some (Ast0.MetaType(_,_,_)) -> None
 	       | Some x ->
                    err exp (Ast0.wrap x) "non-structure type in field ref")
 	| Ast0.RecordPtAccess(exp,ar,field) ->
@@ -272,7 +273,7 @@ let rec propagate_types env =
                      Common.map_option Ast0.unwrap (strip_cv (Some t))
                    with
 		      | Some (Ast0.BaseType(Ast.Unknown, _))
-		      | Some (Ast0.MetaType(_, _))
+		      | Some (Ast0.MetaType(_, _, _))
 		      | Some (Ast0.TypeName(_))
 		      | Some (Ast0.StructUnionName(_, _)) -> None
 		      | Some x ->
@@ -281,7 +282,7 @@ let rec propagate_types env =
                           err exp ty
 			    "non-structure pointer type in field ref"
 		      |	_ -> failwith "not possible")
-               | Some (Ast0.MetaType(_, _))
+               | Some (Ast0.MetaType(_, _, _))
                | Some (Ast0.TypeName(_)) -> None
                | Some x ->
                    let ty = Ast0.wrap x in
@@ -292,9 +293,9 @@ let rec propagate_types env =
 	| Ast0.TypeExp(ty) -> None
 	| Ast0.Constructor(lp,ty,rp,init) -> Some ty
 	| Ast0.MetaErr(name,_,_) -> None
-	| Ast0.MetaExpr(name,_,Some [ty],_,_) -> Some ty
-	| Ast0.MetaExpr(name,_,ty,_,_) -> None
-	| Ast0.MetaExprList(name,_,_) -> None
+	| Ast0.MetaExpr(name,_,Some [ty],_,_,_bitfield) -> Some ty
+	| Ast0.MetaExpr(name,_,ty,_,_,_bitfield) -> None
+	| Ast0.MetaExprList(name,_,_,_) -> None
 	| Ast0.EComma(cm) -> None
 	| Ast0.DisjExpr(_,exp_list,_,_)
 	| Ast0.ConjExpr(_,exp_list,_,_) ->
@@ -367,12 +368,11 @@ let rec propagate_types env =
 
   and process_decl env decl =
     match Ast0.unwrap decl with
-      Ast0.MetaDecl(_,_) | Ast0.MetaField(_,_)
-    | Ast0.MetaFieldList(_,_,_) -> []
-    | Ast0.Init(_,ty,id,_,exp,_) ->
+      Ast0.MetaDecl(_,_,_) -> []
+    | Ast0.Init(_,ty,id,_,_,exp,_) ->
 	let _ = (propagate_types env).VT0.combiner_rec_initialiser exp in
 	List.map (function i -> (i,ty)) (strip id)
-    | Ast0.UnInit(_,ty,id,_) ->
+    | Ast0.UnInit(_,ty,id,_,_) ->
 	List.map (function i -> (i,ty)) (strip id)
     | Ast0.FunProto(fi,nm,lp,params,va,rp,sem) -> []
     | Ast0.MacroDecl(_,_,_,_,_,_) -> []
@@ -383,9 +383,9 @@ let rec propagate_types env =
               (* pad: should handle typedef one day and add a binding *)
     | Ast0.Typedef((a,_,_,_,_,_),b,c,(d,_,_,_,_,_)) ->
 	[]
-    | Ast0.DisjDecl(_,disjs,_,_) ->
+    | Ast0.DisjDecl(_,disjs,_,_)
+    | Ast0.ConjDecl(_,disjs,_,_) ->
 	List.concat(List.map (process_decl env) disjs)
-    | Ast0.Ddots(_,_) -> [] (* not in a statement list anyway *)
     | Ast0.OptDecl(decl) -> process_decl env decl
     | Ast0.AsDecl _ -> failwith "not possible" in
 
@@ -397,7 +397,7 @@ let rec propagate_types env =
       match (Ast0.unwrap exp,Ast0.get_type exp) with
 	(Ast0.Edots(_,_),_) -> None
       | (Ast0.NestExpr(_,_,_,_,_),_) -> None
-      | (Ast0.MetaExpr(_,_,_,_,_),_) ->
+      | (Ast0.MetaExpr(_,_,_,_,_,_bitfield),_) ->
       (* if a type is known, it is specified in the decl *)
 	  None
       | (Ast0.Paren(lp,exp,rp),None) -> process_test exp
