@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -9,7 +9,6 @@ module Ast0 = Ast0_cocci
 module Ast = Ast_cocci
 module V0 = Visitor_ast0
 module VT0 = Visitor_ast0_types
-module TC = Type_cocci
 
 let set_minus s minus = List.filter (function n -> not (List.mem n minus)) s
 
@@ -31,6 +30,7 @@ let plus_checker (nm,_,_,mc,_,_) =
 let get_free checker t =
   let bind x y = x @ y in
   let option_default = [] in
+  let donothing r k e = k e in
 
   (* considers a single list *)
   let collect_unitary_nonunitary free_usage =
@@ -76,27 +76,35 @@ let get_free checker t =
 	detect_unitary_frees(List.map r.VT0.combiner_rec_ident id_list)
     | _ -> k i in
 
-  let rec type_collect res = function
-      TC.ConstVol(_,ty) | TC.Pointer(ty) | TC.FunctionPointer(ty)
-    | TC.Array(ty) -> type_collect res ty
-    | TC.EnumName(TC.MV(tyname,_,_)) ->
-	bind [tyname] res
-    | TC.StructUnionName(_,TC.MV(tyname,_,_)) ->
-	bind [tyname] res
-    | TC.MetaType(tyname,_,_) ->
-	bind [tyname] res
-    | TC.Decimal(e1,e2) ->
-	let e2mv = function TC.MV(mv,_,_) -> [mv] | _ -> [] in
-	bind (e2mv e1) (e2mv e2)
-    | TC.SignedT(_,Some ty) -> type_collect res ty
-    | ty -> res in
+  let type_collect res ty = bind res (Ast0.meta_names_of_typeC ty) in
 
-  let constraints_collect r res = function
-      Ast0.NotExpCstrt(el) ->
-	List.fold_left bind res
-	  (List.map r.VT0.combiner_rec_expression el)
-    | Ast0.SubExpCstrt(names) -> bind names res
-    | _ -> res in
+  let mcode mc =
+    List.fold_left
+      (fun accu e ->
+	match e with
+	  Ast0.MetaPosTag(Ast0.MetaPos(name,constraints,_)) ->
+	    Ast.cstr_fold
+	      { Ast.empty_cstr_transformer with
+		Ast.cstr_script =
+		Some (fun (_,(name,lang,params,_pos,body)) accu ->
+		  (* It seems that position variables are not relevant
+		     for unitaryness, so drop them *)
+		  bind (List.map fst
+		     (List.filter
+			(function
+			    (_,Ast.MetaPosDecl _) -> false
+			  | _ -> true)
+			params)) accu) } constraints accu
+	| _ -> accu)
+      option_default (Ast0.get_pos mc) in
+
+  let constraints_collect r res c =
+    let cstr_expr =
+      Some (fun e res -> bind res (r.VT0.combiner_rec_expression e)) in
+    let cstr_meta_name = Some (fun mn res -> bind [mn] res) in
+    let transformer =
+      { Ast.empty_cstr_transformer with Ast.cstr_expr; Ast.cstr_meta_name } in
+    Ast.cstr_fold transformer c res in
 
   let expression r k e =
     match Ast0.unwrap e with
@@ -104,7 +112,7 @@ let get_free checker t =
 	let constraints =
 	  constraints_collect r option_default constraints in
 	bind (checker name) constraints
-    | Ast0.MetaExpr(name,constraints,type_list,_,_) ->
+    | Ast0.MetaExpr(name,constraints,type_list,_,_,_bitfield) ->
 	let types =
 	  match type_list with
 	    Some type_list ->
@@ -113,7 +121,7 @@ let get_free checker t =
 	let constraints =
 	  constraints_collect r types constraints in
 	bind (checker name) constraints
-    | Ast0.MetaExprList(name,_,_) -> checker name
+    | Ast0.MetaExprList(name,_,_,_) -> checker name
     | Ast0.DisjExpr(starter,expr_list,mids,ender) ->
 	detect_unitary_frees(List.map r.VT0.combiner_rec_expression expr_list)
     | Ast0.ConjExpr(starter,expr_list,mids,ender) ->
@@ -124,22 +132,41 @@ let get_free checker t =
 
   let typeC r k t =
     match Ast0.unwrap t with
-      Ast0.MetaType(name,_) -> checker name
+      Ast0.MetaType(name,_,_) -> checker name
     | Ast0.DisjType(starter,types,mids,ender) ->
 	detect_unitary_frees(List.map r.VT0.combiner_rec_typeC types)
+    | Ast0.ConjType(starter,types,mids,ender) ->
+	List.fold_left
+	  (fun prev cur -> bind (r.VT0.combiner_rec_typeC cur) prev)
+	  option_default types
     | _ -> k t in
 
   let parameter r k p =
     match Ast0.unwrap p with
-      Ast0.MetaParam(name,_) | Ast0.MetaParamList(name,_,_) -> checker name
+      Ast0.MetaParam(name,_,_) | Ast0.MetaParamList(name,_,_,_) -> checker name
     | _ -> k p in
 
   let declaration r k d =
     match Ast0.unwrap d with
-      Ast0.MetaDecl(name,_) | Ast0.MetaField(name,_)
-    | Ast0.MetaFieldList(name,_,_) -> checker name
+      Ast0.MetaDecl(name,_,_) -> checker name
     | Ast0.DisjDecl(starter,decls,mids,ender) ->
 	detect_unitary_frees(List.map r.VT0.combiner_rec_declaration decls)
+    | Ast0.ConjDecl(starter,decls,mids,ender) ->
+	List.fold_left
+	  (fun prev cur -> bind (r.VT0.combiner_rec_declaration cur) prev)
+	  option_default decls
+    | _ -> k d in
+
+  let field r k d =
+    match Ast0.unwrap d with
+      Ast0.MetaField(name,_,_)
+    | Ast0.MetaFieldList(name,_,_,_) -> checker name
+    | Ast0.DisjField(starter,decls,mids,ender) ->
+	detect_unitary_frees(List.map r.VT0.combiner_rec_field decls)
+    | Ast0.ConjField(starter,decls,mids,ender) ->
+	List.fold_left
+	  (fun prev cur -> bind (r.VT0.combiner_rec_field cur) prev)
+	  option_default decls
     | _ -> k d in
 
   let case_line r k c =
@@ -150,7 +177,7 @@ let get_free checker t =
 
   let statement r k s =
     match Ast0.unwrap s with
-      Ast0.MetaStmt(name,_) | Ast0.MetaStmtList(name,_,_) -> checker name
+      Ast0.MetaStmt(name,_,_) | Ast0.MetaStmtList(name,_,_,_) -> checker name
     | Ast0.Disj(starter,stmt_list,mids,ender) ->
 	detect_unitary_frees
 	  (List.map r.VT0.combiner_rec_statement_dots stmt_list)
@@ -176,15 +203,14 @@ let get_free checker t =
 	     whn)
     | _ -> k s in
 
-  let res = V0.combiner bind option_default
-      {V0.combiner_functions with
-	VT0.combiner_identfn = ident;
-	VT0.combiner_exprfn = expression;
-	VT0.combiner_tyfn = typeC;
-	VT0.combiner_paramfn = parameter;
-	VT0.combiner_declfn = declaration;
-	VT0.combiner_stmtfn = statement;
-	VT0.combiner_casefn = case_line} in
+  let res =
+    V0.flat_combiner bind option_default
+      mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
+      mcode mcode mcode mcode
+      donothing donothing donothing donothing donothing donothing donothing
+      donothing
+      ident expression donothing donothing typeC donothing parameter
+      declaration field statement donothing case_line donothing donothing in
 
   collect_unitary_nonunitary
     (List.concat (List.map res.VT0.combiner_rec_top_level t))
@@ -215,32 +241,33 @@ let update_unitary unitary =
     match Ast0.unwrap e with
       Ast0.MetaErr(name,constraints,_) ->
 	Ast0.rewrap e (Ast0.MetaErr(name,constraints,is_unitary name))
-    | Ast0.MetaExpr(name,constraints,ty,form,_) ->
-	Ast0.rewrap e (Ast0.MetaExpr(name,constraints,ty,form,is_unitary name))
-    | Ast0.MetaExprList(name,lenname,_) ->
-	Ast0.rewrap e (Ast0.MetaExprList(name,lenname,is_unitary name))
+    | Ast0.MetaExpr(name,constraints,ty,form,_,bitfield) ->
+	Ast0.rewrap e
+	  (Ast0.MetaExpr(name,constraints,ty,form,is_unitary name,bitfield))
+    | Ast0.MetaExprList(name,lenname,cstr,_) ->
+	Ast0.rewrap e (Ast0.MetaExprList(name,lenname,cstr,is_unitary name))
     | _ -> k e in
 
   let typeC r k t =
     match Ast0.unwrap t with
-      Ast0.MetaType(name,_) ->
-	Ast0.rewrap t (Ast0.MetaType(name,is_unitary name))
+      Ast0.MetaType(name,cstr,_) ->
+	Ast0.rewrap t (Ast0.MetaType(name,cstr,is_unitary name))
     | _ -> k t in
 
   let parameter r k p =
     match Ast0.unwrap p with
-      Ast0.MetaParam(name,_) ->
-	Ast0.rewrap p (Ast0.MetaParam(name,is_unitary name))
-    | Ast0.MetaParamList(name,lenname,_) ->
-	Ast0.rewrap p (Ast0.MetaParamList(name,lenname,is_unitary name))
+      Ast0.MetaParam(name,cstr,_) ->
+	Ast0.rewrap p (Ast0.MetaParam(name,cstr,is_unitary name))
+    | Ast0.MetaParamList(name,lenname,cstr,_) ->
+	Ast0.rewrap p (Ast0.MetaParamList(name,lenname,cstr,is_unitary name))
     | _ -> k p in
 
   let statement r k s =
     match Ast0.unwrap s with
-      Ast0.MetaStmt(name,_) ->
-	Ast0.rewrap s (Ast0.MetaStmt(name,is_unitary name))
-    | Ast0.MetaStmtList(name,lenname,_) ->
-	Ast0.rewrap s (Ast0.MetaStmtList(name,lenname,is_unitary name))
+      Ast0.MetaStmt(name,cstr,_) ->
+	Ast0.rewrap s (Ast0.MetaStmt(name,cstr,is_unitary name))
+    | Ast0.MetaStmtList(name,lenname,cstr,_) ->
+	Ast0.rewrap s (Ast0.MetaStmtList(name,lenname,cstr,is_unitary name))
     | _ -> k s in
 
   let res = V0.rebuilder
@@ -272,12 +299,12 @@ let do_unitary rules =
       [] -> ([],[])
     | (r::rules) ->
       match r with
-        Ast0.ScriptRule (_,_,_,_,_,_)
-      | Ast0.InitialScriptRule (_,_,_,_,_)
-      | Ast0.FinalScriptRule (_,_,_,_,_) ->
+        Ast0.ScriptRule (_,_,_,_,_,_,_)
+      | Ast0.InitialScriptRule (_,_,_,_,_,_)
+      | Ast0.FinalScriptRule (_,_,_,_,_,_) ->
           let (x,rules) = loop rules in
           (x, r::rules)
-      | Ast0.CocciRule ((minus,metavars,chosen_isos),((plus,_) as plusz),rt) ->
+      | Ast0.CocciRule((minus,metavars,chosen_isos),((plus,_) as plusz),inh,rt) ->
           let mm1 = List.map Ast.get_meta_name metavars in
           let (used_after, rest) = loop rules in
           let (m_unitary, m_nonunitary) = get_free minus_checker minus in
@@ -296,7 +323,7 @@ let do_unitary rules =
           let rebuilt = update_unitary m_unitary minus in
           (set_minus (m_nonunitary @ used_after) mm1,
              (Ast0.CocciRule
-		((rebuilt, metavars, chosen_isos),plusz,rt))::rest) in
+		((rebuilt, metavars, chosen_isos),plusz,inh,rt))::rest) in
   let (_,rules) = loop rules in
   rules
 
