@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -13,10 +13,10 @@ let quiet = ref true (* true = no decoration on - context, etc *)
 
 let full_ids = ref false (* true = print rule name as well *)
 
-let start_block str =
+let start_block _str =
   force_newline(); print_string "  "; open_box 0
 
-let end_block str =
+let end_block _str =
   close_box(); force_newline ()
 
 let print_option = Common.do_option
@@ -66,7 +66,7 @@ let mcodekind brackets fn x info mc =
 	else
 	  match brackets with
 	    Some x -> ("[",("]^"^(string_of_int x))) | None -> ("","") in
-      let (plus_streams,t1,t2) = !plus_streams in
+      let (plus_streams,_t1,_t2) = !plus_streams in
       U.print_around
 	(function x ->
 	  print_string lb; fn x; print_string rb)
@@ -85,7 +85,7 @@ let mcodekind brackets fn x info mc =
 	x plus_streams);
   List.iter (function (s,_) -> print s) info.Ast0.strings_after
 
-let mcode fn (x,_,info,mc,pos,adj) =
+let mcode fn (x,_,info,mc,pos,_adj) =
   let fn x = fn x; meta_pos !pos in
   mcodekind (Some info.Ast0.pos_info.Ast0.line_start)(*None*) fn x info mc
 
@@ -117,12 +117,15 @@ let do_disj lst processor sep =
 
 (* --------------------------------------------------------------------- *)
 
+let print_type ty =
+  print_string (Ast.string_of_fullType (Ast0toast.typeC false ty))
+
 let print_types = function
     None -> ()
   | Some ty ->
       print_string "/* ";
       Format.print_flush();
-      print_between (function _ -> print_string ", ") Type_cocci.typeC ty;
+      print_between (function _ -> print_string ", ") print_type ty;
       Format.print_flush();
       print_string " */"
 
@@ -161,7 +164,7 @@ let binaryOp op =
       | Ast0.MetaBinary(name,_,_) -> mcode print_meta name)
 
 let rec expression e =
-  print_option Type_cocci.typeC (Ast0.get_type e);
+  print_option print_type (Ast0.get_type e);
   print_context e
     (function _ ->
       match Ast0.unwrap e with
@@ -231,7 +234,7 @@ let rec expression e =
 	  mcode print_string_box lp; typeC ty; close_box();
 	  mcode print_string rp; initialiser init
       | Ast0.MetaErr(name,_,_) -> mcode print_meta name
-      | Ast0.MetaExpr(name,_,ty,_,pure) ->
+      | Ast0.MetaExpr(name,_,ty,_,_pure,_bitfield) ->
 	  mcode print_meta name; print_types ty(*;
 	  print_string "^";
 	  (match pure with
@@ -239,7 +242,7 @@ let rec expression e =
 	  | Ast0.Impure -> print_string "impure"
 	  | Ast0.Context -> print_string "context"
 	  | Ast0.PureContext -> print_string "pure_context")*)
-      | Ast0.MetaExprList(name,_,_) -> mcode print_meta name
+      | Ast0.MetaExprList(name,_,_,_) -> mcode print_meta name
       | Ast0.EComma(cm) -> mcode print_string cm; print_space()
       | Ast0.DisjExpr(_,exp_list,_,_) -> do_disj exp_list expression "|"
       | Ast0.ConjExpr(_,exp_list,_,_) -> do_disj exp_list expression "&"
@@ -271,7 +274,7 @@ and string_fragment e =
       mcode print_string pct;
       string_format fmt
   | Ast0.Strdots dots -> mcode print_string dots
-  | Ast0.MetaFormatList(pct,name,lenname) ->
+  | Ast0.MetaFormatList(pct,name,_,lenname) ->
       mcode print_string pct;
       mcode print_meta name
 
@@ -324,11 +327,20 @@ and typeC t =
 	  print_option (function x -> ident x; print_string " ") name
       | Ast0.StructUnionDef(ty,lb,decls,rb) ->
 	  typeC ty; mcode print_string lb;
-	  dots force_newline declaration decls;
+	  dots force_newline field decls;
 	  mcode print_string rb
+      | Ast0.TypeOfExpr(tf,lp,exp,rp) ->
+          mcode print_string tf;
+	  mcode print_string_box lp; expression exp; close_box();
+	  mcode print_string rp
+      | Ast0.TypeOfType(tf,lp,ty,rp) ->
+          mcode print_string tf;
+	  mcode print_string_box lp; typeC ty; close_box();
+	  mcode print_string rp
       | Ast0.TypeName(name)-> mcode print_string name; print_string " "
-      | Ast0.MetaType(name,_)-> mcode print_meta name; print_string " "
+      | Ast0.MetaType(name,_,_)-> mcode print_meta name; print_string " "
       | Ast0.DisjType(_,types,_,_) -> do_disj types typeC "|"
+      | Ast0.ConjType(_,types,_,_) -> do_disj types typeC "&"
       | Ast0.OptType(ty) -> print_string "?"; typeC ty
       | Ast0.AsType(ty,asty) -> typeC ty; print_string "@"; typeC asty)
 
@@ -360,17 +372,20 @@ and declaration d =
   print_context d
     (function _ ->
       match Ast0.unwrap d with
-	Ast0.MetaDecl(name,_) | Ast0.MetaField(name,_)
-      | Ast0.MetaFieldList(name,_,_) ->
+	Ast0.MetaDecl(name,_,_) ->
 	  mcode print_meta name
-      |	Ast0.Init(stg,ty,id,eq,ini,sem) ->
+      |	Ast0.Init(stg,ty,id,attr,eq,ini,sem) ->
 	  print_option (mcode U.storage) stg;
 	  print_named_type ty id;
+	  (if not (attr = []) then print_string " ");
+	  print_between (fun _ -> print_string " ") (mcode print_string) attr;
 	  print_string " ";
 	  mcode print_string eq; print_string " "; initialiser ini;
 	  mcode print_string sem
-      | Ast0.UnInit(stg,ty,id,sem) ->
+      | Ast0.UnInit(stg,ty,id,attr,sem) ->
 	  print_option (mcode U.storage) stg; print_named_type ty id;
+	  (if not (attr = []) then print_string " ");
+	  print_between (fun _ -> print_string " ") (mcode print_string) attr;
 	  mcode print_string sem
       | Ast0.FunProto(fninfo,name,lp1,params,va,rp1,sem) ->
 	  List.iter print_fninfo fninfo;
@@ -397,16 +412,47 @@ and declaration d =
 	  mcode print_string sem
       | Ast0.DisjDecl(_,decls,_,_) ->
 	  do_disj decls declaration "|"
-      | Ast0.Ddots(dots,Some (_,_,whencode)) ->
-	  mcode print_string dots; print_string "   when != ";
-	  declaration whencode
-      | Ast0.Ddots(dots,None) -> mcode print_string dots
+      | Ast0.ConjDecl(_,decls,_,_) ->
+	  do_disj decls declaration "&"
       | Ast0.OptDecl(decl) -> print_string "?"; declaration decl
       | Ast0.AsDecl(decl,asdecl) ->
 	  declaration decl; print_string "@"; declaration asdecl)
 
 and declaration_dots l = dots (function _ -> ()) declaration l
 
+(* --------------------------------------------------------------------- *)
+(* Field declaration *)
+
+and field d =
+  print_context d
+    (function _ ->
+      match Ast0.unwrap d with
+	Ast0.MetaField(name,_,_)
+      | Ast0.MetaFieldList(name,_,_,_) ->
+	  mcode print_meta name
+      | Ast0.Field(ty,id,bf,sem) ->
+	  begin
+	    match id with
+	      None -> typeC ty
+	    | Some id -> print_named_type ty id
+	  end;
+	  let bitfield (c, e) =
+	    mcode print_string c;
+	    expression e in
+	  Common.do_option bitfield bf;
+	  mcode print_string sem
+      | Ast0.DisjField(_,fields,_,_) ->
+	  do_disj fields field "|"
+      | Ast0.ConjField(_,fields,_,_) ->
+	  do_disj fields field "&"
+      | Ast0.OptField(decl) -> print_string "?"; field decl
+      | Ast0.Fdots(dots,Some (_,_,whencode)) ->
+	  mcode print_string dots; print_string "   when != ";
+	  field whencode
+      | Ast0.Fdots(dots,None) -> mcode print_string dots)
+
+and field_dots l = dots (function _ -> ()) field l
+
 (* --------------------------------------------------------------------- *)
 (* Initialiser *)
 
@@ -414,8 +460,8 @@ and initialiser i =
   print_context i
     (function _ ->
       match Ast0.unwrap i with
-	Ast0.MetaInit(name,_)-> mcode print_meta name; print_string " "
-      |	Ast0.MetaInitList(name,_,_)-> mcode print_meta name; print_string " "
+	Ast0.MetaInit(name,_,_)-> mcode print_meta name; print_string " "
+      |	Ast0.MetaInitList(name,_,_,_)-> mcode print_meta name; print_string " "
       |	Ast0.InitExpr(exp) -> expression exp
       | Ast0.InitList(lb,initlist,rb,ordered) ->
           (*doesn't show commas dropped in unordered case*)
@@ -456,8 +502,8 @@ and parameterTypeDef p =
 	Ast0.VoidParam(ty) -> typeC ty
       | Ast0.Param(ty,Some id) -> print_named_type ty id
       |	Ast0.Param(ty,None) -> typeC ty
-      | Ast0.MetaParam(name,_) -> mcode print_meta name
-      | Ast0.MetaParamList(name,_,_) -> mcode print_meta name
+      | Ast0.MetaParam(name,_,_) -> mcode print_meta name
+      | Ast0.MetaParamList(name,_,_,_) -> mcode print_meta name
       | Ast0.PComma(cm) -> mcode print_string cm; print_space()
       | Ast0.Pdots(dots) -> mcode print_string dots
       | Ast0.OptParam(param) -> print_string "?"; parameterTypeDef param
@@ -564,7 +610,7 @@ and statement arity s =
 	  mcode print_string lang; print_string " ";
 	  dots (function _ -> print_string " ") exec_code code;
 	  mcode print_string sem
-      | Ast0.MetaStmt(name,pure) ->
+      | Ast0.MetaStmt(name,_,pure) ->
 	  print_string arity; mcode print_meta name;(*
 	  print_string "^";
 	  (match pure with
@@ -572,7 +618,7 @@ and statement arity s =
 	  | Ast0.Impure -> print_string "impure"
 	  | Ast0.Context -> print_string "context"
 	  | Ast0.PureContext -> print_string "pure_context")*)
-      | Ast0.MetaStmtList(name,_,_) ->
+      | Ast0.MetaStmtList(name,_,_,_) ->
 	  print_string arity;  mcode print_meta name
       | Ast0.Disj(starter,statement_dots_list,_,ender) ->
 	  print_string arity;
@@ -614,6 +660,8 @@ and statement arity s =
 	    whn
       | Ast0.Include(inc,s) ->
 	  mcode print_string inc; print_string " "; mcode U.inc_file s
+      | Ast0.MetaInclude(inc,s) ->
+	  mcode print_string inc; print_string " "; expression s
       | Ast0.Undef(def,id) ->
 	  mcode print_string def; print_string " "; ident id
       | Ast0.Define(def,id,params,body) ->
@@ -647,10 +695,13 @@ and print_define_parameters params =
 and print_define_param param =
   match Ast0.unwrap param with
     Ast0.DParam(id) -> ident id
+  | Ast0.MetaDParamList(name,_,_,_) -> mcode print_meta name
   | Ast0.DPComma(comma) -> mcode print_string comma
   | Ast0.DPdots(dots) -> mcode print_string dots
   | Ast0.OptDParam(dp) -> print_string "?"; print_define_param dp
 
+and define_param_dots l = dots (function _ -> ()) print_define_param l
+
 and print_fninfo = function
     Ast0.FStorage(stg) -> mcode U.storage stg
   | Ast0.FType(ty) -> typeC ty
@@ -732,7 +783,9 @@ let rec unparse_anything x =
       print_string "StmDots:"; force_newline();
       statement_dots d
   | Ast0.DotsDeclTag(d) -> declaration_dots d
+  | Ast0.DotsFieldTag(d) -> field_dots d
   | Ast0.DotsCaseTag(d) -> case_dots d
+  | Ast0.DotsDefParamTag(d) -> define_param_dots d
   | Ast0.IdentTag(d)    -> ident d
   | Ast0.ExprTag(d) | Ast0.ArgExprTag(d) | Ast0.TestExprTag(d) ->
       print_string "Exp:"; force_newline();
@@ -745,6 +798,7 @@ let rec unparse_anything x =
   | Ast0.ParamTag(d) -> parameterTypeDef d
   | Ast0.InitTag(d)  -> initialiser d
   | Ast0.DeclTag(d)  -> declaration d
+  | Ast0.FieldTag(d)  -> field d
   | Ast0.StmtTag(d)  ->
       print_string "Stm:"; force_newline();
       statement "" d
@@ -774,6 +828,12 @@ let unparse x =
   print_newline()
 
 let unparse_to_string x = Common.format_to_string (function _ -> unparse x)
+let unparse_x_to_string fn x =
+  let q = !quiet in
+  quiet := true;
+  let res = Common.format_to_string (function _ -> fn x) in
+  quiet := q;
+  res
 
 let show_cocci_parse_tree comment parse_tree =
   Printf.printf "%s\n" comment;
