@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -25,6 +25,7 @@ type 'a combiner =
      combiner_fullType : Ast.fullType -> 'a;
      combiner_typeC : Ast.typeC -> 'a;
      combiner_declaration : Ast.declaration -> 'a;
+     combiner_field : Ast.field -> 'a;
      combiner_initialiser : Ast.initialiser -> 'a;
      combiner_parameter : Ast.parameterTypeDef -> 'a;
      combiner_parameter_list : Ast.parameter_list -> 'a;
@@ -36,6 +37,7 @@ type 'a combiner =
      combiner_expression_dots : Ast.expression Ast.dots -> 'a;
      combiner_statement_dots : Ast.statement Ast.dots -> 'a;
      combiner_anndecl_dots : Ast.annotated_decl Ast.dots -> 'a;
+     combiner_annfield_dots : Ast.annotated_field Ast.dots -> 'a;
      combiner_initialiser_dots : Ast.initialiser Ast.dots -> 'a}
 
 type ('mc,'a) cmcode = 'a combiner -> 'mc Ast_cocci.mcode -> 'a
@@ -47,10 +49,11 @@ let combiner bind option_default
     unary_mcodefn arithop_mcodefn logicalop_mcodefn
     cv_mcodefn sign_mcodefn struct_mcodefn storage_mcodefn
     inc_file_mcodefn
-    expdotsfn paramdotsfn stmtdotsfn anndecldotsfn initdotsfn
+    expdotsfn paramdotsfn stmtdotsfn anndecldotsfn annfielddotsfn initdotsfn
     identfn exprfn fragfn fmtfn assignOpfn binaryOpfn ftfn tyfn initfn
-    paramfn declfn
-    annotated_declfn rulefn stmtfn casefn topfn anyfn =
+    paramfn define_paramfn declfn
+    annotated_declfn fieldfn annotated_fieldfn rulefn stmtfn casefn topfn
+    anyfn =
   let multibind l =
     let rec loop = function
 	[] -> option_default
@@ -83,6 +86,7 @@ let combiner bind option_default
   and iddotsfn all_functions k arg = k arg
   and strdotsfn all_functions k arg = k arg
   and ecdotsfn all_functions k arg = k arg
+  and defpardotsfn all_functions k arg = k arg
 
   and expression_dots d = dotsfn expdotsfn expression all_functions d
   and identifier_dots d = dotsfn iddotsfn ident all_functions d
@@ -90,9 +94,12 @@ let combiner bind option_default
   and statement_dots d = dotsfn stmtdotsfn statement all_functions d
   and annotated_decl_dots d =
     dotsfn anndecldotsfn annotated_decl all_functions d
+  and annotated_field_dots d =
+    dotsfn annfielddotsfn annotated_field all_functions d
   and initialiser_dots d = dotsfn initdotsfn initialiser all_functions d
   and string_fragment_dots d = dotsfn strdotsfn string_fragment all_functions d
   and exec_code_dots d = dotsfn ecdotsfn exec_code all_functions d
+  and define_param_dots d = dotsfn defpardotsfn define_param all_functions d
 
   and ident i =
     let k i =
@@ -210,8 +217,8 @@ let combiner bind option_default
 	  let linit = initialiser init in
 	  multibind [llp; lty; lrp; linit]
       | Ast.MetaErr(name,_,_,_)
-      | Ast.MetaExpr(name,_,_,_,_,_)
-      | Ast.MetaExprList(name,_,_,_) -> meta_mcode name
+      | Ast.MetaExpr(name,_,_,_,_,_,_)
+      | Ast.MetaExprList(name,_,_,_,_) -> meta_mcode name
       |	Ast.AsExpr(exp,asexp) ->
 	  let lexp = expression exp in
 	  let lasexp = expression asexp in
@@ -261,7 +268,7 @@ let combiner bind option_default
 	  let fmt = string_format fmt in
 	  bind pct fmt
       |	Ast.Strdots dots -> string_mcode dots
-      | Ast.MetaFormatList(pct,name,lenname,_,_) ->
+      | Ast.MetaFormatList(pct,name,lenname,_,_,_) ->
 	  let pct = string_mcode pct in
 	  let name = meta_mcode name in
 	  bind pct name in
@@ -285,7 +292,8 @@ let combiner bind option_default
 	  let lty = fullType ty in
 	  let lasty = fullType asty in
 	  bind lty lasty
-      | Ast.DisjType(types) -> multibind (List.map fullType types)
+      | Ast.DisjType(types) | Ast.ConjType(types) ->
+	  multibind (List.map fullType types)
       | Ast.OptType(ty) -> fullType ty in
     ftfn all_functions k ft
 
@@ -350,11 +358,23 @@ let combiner bind option_default
       | Ast.StructUnionDef(ty,lb,decls,rb) ->
 	  let lty = fullType ty in
 	  let llb = string_mcode lb in
-	  let ldecls = annotated_decl_dots decls in
+	  let ldecls = annotated_field_dots decls in
 	  let lrb = string_mcode rb in
 	  multibind [lty; llb; ldecls; lrb]
+      | Ast.TypeOfExpr(tf,lp,e,rp) ->
+	  let ltf = string_mcode tf in
+	  let llp = string_mcode lp in
+	  let le = expression e in
+	  let lrp = string_mcode rp in
+	  multibind [ltf; llp; le; lrp]
+      | Ast.TypeOfType(tf,lp,ty,rp) ->
+	  let ltf = string_mcode tf in
+	  let llp = string_mcode lp in
+	  let lty = fullType ty in
+	  let lrp = string_mcode rp in
+	  multibind [ltf; llp; lty; lrp]
       | Ast.TypeName(name) -> string_mcode name
-      | Ast.MetaType(name,_,_) -> meta_mcode name in
+      | Ast.MetaType(name,_,_,_) -> meta_mcode name in
     tyfn all_functions k ty
 
   and named_type ty id =
@@ -374,25 +394,26 @@ let combiner bind option_default
   and declaration d =
     let k d =
       match Ast.unwrap d with
-	Ast.MetaDecl(name,_,_) | Ast.MetaField(name,_,_)
-      |	Ast.MetaFieldList(name,_,_,_) ->
+	Ast.MetaDecl(name,_,_,_) ->
 	  meta_mcode name
       |	Ast.AsDecl(decl,asdecl) ->
 	  let ldecl = declaration decl in
 	  let lasdecl = declaration asdecl in
 	  bind ldecl lasdecl
-      |	Ast.Init(stg,ty,id,eq,ini,sem) ->
+      |	Ast.Init(stg,ty,id,attr,eq,ini,sem) ->
 	  let lstg = get_option storage_mcode stg in
 	  let lid = named_type ty id in
+	  let lattr = multibind (List.map string_mcode attr) in
 	  let leq = string_mcode eq in
 	  let lini = initialiser ini in
 	  let lsem = string_mcode sem in
-	  multibind [lstg; lid; leq; lini; lsem]
-      | Ast.UnInit(stg,ty,id,sem) ->
+	  multibind [lstg; lid; lattr; leq; lini; lsem]
+      | Ast.UnInit(stg,ty,id,attr,sem) ->
 	  let lstg = get_option storage_mcode stg in
 	  let lid = named_type ty id in
+	  let lattr = multibind (List.map string_mcode attr) in
 	  let lsem = string_mcode sem in
-	  multibind [lstg; lid; lsem]
+	  multibind [lstg; lid; lattr; lsem]
       | Ast.FunProto(fi,name,lp1,params,va,rp1,sem) ->
 	  let lfi = List.map fninfo fi in
 	  let lname = ident name in
@@ -433,25 +454,54 @@ let combiner bind option_default
 	  let lid = typeC id in
 	  let lsem = string_mcode sem in
 	  multibind [lstg; lty; lid; lsem]
-      | Ast.DisjDecl(decls) -> multibind (List.map declaration decls)
+      | Ast.DisjDecl(decls)
+      | Ast.ConjDecl(decls) -> multibind (List.map declaration decls)
       | Ast.OptDecl(decl) -> declaration decl in
     declfn all_functions k d
 
   and annotated_decl d =
     let k d =
       match Ast.unwrap d with
-	Ast.DElem(_,_,d) -> declaration d
-      | Ast.Ddots(dots,whncode) ->
-	  let ldots = string_mcode dots in
-	  let lwhncode = get_option declaration whncode in
-	  bind ldots lwhncode in
+	Ast.DElem(_,_,d) -> declaration d in
     annotated_declfn all_functions k d
 
+  and field d =
+    let k d =
+      match Ast.unwrap d with
+	Ast.MetaField(name,_,_,_) | Ast.MetaFieldList(name,_,_,_,_) ->
+	  meta_mcode name
+      | Ast.Field(ty,id,bf,sem) ->
+	  let lid =
+	    match id with
+	      None -> fullType ty
+	    | Some id -> named_type ty id in
+	  let bitfield (c, e) =
+	    let lc = string_mcode c in
+	    let le = expression e in
+	    [lc; le] in
+	  let lbf = Common.default [] bitfield bf in
+	  let lsem = string_mcode sem in
+	  multibind ([lid] @ lbf @ [lsem])
+      | Ast.DisjField(decls)
+      | Ast.ConjField(decls) -> multibind (List.map field decls)
+      | Ast.OptField(decl) -> field decl in
+    fieldfn all_functions k d
+
+  and annotated_field d =
+    let k d =
+      match Ast.unwrap d with
+	Ast.FElem(_,_,d) -> field d
+      | Ast.Fdots(dots,whncode) ->
+	  let ldots = string_mcode dots in
+	  let lwhncode = get_option field whncode in
+	  bind ldots lwhncode in
+    annotated_fieldfn all_functions k d
+
   and initialiser i =
     let k i =
       match Ast.unwrap i with
-	Ast.MetaInit(name,_,_) -> meta_mcode name
-      |	Ast.MetaInitList(name,_,_,_) -> meta_mcode name
+	Ast.MetaInit(name,_,_,_) -> meta_mcode name
+      |	Ast.MetaInitList(name,_,_,_,_) -> meta_mcode name
       |	Ast.AsInit(init,asinit) ->
 	  let linit = initialiser init in
 	  let lasinit = initialiser asinit in
@@ -510,8 +560,8 @@ let combiner bind option_default
 	Ast.VoidParam(ty) -> fullType ty
       | Ast.Param(ty,Some id) -> named_type ty id
       | Ast.Param(ty,None) -> fullType ty
-      | Ast.MetaParam(name,_,_) -> meta_mcode name
-      | Ast.MetaParamList(name,_,_,_) -> meta_mcode name
+      | Ast.MetaParam(name,_,_,_) -> meta_mcode name
+      | Ast.MetaParamList(name,_,_,_,_) -> meta_mcode name
       |	Ast.AsParam(p,asexp) ->
 	  let lp = parameterTypeDef p in
 	  let lasexp = expression asexp in
@@ -615,9 +665,9 @@ let combiner bind option_default
 	  let lcode = exec_code_dots code in
 	  let lsem = string_mcode sem in
 	  multibind [lexec; lland; lcode; lsem]
-      | Ast.MetaStmt(name,_,_,_) -> meta_mcode name
-      | Ast.MetaStmtList(name,_,_,_) -> meta_mcode name
-      | Ast.MetaRuleElem(name,_,_) -> meta_mcode name
+      | Ast.MetaStmt(name,_,_,_,_) -> meta_mcode name
+      | Ast.MetaStmtList(name,_,_,_,_) -> meta_mcode name
+      | Ast.MetaRuleElem(name,_,_,_) -> meta_mcode name
       | Ast.Exp(exp) -> expression exp
       | Ast.TopExp(exp) -> expression exp
       | Ast.Ty(ty) -> fullType ty
@@ -627,6 +677,10 @@ let combiner bind option_default
 	  let linc = string_mcode inc in
 	  let lname = inc_file_mcode name in
 	  bind linc lname
+      | Ast.MetaInclude(inc,name) ->
+	  let linc = string_mcode inc in
+	  let lname = expression name in
+	  bind linc lname
       |	Ast.Undef(def,id) ->
 	  let ldef = string_mcode def in
 	  let lid = ident id in
@@ -693,16 +747,15 @@ let combiner bind option_default
 	  multibind [llp; lparams; lrp] in
     k p
 
-  and define_param_dots d = multibind (List.map define_param (Ast.unwrap d))
-
   and define_param p =
     let k p =
       match Ast.unwrap p with
 	Ast.DParam(id) -> ident id
+      | Ast.MetaDParamList(name,_,_,_,_) -> meta_mcode name
       | Ast.DPComma(comma) -> string_mcode comma
       | Ast.DPdots(d) -> string_mcode d
       | Ast.OptDParam(dp) -> define_param dp in
-    k p
+    define_paramfn all_functions k p
 
   (* discard the result, because the statement is assumed to be already
      represented elsewhere in the code *)
@@ -841,14 +894,15 @@ let combiner bind option_default
       | Ast.ExpressionTag(exp) -> expression exp
       | Ast.ConstantTag(cst) -> option_default
       | Ast.UnaryOpTag(unop) -> option_default
-      | Ast.AssignOpTag(asgnop) -> option_default
+      | Ast.AssignOpTag(asgnop) -> assignOp asgnop
       | Ast.SimpleAssignOpTag _ -> option_default
       | Ast.OpAssignOpTag _ -> option_default
       | Ast.FixOpTag(fixop) -> option_default
-      | Ast.BinaryOpTag(binop) -> option_default
+      | Ast.BinaryOpTag(binop) -> binaryOp binop
       | Ast.ArithOpTag(arithop) -> option_default
       | Ast.LogicalOpTag(logop) -> option_default
       | Ast.DeclarationTag(decl) -> declaration decl
+      | Ast.FieldTag(decl) -> field decl
       | Ast.InitTag(ini) -> initialiser ini
       | Ast.StorageTag(stg) -> option_default
       | Ast.IncFileTag(stg) -> option_default
@@ -865,6 +919,8 @@ let combiner bind option_default
       | Ast.ParamDotsTag(pd) -> parameter_dots pd
       | Ast.StmtDotsTag(sd) -> statement_dots sd
       | Ast.AnnDeclDotsTag(sd) -> annotated_decl_dots sd
+      | Ast.AnnFieldDotsTag(sd) -> annotated_field_dots sd
+      | Ast.DefParDotsTag(sd) -> define_param_dots sd
       | Ast.TypeCTag(ty) -> typeC ty
       | Ast.ParamTag(param) -> parameterTypeDef param
       | Ast.SgrepStartTag(tok) -> option_default
@@ -881,6 +937,7 @@ let combiner bind option_default
       combiner_fullType = fullType;
       combiner_typeC = typeC;
       combiner_declaration = declaration;
+      combiner_field = field;
       combiner_initialiser = initialiser;
       combiner_parameter = parameterTypeDef;
       combiner_parameter_list = parameter_dots;
@@ -892,6 +949,7 @@ let combiner bind option_default
       combiner_expression_dots = expression_dots;
       combiner_statement_dots = statement_dots;
       combiner_anndecl_dots = annotated_decl_dots;
+      combiner_annfield_dots = annotated_field_dots;
       combiner_initialiser_dots = initialiser_dots} in
   all_functions
 
@@ -909,6 +967,7 @@ type rebuilder =
       rebuilder_fullType : Ast.fullType inout;
       rebuilder_typeC : Ast.typeC inout;
       rebuilder_declaration : Ast.declaration inout;
+      rebuilder_field : Ast.field inout;
       rebuilder_initialiser : Ast.initialiser inout;
       rebuilder_parameter : Ast.parameterTypeDef inout;
       rebuilder_parameter_list : Ast.parameter_list inout;
@@ -919,6 +978,7 @@ type rebuilder =
       rebuilder_expression_dots : Ast.expression Ast.dots inout;
       rebuilder_statement_dots : Ast.statement Ast.dots inout;
       rebuilder_anndecl_dots : Ast.annotated_decl Ast.dots inout;
+      rebuilder_annfield_dots : Ast.annotated_field Ast.dots inout;
       rebuilder_initialiser_dots : Ast.initialiser Ast.dots inout;
       rebuilder_define_param_dots : Ast.define_param Ast.dots inout;
       rebuilder_define_param : Ast.define_param inout;
@@ -934,9 +994,10 @@ let rebuilder
     fix_mcode unary_mcode
     arithop_mcode logicalop_mcode cv_mcode sign_mcode struct_mcode
     storage_mcode inc_file_mcode
-    expdotsfn paramdotsfn stmtdotsfn anndecldotsfn initdotsfn
+    expdotsfn paramdotsfn stmtdotsfn anndecldotsfn annfielddotsfn initdotsfn
     identfn exprfn fragfn fmtfn assignOpfn binaryOpfn ftfn tyfn initfn
-    paramfn declfn annotated_declfn rulefn stmtfn casefn topfn anyfn =
+    paramfn define_paramfn declfn annotated_declfn fieldfn annotated_fieldfn
+    rulefn stmtfn casefn topfn anyfn =
   let get_option f = function
       Some x -> Some (f x)
     | None -> None in
@@ -955,6 +1016,8 @@ let rebuilder
   and statement_dots d = dotsfn stmtdotsfn statement all_functions d
   and annotated_decl_dots d =
     dotsfn anndecldotsfn annotated_decl all_functions d
+  and annotated_field_dots d =
+    dotsfn annfielddotsfn annotated_field all_functions d
   and initialiser_dots d = dotsfn initdotsfn initialiser all_functions d
   and string_fragment_dots d = dotsfn strdotsfn string_fragment all_functions d
   and exec_code_dots d = dotsfn ecdotsfn exec_code all_functions d
@@ -1077,10 +1140,12 @@ let rebuilder
 	    Ast.Constructor(llp, lty, lrp, linit)
 	| Ast.MetaErr(name,constraints,keep,inherited) ->
 	    Ast.MetaErr(meta_mcode name,constraints,keep,inherited)
-	| Ast.MetaExpr(name,constraints,keep,ty,form,inherited) ->
-	    Ast.MetaExpr(meta_mcode name,constraints,keep,ty,form,inherited)
-	| Ast.MetaExprList(name,lenname_inh,keep,inherited) ->
-	    Ast.MetaExprList(meta_mcode name,lenname_inh,keep,inherited)
+	| Ast.MetaExpr(name,constraints,keep,ty,form,inherited,bitfield) ->
+	    Ast.MetaExpr
+	      (meta_mcode name,constraints,keep,ty,form,inherited,bitfield)
+	| Ast.MetaExprList(name,lenname_inh,constraints,keep,inherited) ->
+	    Ast.MetaExprList
+	      (meta_mcode name,lenname_inh,constraints,keep,inherited)
 	| Ast.AsExpr(exp,asexp) ->
 	    let lexp = expression exp in
 	    let lasexp = expression asexp in
@@ -1115,10 +1180,11 @@ let rebuilder
 	    let lfmt = string_format fmt in
 	    Ast.FormatFragment(lpct, lfmt)
 	| Ast.Strdots dots -> Ast.Strdots (string_mcode dots)
-	| Ast.MetaFormatList(pct,name,lenname,keep,inherited) ->
+	| Ast.MetaFormatList(pct,name,lenname,constraints,keep,inherited) ->
 	    let lpct = string_mcode pct in
 	    let lname = meta_mcode name in
-	    Ast.MetaFormatList(lpct, lname, lenname, keep, inherited)) in
+	    Ast.MetaFormatList
+	      (lpct, lname, lenname, constraints, keep, inherited)) in
     fragfn all_functions k e
 
   and string_format e =
@@ -1163,6 +1229,7 @@ let rebuilder
 	    let lasty = fullType asty in
 	    Ast.AsType(lty, lasty)
 	| Ast.DisjType(types) -> Ast.DisjType(List.map fullType types)
+	| Ast.ConjType(types) -> Ast.ConjType(List.map fullType types)
 	| Ast.OptType(ty) -> Ast.OptType(fullType ty)) in
     ftfn all_functions k ft
 
@@ -1220,42 +1287,52 @@ let rebuilder
 	| Ast.StructUnionDef(ty,lb,decls,rb) ->
 	    let lty = fullType ty in
 	    let llb = string_mcode lb in
-	    let ldecls = annotated_decl_dots decls in
+	    let ldecls = annotated_field_dots decls in
 	    let lrb = string_mcode rb in
 	    Ast.StructUnionDef (lty, llb, ldecls, lrb)
+	| Ast.TypeOfExpr(tf,lp,e,rp) ->
+	    let ltf = string_mcode tf in
+	    let llp = string_mcode lp in
+	    let le = expression e in
+	    let lrp = string_mcode rp in
+	    Ast.TypeOfExpr(ltf, llp, le, lrp)
+	| Ast.TypeOfType(tf,lp,ty,rp) ->
+	    let ltf = string_mcode tf in
+	    let llp = string_mcode lp in
+	    let lty = fullType ty in
+	    let lrp = string_mcode rp in
+	    Ast.TypeOfType(ltf, llp, lty, lrp)
 	| Ast.TypeName(name) -> Ast.TypeName(string_mcode name)
-	| Ast.MetaType(name,keep,inherited) ->
-	    Ast.MetaType(meta_mcode name,keep,inherited)) in
+	| Ast.MetaType(name,cstr,keep,inherited) ->
+	    Ast.MetaType(meta_mcode name,cstr,keep,inherited)) in
     tyfn all_functions k ty
 
   and declaration d =
     let k d =
       Ast.rewrap d
 	(match Ast.unwrap d with
-	  Ast.MetaDecl(name,keep,inherited) ->
-	    Ast.MetaDecl(meta_mcode name,keep,inherited)
-	| Ast.MetaField(name,keep,inherited) ->
-	    Ast.MetaField(meta_mcode name,keep,inherited)
-	| Ast.MetaFieldList(name,lenname_inh,keep,inherited) ->
-	    Ast.MetaFieldList(meta_mcode name,lenname_inh,keep,inherited)
+	  Ast.MetaDecl(name,constraints,keep,inherited) ->
+	    Ast.MetaDecl(meta_mcode name,constraints,keep,inherited)
 	| Ast.AsDecl(decl,asdecl) ->
 	    let ldecl = declaration decl in
 	    let lasdecl = declaration asdecl in
 	    Ast.AsDecl(ldecl, lasdecl)
-	| Ast.Init(stg,ty,id,eq,ini,sem) ->
+	| Ast.Init(stg,ty,id,attr,eq,ini,sem) ->
 	    let lstg = get_option storage_mcode stg in
 	    let lty = fullType ty in
 	    let lid = ident id in
+	    let lattr = List.map string_mcode attr in
 	    let leq = string_mcode eq in
 	    let lini = initialiser ini in
 	    let lsem = string_mcode sem in
-	    Ast.Init(lstg, lty, lid, leq, lini, lsem)
-	| Ast.UnInit(stg,ty,id,sem) ->
+	    Ast.Init(lstg, lty, lid, lattr, leq, lini, lsem)
+	| Ast.UnInit(stg,ty,id,attr,sem) ->
 	    let lstg = get_option storage_mcode stg in
 	    let lty = fullType ty in
 	    let lid = ident id in
+	    let lattr = List.map string_mcode attr in
 	    let lsem = string_mcode sem in
-	    Ast.UnInit(lstg, lty, lid, lsem)
+	    Ast.UnInit(lstg, lty, lid, lattr, lsem)
 	| Ast.FunProto(fi,name,lp,params,va,rp,sem) ->
 	    let lfi = List.map fninfo fi in
 	    let lname = ident name in
@@ -1297,6 +1374,7 @@ let rebuilder
 	    let lsem = string_mcode sem in
 	    Ast.Typedef(lstg, lty, lid, lsem)
 	| Ast.DisjDecl(decls) -> Ast.DisjDecl(List.map declaration decls)
+	| Ast.ConjDecl(decls) -> Ast.ConjDecl(List.map declaration decls)
 	| Ast.OptDecl(decl) -> Ast.OptDecl(declaration decl)) in
     declfn all_functions k d
 
@@ -1305,21 +1383,54 @@ let rebuilder
       Ast.rewrap d
 	(match Ast.unwrap d with
 	  Ast.DElem(bef,allminus,decl) ->
-	    Ast.DElem(bef,allminus,declaration decl)
-	| Ast.Ddots(dots,whncode) ->
-	    let ldots = string_mcode dots in
-	    let lwhncode = get_option declaration whncode in
-	    Ast.Ddots(ldots, lwhncode)) in
+	    Ast.DElem(bef,allminus,declaration decl)) in
     annotated_declfn all_functions k d
 
+  and field d =
+    let k d =
+      Ast.rewrap d
+	(match Ast.unwrap d with
+	  Ast.MetaField(name,constraints,keep,inherited) ->
+	    Ast.MetaField(meta_mcode name,constraints,keep,inherited)
+	| Ast.MetaFieldList(name,lenname_inh,constraints,keep,inherited) ->
+	    Ast.MetaFieldList
+	      (meta_mcode name,lenname_inh,constraints,keep,inherited)
+	| Ast.Field(ty,id,bf,sem) ->
+	    let lty = fullType ty in
+	    let lid = Common.map_option ident id in
+	    let bitfield (c, e) =
+	      let lc = string_mcode c in
+	      let le = expression e in
+	      (lc, le) in
+	    let lbf = Common.map_option bitfield bf in
+	    let lsem = string_mcode sem in
+	    Ast.Field(lty, lid, lbf, lsem)
+	| Ast.DisjField(decls) -> Ast.DisjField(List.map field decls)
+	| Ast.ConjField(decls) -> Ast.ConjField(List.map field decls)
+	| Ast.OptField(decl) -> Ast.OptField(field decl)) in
+    fieldfn all_functions k d
+
+  and annotated_field d =
+    let k d =
+      Ast.rewrap d
+	(match Ast.unwrap d with
+	  Ast.FElem(bef, allminus, decl) ->
+	    Ast.FElem(bef, allminus, field decl)
+	| Ast.Fdots(dots,whncode) ->
+	    let ldots = string_mcode dots in
+	    let lwhncode = get_option field whncode in
+	    Ast.Fdots(ldots, lwhncode)) in
+    annotated_fieldfn all_functions k d
+
   and initialiser i =
     let k i =
       Ast.rewrap i
 	(match Ast.unwrap i with
-	  Ast.MetaInit(name,keep,inherited) ->
-	    Ast.MetaInit(meta_mcode name,keep,inherited)
-	| Ast.MetaInitList(name,lenname_inh,keep,inherited) ->
-	    Ast.MetaInitList(meta_mcode name,lenname_inh,keep,inherited)
+	  Ast.MetaInit(name,constraints,keep,inherited) ->
+	    Ast.MetaInit(meta_mcode name,constraints,keep,inherited)
+	| Ast.MetaInitList(name,lenname_inh,constraints,keep,inherited) ->
+	    Ast.MetaInitList
+	      (meta_mcode name,lenname_inh,constraints,keep,inherited)
 	| Ast.AsInit(ini,asini) ->
 	    let lini = initialiser ini in
 	    let lasinit = initialiser asini in
@@ -1378,10 +1489,11 @@ let rebuilder
 	(match Ast.unwrap p with
 	  Ast.VoidParam(ty) -> Ast.VoidParam(fullType ty)
 	| Ast.Param(ty,id) -> Ast.Param(fullType ty, get_option ident id)
-	| Ast.MetaParam(name,keep,inherited) ->
-	    Ast.MetaParam(meta_mcode name,keep,inherited)
-	| Ast.MetaParamList(name,lenname_inh,keep,inherited) ->
-	    Ast.MetaParamList(meta_mcode name,lenname_inh,keep,inherited)
+	| Ast.MetaParam(name,constraints,keep,inherited) ->
+	    Ast.MetaParam(meta_mcode name,constraints,keep,inherited)
+	| Ast.MetaParamList(name,lenname_inh,constraints,keep,inherited) ->
+	    Ast.MetaParamList
+	      (meta_mcode name,lenname_inh,constraints,keep,inherited)
 	| Ast.AsParam(p,asexp) ->
 	    let lp = parameterTypeDef p in
 	    let lasexp = expression asexp in
@@ -1486,12 +1598,13 @@ let rebuilder
 	    let lcode = exec_code_dots code in
 	    let lsem = string_mcode sem in
 	    Ast.Exec(lexec, lland, lcode, lsem)
-	| Ast.MetaStmt(name,keep,seqible,inherited) ->
-	    Ast.MetaStmt(meta_mcode name,keep,seqible,inherited)
-	| Ast.MetaStmtList(name,lenname_inh,keep,inherited) ->
-	    Ast.MetaStmtList(meta_mcode name,lenname_inh,keep,inherited)
-	| Ast.MetaRuleElem(name,keep,inherited) ->
-	    Ast.MetaRuleElem(meta_mcode name,keep,inherited)
+	| Ast.MetaStmt(name,constraints,keep,seqible,inherited) ->
+	    Ast.MetaStmt(meta_mcode name,constraints,keep,seqible,inherited)
+	| Ast.MetaStmtList(name,lenname_inh,constraints,keep,inherited) ->
+	    Ast.MetaStmtList
+	      (meta_mcode name,lenname_inh,constraints,keep,inherited)
+	| Ast.MetaRuleElem(name,constraints,keep,inherited) ->
+	    Ast.MetaRuleElem(meta_mcode name,constraints,keep,inherited)
 	| Ast.Exp(exp) -> Ast.Exp(expression exp)
 	| Ast.TopExp(exp) -> Ast.TopExp(expression exp)
 	| Ast.Ty(ty) -> Ast.Ty(fullType ty)
@@ -1501,6 +1614,10 @@ let rebuilder
 	    let linc = string_mcode inc in
 	    let lname = inc_file_mcode name in
 	    Ast.Include(linc, lname)
+	| Ast.MetaInclude(inc,name) ->
+	    let linc = string_mcode inc in
+	    let lname = expression name in
+	    Ast.MetaInclude(linc, lname)
 	| Ast.Undef(def,id) ->
 	    let ldef = string_mcode def in
 	    let lid = ident id in
@@ -1576,10 +1693,13 @@ let rebuilder
       Ast.rewrap p
 	(match Ast.unwrap p with
 	  Ast.DParam(id) -> Ast.DParam(ident id)
+	| Ast.MetaDParamList(name,lenname_inh,constraints,keep,inherited) ->
+	    Ast.MetaDParamList
+	      (meta_mcode name,lenname_inh,constraints,keep,inherited)
 	| Ast.DPComma(comma) -> Ast.DPComma(string_mcode comma)
 	| Ast.DPdots(d) -> Ast.DPdots(string_mcode d)
 	| Ast.OptDParam(dp) -> Ast.OptDParam(define_param dp)) in
-    k p
+    define_paramfn all_functions k p
 
   and process_bef_aft s =
     Ast.set_dots_bef_aft
@@ -1736,6 +1856,7 @@ let rebuilder
       | Ast.LogicalOpTag(logop) as x -> x
       | Ast.InitTag(decl) -> Ast.InitTag(initialiser decl)
       | Ast.DeclarationTag(decl) -> Ast.DeclarationTag(declaration decl)
+      | Ast.FieldTag(decl) -> Ast.FieldTag(field decl)
       | Ast.StorageTag(stg) as x -> x
       | Ast.IncFileTag(stg) as x -> x
       | Ast.Rule_elemTag(rule) -> Ast.Rule_elemTag(rule_elem rule)
@@ -1752,6 +1873,8 @@ let rebuilder
       | Ast.ParamDotsTag(pd) -> Ast.ParamDotsTag(parameter_dots pd)
       | Ast.StmtDotsTag(sd) -> Ast.StmtDotsTag(statement_dots sd)
       | Ast.AnnDeclDotsTag(sd) -> Ast.AnnDeclDotsTag(annotated_decl_dots sd)
+      | Ast.AnnFieldDotsTag(sd) -> Ast.AnnFieldDotsTag(annotated_field_dots sd)
+      | Ast.DefParDotsTag(sd) -> Ast.DefParDotsTag(define_param_dots sd)
       | Ast.TypeCTag(ty) -> Ast.TypeCTag(typeC ty)
       | Ast.ParamTag(param) -> Ast.ParamTag(parameterTypeDef param)
       | Ast.SgrepStartTag(tok) as x -> x
@@ -1768,6 +1891,7 @@ let rebuilder
       rebuilder_fullType = fullType;
       rebuilder_typeC = typeC;
       rebuilder_declaration = declaration;
+      rebuilder_field = field;
       rebuilder_initialiser = initialiser;
       rebuilder_parameter = parameterTypeDef;
       rebuilder_parameter_list = parameter_dots;
@@ -1778,6 +1902,7 @@ let rebuilder
       rebuilder_expression_dots = expression_dots;
       rebuilder_statement_dots = statement_dots;
       rebuilder_anndecl_dots = annotated_decl_dots;
+      rebuilder_annfield_dots = annotated_field_dots;
       rebuilder_initialiser_dots = initialiser_dots;
       rebuilder_define_param_dots = define_param_dots;
       rebuilder_define_param = define_param;
