@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -69,9 +69,12 @@ let sp_exit _ =
   Py.none
 
 let build_class cname parent fields methods pymodule =
+  let methods =
+    List.map (fun (name, f) -> (name, Py.Callable.of_function_as_tuple f))
+      methods in
   let cx =
-    Py.Class.init (Py.String.of_string cname)
-      ~parents:(Py.Tuple.singleton (pycocci_get_class_type parent))
+    Py.Class.init cname
+      ~parents:[pycocci_get_class_type parent]
       ~fields ~methods in
   Py.Module.set pymodule cname cx;
   cx
@@ -166,13 +169,24 @@ let _pycocci_setargs argv0 =
   let sys_mod = load_module "sys" in
   Py.Module.set sys_mod "argv" argv
 
+let initialize_python_path () =
+  let sep = ":" in
+  let python_libdir = Filename.concat Config.path "python" in
+  match Common.optionise (fun () -> Sys.getenv "PYTHONPATH") with
+    None -> Unix.putenv "PYTHONPATH" python_libdir
+  | Some paths ->
+      let path_list = Common.split sep paths in
+      if not (List.mem python_libdir path_list) then
+	Unix.putenv "PYTHONPATH"
+	  (Printf.sprintf "%s%s%s" paths sep python_libdir)
+
 let pycocci_init () =
   (* initialize *)
   if not !initialised then (
-  initialised := true;
-  let _ = if not (Py.is_initialized ()) then
-  	(if !Flag.show_misc then Common.pr2 "Initializing python\n%!";
-	Py.initialize ()) in
+    initialize_python_path ();
+    let _ = if not (Py.is_initialized ()) then
+      (if !Flag.show_misc then Common.pr2 "Initializing python\n%!";
+	Py.initialize ~interpreter:!Config.python_interpreter ()) in
 
   (* set argv *)
   let argv0 = Sys.executable_name in
@@ -207,6 +221,7 @@ let pycocci_init () =
   let (wrap_ast, unwrap_ast) = Py.Capsule.make "metavar_binding_kind" in
   pywrap_ast := wrap_ast;
   pyunwrap_ast := unwrap_ast;
+  initialised := true;
   ()) else
   ()
 
@@ -287,6 +302,8 @@ let construct_variables mv e =
 	(Py.Tuple.of_pair (str, elements)) in
     let _ = build_variable py repr in () in
 
+  let locals = List.map (function (py,(r,m),_,init) -> py) mv in
+
   List.iter (function (py,(r,m),_,init) ->
     match find_binding (r,m) with
       None ->
@@ -319,11 +336,25 @@ let construct_variables mv e =
     | Some (_, Ast_c.MetaPosValList l) ->
        let locs =
 	 List.map
-	   (function (fname,current_element,(line,col),(line_end,col_end)) ->
-		pycocci_instantiate_class "coccilib.elems.Location"
+	   (function (fname,current_element,current_element_pos,
+		      (line,col),(line_end,col_end)) ->
+	     let (current_element_line,current_element_col,
+		  current_element_line_end,current_element_col_end) =
+	       match current_element_pos with
+		 Some
+		   ((current_element_line,current_element_col),
+		    (current_element_line_end,current_element_col_end)) ->
+		      (current_element_line,current_element_col,
+		       current_element_line_end,current_element_col_end)
+	       | None -> (-1,-1,-1,-1) in
+	     pycocci_instantiate_class "coccilib.elems.Location"
 	       (Py.Tuple.of_list_map Py.String.of_string
 		  [fname;
 		   current_element;
+		   string_of_int current_element_line;
+		   string_of_int current_element_col;
+		   string_of_int current_element_line_end;
+		   string_of_int current_element_col_end;
 		   string_of_int line;
 		   string_of_int col;
 		   string_of_int line_end;
@@ -338,7 +369,9 @@ let construct_variables mv e =
        ()
     ) mv;
 
-  let add_string_literal s = build_variable s (Py.String.of_string s) in
+  let add_string_literal s =
+    if not (List.mem s locals)
+    then build_variable s (Py.String.of_string s) in
   List.iter add_string_literal !Iteration.parsed_virtual_rules;
   List.iter add_string_literal !Iteration.parsed_virtual_identifiers
 
@@ -366,30 +399,53 @@ let set_coccifile cocci_file =
 	cocci_file_name := cocci_file;
 	()
 
+let pickle_variable var =
+  Py.Marshal.dumps (get_variable var)
+
+let unpickle_variable var value =
+  let py_list = Py.List.of_list (List.map Py.Marshal.loads value) in
+  Py.Module.set !coccinelle_module var py_list
+
 let pyrun_simplestring s =
   catch_python_error begin fun () ->
     if not (Py.Run.simple_string s) then
       failwith "Python failure"
   end
 
+let run (file, line) s =
+  try
+    pyrun_simplestring s
+  with Failure msg ->
+    failwith (
+    Printf.sprintf "Error in Python script, line %d, file \"%s\": %s"
+      line file msg)
+
 let py_isinitialized () =
   Py.is_initialized ()
 
-
 let py_finalize () =
   Py.finalize ()
 
-let run_constraint args body =
+let run_constraint args pos body =
   catch_python_error begin fun () ->
     build_classes [];
     let make_arg (name, value) =
       (snd name, name, value, Ast_cocci.NoMVInit) in
     let mv = List.map make_arg args in
     construct_variables mv args;
-    pyrun_simplestring (Printf.sprintf "
+    run pos (Printf.sprintf "
 from coccinelle import *
 from coccilib.iteration import Iteration
 
 coccinelle.result = (%s)" body);
     Py.Bool.to_bool (get_variable "result")
   end
+
+let flush_stdout_and_stderr () =
+  if py_isinitialized () then
+    let _ = pyrun_simplestring "\
+import sys
+sys.stdout.flush()
+sys.stderr.flush()
+" in
+    ()
