@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -10,6 +10,74 @@ open Common
 (* Test framework *)
 (*****************************************************************************)
 
+type redirected_output = {
+    out_file: string;
+    out_channel: out_channel;
+    stdout_backup: Unix.file_descr;
+    show_diff_backup: bool;
+  }
+
+let out_suffix = ".stdout"
+
+let flush_scripts_output () =
+  flush stdout;
+  Pycocci.flush_stdout_and_stderr ()
+
+let begin_redirect_output expected_out =
+  let has_expected_out = Sys.file_exists expected_out in
+  if has_expected_out then
+    let out_file = Common.new_temp_file "redirect" out_suffix in
+    let out_channel = open_out out_file in
+    let out_file_descr = Unix.descr_of_out_channel out_channel in
+    let stdout_backup = Unix.dup Unix.stdout in
+    let show_diff_backup = !Flag_cocci.show_diff in
+    flush_scripts_output ();
+    Unix.dup2 out_file_descr Unix.stdout;
+    Flag_cocci.show_diff := false;
+    Some { out_file; out_channel; stdout_backup; show_diff_backup }
+  else
+    None
+
+let end_redirect_output = Common.map_option (
+  fun { out_file; out_channel; stdout_backup; show_diff_backup } ->
+    flush_scripts_output ();
+    Unix.dup2 stdout_backup Unix.stdout;
+    close_out out_channel;
+    Flag_cocci.show_diff := show_diff_backup;
+    out_file)
+
+let rec test_loop previous_merges cocci_file cfiles =
+  let (cocci_infos,_) = Cocci.pre_engine (cocci_file, !Config.std_iso) in
+  let (res, merges) = Cocci.full_engine cocci_infos cfiles in
+  let merges = Cocci.union_merge_vars previous_merges merges in
+  let next_merges, pending_instance =
+    match Iteration.get_pending_instance () with
+      None ->
+      begin
+        Cocci.post_engine cocci_infos merges;
+        ([], []), Iteration.get_pending_instance ()
+      end
+    | Some instances -> merges, Some instances in
+  match pending_instance with
+    None -> (cocci_infos, res)
+  | Some (cfiles', virt_rules, virt_ids) ->
+      Flag.defined_virtual_rules := virt_rules;
+      Flag.defined_virtual_env := virt_ids;
+      Common.clear_pr2_once ();
+      test_loop next_merges cocci_file cfiles'
+
+let test_with_output_redirected cocci_file cfiles expected_out =
+  Iteration.initialization_stack := [];
+  Postprocess_transinfo.reset_fresh_counter ();
+  let redirected_output = begin_redirect_output expected_out in
+  let (cocci_infos, res) =
+    try test_loop ([], []) cocci_file cfiles
+    with e ->
+      ignore (end_redirect_output redirected_output);
+      raise e in
+  let current_out = end_redirect_output redirected_output in
+  (res, current_out)
+
 (* There can be multiple .c for the same cocci file. The convention
  * is to have one base.cocci and a base.c and some optional
  * base_vernn.[c,res].
@@ -23,11 +91,12 @@ let testone prefix x compare_with_expected_flag =
   let cfile      = prefix ^ x ^ ".c" in
   let cocci_file = prefix ^ base ^ ".cocci" in
 
+  let expected_out = prefix ^ base ^ out_suffix in
+
   let expected_res   = prefix ^ x ^ ".res" in
   begin
-    let (cocci_infos,_) = Cocci.pre_engine (cocci_file, !Config.std_iso) in
-    let res = Cocci.full_engine cocci_infos [cfile] in
-    Cocci.post_engine cocci_infos;
+    let (res, current_out) =
+      test_with_output_redirected cocci_file [cfile] expected_out in
     let generated =
       match Common.optionise (fun () -> List.assoc cfile res) with
       | Some (Some outfile) ->
@@ -46,12 +115,51 @@ let testone prefix x compare_with_expected_flag =
     in
     if compare_with_expected_flag
     then
-      Compare_c.compare_default generated expected_res
-      +> Compare_c.compare_result_to_string
-      +> pr2;
-  end
+      begin
+	Compare_c.compare_default generated expected_res
+	  +> Compare_c.compare_result_to_string
+	  +> pr2;
+	match current_out with
+	  None -> ()
+	| Some current_out' ->
+	    Compare_c.exact_compare current_out' expected_out
+	      +> Compare_c.compare_result_to_string
+	      +> pr2
+      end
+  end;
+  Common.erase_temp_files ()
 
 
+let add_file_to_score score res correct diffxs =
+  (* I don't use Compare_c.compare_result_to_string because
+   * I want to indent a little more the messages.
+   *)
+  match correct with
+    Compare_c.Correct -> Hashtbl.add score res Common.Ok;
+  | Compare_c.Pb s ->
+      let s = Str.global_replace
+	  (Str.regexp "\"/tmp/cocci-output.*\"") "<COCCIOUTPUTFILE>" s
+      in
+      (* on macos the temporary files are stored elsewhere *)
+      let s =
+	Str.global_replace
+	  (Str.regexp "\"/var/folders/.*/cocci-output.*\"")
+	  "<COCCIOUTPUTFILE>" s
+      in
+      let s =
+	"INCORRECT:" ^ s ^ "\n" ^
+	"    diff (result(<) vs expected_result(>)) = \n" ^
+	(diffxs +>
+	 List.map(fun s -> "    "^s^"\n") +> String.concat "")
+      in
+      Hashtbl.add score res (Common.Pb s)
+  | Compare_c.PbOnlyInNotParsedCorrectly s ->
+      let s =
+	"seems incorrect, but only because of code that " ^
+	"was not parsable" ^ s
+      in
+      Hashtbl.add score res (Common.Pb s)
+
 (* ------------------------------------------------------------------------ *)
 (* note: if you get some weird results in -testall, and not in -test,
  * it is possible that a test file work in -test but may not
@@ -87,18 +195,18 @@ let testall_bis extra_test expected_score_file update_score_file =
       let cfile      = "tests/" ^ x ^ ".c" in
       let cocci_file = "tests/" ^ base ^ ".cocci" in
       let expected = "tests/" ^ res in
+      let out = base ^ out_suffix in
+      let expected_out = "tests/" ^ out in
 
       let timeout_testall = 60 in
 
       try (
-        Common.timeout_function timeout_testall  (fun () ->
+        Common.timeout_function "testing" timeout_testall  (fun () ->
 
 	  pr2 res;
 
-	  let (cocci_infos,_) =
-	    Cocci.pre_engine (cocci_file, !Config.std_iso) in
-          let xs = Cocci.full_engine cocci_infos [cfile] in
-	  Cocci.post_engine cocci_infos;
+	  let (xs, current_out) =
+	    test_with_output_redirected cocci_file [cfile] expected_out in
 
           let generated =
             match List.assoc cfile xs with
@@ -120,35 +228,17 @@ let testall_bis extra_test expected_score_file update_score_file =
 		       things that fail on the first test *)
 		    (Compare_c.Correct,[])) in
 
-          (* I don't use Compare_c.compare_result_to_string because
-           * I want to indent a little more the messages.
-           *)
-          (match correct with
-          | Compare_c.Correct -> Hashtbl.add score res Common.Ok;
-          | Compare_c.Pb s ->
-              let s = Str.global_replace
-                (Str.regexp "\"/tmp/cocci-output.*\"") "<COCCIOUTPUTFILE>" s
-              in
-              (* on macos the temporary files are stored elsewhere *)
-              let s =
-		Str.global_replace
-                  (Str.regexp "\"/var/folders/.*/cocci-output.*\"")
-		  "<COCCIOUTPUTFILE>" s
-              in
-              let s =
-                "INCORRECT:" ^ s ^ "\n" ^
-                "    diff (result(<) vs expected_result(>)) = \n" ^
-                (diffxs +>
-		 List.map(fun s -> "    "^s^"\n") +> String.concat "")
-              in
-              Hashtbl.add score res (Common.Pb s)
-          | Compare_c.PbOnlyInNotParsedCorrectly s ->
-              let s =
-                "seems incorrect, but only because of code that " ^
-                "was not parsable" ^ s
-              in
-              Hashtbl.add score res (Common.Pb s)
-          )
+	  add_file_to_score score res correct diffxs;
+
+	  begin
+	    match current_out with
+	      None -> ()
+	    | Some current_out' ->
+		let (correct, diffxs) =
+		  Compare_c.exact_compare current_out' expected_out in
+		add_file_to_score score out correct diffxs
+	  end;
+	  Common.erase_temp_files ();
         )
       )
       with exn ->
@@ -291,13 +381,14 @@ let test_okfailed cocci_file cfiles =
     spf "time: %f" tperfile
   in
 
+  let expected_out = Filename.chop_suffix cocci_file ".cocci" ^ out_suffix in
+
   Common.redirect_stdout_stderr newout (fun () ->
     try (
-      Common.timeout_function_opt !Flag_cocci.timeout (fun () ->
+      Common.timeout_function_opt "testing" !Flag_cocci.timeout (fun () ->
 
-	let (cocci_infos,_) = Cocci.pre_engine (cocci_file, !Config.std_iso) in
-        let outfiles = Cocci.full_engine cocci_infos cfiles in
-	Cocci.post_engine cocci_infos;
+	let (outfiles, current_out) =
+	  test_with_output_redirected cocci_file cfiles expected_out in
 
         let time_str = time_per_file_str () in
 
@@ -352,8 +443,16 @@ let test_okfailed cocci_file cfiles =
                       final_files
                 end
 		else push2 (infile ^ (t_to_s Failed), [s1;time_str]) final_files
-		    )
-	  );
+		    );
+	begin
+	  match current_out with
+	    None -> ()
+	  | Some current_out' ->
+	      let diff = Compare_c.exact_compare current_out' expected_out in
+	      let s = Compare_c.compare_result_to_string diff in
+	      push2 (cocci_file ^ (t_to_s Failed), [s;time_str]) final_files
+	end;
+	 );
       )
     with exn ->
       let clean s =
@@ -463,7 +562,7 @@ let test_parse_cocci file =
   if not (file =~ ".*\\.cocci")
   then pr2 "warning: seems not a .cocci file";
 
-  let (mvs,xs,_,_,_,_,(grep_tokens,query,_,_),_) =
+  let (mvs,xs,_,_,_,_,_,query,_,_) =
     Parse_cocci.process file (Some !Config.std_iso) false in
   xs +> List.iter2 Pretty_print_cocci.unparse mvs;
   Format.print_newline();
@@ -478,16 +577,26 @@ let test_parse_cocci file =
       then Prepare_ocamlcocci.clean_file ocaml_script_file);
       (* Print the list of registered functions *)
       Prepare_ocamlcocci.test ());
-  Printf.printf "grep tokens\n";
-  (match grep_tokens with
-    None -> pr "No query"
-  | Some x -> pr (String.concat " || " x));
-  (* could update to accomodate WTCocciGrep *)
-  (match query with
-    None -> pr "No query"
-  | Some x ->
-      Printf.printf "glimpse tokens\n";
-      pr (String.concat "\nor on glimpse failure\n" x))
+  flush stdout; flush stderr;
+  match (!Flag.scanner,query) with
+    ((Flag.CocciGrep|Flag.NoScanner|Flag.GitGrep),(query,_,_,_)) ->
+      (match query with
+	None -> pr "No grep query"
+      | Some x ->
+	  Printf.printf "Grep query\n";
+	  pr (String.concat " || " x))
+  | (Flag.Glimpse,(_,query,_,_)) ->
+      (match query with
+	None -> pr "No glimpse query"
+      | Some x ->
+	  Printf.printf "Glimpse query\n";
+	  pr (String.concat "\nor on glimpse failure\n" x))
+  | (Flag.IdUtils,(_,_,_,query)) ->
+      (match query with
+	None -> pr "No idutils query"
+      | Some x ->
+	  Printf.printf "Idutils query\n";
+	  pr (Get_constants2.dep2c x))
 
 let print_link t a b =
   if not (a = b)
@@ -501,33 +610,37 @@ let print_dotted_link dst = function
     "" -> ()
   | src -> Printf.printf "  \"%s\" -> \"%s\" [style = dotted];\n" src dst
 
-let rec depto t from = function
-    Ast_cocci.Dep x | Ast_cocci.EverDep x | Ast_cocci.NeverDep x ->
-      print_link t from x
-  | Ast_cocci.AndDep(x,y) | Ast_cocci.OrDep(x,y) ->
-      depto t from x; depto t from y
-  | _ -> ()
+let depto t from d =
+  let rec loop = function
+      Ast_cocci.Dep x | Ast_cocci.EverDep x | Ast_cocci.NeverDep x ->
+	print_link t from x
+    | Ast_cocci.AndDep(x,y) | Ast_cocci.OrDep(x,y) ->
+	loop x; loop y
+    | _ -> () in
+  match d with
+    Ast_cocci.NoDep | Ast_cocci.FailDep -> ()
+  | Ast_cocci.ExistsDep d | Ast_cocci.ForallDep d -> loop d
 
 let test_rule_dependencies file =
   let t = Hashtbl.create 101 in
   if not (file =~ ".*\\.cocci")
   then pr2 "warning: seems not a .cocci file";
   Iso_pattern.verbose_iso := false;
-  let (_,xs,fvs,_,_,_,_,_) =
+  let (_,xs,_,fvs,_,_,_,_,_,_) =
     Parse_cocci.process file (Some !Config.std_iso) false in
   Printf.printf "digraph {\n";
   let prevrule = ref "" in
   List.iter2
     (fun def fvs ->
       match def with
-	Ast_cocci.ScriptRule (nm,_,dep,script_vars,_,_) ->
+	Ast_cocci.ScriptRule (nm,_,dep,script_vars,_,_,_) ->
 	  print_dotted_link nm !prevrule;
 	  prevrule := nm;
 	  depto t nm dep;
 	  List.iter (function (_,(parent,_),_,_) -> print_link t nm parent)
 	    script_vars
-      | Ast_cocci.InitialScriptRule (_,_,_,_,_)
-      | Ast_cocci.FinalScriptRule (_,_,_,_,_) -> ()
+      | Ast_cocci.InitialScriptRule (_,_,_,_,_,_)
+      | Ast_cocci.FinalScriptRule (_,_,_,_,_,_) -> ()
       | Ast_cocci.CocciRule (nm,(dep,_,_),_,_,_) ->
 	  print_dotted_link nm !prevrule;
 	  prevrule := nm;
