@@ -1,5 +1,5 @@
 (*
- * This file is part of Coccinelle, lincensed under the terms of the GPL v2.
+ * This file is part of Coccinelle, licensed under the terms of the GPL v2.
  * See copyright.txt in the Coccinelle source code for more information.
  * The Coccinelle source code can be obtained at http://coccinelle.lip6.fr
  *)
@@ -182,6 +182,7 @@ let rec type_pos t snp
         "pos_gen: Mixed match/patch type disjunctions not supported " ^
         "in position generator."
       )
+  | Ast0.ConjType(lp,tlist,pipelist,rp) -> None (* not sure *)
   | Ast0.ConstVol(const,t) ->
       let constructor ~item = Ast0.ConstVol(const,item) in
       item_wrap ~item:t ~item_posfn:type_pos ~constructor snp
@@ -220,11 +221,18 @@ let rec type_pos t snp
       let c ~item ~mc = Ast0.StructUnionDef(item,lcb,decldots,mc) in
       let alt() = mcode_wrap ~mc:rcb ~constructor:(c ~item:t) snp in
       item_wrap ~item:t ~item_posfn:type_pos ~constructor:(c ~mc:rcb) ~alt snp
+  | Ast0.TypeOfExpr(sizeofmc, lp, exp, rp) ->
+      let constructor ~mc = Ast0.TypeOfExpr(mc, lp, exp, rp) in
+      mcode_wrap ~mc:rp ~constructor snp
+  | Ast0.TypeOfType(sizeofmc, lp, typec, rp) ->
+      let _ = type_pos typec snp in (* sanity check for disj *)
+      let constructor ~mc = Ast0.TypeOfType(mc, lp, typec, rp) in
+      mcode_wrap ~mc:sizeofmc ~constructor snp
   | Ast0.TypeName(nm) ->
       let constructor ~mc = Ast0.TypeName(mc) in
       mcode_wrap ~mc:nm ~constructor snp
-  | Ast0.MetaType(mnmc,pure) ->
-      let constructor ~mc = Ast0.MetaType(mc,pure) in
+  | Ast0.MetaType(mnmc,cstr,pure) ->
+      let constructor ~mc = Ast0.MetaType(mc,cstr,pure) in
       mcode_wrap ~mc:mnmc ~constructor snp
   | Ast0.OptType(t) ->
       let constructor ~item = Ast0.OptType(item) in
@@ -367,8 +375,8 @@ let rec expression_pos exp snp
   | Ast0.MetaErr (mc, co, pu) -> (* is this ever within the rule body? *)
       let constructor ~mc = Ast0.MetaErr (mc, co, pu) in
       mcode_wrap ~mc ~constructor snp
-  | Ast0.MetaExpr(mc, co, ty, fo, pu) ->
-      let constructor ~mc = Ast0.MetaExpr (mc, co, ty, fo, pu) in
+  | Ast0.MetaExpr(mc, co, ty, fo, pu, bitfield) ->
+      let constructor ~mc = Ast0.MetaExpr (mc, co, ty, fo, pu, bitfield) in
       mcode_wrap ~mc ~constructor snp
   | Ast0.OptExp exp ->
       let constructor ~exp = Ast0.OptExp exp in
@@ -385,20 +393,18 @@ let exp_wrap ~exp ~constructor ?(alt = fun _ -> None) snp
 let rec declaration_pos decl snp
 : (Ast0.base_declaration Ast0.wrap * Snap.t) option =
   match Ast0.unwrap decl with
-  | Ast0.DisjDecl _
-  | Ast0.Ddots _
+    Ast0.DisjDecl _
+  | Ast0.ConjDecl _
   | Ast0.MetaDecl _
-  | Ast0.MetaField _
-  | Ast0.MetaFieldList _
   | Ast0.AsDecl _ ->
       None
-  | Ast0.Init(st, ty, id, eq, ini, sem) ->
+  | Ast0.Init(st, ty, id, attr, eq, ini, sem) ->
       let _ = type_pos ty snp in (* sanity check *)
-      let constructor ~id = Ast0.Init(st, ty, id, eq, ini, sem) in
+      let constructor ~id = Ast0.Init(st, ty, id, attr, eq, ini, sem) in
       id_wrap ~id ~constructor snp
-  | Ast0.UnInit(st, ty, id, sem) ->
+  | Ast0.UnInit(st, ty, id, attr, sem) ->
       let _ = type_pos ty snp in (* sanity check *)
-      let constructor ~id = Ast0.UnInit(st, ty, id, sem) in
+      let constructor ~id = Ast0.UnInit(st, ty, id, attr, sem) in
       id_wrap ~id ~constructor snp
   | Ast0.TyDecl (t, sem) ->
       let c ~item ~mc = Ast0.TyDecl(item, mc) in
@@ -420,6 +426,23 @@ let rec declaration_pos decl snp
       let constructor ~id = Ast0.FunProto(fninfo,id,lp1,params,va,rp1,sem) in
       id_wrap ~id ~constructor snp
 
+let rec field_pos decl snp
+: (Ast0.base_field Ast0.wrap * Snap.t) option =
+  match Ast0.unwrap decl with
+    Ast0.DisjField _
+  | Ast0.ConjField _
+  | Ast0.Fdots _
+  | Ast0.MetaField _
+  | Ast0.MetaFieldList _
+  | Ast0.Field(_, None, _, _) -> None
+  | Ast0.Field(ty, Some id, bf, sem) ->
+      let _ = type_pos ty snp in (* sanity check *)
+      let constructor ~id = Ast0.Field(ty, Some id, bf, sem) in
+      id_wrap ~id ~constructor snp
+  | Ast0.OptField(dec) ->
+      let constructor ~item = Ast0.OptField item in
+      item_wrap ~item:dec ~item_posfn:field_pos ~constructor snp
+
 let forinfo_pos f snp
 : (Ast0.base_forinfo Ast0.wrap * Snap.t) option =
   match Ast0.unwrap f with
@@ -457,6 +480,9 @@ let rec statement_pos s snp
   | Ast0.Include (incmc,filemc) ->
       let constructor ~mc = Ast0.Include(incmc, mc) in
       mcode_wrap ~mc:filemc ~constructor snp
+  | Ast0.MetaInclude (incmc,filemc) ->
+      let constructor ~mc = Ast0.MetaInclude(mc, filemc) in
+      mcode_wrap ~mc:incmc ~constructor snp
   | Ast0.Undef (defmc, id) ->
       let constructor ~id = Ast0.Undef(defmc, id) in
       id_wrap ~id ~constructor snp
